#ifndef _CODE_INTERMEDIATE_
#define _CODE_INTERMEDIATE_

/* take notice of the routines in cgoutputabc.c, cgoutputinter.c, and cgwrites.c when altering the values below */

enum t_instr_name{

	/* general instructions */
	
	LDL,
	STL,
	LDNL,
	STNL,
	LBYTE,
	LDC,
	ADC,
	EQC,
	EQD,
	LDLP,
	LDNLP,
	AJW,
	JUMP,
	CJUMP,
	RET,
	LDADDR,
	LDDESC,
	LDPI,
	GT,
	REV,
	DUP,

	ADD,
	SUB,
	MUL,
	DIV,
	REM,
	SUM,
	DIFF,
	PROD,
	AND,
	OR,
	XOR,
	SHL,
	SHR,
	NOT,
	ABS,
		
	LSHL,
	LSHR,
	XDBLE,
	CSNGL,
	
	BSUB,
	WSUB,
	
	MOVE,
	BCNT,
	WCNT,
	MINT,
	
	/* floating point instructions */
	
	FPDUP,
	FPREV,
	FPLDNLDB,
	FPSTNLDB,
	FPADD,
	FPSUB,
	FPMUL,
	FPDIV,
	FPREM,
	FPGT,
	FPEQ,
	
	FPLDZERODB,
	
	FPREMFIRST,
	FPREMSTEP,
	FPUSQRTFIRST,
	FPUSQRTSTEP,
	FPUSQRTLAST,
	
	FPUMULBY2,
	FPUDIVBY2,
	FPUEXPINC32,
	FPUEXPDEC32,
	
	FPINT,
	FPI32TOR64,
	FPSTNLI32,
	FPURM,
	
	/* instructions for flow of control */

	RTN,
	JMP_AP,
	JMP_EVAL,
	
	STOPP,
	RUNP,
	CALL,
	GAJW,

	/****************************************************************************
	 *																			*
	 *		some special instructions											*
	 *																			*
	 ****************************************************************************/
	
	LIBCALL,
	
	/*	instruction consisting of an array of inline instructions (output
		literally) */
	
	INLINE,
	
	/*	directive that groups transputer instruction in sequences that leave
		the transputer evaluation stack in a certain state: GROUP0 means: at
		this point no transputer registers are in use, etc. This enables the
		code generator to insert instructions to schedule locals at these points
	*/
	
	GROUP,

	/****************************************************************************
	 *																			*
	 *		The following instructions are also used as block types				*
	 *																			*
	 ****************************************************************************/
	
	/* block types that may fall through to next block and for which the stack
	   layout has not yet been established */

	JSR_EVAL,
	RELEASE,
	
	SUSPEND_UNEVAL,
	SUSPEND,
	
	CS,
	ENABLE_CS,
	DISABLE_CS,

	/* block types that eventually return and for which the stack layout is fixed */
	
	JSR,
	GCALL,
	JSR_AP,
	SIGNAL,

	/* block types that do not return */
	
	NORMAL_BLOCK,
	INCOMPLETE_BLOCK
};

enum t_dir_name {
	
	WORD_DIR,
	BYTE_DIR,
	ALIGN_DIR,
	EQU_DIR,
	
	EXTERN_DIR,
	GLOBAL_DIR,
	
	ATTRIB_DIR,
	CSEC_DIR,
	DSEC_DIR
};

enum t_inline_argument
{	
	/* expressions */
	
	ACOMMA,
	ABRACKET,
	
	AMONADD,
	AMONSUB,
	
	AMUL,
	ADIV,
	AMOD,
	
	AADD,
	ASUB,
	
	AOR,
	AAND,
	
	/* constants */
	
	AINTEGER,
	AFLOAT,
	ASTRING,					/* must be output as string "..." */

	ALABEL,						/* must be output with label conversion */
	AJLABEL,
	AGLABEL,
	ADLABEL,

	ALITERAL,					/* must be output literally */	

	/* actual parameters */
	
	APARAMETER,					/* index in param list */
	ALOCAL,						/* number of local label */	
	APARAMLIST,					/* param list */
	
	AUNDEFINED					/* used to detect errors */
};

enum t_inline_name {

	/* instructions */
	
	TINSTR,
	TMACRO,
	
	/* definitions */
	
	TWORD_DEF,
	TBYTE_DEF,
	TSTRING_DEF,
	TFLOAT_DEF,
	TLABEL_DEF,

	/* directives */
	
	TALIGN,
	TATTRIB,
	TCSEC,
	TDSEC,
	
	TEXTERN,
	TGLOBAL,
	
	TINCLUDE
};

#endif