
#ifndef _OFFSETS_
#define _OFFSETS_

#define MAX_YET_ARGS_NEEDED_ARITY	2	/* special entry for arity 0,1 and 2 */
#define MAX_EVAL_UPD				20	/* special entry for arity 0..19 */
#define MAX_CLOSURE_ARITY			254	/* and max size = 255 */

typedef enum LOCAL_OFFSET
{
	SAVE5			=-5,
	SAVE4			=-4,
	SAVE3			=-3,
	SAVE2			=-2,
	SAVE1			=-1,
	RETURN_ADDRESS	=0,
	B1				=1,
	B2				=2,
	B3				=3,
	B4				=4,
	B5				=5,
	B6				=6,
	A1				=7,
	A2				=8,
	A3				=9,
	A4				=10,
	EXTRA_LOCAL		=11,
	B_STACK_POINTER	=12,
	A_STACK_POINTER	=13,
	HEAP_START		=14,
	HEAP_END		=15

} LOCAL_OFFSET;

typedef enum DESC_INFO_OFFSET
{
	NIL_OFFSET		=0,

	STRING1_OFFSET	=1,
	STRING2_OFFSET	=2,
	
	FILE_OFFSET		=3,
	REAL_OFFSET		=4,
	
	INT_OFFSET		=5,
	BOOL_OFFSET		=6,
	CHAR_OFFSET		=7,
	
	TUPLE_OFFSET	=8,
	CONS_OFFSET		=9,
	IF_OFFSET		=10
	
} DESC_INFO_OFFSET;

typedef enum CODE_INFO_OFFSET
{
	START_CODE_OFFSET		=0,
	
	RESERVE_CODE_OFFSET		=1,
	RESERVE1_CODE_OFFSET	=2,
	
	REQUEST_CODE_OFFSET		=3,
	REQUEST1_CODE_OFFSET	=4,
	
	RELEASE_PROC_OFFSET		=5,
	RELEASE_PROC1_OFFSET	=6,
	RELEASE_REQUEST_OFFSET	=7,
	RELEASE_REQUEST1_OFFSET	=8,
	
	CHANNEL_CODE_OFFSET		=9,
	
	EA_PROCESS_CODE_OFFSET	=10,
	EA_REDUCE_CODE_OFFSET	=11
	
} CODE_INFO_OFFSET;

typedef enum GLOBAL_OFFSET
{
	DEBUGGER_CHANNEL	=-2,
	DEBUGGER_STATE		=-1,

	HIGH_SYSTEM_STACK	=0,
	LOW_SYSTEM_STACK	=1,

	HEAP_FREE_PTR		=2,
	HEAP_END_PTR		=3,
	CURRENT_HEAP		=4,
	GARBAGE_HEAP		=5,
	HEAP_SIZE			=6,

	COMM_INFO			=7,
	REDUCER_INFO		=8,
	GENERAL_INFO		=9,
	COLLECTOR_INFO		=10,
	DESC_INFO			=11,
	CODE_INFO			=12
		
} GLOBAL_OFFSET;

typedef enum COMM_OFFSET
{
	FIRST_MESS		=0,
	LAST_MESS		=1,
	
	GLOB_TABLE		=2,
	TABLE_SIZE		=3,
	TABLE_FREE		=4,
	
	CURR_PROC		=5,
	N_PROCS			=6,

	ROUT_VECTOR		=7,
	UROUT_VECTOR	=8,
	BROUT_VECTOR	=9,
	
	DEST_INFO		=10

} COMM_OFFSET;

#define N_LOCALS					16
#define N_A_PARAMETER_LOCALS		4
#define N_B_PARAMETER_LOCALS		6

#define N_PARAMETER_LOCALS			N_A_PARAMETER_LOCALS+N_B_PARAMETER_LOCALS+1		/* return + a + b  */

#define B_PAR_FIRST_LOCAL			B1
#define A_PAR_FIRST_LOCAL			A1

#define B_PAR_LAST_LOCAL			(B_PAR_FIRST_LOCAL+N_B_PARAMETER_LOCALS-1)
#define A_PAR_LAST_LOCAL			(A_PAR_FIRST_LOCAL+N_A_PARAMETER_LOCALS-1)

#define N_REGS						3
#define N_FLOAT_REGS				3

#define	CLOS_ARITY	(-2)
#define CLOS_TYPE	(-3)
#define CLOS_SIZE	(-4)

#define DESCRIPTOR_OFFSET	0
#define CODE_OFFSET			0
#define ARG1_OFFSET			1
#define ARG2_OFFSET			2
#define HNF_ARG_PTR_OFFSET	2

#define WAITLIST_OFFSET		1

#define CHAN_NEXT_OFFSET	ARG1_OFFSET
#define CHAN_PREV_OFFSET	ARG2_OFFSET
#define CHAN_TID_OFFSET		3
#define CHAN_TINDEX_OFFSET	4

#define CHANNEL_NODE_SIZE	5	/* code,next,prev,id,index	*/
#define REQUEST_NODE_SIZE	5	/* code,wait,reserved,id,index		*/

typedef enum STRING_OFFSET
{
	STRING_LENGTH_OFFSET 		=0,
	STRING_FIRST_CHAR_OFFSET	=1
	
} STRING_OFFSET;

typedef enum HARD_OFFSET
{
	HARD_STOP_WSP		=28,
	HARD_SIGNAL_CHANNEL	=29,
	HARD_ROOT			=30,
	HARD_DESCRIPTORS	=31

} HARD_OFFSET;

typedef enum ENVIRONMENT_OFFSET {

	ENV_REGS_OFFSET				=-4,				/* registers of reducer */
	ENV_PID_OFFSET				=-3,				/* highest bit set: suspended */

	ENV_ASTART_OFFSET			=-2,				/* bounds of stacks */
	ENV_BSTART_OFFSET			=-1
	
} ENVIRONMENT_OFFSET;

#define LDPI_SIZE			2

#define LAZY_ENTRY_OFFSET	1
#define ARITY_OFFSET		-1

#define LIB_ARG1			-3
#define LIB_ARG2			-4

typedef enum PRELUDE_LOCALS
{
	PRELUDE_RETURN			=0,
	
	/* below: unused words pushed by call */
	
	PRELUDE_SCRATCH1		=1,
	PRELUDE_SCRATCH2		=2,
	PRELUDE_SCRATCH3		=3
	
} PRELUDE_LOCALS;

#endif

