
#include <setjmp.h>

#include "ifinterface.h"

#include "cg.h"
#include "cginputabc.h"
#include "cgmemory.h"
#include "cgopt.h"
#include "cgoutputabc.h"
#include "cglabels.h"
#include "cginputs.h"
#include "cgscanner.h"
#include "cgwrites.h"
#include "cgoutputinter.h"
#include "cggraphs.h"
#include "cgstacks.h"
#include "cgcode.h"

VOID generate_code(CHAR *file_name)
{
	INSTRUCTION_BLOCK *intermediate_code;
	
	options.output_default_attributes=true;
		
	handle_compute_event();
	
	initialise_labels();	
	initialise_graphs();
    initialise_stacks();
	init_inline();
	init_writes();
	initialise_scanner();
    initialize_coding();

	set_code_stage("generating intermediate code");
	set_file_stage("opening \'%s\'",file_name);
	
	handle_compute_event();
	
	if (options.output_intermediate_code1)
	{
		/* write code with virtual locals */
		
		if ((codefile=open_intermediate_file (file_name,options.intermediate_extension1))!=NULL)
		{
			init_write_intermediate();
			intermediate_code=generate_intermediate_code(file_name);
			close_output_file();
			
		} else
		{
			intermediate_code=NULL;
			error("could not open intermediate code file (im1), compilation aborted.\n");
			abort_code_generation();
		}
	
	} else
		intermediate_code=generate_intermediate_code(file_name);
    
    if (nr_errors > 0 || nr_warnings > 0)	error("\n");
	if (nr_errors == 1)						error(" 1 error.");
	if (nr_errors > 1)						error(" %u errors.", nr_errors);
	if (nr_warnings == 1)					error("  1 warning.");
	if (nr_warnings > 1)					error("  %u warnings.", nr_warnings);
	if (nr_errors > 0 || nr_warnings > 0)	error("\n");
	
	if (nr_errors >= options.max_errors)
	{	error("maximum number of parse errors reached, compilation aborted.\n");
		abort_code_generation();
	}
    if (nr_errors > 0)
    {	error("compilation aborted\n");
    	abort_code_generation();
    }
	
	if (options.output_intermediate_code2)
	{
		/* write code with real locals */
		
		if ((codefile=open_intermediate_file (file_name,options.intermediate_extension2))!=NULL)
		{
			set_code_stage("writing intermediate code 2");
			write_intermediate_code(intermediate_code);
			close_output_file();
			
		} else
		{
			error("could not open intermediate code file (im2), compilation aborted.\n");
			abort_code_generation();
		}
	}

    if(options.optimise_jumps)
    {
    	set_code_stage("optimising jumps");
    	intermediate_code=do_jump_optimisation(intermediate_code);    
    }

	if (options.output_intermediate_code3)
	{
		/* write code with optimised jumps */
		
		if ((codefile=open_intermediate_file (file_name,options.intermediate_extension3))!=NULL)
		{
			set_code_stage("writing intermediate code 3");
			write_intermediate_code(intermediate_code);
			close_output_file();
			
		} else
		{
			error("could not open intermediate code file (im3), compilation aborted.\n");
			abort_code_generation();
		}
	}

	set_code_stage("checking labels");

    check_labels();
    
    if (nr_errors>0)
	{	error("compilation aborted\n");
		abort_code_generation();
	}
		
	set_code_stage("writing code");
    out_code(file_name,intermediate_code);    
}