#include "ifinterface.h"

#include "cggraphs.h"
#include "cgmemory.h"
#include "cglin.h"
#include "cglabels.h"

#include "ccnodes.h"
#include "ccmacros.h"

extern LABEL *nil_label;

static INSTRUCTION_GRAPH wait_graph;

/****************************************************************************************
 *                                                                                      *
 *						basic graph construction										*
 *																						*
 ****************************************************************************************/

INSTRUCTION_GRAPH g_new_node (INT16 code,INT16 arity,INT16 arg_size)
{
	register INSTRUCTION_GRAPH instruction;

	instruction=alloc_graph_mem(sizeof (INSTRUCTION_NODE)+arg_size);
	
	instruction->code		=code;
	instruction->arity		=arity;
	instruction->count		=0;
	instruction->used_mark	=0;
	instruction->order_left	=true;			/* slight preference */
	instruction->pref_local =NO_PREF;		/* not known yet */
	instruction->info		=NULL;

	/* the rest is set by calculating routines */
			
	return instruction;
}

static NODE_INFO *new_node_info(LABEL *descriptor, INT16 arity, INT16 size)
{
	register NODE_INFO *info;
	
	info=alloc_graph_mem(sizeof(NODE_INFO));
	
	info->descriptor=descriptor;
	info->node_arity=arity;
	info->node_size=size;
	
	return info;
}

VOID init_heap_pointer(BLOCK_GRAPH *block_graph)
{	
	block_graph->heap_ptr=g_instruction_int16(GLOAD_L,HEAP_END);
}

VOID init_a_stack_pointer(BLOCK_GRAPH *block_graph)
{	
	block_graph->a_stack_ptr=g_instruction_int16(GLOAD_L,A_STACK_POINTER);
}

VOID init_b_stack_pointer(BLOCK_GRAPH *block_graph)
{
	block_graph->b_stack_ptr=g_instruction_int16(GLOAD_L,B_STACK_POINTER);
}

INSTRUCTION_GRAPH g_instruction0 (	INT16 code)
{
	INSTRUCTION_GRAPH instruction;
	
	/* a dummy argument must be included: this node might be overwritten
	   by a 'local' node */
	   
	instruction=g_new_node (code,0,sizeof (PARAMETER));
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction1 (	INT16 code,
									INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,1,sizeof (PARAMETER));
	instruction->parameters[0].pointer=graph;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction2 (	INT16 code,
									INSTRUCTION_GRAPH graph1,
									INSTRUCTION_GRAPH graph2)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,2,2*sizeof (PARAMETER));

	instruction->parameters[0].pointer=graph1;
	instruction->parameters[1].pointer=graph2;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction3 (	INT16 code,
									INSTRUCTION_GRAPH graph1,
									INSTRUCTION_GRAPH graph2,
									INSTRUCTION_GRAPH graph3)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,3,3*sizeof (PARAMETER));

	instruction->parameters[0].pointer=graph1;
	instruction->parameters[1].pointer=graph2;
	instruction->parameters[2].pointer=graph3;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction4 (	INT16 code,
											INSTRUCTION_GRAPH graph1,
											INSTRUCTION_GRAPH graph2,
											INSTRUCTION_GRAPH graph3,
											INSTRUCTION_GRAPH graph4)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,4,4*sizeof (PARAMETER));

	instruction->parameters[0].pointer=graph1;
	instruction->parameters[1].pointer=graph2;
	instruction->parameters[2].pointer=graph3;
	instruction->parameters[3].pointer=graph4;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction5 (	INT16 code,
											INSTRUCTION_GRAPH graph1,
											INSTRUCTION_GRAPH graph2,
											INSTRUCTION_GRAPH graph3,
											INSTRUCTION_GRAPH graph4,
											INSTRUCTION_GRAPH graph5)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,5,5*sizeof (PARAMETER));

	instruction->parameters[0].pointer=graph1;
	instruction->parameters[1].pointer=graph2;
	instruction->parameters[2].pointer=graph3;
	instruction->parameters[3].pointer=graph4;
	instruction->parameters[4].pointer=graph5;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_int32 (INT16 code, INT32 value)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,0,sizeof (PARAMETER));
	instruction->parameters[0].int32=value;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_int16 (INT16 code, INT16 value)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,0,sizeof (PARAMETER));
	instruction->parameters[0].int16=value;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_label (INT16 code, LABEL *label)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,0,sizeof (PARAMETER));
	instruction->parameters[0].label=label;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_float (INT16 code, FLOAT *r)
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,0,sizeof (FLOAT));
	((FLOAT*)&instruction->parameters[0])->f = r->f;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_int16_int16 (INT16 code, INT16 n1, INT16 n2)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code ,0,2*sizeof (PARAMETER));
	instruction->parameters[0].int16=n1;
	instruction->parameters[1].int16=n2;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_p_int32 (INT16 code, INSTRUCTION_GRAPH p, INT32 i)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,0,2*sizeof (PARAMETER));
	instruction->parameters[0].pointer=p;
	instruction->parameters[1].int32=i;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_p_int16 (INT16 code, INSTRUCTION_GRAPH p, INT16 i)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,0,2*sizeof (PARAMETER));
	instruction->parameters[0].pointer=p;
	instruction->parameters[1].int16=i;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_label_int16 (INT16 code, LABEL *label,INT16 i)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,0,2*sizeof (PARAMETER));
	instruction->parameters[0].label=label;
	instruction->parameters[1].int16=i;
	
	return instruction;
}

INSTRUCTION_GRAPH g_instruction_p_label_int16 (INT16 code, INSTRUCTION_GRAPH graph, LABEL *label, INT16 i)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (code,0,3*sizeof (PARAMETER));
	instruction->parameters[0].pointer=graph;
	instruction->parameters[1].label=label;
	instruction->parameters[2].int16=i;
	
	return instruction;
}

/********************************************************************************
 *																				*
 *		arithmetic																*
 *																				*
 ********************************************************************************/

INSTRUCTION_GRAPH g_negate2 (INSTRUCTION_GRAPH graph)
{
	/* compute checked -graph with 2 registers */
	
	register INSTRUCTION_GRAPH zero, instruction;

	zero = g_load_const ((INT32)0);
	instruction = g_sub (zero,graph);
	
	return instruction;
}

INSTRUCTION_GRAPH g_negate_no_check (INSTRUCTION_GRAPH graph)
{
	/* just compute 0 - graph, with unchecked subtraction */
	
	register INSTRUCTION_GRAPH zero, instruction;

	zero = g_load_const ((INT32)0);
	instruction = g_diff (zero,graph);
	
	return instruction;
}

INSTRUCTION_GRAPH g_prod (INSTRUCTION_GRAPH graph_1, INSTRUCTION_GRAPH graph_2)
{
	if (options.check_arithmetic)
		return g_mul(graph_1,graph_2);
	else
		return g_instruction2(GPROD,graph_1,graph_2);
}

INSTRUCTION_GRAPH g_addI (INSTRUCTION_GRAPH graph_1, INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions and replace ldc, add by adc */
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
				return g_load_const(graph_2->parameters[0].int32 + constant);
				
			if (constant == (INT32)0)
				return graph_2;
			
			return g_adc (graph_2, constant);
		}
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == (INT32)0)
				return graph_1;
				
			return g_adc (graph_1, constant);
		}
	}
	
	return g_add (graph_1,graph_2);	
}

INSTRUCTION_GRAPH g_subI (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{	
	if (options.optimise_arith)
	{
		/* evaluate constant expressions and replace ldc, add by adc */
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (graph_1->code == GLOAD_CONST)
				return g_load_const(graph_1->parameters[0].int32 - constant);
				
			if (constant == (INT32)0)
				return graph_1;
			
			return g_adc (graph_1, -constant);
		
		}
		
		if (graph_1->code == GLOAD_CONST && (graph_1->parameters[0].int32 == (INT32)0))
			return g_negate1(graph_2);
	}
	
	return g_sub (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_mulI (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/*	evaluate constant expressions and rearrange operands if this
			is probably more efficient: mul will be replaced by prod,
			unless arithmetic must be checked */
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
				return g_load_const(graph_2->parameters[0].int32 * constant);
				
			if (constant > -SMALL_CONSTANT && constant <= SMALL_CONSTANT)
			{	if (constant == (INT32)0)
					return graph_1;
				if (constant == (INT32)1)
					return graph_2;
				if (constant == (INT32)-1)
					return g_negate1(graph_2);
				return g_prod (graph_2, graph_1);
			}
			return g_prod (graph_1,graph_2); /* assume second arg smaller */		
		}
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant > -SMALL_CONSTANT && constant <= SMALL_CONSTANT)
			{	if (constant == (INT32)0)
					return graph_2;
				if (constant == (INT32)1)
					return graph_1;
				if (constant == (INT32)-1)
					return g_negate1(graph_1);
				return g_prod(graph_1,graph_2);
			}
			return g_prod (graph_2, graph_1); /* assume first arg smaller */
		}
	}	
	return g_prod (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_divI (INSTRUCTION_GRAPH graph_1, INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == 0)
			{
				error("divI: division by 0 (ignored)\n");
				return graph_1;
			}
			
			if (graph_1->code == GLOAD_CONST)
				return g_load_const(graph_1->parameters[0].int32 / constant);
				
			if (constant == 1)
				return graph_1;
			if (constant == -1)
				return g_negate1(graph_1);
			return g_div(graph_1,graph_2);
		}
		
		if (graph_1->code == GLOAD_CONST && (graph_1->parameters[0].int32 == (INT32)0))
			return graph_1;
	}
	return g_div (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_modI (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == 0)
			{
				error("modI: division by 0 (instruction ignored)\n");
				return graph_1;
			}
			if (graph_1->code == GLOAD_CONST)
				return g_load_const(graph_1->parameters[0].int32 % constant);
			
			if (constant == 1 || constant == -1)
				return g_load_const((INT32) 0);
			
			return g_mod(graph_1,graph_2);			
		}
		
		if (graph_1->code == GLOAD_CONST && (graph_1->parameters[0].int32 == (INT32)0))
			return graph_1;
	}
	return g_mod (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_andI (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
				return g_load_const(graph_2->parameters[0].int32 & constant);
			if (constant == (INT32)0)
				return graph_1;
			if (constant == (INT32)-1)
				return graph_2;
			
			return g_and (graph_1, graph_2);
	
		}
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == (INT32)0)
				return graph_2;
			if (constant == (INT32)-1)
				return graph_1;
		}
	}
	return g_and (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_andB (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
				return g_load_const(graph_2->parameters[0].int32 & constant);
			if (constant == (INT32)false)
				return graph_1;
			if (constant == (INT32)true)
				return graph_2;
			return g_and (graph_1, graph_2);
	
		}
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == (INT32)false)
				return graph_2;
			if (constant == (INT32)true)
				return graph_1;
		}
	}
	return g_and (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_orI (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
				return g_load_const(graph_2->parameters[0].int32 | constant);
			if (constant == (INT32)0)
				return graph_2;
			if (constant == (INT32)-1)
				return graph_1;
			return g_or (graph_1, graph_2);
		}
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == (INT32)0)
				return graph_1;
			if (constant == (INT32)-1)
				return graph_2;
			return g_or (graph_1, graph_2);
		}
	}
	return g_or (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_orB (INSTRUCTION_GRAPH graph_1, INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
				return g_load_const(graph_2->parameters[0].int32 | constant);
			if (constant == (INT32)false)
				return graph_2;
			if (constant == (INT32)true)
				return graph_1;
			return g_or (graph_1, graph_2);
		}
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == (INT32)false)
				return graph_1;
			if (constant == (INT32)true)
				return graph_2;
			return g_or (graph_1, graph_2);
		}
	}
	return g_or (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_xorI (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
				return g_load_const(graph_2->parameters[0].int32 ^ constant);
			if (constant == (INT32)0)
				return graph_2;
			if (constant == (INT32)-1)
				return g_not(graph_2);
			return g_xor (graph_1, graph_2);
	
		}
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == (INT32)0)
				return graph_1;
			if (constant == (INT32)-1)
				return g_not(graph_1);
		}
	}
	return g_xor (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_xorB (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
				return g_load_const(graph_2->parameters[0].int32 ^ constant);
			if (constant == (INT32)false)
				return graph_2;
			if (constant == (INT32)true)
				return g_eqc(graph_2,(INT32)false);
			return g_xor (graph_1, graph_2);
		}
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant == (INT32)false)
				return graph_1;
			if (constant == (INT32)true)
				return g_eqc(graph_1,(INT32)false);		
		}
	}
	return g_xor (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_notI (INSTRUCTION_GRAPH graph_1)
{
	if (options.optimise_arith && graph_1->code == GLOAD_CONST)
		return g_load_const(~graph_1->parameters[0].int32);
	
	return g_not(graph_1);
}

INSTRUCTION_GRAPH g_notB (INSTRUCTION_GRAPH graph_1)
{
	if (options.optimise_arith && graph_1->code == GLOAD_CONST)
		return g_load_const(!graph_1->parameters[0].int32);
	
	return g_eqc(graph_1,(INT32)false);
}

INSTRUCTION_GRAPH g_L_shiftl (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant < 0 || constant >= 32)
			{
				error ("L_shiftI: logic shift left by invalid amount (zeroed)\n");
				return g_load_const((INT32)0);
			}
			if (graph_1->code == GLOAD_CONST)
				return g_load_const((U_INT32)graph_1->parameters[0].int32 << constant);
			if (constant == (INT32)0)
				return graph_1;
			return g_lsl (graph_1, graph_2);
		}
		
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (constant == (INT32)0)
				return graph_1;
		}
	}
	return g_lsl (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_L_shiftr (INSTRUCTION_GRAPH graph_1, INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant < 0 || constant >= 32)
			{
				error ("L_shiftr: logic shift right by invalid amount (zeroed)\n");
				return g_load_const((INT32)0);
			}
			if (graph_1->code == GLOAD_CONST)
				return g_load_const((U_INT32)graph_1->parameters[0].int32 >> constant);
			if (constant == (INT32)0)
				return graph_1;
			return g_lsr (graph_1, graph_2);
		}
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (constant == (INT32)0)
				return graph_1;		
		}
	}
	return g_lsr (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_A_shiftl (INSTRUCTION_GRAPH graph_1, INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant < 0 || constant >= 32)
			{
				error ("A_shiftI: arithmetic shift left by invalid amount (zeroed)\n");
				return g_load_const((INT32)0);	
			}
			if (graph_1->code == GLOAD_CONST)
				return g_load_const((INT32)graph_1->parameters[0].int32 << constant);
			if (constant == (INT32)0)
				return graph_1;
			return g_asl (graph_1, graph_2);
		}
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (constant == (INT32)0)
				return graph_1;		
		}
	}
	return g_asl (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_A_shiftr (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			
			if (constant < 0 || constant >= 32)
			{
				error ("A_shiftr: arithmetic shift right by invalid amount (zeroed)\n");
				return g_load_const((INT32)0);
			}
			if (graph_1->code == GLOAD_CONST)
				return g_load_const((INT32)graph_1->parameters[0].int32 >> constant);
			if (constant == (INT32)0)
				return graph_1;
			return g_asr (graph_1, graph_2);
		}
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_1->parameters[0].int32;
			
			if (constant == (INT32)0)
				return graph_1;		
		}
	}
	return g_asr (graph_1,graph_2);
}

INSTRUCTION_GRAPH g_eqI (INSTRUCTION_GRAPH graph_1, INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		if (graph_1->code == GLOAD_CONST)
		{	INT32 constant1;
		
			constant1 = graph_1->parameters[0].int32;
			
			if (graph_2->code == GLOAD_CONST)
			{
				INT32 constant2;
				
				constant2=graph_2->parameters[0].int32;
				
				if (constant1==constant2)
					return g_load_const((INT32)true);
				return g_load_const((INT32)false);
			
			}
			return g_eqc (graph_2, constant1);
	
		}
		if (graph_2->code == GLOAD_CONST)
		{	INT32 constant;
		
			constant = graph_2->parameters[0].int32;
			return g_eqc (graph_1, constant);
		
		}
	}
	return g_eq (graph_2,graph_1);
}

INSTRUCTION_GRAPH g_gtI (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith && graph_1->code==GLOAD_CONST && graph_2->code==GLOAD_CONST)
	{
		if (graph_1->parameters[0].int32 > graph_2->parameters[0].int32)
			return g_load_const((INT32)true);
		return g_load_const((INT32)false);
	}
	return g_gt (graph_1,graph_2);
}

/********************************************************************************
 *																				*
 *		floating point arithmetic												*
 *																				*
 ********************************************************************************/

INSTRUCTION_GRAPH g_fload_const (FLOAT *f)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (GFLOAD_CONST,0,2*sizeof (PARAMETER));
	instruction->parameters[0].int32=f->join.high;
	instruction->parameters[1].int32=f->join.low;

	return instruction;
}

INSTRUCTION_GRAPH g_addR (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_1->code == GFLOAD_CONST)
		{	FLOAT constant1;
		
			constant1.join.high	= graph_1->parameters[0].int32;
			constant1.join.low	= graph_1->parameters[1].int32;
			
			if (graph_2->code == GFLOAD_CONST)
			{	FLOAT constant2, result;
			
				constant2.join.high	= graph_2->parameters[0].int32;
				constant2.join.low	= graph_2->parameters[1].int32;
			
				result.f=constant1.f + constant2.f;
				
				return g_fload_const(&result);
			}
			if (constant1.f == (DOUBLE)0.0)
				return graph_2;
				
			return g_fadd(graph_1,graph_2);
		}
		
		if (graph_2->code == GFLOAD_CONST)
		{	FLOAT constant2;

			constant2.join.high	= graph_2->parameters[0].int32;
			constant2.join.low	= graph_2->parameters[1].int32;
					
			if (constant2.f == (DOUBLE)0.0)
				return graph_1;
		}
	}
	return g_fadd(graph_1,graph_2);
}

INSTRUCTION_GRAPH g_subR (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/* evaluate constant expressions */
		
		if (graph_2->code == GFLOAD_CONST)
		{	FLOAT constant2;
		
			constant2.join.high	= graph_2->parameters[0].int32;
			constant2.join.low	= graph_2->parameters[1].int32;
			
			if (graph_1->code == GFLOAD_CONST)
			{	FLOAT constant1, result;
			
				constant1.join.high	= graph_1->parameters[0].int32;
				constant1.join.low	= graph_1->parameters[1].int32;
			
				result.f=constant1.f - constant2.f;
				
				return g_fload_const(&result);
			}
			if (constant2.f == (DOUBLE)0.0)
				return graph_1;
		}		
	}
	return g_fsub(graph_1,graph_2);
}

INSTRUCTION_GRAPH g_mulR (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/*	evaluate constant expressions and generate mulby2, mulby2^32,
			divby2 and divby2^32 instructions */
		
		if (graph_1->code == GFLOAD_CONST)
		{	FLOAT constant1;
		
			constant1.join.high	= graph_1->parameters[0].int32;
			constant1.join.low	= graph_1->parameters[1].int32;
			
			if (graph_2->code == GFLOAD_CONST)
			{	FLOAT constant2, result;
			
				constant2.join.high	= graph_2->parameters[0].int32;
				constant2.join.low	= graph_2->parameters[1].int32;
			
				result.f=constant1.f * constant2.f;
				
				return g_fload_const(&result);
			}
			
			/* first argument is a constant, check for: 0, 1, -1, 2, 0.5, (2^32, 2^-32) */
			
			if (constant1.f == (DOUBLE)0.0)
				return graph_1;
			if (constant1.f == (DOUBLE)1.0)
				return graph_2;
			if (constant1.f == (DOUBLE)-1.0)
			{	FLOAT zero;
				zero.f=0.0;
				return g_fsub(g_fload_const(&zero),graph_2);
			}
			if (constant1.f == (DOUBLE)2.0)
				return g_fmulby2 (graph_2);
			if (constant1.f == (DOUBLE)0.5)
				return g_fdivby2 (graph_2);
			/*
			if (constant1.f == (DOUBLE)2^32)
				return g_fexpinc32 (graph_2);
			if (constant1.f == (DOUBLE)2^-32)
				return g_fexpdec32 (graph_2);
			*/
			
			return g_fmul(graph_1,graph_2);
		}
		
		if (graph_2->code == GFLOAD_CONST)
		{	FLOAT constant2;

			constant2.join.high	= graph_2->parameters[0].int32;
			constant2.join.low	= graph_2->parameters[1].int32;
			
			if (constant2.f == (DOUBLE)0.0)
				return graph_2;
			if (constant2.f == (DOUBLE)1.0)
				return graph_1;
			if (constant2.f == (DOUBLE)-1.0)
			{	FLOAT zero;
				zero.f=0.0;
				return g_fsub(g_fload_const(&zero),graph_1);
			}
			if (constant2.f == (DOUBLE)2.0)
				return g_fmulby2 (graph_1);
			if (constant2.f == (DOUBLE)0.5)
				return g_fdivby2 (graph_1);
		}
	}

	return g_fmul(graph_1,graph_2);
}

INSTRUCTION_GRAPH g_divR (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/*	evaluate constant expressions and generate mulby2, mulby2^32,
			divby2 and divby2^32 instructions */
		
		if (graph_2->code == GFLOAD_CONST)
		{	FLOAT constant2;
		
			/* second argument is a constant, check for: 0, 1, -1, 2, 0.5, (2^32, 2^-32) */

			constant2.join.high	= graph_2->parameters[0].int32;
			constant2.join.low	= graph_2->parameters[1].int32;
			
			if (constant2.f == (DOUBLE)0.0)
			{
				error("divR: division by zero (ignored)\n");
				return graph_1;
			}

			if (graph_1->code == GFLOAD_CONST)
			{	FLOAT constant1, result;
			
				constant1.join.high	= graph_1->parameters[0].int32;
				constant1.join.low	= graph_1->parameters[1].int32;
			
				result.f=constant1.f / constant2.f;
				
				return g_fload_const(&result);
			}

			if (constant2.f == (DOUBLE)1.0)
				return graph_1;
			if (constant2.f == (DOUBLE)-1.0)
			{	FLOAT zero;
				zero.f=0.0;
				return g_fsub(g_fload_const(&zero),graph_1);
			}
			if (constant2.f == (DOUBLE)2.0)
				return g_fdivby2 (graph_1);
			if (constant2.f == (DOUBLE)0.5)
				return g_fmulby2 (graph_1);
		
			return g_fdiv(graph_1,graph_2);
		}

		if (graph_1->code == GFLOAD_CONST)
		{	FLOAT constant1;
		
			constant1.join.high	= graph_1->parameters[0].int32;
			constant1.join.low	= graph_1->parameters[1].int32;

			if (constant1.f == (DOUBLE)0.0)
				return graph_1;
		}
	}

	return g_fdiv(graph_1,graph_2);
}

INSTRUCTION_GRAPH g_remR (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (options.optimise_arith)
	{
		/*	evaluate constant expressions */
		
		if (graph_2->code == GFLOAD_CONST)
		{	FLOAT constant2;
		
			/* second argument is a constant, check for: 0 */

			constant2.join.high	= graph_2->parameters[0].int32;
			constant2.join.low	= graph_2->parameters[1].int32;
			
			if (constant2.f == (DOUBLE)0.0)
			{
				error("remR: division by zero (ignored)\n");
				return graph_1;
			}

			return g_frem(graph_1,graph_2);
		}

		if (graph_1->code == GFLOAD_CONST)
		{	FLOAT constant1;
		
			constant1.join.high	= graph_1->parameters[0].int32;
			constant1.join.low	= graph_1->parameters[1].int32;

			if (constant1.f == (DOUBLE)0.0)
				return graph_1;
		}
	}

	return g_frem(graph_1,graph_2);
}

INSTRUCTION_GRAPH g_eqR (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (graph_1->code == GFLOAD_CONST && graph_2->code == GFLOAD_CONST)
	{
		FLOAT constant1, constant2;
		
		constant1.join.high	= graph_1->parameters[0].int32;
		constant1.join.low	= graph_1->parameters[1].int32;
		constant2.join.high	= graph_2->parameters[0].int32;
		constant2.join.low	= graph_2->parameters[1].int32;

		if (constant1.f == constant2.f)
			return g_load_const((INT32)true);
		return g_load_const((INT32)false);
	}
	return g_feq(graph_1,graph_2);
}

INSTRUCTION_GRAPH g_gtR (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (graph_1->code == GFLOAD_CONST && graph_2->code == GFLOAD_CONST)
	{
		FLOAT constant1, constant2;
		
		constant1.join.high	= graph_1->parameters[0].int32;
		constant1.join.low	= graph_1->parameters[1].int32;
		constant2.join.high	= graph_2->parameters[0].int32;
		constant2.join.low	= graph_2->parameters[1].int32;

		if (constant1.f > constant2.f)
			return g_load_const((INT32)true);
		return g_load_const((INT32)false);
	}
	return g_fgt(graph_1,graph_2);
}

/************************************************************************************
 *																					*
 *		descriptor operations														*
 *																					*
 ************************************************************************************/

INSTRUCTION_GRAPH g_load_desc_id (LABEL *desc, INT16 arity)
{	
	if (desc==nil_label && arity==0)
		return g_load_const(0);

	return g_instruction_label_int16 (GLOAD_DESC,desc,arity);
}

INSTRUCTION_GRAPH g_equal_desc_id (INSTRUCTION_GRAPH graph, LABEL *desc, INT16 arity)
{
	if (desc==nil_label && arity==0)
		return g_eqc(graph,0);

	return g_instruction_p_label_int16 (GEQD,graph,desc,arity);
}

/************************************************************************************
 *																					*
 *		create and fill graphs														*
 *																					*
 ************************************************************************************/

INSTRUCTION_GRAPH g_create_string (STRING *string)
{
	INSTRUCTION_GRAPH graph_1, graph_2;
	LABEL *str_label;
	
	str_label=define_string (string);
	
	graph_1=g_load_addr (str_label);
	graph_2=g_instruction_p_int16(GCREATE_S,graph_1,string->length);
	
	return graph_2;
}

INSTRUCTION_GRAPH g_create_3(INSTRUCTION_GRAPH g1,INSTRUCTION_GRAPH g2,INSTRUCTION_GRAPH g3,INT16 arity,INT16 size,LABEL *desc)
{
	INSTRUCTION_GRAPH graph_1;
	
	graph_1=g_instruction3(GCREATE,g1,g2,g3);
	graph_1->info=new_node_info(desc,arity,size);
	
	return graph_1;
}

INSTRUCTION_GRAPH g_create_4(INSTRUCTION_GRAPH g1,INSTRUCTION_GRAPH g2,INSTRUCTION_GRAPH g3,INSTRUCTION_GRAPH g4,INT16 arity,INT16 size,LABEL *desc)
{
	INSTRUCTION_GRAPH graph_1;
	
	graph_1=g_instruction4(GCREATE,g1,g2,g3,g4);
	graph_1->info=new_node_info(desc,arity,size);
	
	return graph_1;
}

INSTRUCTION_GRAPH g_create_m(INT16 m,INT16 arity,INT16 size,LABEL *desc)
{
	INSTRUCTION_GRAPH graph_1;
	
	graph_1=g_new_node(GCREATE,(m),(m)*sizeof (PARAMETER));
	graph_1->info=new_node_info(desc,arity,size);
	
	return graph_1;
}

INSTRUCTION_GRAPH g_fill_3 (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph, INSTRUCTION_GRAPH graph_2,INSTRUCTION_GRAPH graph_3, INT16 arity,INT16 size,LABEL *desc)
{
	register INSTRUCTION_GRAPH instruction;

	/*	if the wait graph is a constant, it does not have to be extracted from
		this node, so the save field can stay NULL */
	
	if (wait_graph->code==GLOAD_CONST)
		wait_graph=NULL;
	
	if (graph_1->code==GFILL)
	{
		/*	if address is another FILL node, some closure is overwritten in the
			same basic block. The arity of this closure should be at least 4, else
			the node is too small. The passed waitgraph should point to a graph that
			yields the waiting list of the node that is about to be overwritten.
			The underlying node either already has such graph, or must be provided
			with this one. If no waitgraph is present it apparantly does not need
			to be saved.
		*/

		if (graph_1->arity >= 4)
		{
			/* if original waitlist isn't saved yet, do it now */

			if (graph_1->parameters[1].pointer==NULL)
				graph_1->parameters[1].pointer=wait_graph;
			
			/* overwrite the old node */
			
			graph_1->parameters[2].pointer=graph_2;
			graph_1->parameters[3].pointer=graph_3;
			graph_1->arity=4;
			
			graph_1->info->node_arity=arity;
			graph_1->info->node_size=size;
			graph_1->info->descriptor=desc;

			instruction	=graph_1;
			graph_1		=graph_1->parameters[0].pointer;
						
		} else
			internal_error("g_fill_3: small node filled by large one");
	} else
	{
		instruction=g_instruction4(GFILL,graph_1,wait_graph,graph_2,graph_3);
		instruction->info=new_node_info(desc,arity,size);
	}

	/*	if the underlying node is a create node, set nodes that are overwritten
		to NULL. Note that fill_3 fills only 2 fields
	*/
	
	if (graph_1->code==GCREATE)
	{
		INT16 arity=graph_1->arity;
		if (arity>2)
			arity=2;
		
		while (arity>0)
			graph_1->parameters[--arity].pointer=NULL;
	}
	
	return instruction;	 
}

INSTRUCTION_GRAPH g_fill_4 (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph, INSTRUCTION_GRAPH graph_2,INSTRUCTION_GRAPH graph_3,INSTRUCTION_GRAPH graph_4, INT16 arity,INT16 size,LABEL *desc)
{
	register INSTRUCTION_GRAPH instruction;

	if (wait_graph->code==GLOAD_CONST)
		wait_graph=NULL;

	if (graph_1->code==GFILL)
	{	if (graph_1->arity >= 5)
		{
			/* if original waitlist isn't saved yet, do it now */
			
			if (graph_1->parameters[1].pointer==NULL)
				graph_1->parameters[1].pointer=wait_graph;

			graph_1->parameters[2].pointer=graph_2;
			graph_1->parameters[3].pointer=graph_3;
			graph_1->parameters[4].pointer=graph_4;
			graph_1->arity=5;

			graph_1->info->node_arity=arity;
			graph_1->info->node_size=size;
			graph_1->info->descriptor=desc;

			instruction	=graph_1;
			graph_1		=graph_1->parameters[0].pointer;
			
		} else
			internal_error("g_fill_4: small node filled by large one");
	} else
	{
		instruction=g_instruction5(GFILL,graph_1,wait_graph,graph_2,graph_3,graph_4);
		instruction->info=new_node_info(desc,arity,size);
	}
	
	if (graph_1->code==GCREATE)
	{
		INT16 arity=graph_1->arity;
		if (arity>3)
			arity=3;
					
		while (arity>0)
			graph_1->parameters[--arity].pointer=NULL;
	}

	return instruction;	 
}

INSTRUCTION_GRAPH g_fill_m (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph, INT16 n_args, INT16 arity,INT16 size,LABEL *desc)
{
	register INSTRUCTION_GRAPH instruction;

	if (wait_graph->code==GLOAD_CONST)
		wait_graph=NULL;

	if (graph_1->code==GFILL)
	{	
		if(graph_1->arity < n_args+2)
			internal_error("g_fill_m: small node filled by larger one");
		
		if (graph_1->parameters[1].pointer==NULL)
			graph_1->parameters[1].pointer=wait_graph;
		
		graph_1->arity=n_args+2;

		graph_1->info->node_arity=arity;
		graph_1->info->node_size=size;
		graph_1->info->descriptor=desc;

		return graph_1;
	}

	instruction=g_new_node (GFILL,n_args+2,(n_args+2)*sizeof (PARAMETER));
	instruction->parameters[0].pointer=graph_1;
	instruction->parameters[1].pointer=wait_graph;
	instruction->info=new_node_info(desc,arity,size);

	return instruction;
}

VOID optimise_fill_m (INSTRUCTION_GRAPH instruction, INT16 n_args)
{
	register INSTRUCTION_GRAPH graph_1;

	graph_1=instruction->parameters[0].pointer;

	if (graph_1->code==GCREATE)
	{	INT16 arity;
	
		arity=graph_1->arity;
		if (arity > n_args)
			arity = n_args;
		
		while (arity>0)
			graph_1->parameters[--arity].pointer=NULL;
	}
}

INSTRUCTION_GRAPH g_fill_copy(	INSTRUCTION_GRAPH source,
								INSTRUCTION_GRAPH dest,
								INSTRUCTION_GRAPH wait_graph,
								INT16 n_bytes)
{
	register INSTRUCTION_GRAPH instruction;
	
	if (wait_graph->code==GLOAD_CONST)
		wait_graph=NULL;

	instruction=g_new_node (GFILL_COPY,0,4*sizeof (PARAMETER));
	instruction->parameters[0].pointer=source;
	instruction->parameters[1].pointer=dest;
	instruction->parameters[2].pointer=wait_graph;
	instruction->parameters[3].int16=n_bytes;
	
	return instruction;
}

INSTRUCTION_GRAPH g_fill_r (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph,INSTRUCTION_GRAPH graph_2,INSTRUCTION_GRAPH graph_3)
{
	register INSTRUCTION_GRAPH instruction;

	if (wait_graph->code==GLOAD_CONST)
		wait_graph=NULL;

	instruction=g_instruction4(GFILL_R,graph_1,wait_graph,graph_2,graph_3);
	
	if (graph_1->code==GCREATE)
	{	INT16 arity=graph_1->arity;
		if (arity>3)
			arity=3;
		
		while (arity>0)
			graph_1->parameters[--arity].pointer=NULL;
	}

	return instruction;	 
}

/********************************************************************************
 *																				*
 *		transparant fill (leaves undefined fields intact)						*
 *																				*
 ********************************************************************************/
 
/*	this routine leaves existing arguments of the node intact for each new
	argument that is NULL (arity will not be changed). This is mainly used
	to put a waiting list in a (locked) node. The size of the underlying
	nodes should be at least 2. */

INSTRUCTION_GRAPH g_transparant_fill_4 (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph,INSTRUCTION_GRAPH graph_2,INSTRUCTION_GRAPH graph_3,INSTRUCTION_GRAPH graph_4,INT16 arity,INT16 size,LABEL *desc)
{
	register INSTRUCTION_GRAPH instruction;

	if (wait_graph->code==GLOAD_CONST)
		wait_graph=NULL;

	if (graph_1->code==GFILL)
	{
		if (graph_4!=NULL && graph_1->arity < 5)
			internal_error("g_transparant_fill_4: small node filled by larger (size 3) one");
		else if (graph_1->arity < 4)
			internal_error("g_transparant_fill_4: small node filled by larger (size 2) one");
		else
		{
			if (graph_1->parameters[1].pointer==NULL)
				graph_1->parameters[1].pointer=wait_graph;

			if (graph_2!=NULL)
				graph_1->parameters[2].pointer=graph_2;
			if (graph_3!=NULL)
				graph_1->parameters[3].pointer=graph_3;
			if (graph_4!=NULL)
				graph_1->parameters[4].pointer=graph_4;
		}

		return graph_1;
	}
	
	/*	underlying node not available directly, build new fill node. NULL
		entries will not result in code.
	*/
	
	instruction=g_instruction5(GFILL,graph_1,wait_graph,graph_2,graph_3,graph_4);
	instruction->info=new_node_info(desc,arity,size);
	
	/*	Using this routine on a newly created node is not appreciated: in this
		case one should build the whole node.
	*/
	
	if (graph_1->code==GCREATE)
		internal_error("transparant fill: a create node must not be overwritten");
	
	return instruction;	 
}

/********************************************************************************
 *																				*
 *		duplication of create/fill nodes										*
 *																				*
 ********************************************************************************/

INSTRUCTION_GRAPH g_duplicate_closure(INSTRUCTION_GRAPH original)
{
	INSTRUCTION_GRAPH duplicate;
	INT16 i,arity;
	
	/*	the original should be a fill/create node that is not in rnf. This
		routine delivers a new create node that has the same arguments.
	*/
	
	switch (original->code)
	{
		case GFILL:
		{
			INSTRUCTION_GRAPH code;
			
			code=original->parameters[2].pointer;
			
			if (code==NULL || code->code!=GLOAD_ADDR)
			{
				error("g_duplicate_closure: original is not a closure\n");
				return original;
			}
			
			/*	Create a copy of the closure, the create node has 2 fields less
				as the fill node: no fill address and no waiting list to be saved.
				The original should have at least 3 true arguments. */
			
			arity=original->arity-2;
			
			if (original->arity<3)
				internal_error("g_duplicate_closure: original fill node too small");
			
			duplicate=g_new_node(GCREATE,arity,arity*sizeof(PARAMETER));
			duplicate->info=original->info;
			
			/*	fill the create node with the arguments of the fill node. Note that
				it will not have a 'save waitlist' field.
			*/
			
			for (i=0; i<arity; ++i)
				duplicate->parameters[i].pointer=original->parameters[i+2].pointer;
			
			return duplicate;
		}
		case GCREATE:
		{
			INSTRUCTION_GRAPH code;
			
			code=original->parameters[0].pointer;
			
			if (code==NULL || code->code!=GLOAD_ADDR)
			{
				error("g_duplicate_closure: original is not a closure\n");
				return original;
			}
			
			/*	Create a copy of the closure,
				The original should have at least 3 arguments. */
			
			arity=original->arity;
			
			if (original->arity<3)
				internal_error("g_duplicate_closure: original create node too small");
			
			duplicate=g_new_node(GCREATE,arity,arity*sizeof(PARAMETER));
			duplicate->info=original->info;
			
			for (i=0; i<arity; ++i)
				duplicate->parameters[i].pointer=original->parameters[i].pointer;

			return duplicate;
		}
		default:
		{
			error("g_duplicate_closure: no newly built closure available\n");
			return original;
		}
	}
}

INSTRUCTION_GRAPH g_reserve_closure(INSTRUCTION_GRAPH closure)
{
	/*	the closure must be transformed into a reserved node
		If the closure has been created in this block, it can be
		filled with an empty waiting list. If it was not created in
		this block, it must be an empty (reserved) node already.
		
		All this can be done by ignoring fill instructions, and
		setting the underlying create node (if any) to a reserved
		node.
	*/
	
	if (closure->code == GFILL)
	{
		INT16 i,arity;
		INSTRUCTION_GRAPH code;

		/*	Remove all argument of the fill node. The waiting list does
			not have to be saved, as it will not get overwritten */
					
		code=closure->parameters[2].pointer;
		
		if (code==NULL || code->code!=GLOAD_ADDR)
		{
			error("g_reserve_closure: not a closure\n");
			return closure;
		}
		
		arity=closure->arity;
		
		for(i=1;i<arity;++i)
			closure->parameters[i].pointer=NULL;
		
		closure=closure->parameters[0].pointer;
	}
	
	if (closure->code == GCREATE)
	{
		if (closure->arity < 3)
			internal_error("g_reserve_closure: create node too small");
		
		closure->arity=3;
		closure->parameters[0].pointer=g_load_addr(enter_reserve_label(3));
		closure->parameters[1].pointer=g_load_const((INT32)0);
		closure->parameters[2].pointer=NULL;
		
		closure->info=new_node_info(nil_label,0,3);
	}
	
	/* if closure not created here, leave it as it is */
	
	return closure;
}

/********************************************************************************
 *																				*
 *		load and store graphs													*
 *																				*
 ********************************************************************************/

INSTRUCTION_GRAPH g_load_nl(INSTRUCTION_GRAPH address, INT16 offset)
{
	/* if a node field = NULL, the field has not been filled in this basic block,
	   so get it out of the heap.
	*/

	if (address->code == GFILL &&
		address->arity>offset+2)
	{
		INSTRUCTION_GRAPH result;
		
		result=address->parameters[offset+2].pointer;
		
		if (result == NULL)
		{
			/* do not use the fill node for the address of the load_nl node:
			   if this instruction is used to get the waiting list of an existing
			   node the address of waiting list (ldnl fill...) will be put in the
			   node itself, which would introduce a cycle... Use the address of the
			   fill node instead.
			*/
		
			return g_instruction_p_int16(GLOAD_NL,address->parameters[0].pointer ,offset);
		}
		return result;
	}
	
	if (address->code == GCREATE &&
		address->arity>offset)
	{
		INSTRUCTION_GRAPH result;
		
		result=address->parameters[offset].pointer;
		
		if (result == NULL)
			internal_error("g_load_nl: NULL field of create node selected");
			
		return result;
	}
		
	return g_instruction_p_int16(GLOAD_NL,address,offset);
}

INSTRUCTION_GRAPH g_store_nl (	INSTRUCTION_GRAPH graph,
								INSTRUCTION_GRAPH address,
								INT16 offset,
								INSTRUCTION_GRAPH load_graph)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (GSTORE_NL,0,4*sizeof (PARAMETER));
	instruction->parameters[0].pointer=graph;
	instruction->parameters[1].pointer=address;
	instruction->parameters[2].int16=offset;
	
	if (load_graph && (load_graph->code==GFLOW || load_graph->code==GFHIGH))
		instruction->parameters[3].pointer=load_graph->parameters[0].pointer;
	else
		instruction->parameters[3].pointer=load_graph;
	
	return instruction;
}

INSTRUCTION_GRAPH g_fstore_nl (	INSTRUCTION_GRAPH graph,
								INSTRUCTION_GRAPH address,
								INT16 offset,
								INSTRUCTION_GRAPH load_graph_1,
								INSTRUCTION_GRAPH load_graph_2)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction=g_new_node (GFSTORE_NL,0,5*sizeof (PARAMETER));
	instruction->parameters[0].pointer=graph;
	instruction->parameters[1].pointer=address;
	instruction->parameters[2].int16=offset;
	
	if (load_graph_1 && (load_graph_1->code==GFLOW || load_graph_1->code==GFHIGH))
		instruction->parameters[3].pointer=load_graph_1->parameters[0].pointer;
	else
		instruction->parameters[3].pointer=load_graph_1;
		
	if (load_graph_2 && (load_graph_2->code==GFLOW || load_graph_2->code==GFHIGH))
		instruction->parameters[4].pointer=load_graph_2->parameters[0].pointer;
	else
		instruction->parameters[4].pointer=load_graph_2;
		
	return instruction;
}

/********************************************************************************
 *																				*
 *		local graphs															*
 *																				*
 ********************************************************************************/

INSTRUCTION_GRAPH g_local (INT16 offset, INSTRUCTION_GRAPH local_parameter_node[])
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_instruction_int16(GLOCAL, offset);
	
	if (offset < N_LOCALS)
		local_parameter_node[offset]=instruction;
	else
		internal_error ("g_local with invalid offset");

	return instruction;
}

INSTRUCTION_GRAPH g_flocal (INT16 offset,INSTRUCTION_GRAPH local_parameter_node[])
{
	INSTRUCTION_GRAPH instruction;
	
	instruction=g_instruction_int16(GFLOAT_LOCAL, offset);
	
	if (offset < N_LOCALS-1)
	{	local_parameter_node[offset]=instruction;
		local_parameter_node[offset+1]=instruction;
	} else
		internal_error ("g_flocal with invalid offset");

	return instruction;
}

/********************************************************************************
 *																				*
 *		stack graphs															*
 *																				*
 ********************************************************************************/
 
INSTRUCTION_GRAPH g_load_as(BLOCK_GRAPH *block_graph,INT16 offset)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction = g_load_nl (block_graph->a_stack_ptr, offset);
	
	return instruction;
}

INSTRUCTION_GRAPH g_load_bs(BLOCK_GRAPH *block_graph,INT16 offset)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction = g_load_nl (block_graph->b_stack_ptr, offset);
	
	return instruction;
}

INSTRUCTION_GRAPH g_store_as (	BLOCK_GRAPH *block_graph,
								INSTRUCTION_GRAPH graph,
								INT16 offset,
								INSTRUCTION_GRAPH load_graph)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction = g_store_nl (graph,block_graph->a_stack_ptr,offset,load_graph);
	
	return instruction;
}

INSTRUCTION_GRAPH g_store_bs (	BLOCK_GRAPH *block_graph,
								INSTRUCTION_GRAPH graph,
								INT16 offset,
								INSTRUCTION_GRAPH load_graph)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction = g_store_nl (graph,block_graph->b_stack_ptr,offset,load_graph);
	
	return instruction;
}

INSTRUCTION_GRAPH g_fload_bs(BLOCK_GRAPH *block_graph,INT16 offset)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction = g_fload_nl (block_graph->b_stack_ptr,offset);
	
	return instruction;
}

INSTRUCTION_GRAPH g_fstore_bs (	BLOCK_GRAPH *block_graph,
								INSTRUCTION_GRAPH graph,
								INT16 offset,
								INSTRUCTION_GRAPH load_graph1,
								INSTRUCTION_GRAPH load_graph2)
{
	register INSTRUCTION_GRAPH instruction;
	
	instruction = g_fstore_nl (graph,block_graph->b_stack_ptr,offset,load_graph1,load_graph2);
	
	return instruction;
}

/********************************************************************************
 *																				*
 *		misc graphs																*
 *																				*
 ********************************************************************************/

INSTRUCTION_GRAPH g_fjoin (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2)
{
	if (graph_1->code==GFLOW
		&& graph_2->code==GFHIGH
		&& graph_1->parameters[0].pointer==graph_2->parameters[0].pointer)
			return graph_1->parameters[0].pointer;
	
	if (graph_1->code==GLOAD_NL
		&& graph_2->code==GLOAD_NL)
	{
		INSTRUCTION_GRAPH address_1, address_2;
				
		address_1 = graph_1->parameters[0].pointer;
		address_2 = graph_2->parameters[0].pointer;
		
		if (address_1->code==GLOAD_L
			&& address_2->code==GLOAD_L
			&& address_1->parameters[0].int16==address_2->parameters[0].int16
			&& graph_1->parameters[1].int16+1==graph_2->parameters[1].int16)
		{	INSTRUCTION_GRAPH fload_graph;
		
			fload_graph=g_fload_nl (graph_1->parameters[0].pointer,
									graph_1->parameters[1].int16);
									
			graph_1->code=GFLOW;
			graph_2->code=GFHIGH;
			graph_1->parameters[0].pointer=fload_graph;
			graph_2->parameters[0].pointer=fload_graph;
			return fload_graph;
		}
	}
	
	return g_instruction2 (GFJOIN,graph_1,graph_2);
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

VOID initialise_graphs(VOID)
{	
	wait_graph=NULL;
}