#ifndef _GRAPHS_
#define _GRAPHS_

#ifndef _TYPES_
#include "cctypes.h"
#endif

/* initialising routines */

extern VOID init_heap_pointer(BLOCK_GRAPH *block_graph);
extern VOID init_a_stack_pointer(BLOCK_GRAPH *block_graph);
extern VOID init_b_stack_pointer(BLOCK_GRAPH *block_graph);

/* waitlist routines */

INSTRUCTION_GRAPH	get_node_waitlist	(INSTRUCTION_GRAPH graph_1);
INSTRUCTION_GRAPH	get_waitlist		(VOID);
VOID				put_waitlist		(INSTRUCTION_GRAPH wait_list);

/* graph routines */

extern INSTRUCTION_GRAPH g_new_node (INT16 instruction_code,INT16 arity,INT16 arg_size);

extern INSTRUCTION_GRAPH g_instruction0 		(INT16 instruction_code);
extern INSTRUCTION_GRAPH g_instruction1			(INT16 instruction_code,INSTRUCTION_GRAPH graph);
extern INSTRUCTION_GRAPH g_instruction2			(INT16 instruction_code,INSTRUCTION_GRAPH graph1,INSTRUCTION_GRAPH graph2);
extern INSTRUCTION_GRAPH g_instruction3			(INT16 instruction_code,INSTRUCTION_GRAPH graph1,INSTRUCTION_GRAPH graph2,INSTRUCTION_GRAPH graph3);
extern INSTRUCTION_GRAPH g_instruction4			(INT16 instruction_code,INSTRUCTION_GRAPH graph1,INSTRUCTION_GRAPH graph2,INSTRUCTION_GRAPH graph3,INSTRUCTION_GRAPH graph4);
extern INSTRUCTION_GRAPH g_instruction5			(INT16 instruction_code,INSTRUCTION_GRAPH graph1,INSTRUCTION_GRAPH graph2,INSTRUCTION_GRAPH graph3,INSTRUCTION_GRAPH graph4,INSTRUCTION_GRAPH graph5);
											
extern INSTRUCTION_GRAPH g_instruction_int16			(INT16 instruction_code, INT16 n);
extern INSTRUCTION_GRAPH g_instruction_int16_int16		(INT16 instruction_code, INT16 n1,INT16 n2);
extern INSTRUCTION_GRAPH g_instruction_p_int16			(INT16 instruction_code, INSTRUCTION_GRAPH p, INT16 n);
extern INSTRUCTION_GRAPH g_instruction_label_int16		(INT16 instruction_code, LABEL *label,INT16 n);
extern INSTRUCTION_GRAPH g_instruction_p_label_int16	(INT16 instruction_code, INSTRUCTION_GRAPH graph, LABEL *label, INT16 n);

extern INSTRUCTION_GRAPH g_instruction_label	(INT16 instruction_code, LABEL *label);
extern INSTRUCTION_GRAPH g_instruction_float	(INT16 instruction_code, FLOAT *f);

extern INSTRUCTION_GRAPH g_instruction_p_int32	(INT16 instruction_code, INSTRUCTION_GRAPH p, INT32 n);
extern INSTRUCTION_GRAPH g_instruction_int32	(INT16 instruction_code, INT32 n);

#define g_keep(g1,g2)			g_instruction2(GKEEP,(g1),(g2))
#define g_save(g1,g2)			g_instruction2(GSAVE,(g1),(g2))

#define g_not(g1) 				g_instruction1(GNOT,(g1))
#define g_abs(g1)				g_instruction1(GABS,(g1))

#define g_adc(g1,const)			g_instruction_p_int32(GADC, (g1),(const))
#define g_eqc(g1,const)			g_instruction_p_int32(GEQC, (g1),(const))

#define g_bcnt(g1)				g_instruction1(G_BCOUNT, (g1))
#define g_and(g1,g2)			g_instruction2(GAND,(g1),(g2))
#define g_or(g1,g2)				g_instruction2(GOR,	(g1),(g2))
#define g_xor(g1,g2)			g_instruction2(GXOR,(g1),(g2))
#define g_lsl(g1,g2)			g_instruction2(GLSL,(g1),(g2))
#define g_lsr(g1,g2)			g_instruction2(GLSR,(g1),(g2))
#define g_asl(g1,g2)			g_instruction2(GASL,(g1),(g2))
#define g_asr(g1,g2)			g_instruction2(GASR,(g1),(g2))

#define g_eq(g1,g2) 			g_instruction2(GEQ,(g1),(g2))
#define g_gt(g1,g2) 			g_instruction2(GGT,(g1),(g2))

#define g_feq(g1,g2)			g_instruction2(GFEQ,(g1),(g2))
#define g_fgt(g1,g2)			g_instruction2(GFGT,(g1),(g2))

#define g_add(g1,g2)			g_instruction2(GADD,(g1),(g2))
#define g_sub(g1,g2)			g_instruction2(GSUB,(g1),(g2))
#define g_mul(g1,g2)			g_instruction2(GMUL,(g1),(g2))
#define g_div(g1,g2)			g_instruction2(GDIV,(g1),(g2))
#define g_mod(g1,g2)			g_instruction2(GMOD,(g1),(g2))
#define g_diff(g1,g2)			g_instruction2(GDIFF,(g1),(g2))

#define g_byte_sub(g1,g2)		g_instruction2(GBSUB,	(g1),(g2))
#define g_word_sub(g1,g2)		g_instruction2(GWSUB,	(g1),(g2))

#define g_fadd(g1,g2)			g_instruction2(GFADD,(g1),(g2))
#define g_fsub(g1,g2)			g_instruction2(GFSUB,(g1),(g2))
#define g_fmul(g1,g2)			g_instruction2(GFMUL,(g1),(g2))
#define g_fdiv(g1,g2)			g_instruction2(GFDIV,(g1),(g2))
#define g_frem(g1,g2)			g_instruction2(GFREM,(g1),(g2))

#define g_fmulby2(g1)			g_instruction1(GFPUMULBY2,(g1))
#define g_fdivby2(g1)			g_instruction1(GFPUDIVBY2,(g1))
#define g_fexpinc32(g1)			g_instruction1(GFPUEXPINC32,(g1))
#define g_fexpdec32(g1)			g_instruction1(GFPUEXPDEC32,(g1))

#define g_fitor(g1)				g_instruction1(GFITOR,	(g1))
#define g_frtoi(g1)				g_instruction1(GFRTOI,	(g1))

#define g_flow(g1)				g_instruction1(GFLOW,	(g1))
#define g_fhigh(g1)				g_instruction1(GFHIGH,	(g1))

#define g_load_const(n)			g_instruction_int32		(GLOAD_CONST,	(n))
#define g_mint()				g_instruction0			(GMINT)

#define g_load_addr(label)					g_instruction_label_int16	(GLOAD_ADDR,	(label),0)
#define g_load_addr_offset(label,offset)	g_instruction_label_int16	(GLOAD_ADDR,	(label),(offset))

extern INSTRUCTION_GRAPH g_load_desc_id		(LABEL *desc, INT16 arity);
extern INSTRUCTION_GRAPH g_equal_desc_id	(INSTRUCTION_GRAPH graph, LABEL *desc, INT16 arity);

#define g_load_b(address)				g_instruction1			(GLOAD_B,	(address))
#define g_load_nlp(address,offset)		g_instruction_p_int16	(GLOAD_NLP,	(address),(offset))

#define g_store_l(graph,offset)			g_instruction_p_int16(GSTORE_L,		(graph),(offset))
#define g_fstore_l(graph,offset)		g_instruction_p_int16(GFSTORE_L,	(graph),(offset))

#define g_fload_nl(address,offset)		g_instruction_p_int16(GFLOAD_NL,		(address),(offset))

#define g_random()						g_instruction0(GRANDOM)
#define g_processors()					g_instruction0(GPROC_INFO)
#define g_desc_table()					g_instruction0(GDESC_TABLE)
#define g_id()							g_instruction0(GID)

#define g_negate1(graph)				g_adc(g_not(graph),(INT32)1)

extern INSTRUCTION_GRAPH g_negate2			(INSTRUCTION_GRAPH graph);
extern INSTRUCTION_GRAPH g_negate_no_check	(INSTRUCTION_GRAPH graph);

extern INSTRUCTION_GRAPH g_addI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_subI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_mulI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_divI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_modI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_andI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_andB		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_orI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_orB		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_xorI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_xorB		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_notI		(INSTRUCTION_GRAPH graph_1);
extern INSTRUCTION_GRAPH g_notB		(INSTRUCTION_GRAPH graph_1);
extern INSTRUCTION_GRAPH g_L_shiftl	(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_L_shiftr	(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_A_shiftl	(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_A_shiftr	(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_eqI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_gtI		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);

extern INSTRUCTION_GRAPH g_addR		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_subR		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_mulR		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_divR		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_remR		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_eqR		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_gtR		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);

extern INSTRUCTION_GRAPH g_create_string (STRING *string);

extern INSTRUCTION_GRAPH g_create_3(INSTRUCTION_GRAPH g1,INSTRUCTION_GRAPH g2,INSTRUCTION_GRAPH g3,INT16 arity,INT16 size,LABEL *desc);
extern INSTRUCTION_GRAPH g_create_4(INSTRUCTION_GRAPH g1,INSTRUCTION_GRAPH g2,INSTRUCTION_GRAPH g3,INSTRUCTION_GRAPH g4,INT16 arity,INT16 size,LABEL *desc);
extern INSTRUCTION_GRAPH g_create_m(INT16 m,INT16 arity,INT16 size,LABEL *desc);

extern INSTRUCTION_GRAPH g_fill_3 (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph,INSTRUCTION_GRAPH graph_2,INSTRUCTION_GRAPH graph_3, INT16 arity,INT16 size,LABEL *desc);
extern INSTRUCTION_GRAPH g_fill_4 (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph,INSTRUCTION_GRAPH graph_2,INSTRUCTION_GRAPH graph_3,INSTRUCTION_GRAPH graph_4, INT16 arity,INT16 size,LABEL *desc);
extern INSTRUCTION_GRAPH g_transparant_fill_4 (INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph,INSTRUCTION_GRAPH graph_2,INSTRUCTION_GRAPH graph_3,INSTRUCTION_GRAPH graph_4, INT16 arity,INT16 size,LABEL *desc);

extern VOID optimise_fill_m				(INSTRUCTION_GRAPH instruction,INT16 n_args);
extern INSTRUCTION_GRAPH g_fill_m		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph, INT16 n_args, INT16 arity,INT16 size,LABEL *desc);
extern INSTRUCTION_GRAPH g_fill_r		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH wait_graph,INSTRUCTION_GRAPH graph_2,INSTRUCTION_GRAPH graph_3);
extern INSTRUCTION_GRAPH g_fill_copy	(INSTRUCTION_GRAPH source,INSTRUCTION_GRAPH dest,INSTRUCTION_GRAPH wait_graph,INT16 n_bytes);

extern INSTRUCTION_GRAPH g_duplicate_closure	(INSTRUCTION_GRAPH original);
extern INSTRUCTION_GRAPH g_reserve_closure		(INSTRUCTION_GRAPH closure);

extern INSTRUCTION_GRAPH g_load_nl		(INSTRUCTION_GRAPH address, INT16 offset);
extern INSTRUCTION_GRAPH g_fjoin		(INSTRUCTION_GRAPH graph_1,INSTRUCTION_GRAPH graph_2);
extern INSTRUCTION_GRAPH g_fload_const	(FLOAT *f);
extern INSTRUCTION_GRAPH g_fstore_nl	(INSTRUCTION_GRAPH graph,INSTRUCTION_GRAPH address,INT16 offset,INSTRUCTION_GRAPH load_graph_1,INSTRUCTION_GRAPH load_graph_2);
extern INSTRUCTION_GRAPH g_store_nl		(INSTRUCTION_GRAPH graph,INSTRUCTION_GRAPH address,INT16 offset,INSTRUCTION_GRAPH load_graph);
extern INSTRUCTION_GRAPH g_local		(INT16 offset, INSTRUCTION_GRAPH local_parameter_node[]);
extern INSTRUCTION_GRAPH g_flocal		(INT16 offset,INSTRUCTION_GRAPH local_parameter_node[]);

extern INSTRUCTION_GRAPH g_load_as		(BLOCK_GRAPH *block_graph,INT16 offset);
extern INSTRUCTION_GRAPH g_load_bs		(BLOCK_GRAPH *block_graph,INT16 offset);

extern INSTRUCTION_GRAPH g_store_as		(BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph,INT16 offset,INSTRUCTION_GRAPH load_graph);
extern INSTRUCTION_GRAPH g_store_bs		(BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph,INT16 offset,INSTRUCTION_GRAPH load_graph);

extern INSTRUCTION_GRAPH g_fload_bs		(BLOCK_GRAPH *block_graph,INT16 offset);
extern INSTRUCTION_GRAPH g_fstore_bs	(BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph,INT16 offset,INSTRUCTION_GRAPH load_graph1,INSTRUCTION_GRAPH load_graph2);

extern INSTRUCTION_GRAPH g_prod			(INSTRUCTION_GRAPH g1, INSTRUCTION_GRAPH g2);
						
extern VOID initialise_graphs(VOID);

#endif
