#include "ifinterface.h"

#include <string.h>

#include "ccmacros.h"

#include "cglabels.h"
#include "cgmemory.h"
#include "cgoutputabc.h"
#include "cgstacks.h"

static LABEL_NODE	*labels;
static LOC_LABEL	*local_labels;
static ENTRY_LABEL	*defered_labels;
static ENTRY_LABEL	*reserved_labels;
static LABEL_NODE	*macro_labels;

static INT16 new;					/* current label id */

/********************************************************************************
 *																				*
 *		symbols																	*
 *																				*
 ********************************************************************************/

CHAR		*init_hnf_symbol		= "_init_hnf";

CHAR		*start_symbol			= "_start";
CHAR		*nostart_symbol			= "_nostart_";
CHAR		*ea_process_symbol		= "_ea_process_code";
CHAR		*ea_reduce_symbol		= "_ea_reduce_code";

CHAR		*cycle_in_spine_symbol	= "_cycle_in_spine";
CHAR		*hnf_symbol				= "_hnf";

CHAR		*reserve_symbol			= "_reserve";
CHAR		*reserve1_symbol		= "_reserve1";
CHAR		*request_symbol			= "_request";
CHAR		*request1_symbol		= "_request1";
CHAR		*releasep_symbol		= "_release_proc";
CHAR		*releasep1_symbol		= "_release_proc1";
CHAR		*releaser_symbol		= "_release_request";
CHAR		*releaser1_symbol		= "_release_request1";

CHAR		*driver_symbol			= "_driver";
CHAR		*hnf_reducer_symbol		= "_hnf_reducer";
CHAR		*match_error_symbol		= "_match_error";
CHAR		*add_arg_symbol			= "_add_arg";
CHAR		*e_system_sAP_symbol	= "e_system_sAP";

CHAR		*Defer_symbol			= "_Defer";
CHAR		*defer_code_symbol		= "_defer_code";
CHAR		*channel_code_symbol	= "_channel_code";

CHAR		*int_symbol				= "INT";
CHAR		*char_symbol			= "CHAR";
CHAR		*bool_symbol			= "BOOL";
CHAR		*real_symbol			= "REAL";
CHAR		*file_symbol			= "FILE";
CHAR		*string1_symbol			= "_STRING_";		/* not used yet */
CHAR		*string2_symbol			= "STRING";

CHAR		*nil_symbol				= "_Nil";
CHAR		*cons_symbol			= "_Cons";
CHAR		*tuple_symbol			= "_Tuple";

/****************************************************************************************
 *                                                                                      *
 *						declaration of labels        									*
 *																						*
 ****************************************************************************************/

LABEL			*reserve_label, *reserve1_label,
				*request_label, *request1_label,
				*release_proc_label, *release_proc1_label,
				*release_request_label, *release_request1_label,
				*repl_args_label,
				*start_label, *ea_process_label, *ea_reduce_label,
				*channel_label;

LABEL			*eq_symbol_label,
				*equal_string_label, *cmp_string_label;
				
LABEL			*ln_label, *log10_label, *exp_label,
				*sin_label, *cos_label, *tan_label,
				*asin_label, *acos_label, *atan_label,
				*power_label, *random_label;
				
LABEL			*yet_args_needed_label,
				*yet_args_needed_labels[MAX_YET_ARGS_NEEDED_ARITY+1],
				*init_yet_args_needed_label,
				*init_yet_args_needed_labels[MAX_YET_ARGS_NEEDED_ARITY+1],
				*hnf_label,
				*init_hnf_label;

LABEL			*eval_fill_label,*eval_upd_labels[MAX_EVAL_UPD+1];

LABEL			*realloc_label, *extend_label, *signal_label;

LABEL			*INT_label,*BOOL_label,*CHAR_label,*REAL_label,*FILE_label,
				*nil_label, *STRING1_label, *STRING2_label,
				*tuple_label, *cons_label;

/********************************************************************************
 *																				*
 *		low level label routines												*
 *																				*
 ********************************************************************************/

INT16 new_label_id (VOID)
{
    return (new++);
}

LABEL *enter_label2 (register CHAR *name,U_INT16 flags)
{
	register LABEL_NODE **label_p,*new_label;

	label_p=&labels;
	while (*label_p!=NULL)
	{	LABEL_NODE *label=*label_p;
		INT16 r=strcmp (name,label->label.name);
		if (r==0){
			label->label.flags |= flags;
			return &label->label;
		}
		if (r<0)
			label_p=&label->left;
		else
			label_p=&label->right;
	}
	
	new_label=alloc_heap_mem(sizeof (LABEL_NODE));
		
	new_label->left		=NULL;
	new_label->right	=NULL;
	
	new_label->label.def_type	=D_NODATA;
	new_label->label.def.any	=NULL;
	
	new_label->label.number		=new_label_id();
	new_label->label.name		=new_chars_from_chars(name);
	
	new_label->label.flags		=flags;
	new_label->label.entry_type	=NORMAL_NODE;
	
	new_label->label.a.stack_size=0;
	new_label->label.b.stack_size=0;

	/* vectors set by other routines */
	
	*label_p=new_label;
	return &new_label->label;
}

LABEL *enter_label (register STRING *name,U_INT16 flags)
{
	register LABEL_NODE **label_p,*new_label;

	label_p=&labels;
	while (*label_p!=NULL)
	{	LABEL_NODE *label=*label_p;
		INT16 r=strcmp (name->string,label->label.name);
		if (r==0)
		{
			label->label.flags |= flags;
			return &label->label;
		}
		if (r<0)
			label_p=&label->left;
		else
			label_p=&label->right;
	}
	
	new_label=alloc_heap_mem(sizeof (LABEL_NODE));
		
	new_label->left		=NULL;
	new_label->right	=NULL;
	
	new_label->label.def_type	=D_NODATA;
	new_label->label.def.any	=NULL;
	
	new_label->label.number		=new_label_id();
	new_label->label.name		=new_chars_from_string(name);
	
	new_label->label.flags		=flags;
	new_label->label.entry_type	=NORMAL_NODE;
	
	new_label->label.a.stack_size=0;
	new_label->label.b.stack_size=0;

	/* vectors set by other routines */
	
	*label_p=new_label;
	return &new_label->label;
}

LABEL *enter_local_label(U_INT16 flags)
{	LOC_LABEL *label;

	/* allocates a label which is not in label tree (used for data (strings,reals) in
	   code */
	
	label=alloc_heap_mem(sizeof (LOC_LABEL));
	
	label->label.def_type	=D_NODATA;
	label->label.def.any	=NULL;
	
	label->label.number		=new_label_id();
	label->label.name		=NULL;
	
	label->label.flags		=flags;
	label->label.entry_type	=NORMAL_NODE;
	
	label->label.a.stack_size=0;
	label->label.b.stack_size=0;

	/* vectors set by other routines */
	
	label->next=local_labels;
	local_labels=label;
	
	return &label->label;
}

LABEL *enter_defer_label(NODE_INFO *node_info, LABEL *entry)
{
	/* returns a defer entry for the given entry. The arity should be
	   the correct arity of the entry. The defer entries are sorted on their
	   original entry labels, so that it's easy to prevent duplicates */
	
	register ENTRY_LABEL **label_p,*new_label;

	if (node_info == NULL)
		internal_error("enter_defer_label: node info is undefined");
	
	if (entry->entry_type & DEFERED_NODE)
	{
		/*	the entry is already defered, give a warning and return
			the defered entry itself
		*/
		
		error("enter_defer_label: entry already defered");
		return entry;
	} 
	
	label_p=&defered_labels;
	while (*label_p!=NULL)
	{	ENTRY_LABEL *label=*label_p;
	
		if (label->entry==entry)
			return &label->label;
		
		if (label->entry<entry)
			label_p=&label->left;
		else
			label_p=&label->right;
	}
	
	new_label=alloc_heap_mem(sizeof(ENTRY_LABEL));
	new_label->left=NULL;
	new_label->right=NULL;
	
	new_label->label.def_type	=D_NODATA;
	new_label->label.def.any	=NULL;
	
	new_label->label.number		=new_label_id();
	new_label->label.name		=NULL;
	
	new_label->label.flags=USE_LABEL | GCALL_LABEL | JSR_LABEL;
	new_label->label.entry_type=(entry->entry_type | DEFERED_NODE);
	
	new_label->label.a.stack_size=0;
	new_label->label.b.stack_size=0;

	/* vectors set by other routines */
	
	new_label->descriptor	=node_info->descriptor;
	new_label->arity		=node_info->node_arity;
	new_label->size			=node_info->node_size;
	new_label->entry		=entry;
	
	*label_p=new_label;
	return &new_label->label;
}

LABEL *enter_reserve_label(INT16 size)
{
	/*	returns a special reserve entry for the given size. This is used
		for 'big' empty nodes. The reserved entries are
		sorted on size to avoid duplication
	*/
	
	register ENTRY_LABEL **label_p,*new_label;

	reserve_label->flags		|=USE_LABEL;

	if (size<=3)
		return reserve_label;
	
	label_p=&reserved_labels;
	while (*label_p!=NULL)
	{	ENTRY_LABEL *label=*label_p;
	
		if (label->arity==size)
			return &label->label;
		
		if (label->arity<size)
			label_p=&label->left;
		else
			label_p=&label->right;
	}
	
	new_label=alloc_heap_mem(sizeof(ENTRY_LABEL));
	new_label->left=NULL;
	new_label->right=NULL;
	
	new_label->label.def_type	=D_NODATA;
	new_label->label.def.any	=NULL;
	
	new_label->label.number		=new_label_id();
	new_label->label.name		=NULL;
	
	new_label->label.flags=USE_LABEL | GCALL_LABEL | JSR_LABEL;
	new_label->label.entry_type=(RESERVED_NODE | DEFERED_NODE);
	
	new_label->label.a.stack_size=0;
	new_label->label.b.stack_size=0;

	/* vectors set by other routines */
	
	new_label->descriptor	=nil_label;
	new_label->arity		=0;
	new_label->size			=size;
	new_label->entry		=reserve_label;
	
	*label_p=new_label;
	return &new_label->label;
}

/********************************************************************************
 *																				*
 *		routines used during parsing											*
 *																				*
 ********************************************************************************/

VOID define_label (LABEL *label)
{
	/* marks label as defined */
	
	if (label->flags & DEF_LABEL)
		error ("Label %s defined twice\n",label->name);
	label->flags|=DEF_LABEL;
}

VOID define_inline_label (LABEL *label, U_INT16 new_flags)
{
	/* see routine above. macro labels are somewhat more complicated:
	   a label may be defined as JUMP label, but also as GCALL label...
	   double defs of these labels are not detected, so use with care */
	
	if (!(label->flags & DEF_LABEL))
	{
		label->flags |= DEF_LABEL;
		return;
	}
	
	if ((label->flags & (JUMP_LABEL | GCALL_LABEL)) &&
		(new_flags & (JUMP_LABEL | GCALL_LABEL)))
	{
		error("warning: label %s might be defined twice\n",label->name);
		return;
	}
	error ("label %s defined twice\n",label->name);
}

/********************************************************************************
 *																				*
 *		routines used for macro labels and parameters							*
 *																				*
 ********************************************************************************/

LABEL *enter_macro_label (register STRING *name, int flags, int arg_number)
{
	register LABEL_NODE **label_p,*new_label;
	
	label_p=&macro_labels;
	while (*label_p!=NULL)
	{
		LABEL_NODE *label=*label_p;
		INT16 r=strcmp (name->string,label->label.name);
		if (r==0)
		{
			label->label.flags |= flags;			
			return &label->label;
		}
		if (r<0)
			label_p=&label->left;
		else
			label_p=&label->right;
	}
	
	new_label=alloc_macro_mem (sizeof(LABEL_NODE));
	
	new_label->left		=NULL;
	new_label->right	=NULL;
	
	new_label->label.def_type	=D_NODATA;
	new_label->label.def.any	=NULL;
	
	new_label->label.number		=arg_number;
	new_label->label.name		=alloc_macro_string(name);
	new_label->label.flags		=flags;

	*label_p=new_label;
	return &new_label->label;
}

LABEL *find_macro_label (register STRING *name)
{
	register LABEL_NODE *label_p;
	
	label_p=macro_labels;
	while (label_p!=NULL)
	{	INT16 r=strcmp (name->string,label_p->label.name);
		if (r==0)
			return &label_p->label;
		if (r<0)
			label_p=label_p->left;
		else
			label_p=label_p->right;
	}
	return NULL;
}

VOID free_macro_labels (VOID)
{
	macro_labels=NULL;
	free_macro_mem();
}

/********************************************************************************
 *																				*
 *		routines used after parsing												*
 *																				*
 ********************************************************************************/

INT16 get_ctop_offset(INT16 b_stack_size,U_INT32 *vector)
{
	register INT16 number_of_b_parameters;
	register INT16 stack_displacement;
	register BOOL all_parameters_in_locals;
	register INT16 offset;
	
	/* first determine amount of parameters that are NOT in locals */

	number_of_b_parameters=0;
	stack_displacement=0;
	all_parameters_in_locals=true;
		
	for (offset=0; offset<b_stack_size;++offset)
	{	
		if (is_vector_bit_set(vector,offset))
		{
			/* a float parameter */
			
			if (number_of_b_parameters<N_B_PARAMETER_LOCALS-1)
			{	
				number_of_b_parameters+=2;
				if (all_parameters_in_locals)
					stack_displacement-=2;
			} else
				all_parameters_in_locals=false;
			
			++offset;
		} else
		{
			/* a basic value parameter */
			
			if (number_of_b_parameters<N_B_PARAMETER_LOCALS)
			{				
				++number_of_b_parameters;
				
				if (all_parameters_in_locals)
					--stack_displacement;
			} else
				all_parameters_in_locals=false;
		}
	}	
	return b_stack_size+stack_displacement;
}

static LABEL_NODE *export_inline;
static LABEL_NODE *export_data;
static LABEL_NODE *export_code;
static LABEL_NODE *import_inline;
static LABEL_NODE *import_data;
static LABEL_NODE *import_code;

static LABEL_NODE *local_inline;
static LABEL_NODE *local_data;
static LABEL_NODE *local_code;

static VOID check_label(register LABEL_NODE *label_node)
{
	register LABEL *label;
	register LABEL_NODE *left_label;
	
	/* checks all labels in reverse alphabetical order */
	
	if (label_node==NULL)
		return;
	
	label=&label_node->label;
	left_label=label_node->left;

	check_label(label_node->right);
	
	/* determine stack layout (overwriting stack size fields!) */
		
	if (label->a.stack_size > N_A_PARAMETER_LOCALS)
		label->a.n_locals = N_A_PARAMETER_LOCALS;

	/* if the label is a jsr_label, b_stack_size is used to denote the offset to
	   the ctop at that moment (= amount of parameters that are NOT in locals)
	   If stack size is 0 nothing has to be done, as offset is 0 too in this case */
	
	if (label->flags & JSR_LABEL)
	{
		label->flags |= FIXED_CTOP_OFFSET;
		if (label->b.stack_size>0)
			label->b.c_offset=get_ctop_offset(label->b.stack_size,label->vector);
	}
	
	if (label->flags & DEF_LABEL)
	{	if (label->flags & IMPORT_LABEL)
		{
			/*	label is defined in this module and imported. This results in
				a warning (except for some system routines). It is not considered
				an error as this can happen when including macros that import
				a label, while this label is defined in this module as well.
				In other words: the import statement is ignored */

			if ((strcmp (label->name,nil_symbol)==0) ||
				(strcmp (label->name,cons_symbol)==0) ||
				(strcmp (label->name,tuple_symbol)==0) ||
				
				(strcmp (label->name,int_symbol)==0) ||
				(strcmp (label->name,bool_symbol)==0) ||
				(strcmp (label->name,char_symbol)==0) ||
				(strcmp (label->name,file_symbol)==0) ||
				(strcmp (label->name,real_symbol)==0) ||
				(strcmp (label->name,string1_symbol)==0) ||
				(strcmp (label->name,string2_symbol)==0) ||
				
				(strcmp (label->name,Defer_symbol)==0))
			{
				label->flags|=(DATA_LABEL | EXPORT_LABEL);
				
			} else if ((strcmp (label->name,driver_symbol)==0) ||
						(strcmp (label->name,match_error_symbol)==0))
			{
				label->flags|=(JUMP_LABEL | EXPORT_LABEL);

				label->flags|=FIXED_CTOP;
				label->flags&=~CTOP_STACK;
			
			} else if ((strcmp (label->name,cycle_in_spine_symbol)==0) ||
						(strcmp (label->name,reserve_symbol)==0) ||
						(strcmp (label->name,reserve1_symbol)==0) ||
						(strcmp (label->name,request_symbol)==0) ||
						(strcmp (label->name,request1_symbol)==0) ||
						(strcmp (label->name,releasep_symbol)==0) ||
						(strcmp (label->name,releasep1_symbol)==0) ||
						(strcmp (label->name,releaser_symbol)==0) ||
						(strcmp (label->name,releaser1_symbol)==0) ||
						(strcmp (label->name,hnf_symbol)==0) ||
						(strcmp (label->name,hnf_reducer_symbol)==0) ||
						(strcmp (label->name,ea_process_symbol)==0) ||
						(strcmp (label->name,ea_reduce_symbol)==0) ||
						(strcmp (label->name,channel_code_symbol)==0) ||
						(strcmp (label->name,defer_code_symbol)==0))
			{
				label->flags|=(GCALL_LABEL | EXPORT_LABEL);
			
			} else if ((strcmp (label->name,"REALLOC_STACK")==0) ||
						(strcmp (label->name,"EXTEND_HEAP")==0) ||
						(strcmp (label->name,"BREAK")==0))
			{
				label->flags|=(INLINE_LABEL | EXPORT_LABEL);
			} else
				error ("warning: label %s defined and imported\n",label->name);
		}
		
		/*	label is defined. If it is exported, put it in the set of exported
			labels, else put it in the set of local labels */
		
		if (label->flags & EXPORT_LABEL)
		{	if(label->flags & DATA_LABEL)
			{	if (label->flags & (JUMP_LABEL | GCALL_LABEL | INLINE_LABEL))
					label_error(label,"exported data label also used as code/inline label");
				else
				{	label_node->left=NULL;
					label_node->right=export_data;
					export_data=label_node;
				}
			} else if (label->flags & INLINE_LABEL)
			{	if (label->flags & (JUMP_LABEL | GCALL_LABEL))
					label_error(label,"exported inline label also used as code label");
				else
				{	label_node->left=NULL;
					label_node->right=export_inline;
					export_inline=label_node;
				}
			} else
			{	
				/*	export code label
				
					if not gcall nor inline label, assume jump label (gcall labels are
				 	recognised by the descriptor definition, jump labels cannot
					be recognised as such if they aren't used within this module)
					For labels with the attribute 'JUMP' it is assumed the return
					address must be stored in a local. For other labels such a
					convention is unnecessary as the called routine will store
					the return address itself.
				*/
				
				if (!(label->flags & GCALL_LABEL))
					label->flags|=JUMP_LABEL;
				
				if (label->flags & JUMP_LABEL)
				{	label->flags|=FIXED_CTOP;
					label->flags&=~CTOP_STACK;
				}
						
				label_node->left=NULL;
				label_node->right=export_code;
				export_code=label_node;
			}
		} else
		{	if(label->flags & DATA_LABEL)
			{	if(label->flags & (JUMP_LABEL | GCALL_LABEL | INLINE_LABEL))
					label_error(label,"local data label also used as code/inline label");
				else
				{	label_node->left=NULL;
					label_node->right=local_data;
					local_data=label_node;
				}
			} else if (label->flags & INLINE_LABEL)
			{	if (label->flags & (JUMP_LABEL | GCALL_LABEL))
					label_error(label,"local inline label also used as code label");
				else
				{	label_node->left=NULL;
					label_node->right=local_inline;
					local_inline=label_node;
				}
			} else
			{	label_node->left=NULL;
				label_node->right=local_code;
				local_code=label_node;
			}
		}
	} else
	{
		/* label is not defined in this module */
		
		if (label->flags & EXPORT_LABEL)
			label_error (label,"undefined label exported");
		else
		{	if (options.implicit_imports || label->flags & IMPORT_LABEL)
			{
				/* label is (implicitly) imported, put in list of imported labels, but
				   only if it has actually been used */
			
				if (label->flags & USE_LABEL)
				{
					label->flags|=IMPORT_LABEL;
					
					if (label->flags & DATA_LABEL)
					{	if (label->flags & (JUMP_LABEL | GCALL_LABEL | INLINE_LABEL))
							label_error(label,"imported data label also used as code/inline label");
						else
						{	label_node->left=NULL;
							label_node->right=import_data;
							import_data=label_node;
						}
					} else if (label->flags &INLINE_LABEL)
					{	if (label->flags & (JUMP_LABEL | GCALL_LABEL))
							label_error(label,"imported inline label also used as code label");
						else
						{	label_node->left=NULL;
							label_node->right=import_inline;
							import_inline=label_node;
						}
					} else
					{
						/*	if label is used as JUMP label, ctop is in local by convention, note that
							if a label is used as GCALL label and this is sequential code, it should
							also be imported as JUMP label */
						
						if (!options.parallel_code && label->flags & GCALL_LABEL)
							label->flags|=JUMP_LABEL;
						
						if (label->flags & JUMP_LABEL)
						{	label->flags|=FIXED_CTOP;
							label->flags&=~CTOP_STACK;
						}
						
						label_node->left=NULL;
						label_node->right=import_code;
						import_code=label_node;
					}
				} else
				{
					/*	else: no need to import: not used
						this is not possible for implicit imports, else
						the label would have been entered without def,
						use or import flag set */
					
					if (!(label->flags & IMPORT_LABEL))
						internal_error ("check labels: label \"%s\" undefined, unused and not imported",label->name);
				}
			} else
			{
				/*	label isn't defined, nor imported
					it should be used however (something must have been done with it:
					it is in the label 'database').
					
					if the 'implicit-import' option is true, the label must be
					imported anyway: this is done above
				*/
				
				if (label->flags & USE_LABEL)
					label_error (label,"undefined label");
				else
					internal_error ("check labels: label \"%s\" undefined, unused and not imported",label->name);
			}
		}
	}

	check_label (left_label);
}

VOID port_labels(VOID)
{	
	if (import_inline!=NULL)
		import_inline_labels(import_inline);
	if (import_data!=NULL)
		import_data_labels(import_data);
	if (import_code!=NULL)
		import_code_labels(import_code);
	
	if (export_inline!=NULL)
		export_inline_labels(export_inline);
	if (export_data!=NULL)
		export_data_labels(export_data);
	if (export_code!=NULL)
		export_code_labels(export_code);
}

VOID define_entry_labels(VOID)
{
	output_alt_entries(defered_labels,"defered entries");
	output_alt_entries(reserved_labels,"reserved entries");
}

VOID check_labels (VOID)
{		
	check_label(labels);
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

VOID initialise_labels (VOID)
{
	INT16 n;
	LABEL *label;
	
	new=0;

	labels			=NULL;
	local_labels	=NULL;
	defered_labels	=NULL;
	reserved_labels	=NULL;
	macro_labels	=NULL;
	
	export_inline	=NULL;
	export_data		=NULL;
	export_code		=NULL;
	import_inline	=NULL;
	import_data		=NULL;
	import_code		=NULL;
	
	local_inline	=NULL;
	local_data		=NULL;
	local_code		=NULL;
	
	/* set labels to NULL */
	
	cmp_string_label		=NULL;
	random_label			=NULL;
	equal_string_label		=NULL;
	eq_symbol_label			=NULL;
	repl_args_label			=NULL;
	
	ln_label			=NULL;
	log10_label			=NULL;
	exp_label			=NULL;
	sin_label			=NULL;
	cos_label			=NULL;
	tan_label			=NULL;
	asin_label			=NULL;
	acos_label			=NULL;
	atan_label			=NULL;
	power_label			=NULL;

	yet_args_needed_label		=NULL;
	init_yet_args_needed_label	=NULL;
	hnf_label					=NULL;
	init_hnf_label				=NULL;
	
	for (n=0; n<=MAX_YET_ARGS_NEEDED_ARITY; ++n)
	{	yet_args_needed_labels[n]		=NULL;
		init_yet_args_needed_labels[n]	=NULL;
	}
	
	eval_fill_label=NULL;
	for (n=0; n<=MAX_EVAL_UPD; ++n)
		eval_upd_labels[n]=NULL;

	start_label			=NULL;

	/************************************************************************************
	 *																					*
	 *		channel labels																*
	 *																					*
	 ************************************************************************************/
	
	channel_label=enter_label2 (channel_code_symbol,IMPORT_LABEL | GCALL_LABEL | JSR_LABEL);

	/************************************************************************************
	 *																					*
	 *		system evaluation codes														*
	 *																					*
	 ************************************************************************************/

	label=enter_label2(defer_code_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	label=enter_label2(Defer_symbol,DATA_LABEL | IMPORT_LABEL);

	/************************************************************************************
	 *																					*
	 *		system hnf descriptors														*
	 *																					*
	 ************************************************************************************/
 	
	nil_label	=enter_label2 (nil_symbol,IMPORT_LABEL | DATA_LABEL | USE_LABEL);
	cons_label	=enter_label2 (cons_symbol,IMPORT_LABEL | DATA_LABEL);
	tuple_label	=enter_label2 (tuple_symbol,IMPORT_LABEL | DATA_LABEL);

	INT_label=enter_label2 (int_symbol,DATA_LABEL | IMPORT_LABEL);
	BOOL_label=enter_label2 (bool_symbol,DATA_LABEL | IMPORT_LABEL);
	CHAR_label=enter_label2 (char_symbol,DATA_LABEL | IMPORT_LABEL);
	REAL_label=enter_label2 (real_symbol,DATA_LABEL | IMPORT_LABEL);
	FILE_label=enter_label2 (file_symbol,DATA_LABEL | IMPORT_LABEL);
	STRING1_label=enter_label2 (string1_symbol,DATA_LABEL | IMPORT_LABEL);
	STRING2_label=enter_label2 (string2_symbol,DATA_LABEL | IMPORT_LABEL);

	/************************************************************************************
	 *																					*
	 *		reserve / cycle in spine													*
	 *																					*
	 ************************************************************************************/

	label=enter_label2(reserve_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(label,1,0,e_vector);
	label->entry_type|=RESERVED_NODE;

	if (options.parallel_code)
		reserve_label=label;

	label=enter_label2(cycle_in_spine_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(label,1,0,e_vector);

	if (!options.parallel_code)
		reserve_label=label;

	reserve1_label=enter_label2(reserve1_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(reserve1_label,1,0,e_vector);

	/************************************************************************************
	 *																					*
	 *		request nodes																*
	 *																					*
	 ************************************************************************************/
	
	request_label=enter_label2 (request_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(request_label,1,0,e_vector);
 
	request1_label=enter_label2(request1_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(request1_label,1,0,e_vector);

	/************************************************************************************
	 *																					*
	 *		request nodes																*
	 *																					*
	 ************************************************************************************/

	release_proc_label=enter_label2 (releasep_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(release_proc_label,1,0,e_vector);
 
	release_proc1_label=enter_label2(releasep1_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(release_proc1_label,1,0,e_vector);

	release_request_label=enter_label2 (releaser_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(release_request_label,1,0,e_vector);
 
	release_request1_label=enter_label2(releaser1_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(release_request1_label,1,0,e_vector);

	/************************************************************************************
	 *																					*
	 *		system code labels															*
	 *																					*
	 ************************************************************************************/
 
	label=enter_label2 (driver_symbol,IMPORT_LABEL | JSR_LABEL);
	set_label_layout(label,1,0,e_vector);
	
	label=enter_label2 (hnf_reducer_symbol,GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
	set_label_layout(label,1,0,e_vector);
	
	label=enter_label2 (match_error_symbol,IMPORT_LABEL | JSR_LABEL);
	set_label_layout(label,0,2,i_i_vector);
	
	ea_process_label=enter_label2(ea_process_symbol,IMPORT_LABEL | JSR_LABEL | GCALL_LABEL | JUMP_LABEL);
	set_label_layout(ea_process_label,2,0,e_vector);
	
	ea_reduce_label=enter_label2(ea_reduce_symbol,IMPORT_LABEL | JSR_LABEL | GCALL_LABEL | JUMP_LABEL);
	set_label_layout(ea_reduce_label,2,0,e_vector);

	/************************************************************************************
	 *																					*
	 *		low level labels															*
	 *																					*
	 ************************************************************************************/
 
	extend_label	=enter_label2("EXTEND_HEAP", INLINE_LABEL | IMPORT_LABEL | USE_LABEL);
	realloc_label	=enter_label2("REALLOC_STACK", INLINE_LABEL | IMPORT_LABEL | USE_LABEL);
	signal_label	=enter_label2("SIGNAL", INLINE_LABEL | IMPORT_LABEL | USE_LABEL);
}
