#ifndef _LABELS_
#define _LABELS_

#ifndef _TYPES_
#include "cctypes.h"
#endif

/********************************************************************************
 *																				*
 *		exported symbols														*
 *																				*
 ********************************************************************************/

extern CHAR		*init_hnf_symbol;

extern CHAR		*start_symbol;
extern CHAR		*nostart_symbol;
extern CHAR		*ea_process_symbol;
extern CHAR		*ea_reduce_symbol;

extern CHAR		*cycle_in_spine_symbol;
extern CHAR		*hnf_symbol;

extern CHAR		*reserve_symbol;
extern CHAR		*reserve1_symbol;
extern CHAR		*request_symbol;
extern CHAR		*request1_symbol;
extern CHAR		*releasep_symbol;
extern CHAR		*releasep1_symbol;
extern CHAR		*releaser_symbol;
extern CHAR		*releaser1_symbol;

extern CHAR		*reserve_symbol;
extern CHAR		*wait_symbol;
extern CHAR		*reserve5_symbol;
extern CHAR		*wait5_symbol;

extern CHAR		*driver_symbol;
extern CHAR		*hnf_reducer_symbol;
extern CHAR		*match_error_symbol;
extern CHAR		*add_arg_symbol;
extern CHAR		*e_system_sAP_symbol;

extern CHAR		*Defer_symbol;
extern CHAR		*defer_code_symbol;
extern CHAR		*channel_code_symbol;

extern CHAR		*int_symbol;
extern CHAR		*char_symbol;
extern CHAR		*bool_symbol;
extern CHAR		*real_symbol;
extern CHAR		*file_symbol;
extern CHAR		*string1_symbol;
extern CHAR		*string2_symbol;

extern CHAR		*nil_symbol;
extern CHAR		*cons_symbol;
extern CHAR		*tuple_symbol;

/********************************************************************************
 *																				*
 *		exported labels															*
 *																				*
 ********************************************************************************/

extern LABEL	*ln_label;
extern LABEL	*log10_label;
extern LABEL	*exp_label;
extern LABEL	*sin_label;
extern LABEL	*cos_label;
extern LABEL	*tan_label;
extern LABEL	*asin_label;
extern LABEL	*acos_label;
extern LABEL	*atan_label;
extern LABEL	*power_label;
extern LABEL	*random_label;

extern LABEL	*INT_label;
extern LABEL	*BOOL_label;
extern LABEL	*CHAR_label;
extern LABEL	*REAL_label;
extern LABEL	*FILE_label;
extern LABEL	*nil_label;
extern LABEL	*STRING1_label;
extern LABEL	*STRING2_label;
extern LABEL	*tuple_label;
extern LABEL	*cons_label;

extern LABEL	*reserve_label;
extern LABEL	*reserve1_label;
extern LABEL	*request_label;
extern LABEL	*request1_label;

extern LABEL	*release_proc_label;
extern LABEL	*release_proc1_label;
extern LABEL	*release_request_label;
extern LABEL	*release_request1_label;

extern LABEL	*repl_args_label;
extern LABEL	*start_label;
extern LABEL	*ea_process_label;
extern LABEL	*ea_reduce_label;
extern LABEL	*channel_label;

extern LABEL	*eq_symbol_label;
extern LABEL	*equal_string_label;
extern LABEL	*cmp_string_label;

extern LABEL	*yet_args_needed_label;
extern LABEL	*yet_args_needed_labels[MAX_YET_ARGS_NEEDED_ARITY+1];
extern LABEL	*init_yet_args_needed_label;
extern LABEL	*init_yet_args_needed_labels[MAX_YET_ARGS_NEEDED_ARITY+1];
extern LABEL	*eval_fill_label;
extern LABEL	*eval_upd_labels[MAX_EVAL_UPD+1];
extern LABEL	*hnf_label;
extern LABEL	*init_hnf_label;

extern LABEL	*realloc_label;
extern LABEL	*extend_label;
extern LABEL	*signal_label;

typedef struct LABEL_NODE
{
	struct LABEL_NODE *	left;
	struct LABEL_NODE *	right;
	LABEL				label;

} LABEL_NODE;


typedef struct LOC_LABEL
{
	struct LOC_LABEL 	*next;
	LABEL				label;

} LOC_LABEL;

typedef struct ENTRY_LABEL
{
	/* sorted on entry pointers */
	
	struct ENTRY_LABEL	*left;				/* left branch of sorted tree */
	struct ENTRY_LABEL	*right;				/* right branch of sorted tree */
	INT16				arity;				/* arity of node */
	INT16				size;				/* size of node */
	LABEL				*descriptor;		/* descriptor of entry */
	LABEL				*entry;				/* label that entry should jump to */
	LABEL				label;				/* label of entry itself */

} ENTRY_LABEL;

typedef enum LABEL_FLAGS{

	/* flags for imported/exported labels */

	IMPORT_LABEL		=1,			/* true if label imported */
	EXPORT_LABEL		=2,			/* true if label exported */

	/* flags for defined/ used labels */

	DEF_LABEL			=4,			/* true if label defined in this module */
	USE_LABEL			=8,			/* true if label used */
	
	DATA_LABEL			=16,		/* true if used as data */
	GCALL_LABEL			=32,		/* true if used for gcall instruction */
	JUMP_LABEL			=64,		/* true if used for jump instruction */
	INLINE_LABEL		=128,		/* true if used as inline label */
	
	JSR_LABEL			=512,		/* true if used for some sort of jsr (c top can be determined at this label) */

	/* flags for local  allocation */

	FIXED_LOCALS		=1024,		/* true if stack layout is fixed */
	FIXED_CTOP			=2048,		/* true if location of ctop is fixed */
	FIXED_CTOP_OFFSET	=4096,		/* true if offset of ctop is fixed */
	
	CTOP_STACK			=8192,		/* true if c top is on stack (versus local) */

/* flags for reachability of label */

	REACHABLE			=16384		/* true if label is reachable from previous block */

} LABEL_FLAGS;

typedef enum MACRO_LABEL_FLAGS {

	/* for locally defined macro labels only a few separare flags are needed:
	   they only have to indicate the type of the label used:
	   -	a locally defined labels
	   -	an argument
	*/
	
	MACRO_ARGUMENT		=1,
	MACRO_LOCAL			=2
	
} MACRO_LABEL_FLAGS;

/* label routines */

extern INT16	new_label_id		(VOID);
extern LABEL	*enter_label		(register STRING *label_name,U_INT16 label_flags);
extern LABEL	*enter_label2		(register CHAR *name,U_INT16 flags);

extern VOID		define_label		(LABEL *label);
extern VOID		define_inline_label (LABEL *label, U_INT16 new_flags);

extern LABEL	*enter_local_label	(U_INT16 label_flags);
extern LABEL	*enter_defer_label	(NODE_INFO *node_info, LABEL *entry);
extern LABEL	*enter_reserve_label(INT16 arity);

extern LABEL	*enter_macro_label	(register STRING *name, int flags, int arg_number);
extern LABEL	*find_macro_label	(register STRING *name);
extern VOID		free_macro_labels	(VOID);

extern INT16 get_ctop_offset(INT16 b_stack_size,U_INT32 *vector);

extern VOID initialise_labels(VOID);
extern VOID check_labels(VOID);

/* routines to output labels */

extern VOID port_labels(VOID);
extern VOID define_entry_labels(VOID);

#endif