#include <string.h>

#include "ifinterface.h"

#include "cglin.h"

#include "cgmemory.h"
#include "cglabels.h"
#include "cgcalc.h"
#include "cggraphs.h"
#include "cglocal.h"
#include "cgoutputinter.h"
#include "cgstacks.h"

#include "ccmacros.h"
#include "ccoffsets.h"
#include "ccintermediate.h"
#include "ccnodes.h"

extern LABEL	*ln_label, *log10_label, *exp_label,
				*sin_label, *cos_label, *tan_label,
				*asin_label, *acos_label, *atan_label,
				*power_label, *random_label, *newp_label, *eq_symbol_label,
				*cat_string_label, *equal_string_label, *cmp_string_label,
				*push_arg_b_label;

/* global variables to keep track of blocks */

static INSTRUCTION_BLOCK *first_block;
static INSTRUCTION_BLOCK *current_block;

static BLOCK_GRAPH *curr_block_graph;

/* global variables to keep track of amount of virtual locals used */

static INT16 n_virtual_locals;

/* global variables to keep track of group instructions */

static INSTRUCTION_GROUP *first_group;
static INSTRUCTION_GROUP *last_group;

/* global variables to keep track of locals, registers and data */

static INT16 n_free_registers;
static INT16 n_free_float_registers;

/* data placed in code segment */

static DSTRING	*uncoded_strings;
static DFLOAT	*uncoded_floats;

static DSYMBOL	*first_uncoded_symbol;
static DSYMBOL	*last_uncoded_symbol;

/* data placed in data segment */

DDESCRIPTOR	*first_uncoded_descriptor;
static DDESCRIPTOR	*last_uncoded_descriptor;

INT32	number_of_blocks;

/* virtual local administration */

typedef struct LOCAL_LIST {
	U_INT32			vector;
	struct LOCAL_LIST	*next;
} LOCAL_LIST;

LOCAL_LIST free_locals_list;

INT32 first_free_local;
INT32 last_free_local;

/* instruction grouping */

#define pop_register()		{++n_free_registers;if (n_free_registers==N_REGS) group(N_REGS);}
#define pop_registers(n)	{n_free_registers+=n;if (n_free_registers==N_REGS) group(N_REGS);}
#define empty_registers()	{n_free_registers=N_REGS; group(N_REGS);}

#define group_if_free(n)	{if (n_free_registers==n) group(n);}

/****************************************************************************
 *																			*
 *		block administration												*
 *																			*
 ****************************************************************************/

static VOID new_instruction_block (VOID)
{
	register INSTRUCTION_BLOCK *block;

	n_virtual_locals=0;
	reset_end_locals();
	
	block=alloc_heap_mem(sizeof(INSTRUCTION_BLOCK));

	block->next					=NULL;
	block->instructions			=NULL;
	block->last_instruction		=NULL;
	block->data					=NULL;
	block->start_label			=NULL;
	block->node_descriptor		=NULL;
	block->node_arity			=-1;
	block->heap_use				=0;
	block->a_increase			=0;
	block->b_increase			=0;
	block->begin_n_a_locals		=0;
	block->begin_n_b_locals		=0;
	block->check_size			=0;
	block->ready				=false;

	if(current_block!=NULL)
		current_block->next=block;
	else
		first_block=block;
	
	current_block=block;
	
	++number_of_blocks;	
}

/********************************************************************
 *																	*
 *		administration of free locals								*
 *																	*
 *		1: local is free											*
 *		0: local is allocated										*
 *																	*
 ********************************************************************/

/********************************************************************
 *																	*
 *		low level routines											*
 *																	*
 ********************************************************************/
 
static INT16 alloc1_free_local (VOID)
{
	LOCAL_LIST *list;
	register U_INT16 bit_n,reg_n;
	
	reg_n=first_free_local;
	list=&free_locals_list;
	
	bit_n=reg_n;

	while (bit_n > VECTOR_ELEMENT_SIZE)
	{
		bit_n-=VECTOR_ELEMENT_SIZE;
		list = list->next;
		
		if (list == NULL)
			internal_error("alloc1_free_local: first free local (%ld) has no vector",first_free_local);
	}
	if (bit_n == VECTOR_ELEMENT_SIZE)
	{
		/* bit_n point to first elem of next vector, allocate if none */
		
		if (list->next == NULL)
		{	list=list->next=alloc_block_mem(sizeof(LOCAL_LIST));
			list->next=NULL;
		} else
			list=list->next;
		bit_n = 0;
	}
	
	/* first free local is always lower or equal to real first free local */
	
	while (reg_n<last_free_local)
	{
		if (list->vector is_bit_set bit(bit_n))
		{	list->vector clear_bit (bit(bit_n));
			first_free_local=reg_n+1;
			return reg_n;
		}
		++reg_n;
		++bit_n;
		
		if (bit_n==VECTOR_ELEMENT_SIZE)
		{
			/* next bit to inspect is the first one of the next vector */
				
			if (list->next == NULL)
			{
				/*	no vector exists yet, allocate one. Note that reg_n should
					be equal to last_free_local: else this is impossible.
				*/
				
				if (reg_n != last_free_local)
					internal_error("alloc1_free_local: local (%d) < last (%ld) has no vector",reg_n,last_free_local);
				
				list=list->next=alloc_block_mem(sizeof(LOCAL_LIST));
				list->next=NULL;

			} else
				list=list->next;
			bit_n=0;			
		}	
	}
	
	/* last free local hit, vector bit has not yet been initialised */
	
	list->vector clear_bit (bit(bit_n));
	
	last_free_local++;
	first_free_local=last_free_local;
	return reg_n;
}

static INT16 alloc1_free_local_from (INT16 search_from)
{
	LOCAL_LIST *list;
	register U_INT16 bit_n,reg_n;
	
	if (search_from>last_free_local)
		internal_error("alloc1_free_local_from: search (%d) starts above highest free local (%ld)",search_from,last_free_local);
	
	reg_n=search_from;
	list=&free_locals_list;
	
	bit_n=reg_n;

	while (bit_n > VECTOR_ELEMENT_SIZE)
	{
		bit_n-=VECTOR_ELEMENT_SIZE;
		list = list->next;
		
		if (list == NULL)
			internal_error("alloc1_free_local_from: start value (%d) has no vector",search_from);
	}
	if (bit_n == VECTOR_ELEMENT_SIZE)
	{
		if (list->next == NULL)
		{	list=list->next=alloc_block_mem(sizeof(LOCAL_LIST));
			list->next=NULL;
		} else
			list=list->next;
		bit_n = 0;
	}
	
	/*	first vector to inspect has been found, take care that the bit and the
		vector of the last free local might be uninitialised.
	*/
	
	while (reg_n<last_free_local)
	{		
		if (list->vector is_bit_set bit(bit_n))
		{
			list->vector clear_bit (bit(bit_n));
			return reg_n;
		}
		++reg_n;
		++bit_n;
		
		if (bit_n==VECTOR_ELEMENT_SIZE)
		{
			/* next bit to inspect is the first one of the next vector */
						
			if (list->next == NULL)
			{
				/*	no vector exists yet, allocate one. Note that reg_n should
					be equal to last_free_local: else this is impossible.
				*/
				
				if (reg_n != last_free_local)
					internal_error("alloc1_free_local_from: local (%d) < last (%ld) has no vector",reg_n,last_free_local);
				
				list=list->next=alloc_block_mem(sizeof(LOCAL_LIST));
				list->next=NULL;

			} else
				list=list->next;
			bit_n=0;
		}	
	}
	
	list->vector clear_bit (bit(bit_n));
	
	last_free_local++;
	return reg_n;
}

/********************************************************************************
 *																				*
 *		main local allocation routines											*
 *																				*
 ********************************************************************************/

static VOID free_local (INT16 n)
{
	LOCAL_LIST *list;
	register U_INT16 bit_n;
	
	/* assumes initialisation of bit at position 'n' */
	
	if (n < 0)
		internal_error ("free_local: free local at negative offset (%d)",n);

	if (n < N_LOCALS)
		curr_block_graph->local_parameter_node[n]=NULL;
	
	if (n < first_free_local)
		first_free_local=n;
	
	bit_n=n;
	list=&free_locals_list;
	
	while (bit_n>=VECTOR_ELEMENT_SIZE)
	{	
		bit_n-=VECTOR_ELEMENT_SIZE;
		list=list->next;
	}
	
	if (list->vector is_bit_set (bit(bit_n)))
		internal_error ("free_local: local %d is already free",n);
	else
		list->vector set_bit (bit(bit_n));
}

static VOID alloc_parameter_local (U_INT16 n)
{
	LOCAL_LIST *list;
	register U_INT16 bit_n;
	
	/* allocates a certain parameter local (used at start and end of block) */
	
	if (n>=N_LOCALS)
		internal_error("alloc_parameter_locals: local %d cannot be used for parameters",n);
	
	bit_n=n;
	list=&free_locals_list;
	
	/* this loop is actually not needed as n should be small! */
	
	while (bit_n>=VECTOR_ELEMENT_SIZE)
	{	bit_n-=VECTOR_ELEMENT_SIZE;
		list=list->next;
	}
	
	if (list->vector is_bit_set bit(bit_n))
		list->vector clear_bit (bit(bit_n));
	else
		internal_error("allocate_parameter_local: parameter local (%d) is already allocated",n);
}

static INT16 alloc_free_local (VOID)
{
	INT16 local;
	
	/* allocates the first free local */
	
	local=alloc1_free_local();
	
	/* adjust amount of virtual locals needed */
	
	if (local >= n_virtual_locals)
		n_virtual_locals=local+1;	
	
	return local;
}

static INT16 alloc_pref_free_local (INT16 n)
{
	register LOCAL_LIST *list;
	register U_INT16 bit_n;
	
	/* allocates free local with preference to a certain parameter local (extra local
	   may also be indicated */
	
	if (n>N_PARAMETER_LOCALS)
		internal_error("preferred local %d is not a parameter local",n);
	
	if (n==NO_PREF)
		return alloc_free_local();
	
	list			=&free_locals_list;
	bit_n			=n;
	
	while (bit_n>=VECTOR_ELEMENT_SIZE)
	{
		bit_n-=VECTOR_ELEMENT_SIZE;
		list=list->next;
		
		if (list == NULL)
			internal_error("alloc_pref_free_local: prefered local (%d) has no vector",n);
	}
	
	if (list->vector is_bit_set (bit(bit_n)))
	{	list->vector clear_bit (bit(bit_n));
		return n;
	}
		
	/* failed to get preferred local */
	
	return alloc_free_local();
}

static INT16 alloc_free_float_local(VOID)
{
	register INT16 low_local, high_local;
	
	/* allocates the first free float local */
	
	for(low_local=alloc_free_local(); ;free_local(low_local),low_local=high_local)
	{
		high_local=alloc1_free_local_from(low_local);
		
		if (low_local+1==high_local)
			break;
	}
	
	/* adjust amount of virtual locals needed */
	
	if (low_local >= n_virtual_locals-1)
		n_virtual_locals=low_local+2;
	
	return low_local;
}

static INT16 alloc_pref_free_float_local(INT16 n)
{
	register LOCAL_LIST *list;
	register U_INT16 bit_n;
	
	/* allocates free float local with preference to certain parameter local */

	if (n>=N_PARAMETER_LOCALS-1)
		internal_error("preferred float local (%d) is not a parameter local",n);
		
	if (n==NO_PREF)
		return alloc_free_float_local();
	
	list			=&free_locals_list;
	bit_n			=n;
	
	while (bit_n>=VECTOR_ELEMENT_SIZE)
	{	bit_n-=VECTOR_ELEMENT_SIZE;
		list=list->next;
		
		if (list == NULL)
			internal_error("alloc_pref_free_float_local: prefered low local (%d) has no vector",n);
	}

	if (list->vector is_bit_set (bit(bit_n)))
	{
		/* low word available, try allocate high word */
		
		if(bit_n==VECTOR_ELEMENT_SIZE-1)
		{
			if (list->next == NULL)
				internal_error("alloc_pref_free_float_local: prefered high local (%d) has no vector",n+1);

			if(list->next->vector is_bit_set (bit(0)))
			{	list->vector		clear_bit (bit(bit_n));
				list->next->vector	clear_bit (bit(0));
				return n;
			}
		} else
		{
			if(list->vector is_bit_set (bit(bit_n+1)))
			{	list->vector clear_bit (bit(bit_n));
				list->vector clear_bit (bit(bit_n+1));
				return n;
			}
		}
	}
	
	/* failed to get preferred float local */
	
	return alloc_free_float_local();
}

static VOID free_all_locals (VOID)
{		
	/* free all extra allocated vectors */
		
	first_free_local=B1;
	last_free_local=N_LOCALS;

	free_locals_list.vector=
						  (bit (B1))
						| (bit (B2))
						| (bit (B3))
						| (bit (B4))
						| (bit (B5))
						| (bit (B6))
						| (bit (A1))
						| (bit (A2))
						| (bit (A3))
						| (bit (A4))
						| (bit (EXTRA_LOCAL))
						| (bit (N_LOCALS));
	free_locals_list.next=NULL;
}

/********************************************************************
 *																	*
 *		single instruction construction								*
 *																	*
 ********************************************************************/

static INSTRUCTION *new_instruction (register INSTRUCTION_BLOCK *block, INT16 instruction_code, INT16 arg_size)
{
	register INSTRUCTION *instruction;
	
	instruction=alloc_heap_mem(sizeof (INSTRUCTION) + arg_size);
	
	instruction->next=NULL;
	instruction->prev=block->last_instruction;
	
	instruction->code=instruction_code;
	
	if (block->instructions==NULL)
		block->instructions=instruction;
	else 
		block->last_instruction->next=instruction;
	block->last_instruction=instruction;
	
	return instruction;
}

static VOID instruction_label_short (INT16 instruction_code,LABEL *label, INT16 i)
{
	register INSTRUCTION *instruction;
	
	instruction=new_instruction (current_block,instruction_code,2*sizeof(PARAMETER));
	
	instruction->parameters[0].label=label;
	instruction->parameters[1].int16=i;
}

static VOID instruction_void (INT16 instruction_code)
{	
	new_instruction (current_block,instruction_code, 0);
}

static VOID instruction_integer (INT16 instruction_code, INT32 constant)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (current_block,instruction_code, sizeof(PARAMETER));
	
	instruction->parameters[0].int32 = constant;
}

static VOID instruction_short (INT16 instruction_code, INT16 constant)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (current_block,instruction_code, sizeof(PARAMETER));
	
	instruction->parameters[0].int16 = constant;
}

static VOID instruction_label (INT16 instruction_code, LABEL *label)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (current_block,instruction_code, sizeof(PARAMETER));
	
	instruction->parameters[0].label = label;
}

INSTRUCTION *append_instr_void (INSTRUCTION_BLOCK *block,INT16 instruction_code)
{	
	return new_instruction (block,instruction_code, 0);
}

INSTRUCTION *append_instr_integer (INSTRUCTION_BLOCK *block,INT16 instruction_code, INT32 constant)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (block,instruction_code, sizeof(PARAMETER));
	
	instruction->parameters[0].int32 = constant;
	
	return instruction;
}

INSTRUCTION *append_instr_short (INSTRUCTION_BLOCK *block,INT16 instruction_code, INT16 constant)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (block,instruction_code, sizeof(PARAMETER));
	
	instruction->parameters[0].int16 = constant;
	
	return instruction;
}

INSTRUCTION *append_instr_short2 (INSTRUCTION_BLOCK *block,INT16 instruction_code, INT16 n1,INT16 n2)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (block,instruction_code, 2*sizeof(PARAMETER));
	
	instruction->parameters[0].int16 = n1;
	instruction->parameters[1].int16 = n2;
	
	return instruction;
}

INSTRUCTION *append_instr_short4 (INSTRUCTION_BLOCK *block,INT16 instruction_code, INT16 c1,INT16 c2,INT16 c3,INT16 c4)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (block,instruction_code, 4*sizeof(PARAMETER));
	
	instruction->parameters[0].int16 = c1;
	instruction->parameters[1].int16 = c2;
	instruction->parameters[2].int16 = c3;
	instruction->parameters[3].int16 = c4;
	
	return instruction;
}

INSTRUCTION *append_instr_label (INSTRUCTION_BLOCK *block,INT16 instruction_code, LABEL *label)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (block,instruction_code, sizeof(PARAMETER));
	
	instruction->parameters[0].label = label;
	
	return instruction;
}

INSTRUCTION *append_instr_label_int16 (INSTRUCTION_BLOCK *block,INT16 instruction_code, LABEL *label,INT16 n)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (block,instruction_code, 2*sizeof(PARAMETER));
	
	instruction->parameters[0].label = label;
	instruction->parameters[1].int16 = n;
	
	return instruction;
}

TINSTRUCTION **append_inline_block (INSTRUCTION_BLOCK *block)
{
	register INSTRUCTION *instruction;
	
	instruction = new_instruction (block,INLINE, sizeof(PARAMETER));
	
	instruction->parameters[0].inline_block=NULL;
	
	return &(instruction->parameters[0].inline_block);
}

/************************************************************************************
 *																					*
 *		instruction grouping														*
 *																					*
 ************************************************************************************/

static VOID group (INT16 free_regs)
{
	/*	only generate group instruction if last one isn't one already. If old
		group instruction has a lower number of regs free, use new number of
		regs. group instructions are only allocated for the duration of the
		block and are linked to each other, so that they form some sort
		of super-instructions */
	
	register INSTRUCTION_GROUP *group;
	
	if (current_block->instructions==NULL)
	{
		/* this is the first instruction of the block */
		
		group=alloc_heap_mem(sizeof (INSTRUCTION_GROUP));
	
		group->next=NULL;
		group->prev=NULL;;
	
		group->code=GROUP;
		group->free=free_regs;
	
		current_block->instructions		=(INSTRUCTION *)group;
		current_block->last_instruction	=(INSTRUCTION *)group;
		
		/* this is also the first group of this block */
		
		group->prev_group=NULL;
		group->next_group=NULL;
		
		first_group	=group;
		last_group	=group;
		
	} else
	{
		if (current_block->last_instruction->code == GROUP)
		{
			INSTRUCTION_GROUP *prev_group;
			
			/* only adjust free number of regs if new one has more */
		
			prev_group=(INSTRUCTION_GROUP *)current_block->last_instruction;
			if (free_regs > prev_group->free)
				prev_group->free=free_regs;
			
			/* linking of groups does not alter */
			
			return;
		}
		
		group=alloc_heap_mem(sizeof (INSTRUCTION_GROUP));
	
		group->next=NULL;
		group->prev=current_block->last_instruction;
	
		group->code=GROUP;
		group->free=free_regs;
	
		current_block->last_instruction->next	=(INSTRUCTION *)group;
		current_block->last_instruction			=(INSTRUCTION *)group;
		
		/* link this group in the list of groups */

		group->next_group=NULL;
		group->prev_group=last_group;
		
		if (first_group == NULL)
			first_group=group;
		else
			last_group->next_group=group;
		last_group=group;	
	}
}

static VOID ungroup_block (INSTRUCTION_BLOCK *block)
{
	register INSTRUCTION_GROUP *group;
	
	/* remove all group instructions */
	
	for (group=first_group; group != NULL; group=group->next_group)
	{
		register INSTRUCTION *next, *prev;
		
		next=group->next;
		prev=group->prev;

		if (prev != NULL)
			prev->next=next;
		else
			block->instructions=next;
		
		if (next != NULL)
			next->prev=prev;
	}		
		
	first_group	=NULL;
	last_group	=NULL;
}

/************************************************************************************
 *																					*
 *		data declaration															*
 *																					*
 ************************************************************************************/

static LABEL *define_float (FLOAT *number)
{
	LABEL *label;
	register DFLOAT **float_p, *new_float;
	
	/* check if float has been defined for this block already, if not define new one */

	float_p=&uncoded_floats;
	while (*float_p!=NULL)
	{
		new_float=*float_p;
		
		if (number->f == new_float->f.f)
		{
			label=new_float->label;
			label->flags |= (USE_LABEL | DEF_LABEL | DATA_LABEL);
			return label;
		}
		
		if (number->f < new_float->f.f)
			float_p=&new_float->left;
		else
			float_p=&new_float->right;
	}

	/* float not yet defined, define new one */

	label=enter_local_label(USE_LABEL | DEF_LABEL | DATA_LABEL);	
	new_float=alloc_heap_mem(sizeof(DFLOAT));

	label->def_type	= D_FLOAT;
	label->def.f	= new_float;
	
	new_float->left			= NULL;
	new_float->right		= NULL;
	new_float->label		= label;
	new_float->f.f			= number->f;

	*float_p=new_float;
	return label;
}

VOID define_label_string (LABEL *label, STRING *string)
{
	register DSTRING **string_p, *new_string;
	
	/* just enter this string in the string tree. Do not care for multiple definitions */
	
	string_p=&uncoded_strings;
	while (*string_p!=NULL)
	{
		INT16 r;

		new_string=*string_p;
		
		r=strcmp (string->string,new_string->s.string);
		
		if (r < 0)		
			string_p=&new_string->left;
		else
			string_p=&new_string->right;
	}

	/* define new string */

	new_string=alloc_heap_mem(sizeof(DSTRING));

	label->def_type		= D_STRING;
	label->def.string	= new_string;
	
	new_string->left		= NULL;
	new_string->right		= NULL;
	new_string->label		= label;

	put_new_string_from_string(string,&new_string->s);
	
	*string_p=new_string;
}

LABEL *define_string (STRING *string)
{
	LABEL *label;
	register DSTRING **string_p, *new_string;
	
	/* check if the string has already been defined in this block, else define new one */
	
	string_p=&uncoded_strings;
	while (*string_p!=NULL)
	{
		INT16 r;

		new_string=*string_p;
		
		r=strcmp (string->string,new_string->s.string);
		
		if (r == 0)
		{
			label=new_string->label;
			label->flags |= (USE_LABEL | DEF_LABEL | DATA_LABEL);
			return label;
		}
		
		if (r < 0)		
			string_p=&new_string->left;
		else
			string_p=&new_string->right;
	}

	/* string not defined already, define new one */

	label=enter_local_label(USE_LABEL | DEF_LABEL | DATA_LABEL);
	new_string=alloc_heap_mem(sizeof(DSTRING));

	label->def_type		= D_STRING;
	label->def.string	= new_string;
	
	new_string->left		= NULL;
	new_string->right		= NULL;
	new_string->label		= label;

	put_new_string_from_string(string,&new_string->s);
	
	*string_p=new_string;
	return label;
}

VOID define_label_descriptor (LABEL *label, LABEL *node_entry_label, LABEL *lazy_entry_label, INT16 arity, STRING *descriptor_name)
{
	/* descriptors should be placed in data segment */
	
	register DDESCRIPTOR *descriptor;
	
	descriptor = alloc_heap_mem(sizeof (DDESCRIPTOR));

	label->def_type					=D_DESCRIPTOR;
	label->def.descriptor			=descriptor;
	
	descriptor->next				= NULL;
	
	descriptor->label				= label;
	descriptor->string_label		= enter_local_label(DEF_LABEL | DATA_LABEL);
	
	descriptor->d.node_entry_label	= node_entry_label;
	descriptor->d.lazy_entry_label	= lazy_entry_label;
	
	descriptor->d.arity				= arity;
	
	descriptor->d.string_length		= descriptor_name->length;
	descriptor->d.string			= new_chars_from_string (descriptor_name);
	
	if (first_uncoded_descriptor == NULL)
		first_uncoded_descriptor = descriptor;
	else
		last_uncoded_descriptor->next = descriptor;

	last_uncoded_descriptor = descriptor;
}

VOID define_symbol (LABEL *label)
{
	register DSYMBOL *symbol;
	
	if (!options.output_debug_symbols)
		return;
	
	symbol = alloc_heap_mem(sizeof (DSYMBOL));
	symbol->next = NULL;
	symbol->label = label;
	
	if (first_uncoded_symbol == NULL)
		first_uncoded_symbol = symbol;
	else
		last_uncoded_symbol->next = symbol;
	
	last_uncoded_symbol = symbol;
}

static VOID lin_floats (DFLOAT **first, DFLOAT **last, DFLOAT *f)
{
	DFLOAT *left, *right;
	
	if (f == NULL)
		return;
	
	left	=f->left;
	right	=f->right;
	
	if (*first == NULL)
		*first = f;
	else
		(*last)->right = f;
	*last=f;
	
	lin_floats(first,last,left);
	lin_floats(first,last,right);
}

static VOID lin_strings (DSTRING **first, DSTRING **last, DSTRING *s)
{
	DSTRING *left, *right;
	
	if (s == NULL)
		return;
	
	left	=s->left;
	right	=s->right;
	
	if (*first == NULL)
		*first = s;
	else
		(*last)->right = s;
	*last=s;
	
	lin_strings(first,last,left);
	lin_strings(first,last,right);
}

/************************************************************************************
 *																					*
 *		routine to add declared data to block										*
 *																					*
 ************************************************************************************/

VOID lin_cdata(INSTRUCTION_BLOCK *block)
{
	/* used for data that should be placed in the code area. Data that is
	   placed in the data area is output at the very start of the assembly/
	   object file.
	*/

	/*	This routine will only have effect if there is some data to be
		output. if a block does not hold a data block yet, a new data
		block must be allocated, else the data can be appended to the
		existing data block
	*/	

	BLOCK_DATA *data;

	if (uncoded_strings			== NULL &&
		uncoded_floats			== NULL &&
		first_uncoded_symbol	== NULL)
		return;
	
	/* some data must be appended to the block */
	
	if (block->data == NULL)
	{
		data=alloc_heap_mem(sizeof(BLOCK_DATA));

		data->first_string	=NULL;
		data->last_string	=NULL;
		lin_strings(&data->first_string,&data->last_string,uncoded_strings);
		
		data->first_float	=NULL;
		data->last_float	=NULL;
		lin_floats(&data->first_float,&data->last_float,uncoded_floats);
		
		data->first_symbol	=first_uncoded_symbol;
		data->last_symbol	=last_uncoded_symbol;

		block->data=data;
		
	} else
	{
		data=block->data;
		
		lin_strings(&data->first_string,&data->last_string,uncoded_strings);
		lin_floats(&data->first_float,&data->last_float,uncoded_floats);
		
		if (first_uncoded_symbol != NULL)
		{
			if (data->first_symbol == NULL)
				data->first_symbol = first_uncoded_symbol;
			else
				data->last_symbol->next = first_uncoded_symbol;
			data->last_symbol = last_uncoded_symbol;
		}
	}
	
	first_uncoded_symbol	=NULL;
	last_uncoded_symbol		=NULL;
	
	uncoded_strings			=NULL;
	uncoded_floats			=NULL;
}

/****************************************************************************
 *																			*
 *		graph manipulating functions (work on current block_graph)			*
 *																			*
 ****************************************************************************/

static INSTRUCTION_GRAPH skip_indirection_nodes (register INSTRUCTION_GRAPH graph)
{
	while(graph!=NULL)
	{	if (graph->code!=GINDIRECT)
			return graph;
		graph = graph->parameters[0].pointer;
	}
	internal_error("skip indirection nodes: graph is empty");
	
	return NULL;
}

static INSTRUCTION_GRAPH change_to_indirection_node (register INSTRUCTION_GRAPH graph, INSTRUCTION_GRAPH new_graph)
{
	if(graph->code==GINDIRECT)
		internal_error("change to indirection node: source is already an indirection node");
	if(new_graph->code==GINDIRECT)
		internal_error("change to indirection node: dest is already an indirection node");
	
	graph->code=GINDIRECT;
	graph->count=1;								/* exact count of indirection node should be of no concern */
	graph->arity=1;
	graph->parameters[0].pointer=new_graph;
	
	return new_graph;
}

static INSTRUCTION_GRAPH change_to_local_node (register INSTRUCTION_GRAPH graph, INT16 local)
{
	graph->code=GLOCAL;
	graph->parameters[0].int16=local;
	graph->uregs=1;						/* now only 1 reg needed to compute result */
	graph->ufloat_regs=0;				/* and zero float regs */
		
	if (local<N_LOCALS)
		curr_block_graph->local_parameter_node[local]=graph;
	
	/*	else no need to store graph: this is only needed to save values when
		a parameter local is written */
		
	return graph;
}

static INSTRUCTION_GRAPH change_to_float_local_node (register INSTRUCTION_GRAPH graph, INT16 local)
{
	graph->code=GFLOAT_LOCAL;
	graph->parameters[0].int16=local;
	graph->uregs=1;						/* 1 reg needed for loading address */
	graph->ufloat_regs=1;				/* 1 float reg needed to compute result */
	
	if (local<N_LOCALS-1)
	{	curr_block_graph->local_parameter_node[local]=graph;
		curr_block_graph->local_parameter_node[local+1]=graph;
	}
	/* else no need to store graph */
	
	return graph;
}

/********************************************************************************
 *																				*
 *		intermediate code generation from graphs (working on current			*
 *		block graph)															*
 *																				*
 ********************************************************************************/

/* the reference count of each graph will not be decreased as soon as linearization
   of this graph has delivered a result, but at the moment this result is actually
   used by someone i.e. if it is loaded into Areg or FAreg. In this way updating of
   the reference count is done more or less implicitly by the in_Areg routines
*/

/* The reference count of local nodes may become negative. This means this
   result is 'fixed: it must not be thrown away or moved. Furthermore: load_l
   nodes must be handled in a separate way as they may not be overwritten nor
   removed */

/*	the compiler cannot handle all possible PABC codes yet. If a group would be
	generated that would use more than the available amount of real locals, the
	local scheduler will report an error. Note that this is VERY unlikely. If
	it happens, it could be resolved by splitting this group into smaller
	pieces and allocate locals for these parts. Note that this is slightly
	more complex as then less registers will be available.
	
	The local scheduler will not do much effort to optimise local scheduling:
	e.g. ldl(virtual local) could result in ldl(virtual mem), ldnl (virtual local),
	stl (real local), ldl (real local). Better would be to generate ldl(virtual mem),
	ldnl(virtual local). Take care when the local is used more frequently!! Note also
	that optimising a local store into a non-local one requires an extra register,
	so this is not trivial.
*/

static INSTRUCTION_GRAPH lin_graph			(register INSTRUCTION_GRAPH graph);
static INSTRUCTION_GRAPH lin_float_graph	(register INSTRUCTION_GRAPH graph);

/************************************************************************************
 *																					*
 *		lin aux																		*
 *																					*
 ************************************************************************************/

static VOID decrement_ref_count(register INSTRUCTION_GRAPH graph)
{
	--graph->count;
	
	if(graph->count==0)
	{	switch(graph->code)
		{	case GLOCAL:
				free_local(graph->parameters[0].int16);
				break;
			case GFLOAT_LOCAL:
				free_local(graph->parameters[0].int16);
				free_local(graph->parameters[0].int16+1);
				break;
			case GFLOW:
			{	register INSTRUCTION_GRAPH float_graph;
			
				float_graph=graph->parameters[0].pointer;
				if(float_graph->code!=GFLOAT_LOCAL)
					internal_error("decrement ref count: a float local was expected");
				
				--float_graph->count;
				if (float_graph->count==0)
				{	free_local(float_graph->parameters[0].int16);
					free_local(float_graph->parameters[0].int16+1);
				}
				break;
			}
			case GFHIGH:
			{	register INSTRUCTION_GRAPH float_graph;
			
				float_graph=graph->parameters[0].pointer;
				if(float_graph->code!=GFLOAT_LOCAL)
					internal_error("decrement ref count: a float local was expected");

				--float_graph->count;
				if (float_graph->count==0)
				{	free_local(float_graph->parameters[0].int16);
					free_local(float_graph->parameters[0].int16+1);
				}
				break;
			}
			default:
				break;
		}
	}
}

/* be carefull NOT to declare nodes 'a small computation' if they've got references
   to others: the reference count of these others isn't increased, while multiple
   acces would become possible, resulting in access to values that have reference count
   zero, so that these values are probably invalid !! */

static BOOL is_small_comp (register INSTRUCTION_GRAPH graph)
{
	if (graph->code==GLOAD_CONST &&
		graph->parameters[0].int32<256 &&
		graph->parameters[0].int32>-16)
		return true;
	return false;
}

static VOID discard_result (register INSTRUCTION_GRAPH graph)
{
	INT16 local;
	
	/* discards result of graph (after linearisation) when it isn't needed
	   after all. This happens when a node must be built that won't be stored
	   on the a-stack for instance. At the moment this is only usefull in
	   parallel programs where other processes may access a node that won't be
	   put on the a-stack. */
	
	--graph->count;
	
	switch (graph->code)
	{
		case GLOCAL:
		{
			/* if it's in a local: free this local if possible */
			
			if (graph->count==0)
				free_local(graph->parameters[0].int16);
			
			return;
		}
		case GFLOAT_LOCAL:
			internal_error ("a float (in locals) cannot be discared by discard_result");
		
		case GFLOW:
		{	INSTRUCTION_GRAPH float_graph;
		
			/* discard low word from float local */
			
			float_graph=graph->parameters[0].pointer;
			if (float_graph->code!=GFLOAT_LOCAL)
				internal_error("discard_result: a float local node was expected");
						
			if(graph->count==0)
			{	--float_graph->count;
				if(float_graph->count==0)
				{	free_local(float_graph->parameters[0].int16);
					free_local(float_graph->parameters[0].int16+1);
				}
			}
			return;
		}
		case GFHIGH:
		{	INSTRUCTION_GRAPH float_graph;
		
			/* discard high word from float local */
			
			float_graph=graph->parameters[0].pointer;
			if (float_graph->code!=GFLOAT_LOCAL)
				internal_error("discard_result: a float local node was expected");
						
			if(graph->count==0)
			{	--float_graph->count;
				if(float_graph->count==0)
				{	free_local(float_graph->parameters[0].int16);
					free_local(float_graph->parameters[0].int16+1);
				}
			}
			return;
		}
		case GEMPTY:
			internal_error("discard_result: no result available to discard");
		
		case GINDIRECT:
			internal_error("discard_result: indirect node should not be returned for result");
		
		case GLOAD_L:
		
			/* Result is in a fixed local, do nothing */
			
			return;
			
		default:
		{
			/* default it is assumed the result of the graph is in Areg. If not it
			   will have been overwritten by one of the nodes above */
			
			if (graph->count == 0)
			{
				/* no saving necessary for later use, the code will be set to EMPTY
				   to detect invalid access to this node and the value will be removed
				   from the register */
				
				graph->code=GEMPTY;
				
				pop_register();
				
				/*	if some registers are in use, shift'm by storing Areg, if
					all regs unused, shifting is not necessary */
				
				if (n_free_registers < N_REGS)
					instruction_short (STL, SAVE1);
				
				return;
			}

			if (is_small_comp(graph))
			{
				/* no saving necessary, compute result again if needed */
				
				pop_register();

				if (n_free_registers < N_REGS)
					instruction_short (STL, SAVE1);

				return;
			}

			/* result is needed lateron: store in local without dup */
			
			local = alloc_pref_free_local(graph->pref_local);
			
			pop_register();
			instruction_short (STL, local);
			
			change_to_local_node (graph, local);
			return;
		}
	}
}

static VOID in_Areg (register INSTRUCTION_GRAPH graph)
{
	INT16 local;
	
	/* puts result of graph (after linearisation) in Areg */
	
	--graph->count;
	
	switch (graph->code)
	{
		case GLOCAL:
		{	
			instruction_short(LDL, graph->parameters[0].int16);
			--n_free_registers;
			
			if (graph->count==0)
				free_local(graph->parameters[0].int16);
			
			return;
		}
		case GFLOAT_LOCAL:
			internal_error ("a float (in locals) cannot be loaded in Areg");
		
		case GFLOW:
		{	INSTRUCTION_GRAPH float_graph;
		
			/* get low word from float local */
			
			float_graph=graph->parameters[0].pointer;
			if (float_graph->code!=GFLOAT_LOCAL)
				internal_error("in Areg: a float local node was expected");
			
			instruction_short (LDL, float_graph->parameters[0].int16);
			--n_free_registers;
			
			if(graph->count==0)
			{	--float_graph->count;
				if(float_graph->count==0)
				{	free_local(float_graph->parameters[0].int16);
					free_local(float_graph->parameters[0].int16+1);
				}
			}
			return;
		}
		case GFHIGH:
		{	INSTRUCTION_GRAPH float_graph;
		
			/* get high word from float local */
			
			float_graph=graph->parameters[0].pointer;
			if (float_graph->code!=GFLOAT_LOCAL)
				internal_error("in Areg: a float local node was expected");
			
			instruction_short (LDL, float_graph->parameters[0].int16+1);
			--n_free_registers;
			
			if(graph->count==0)
			{	--float_graph->count;
				if(float_graph->count==0)
				{	free_local(float_graph->parameters[0].int16);
					free_local(float_graph->parameters[0].int16+1);
				}
			}
			return;
		}
		case GEMPTY:
			internal_error("in Areg: no result available to store in Areg");
		
		case GINDIRECT:
			internal_error("in Areg: indirect node should not be return for result");
		
		case GLOAD_L:
		
			/* result is in a fixed local, get it out of this local */
			
			instruction_short(LDL, graph->parameters[0].int16);
			--n_free_registers;
						
			return;
			
		default:
		{
			/* default it is assumed the result of the graph is in Areg. If not it
			   will have been overwritten by one of the nodes above */
			
			if (graph->count == 0)
			{
				/* no saving necessary, the code will be set to EMPTY to detect
				   invalid access to this node */
				
				graph->code=GEMPTY;
				return;
			}

			if (is_small_comp(graph))
			{
				/* no saving necessary, compute result again if needed */
				
				return;
			}

			/* result is needed lateron: store in local */
			
			local = alloc_pref_free_local(graph->pref_local);
			if (n_free_registers > 0)
			{	instruction_void (DUP);
				instruction_short (STL, local);
			} else
			{	instruction_short (STL, local);
				instruction_short (LDL, local);
			}
			change_to_local_node (graph, local);
			return;
		}
	}
}

static INT16 in_local (register INSTRUCTION_GRAPH graph)
{
	INT16 local;
	
	/* puts result of graph (after linearisation) in a local */
	
	switch (graph->code)
	{	case GLOCAL:
		{
			/* result already in local return local offset */
			
			return graph->parameters[0].int16;
		}
		case GFLOAT_LOCAL:
			internal_error("in local: a float local cannot be loaded into a single local");
		
		case GFLOW:
		{	INSTRUCTION_GRAPH float_graph;
		
			/* result is in low word of float local, return local offset of first word */
			
			float_graph=graph->parameters[0].pointer;
			if(float_graph->code!=GFLOAT_LOCAL)
				internal_error("in local: a float local node was expected");
			
			return float_graph->parameters[0].int16;
		}
		case GFHIGH:
		{	INSTRUCTION_GRAPH float_graph;
		
			/* result is in high word of float local, return local offset of second word */
			
			float_graph=graph->parameters[0].pointer;
			if(float_graph->code!=GFLOAT_LOCAL)
				internal_error("in local: a float local node was expected");
			
			return float_graph->parameters[0].int16+1;
		}
		case GEMPTY:
			internal_error("in local: no value available to put in local");
		
		case GINDIRECT:
			internal_error("in local: indirect nodes should not be returned for result");
		
		case GLOAD_L:

			/* result already in fixed local, return local offset */
			
			return graph->parameters[0].int16;
		
		default:
		{
			/* by default the result is assumed to be in Areg */	
			
			local=alloc_pref_free_local(graph->pref_local);
			instruction_short (STL, local);
			pop_register();
			change_to_local_node (graph, local);
			return local;
		}
	}
}

static VOID in_float_Areg (register INSTRUCTION_GRAPH graph)
{
	INT16 local;
	
	/* puts result of float graph (after linearisation) in fAreg */
	
	--graph->count;
	
	switch (graph->code)
	{	case GLOCAL:
			internal_error("in fAreg: a single local value cannot be loaded in fAreg");
		
		case GFLOAT_LOCAL:
		{
			/* result must be loaded from locals */
			
			if (n_free_registers > 0)
			{
				instruction_short (LDLP, graph->parameters[0].int16);
				instruction_void (FPLDNLDB);
				group_if_free(N_REGS);
				
			} else
			{
				/* not enough regs free to form group */
				
				instruction_short (STL,SAVE1);
				instruction_short (LDLP, graph->parameters[0].int16);
				instruction_void (FPLDNLDB);
				instruction_short (LDL,SAVE1);
			}
			--n_free_float_registers;
			
			if (graph->count==0)
			{	free_local(graph->parameters[0].int16);
				free_local(graph->parameters[0].int16+1);
			}
			
			return;
		}
		case GFLOW:
			internal_error("in fAreg: the low word of a float cannot be loaded in fAreg");
			
		case GFHIGH:
			internal_error("in fAreg: the high word of a float cannot be loaded in fAreg");
			
		case GEMPTY:
			internal_error("in fAreg: no value available to store in fAreg");
		
		case GINDIRECT:
			internal_error("in fAreg: indirect nodes should not be returned for result");
		
		case GLOAD_L:
			internal_error("in fAreg: load local cannot deliver float result");
		
		default:
		{
			/* default the result is assumed to be in fAreg */
			
			if (graph->count == 0)
			{
				graph->code=GEMPTY;
				return;
			}
			
			/* result is used lateron and must be saved */
			
			local = alloc_pref_free_float_local(graph->pref_local);
			if (n_free_float_registers > 0)
			{	instruction_void (FPDUP);
				if (n_free_registers > 0)
				{
					instruction_short (LDLP, local);
					instruction_void (FPSTNLDB);
					group_if_free(N_REGS);
					
				} else
				{	instruction_short (STL,SAVE1);
					instruction_short (LDLP,local);
					instruction_void (FPSTNLDB);
					instruction_short (LDL,SAVE1);
				}
			} else
			{
				/* floating point eval stack full */
				
				if (n_free_registers > 0)
				{
					instruction_short (LDLP, local);
					instruction_void (FPSTNLDB);
					group_if_free(N_REGS);
					instruction_short (LDLP, local);
					instruction_void (FPLDNLDB);
					group_if_free(N_REGS);
					
				} else
				{
					instruction_short (STL,SAVE1);
					instruction_short (LDLP,local);
					instruction_void (FPSTNLDB);
					instruction_short (LDLP,local);
					instruction_void (FPLDNLDB);
					instruction_short (LDL,SAVE1);
				}
			}
			change_to_float_local_node (graph, local);
			return;
		}
	}
}

static INT16 in_float_local (register INSTRUCTION_GRAPH graph)
{
	INT16 local;
	
	/* puts result of float graph (after linearisation) in locals */
	
	switch (graph->code)
	{	case GLOCAL:
			internal_error("a single local cannot be put in a float local");
		
		case GFLOAT_LOCAL:
		{
			/* float already in local, return local offset */
			
			return graph->parameters[0].int16;
		}
		case GFLOW:
			internal_error("the low word of a float cannot be put in a float local");
		
		case GFHIGH:
			internal_error("the high word of a float cannot be put in a float local");
		
		case GEMPTY:
			internal_error("no value availabel to put in a float local");
		
		case GINDIRECT:
			internal_error("in float local: indirect nodes should not be returned for result");
		
		case GLOAD_L:
			internal_error("in float local: load local cannot deliver float result");
		
		default:
		{
			/* default the result is assumed to be in fAreg */
			
			local=alloc_pref_free_float_local(graph->pref_local);
			if (n_free_registers > 0)
			{
				instruction_short (LDLP, local);
				instruction_void (FPSTNLDB);
				group_if_free(N_REGS);
				
			} else
			{	instruction_short (STL,SAVE1);
				instruction_short (LDLP, local);
				instruction_void (FPSTNLDB);
				instruction_short (LDL,SAVE1);
			}
			++n_free_float_registers;
			change_to_float_local_node (graph, local);
			
			return local;
		}
	}
}

/********************************************************************************
 *																				*
 *		arithmetic operations													*
 *																				*
 ********************************************************************************/

#if _COMPILE_UNUSED_

static INSTRUCTION_GRAPH lin_monadic_operator (INT16 instruction_code,INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH graph_1;
		
	graph_1 = lin_graph (graph->parameters[0].pointer);
	in_Areg (graph_1);
	
	instruction_void (instruction_code);
	
	return graph;
}
#endif

static INSTRUCTION_GRAPH lin_monadic_float_operator (INT16 instruction_code, INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH argument;

	argument=graph->parameters[0].pointer;
	
	argument=lin_float_graph(argument);
	in_float_Areg(argument);
	instruction_void(instruction_code);
		
	return graph;
}

static VOID commutative_subgraphs (	register INSTRUCTION_GRAPH graph_1,
									register INSTRUCTION_GRAPH graph_2,
									BOOL order_left)
{
	if (order_left)
	{	graph_1 = lin_graph (graph_1);
		if (graph_2->uregs >= N_REGS)
		{	in_local(graph_1);
			graph_2 = lin_graph (graph_2);
			in_Areg(graph_2);
			in_Areg(graph_1);
		} else
		{	in_Areg(graph_1);
			graph_2 = lin_graph (graph_2);
			in_Areg(graph_2);
		}
	} else
	{	graph_2 = lin_graph (graph_2);
		if (graph_1->uregs >= N_REGS)
		{	in_local(graph_2);
			graph_1 = lin_graph (graph_1);
			in_Areg(graph_1);
			in_Areg(graph_2);
		} else
		{	in_Areg(graph_2);
			graph_1 = lin_graph (graph_1);
			in_Areg(graph_1);
		}
	}
}

static VOID commutative_float_subgraphs (	register INSTRUCTION_GRAPH graph_1,
											register INSTRUCTION_GRAPH graph_2,
											BOOL order_left)
{
	if (order_left)
	{	graph_1 = lin_float_graph (graph_1);
		if (graph_2->ufloat_regs >= N_FLOAT_REGS)
		{	in_float_local(graph_1);
			graph_2 = lin_float_graph (graph_2);
			in_float_Areg(graph_2);
			in_float_Areg(graph_1);
		} else
		{	in_float_Areg(graph_1);
			graph_2 = lin_float_graph (graph_2);
			in_float_Areg(graph_2);
		}
	} else
	{	graph_2 = lin_float_graph (graph_2);
		if (graph_1->ufloat_regs >= N_FLOAT_REGS)
		{	in_float_local(graph_2);
			graph_1 = lin_float_graph (graph_1);
			in_float_Areg(graph_1);
			in_float_Areg(graph_2);
		} else
		{	in_float_Areg(graph_2);
			graph_1 = lin_float_graph (graph_1);
			in_float_Areg(graph_1);
		}
	}
}

static VOID non_commutative_subgraphs (	register INSTRUCTION_GRAPH graph_1,
										register INSTRUCTION_GRAPH graph_2,
										BOOL order_left)
{
	if (order_left)
	{	graph_1 = lin_graph (graph_1);
		if (graph_2->uregs >= N_REGS)
		{	in_local(graph_1);
			graph_2 = lin_graph (graph_2);
			in_Areg(graph_2);
			in_Areg(graph_1);
			instruction_void (REV);
		} else
		{	in_Areg(graph_1);
			graph_2 = lin_graph (graph_2);
			in_Areg(graph_2);
		}
	} else
	{	graph_2 = lin_graph (graph_2);
		if (graph_1->uregs >= N_REGS)
		{	in_local(graph_2);
			graph_1 = lin_graph (graph_1);
			in_Areg(graph_1);
			in_Areg(graph_2);
		} else
		{	in_Areg(graph_2);
			graph_1 = lin_graph (graph_1);
			in_Areg(graph_1);
			instruction_void (REV);
		}
	}
}

static VOID non_commutative_float_subgraphs (	register INSTRUCTION_GRAPH graph_1,
												register INSTRUCTION_GRAPH graph_2,
												BOOL order_left)
{
	if (order_left)
	{	graph_1 = lin_float_graph (graph_1);
		if (graph_2->ufloat_regs >= N_FLOAT_REGS)
		{	in_float_local(graph_1);
			graph_2 = lin_float_graph (graph_2);
			in_float_Areg(graph_2);
			in_float_Areg(graph_1);
			instruction_void (FPREV);
		} else
		{	in_float_Areg(graph_1);
			graph_2 = lin_float_graph (graph_2);
			in_float_Areg(graph_2);
		}
	} else
	{	graph_2 = lin_float_graph (graph_2);
		if (graph_1->ufloat_regs >= N_FLOAT_REGS)
		{	in_float_local(graph_2);
			graph_1 = lin_float_graph (graph_1);
			in_float_Areg(graph_1);
			in_float_Areg(graph_2);
		} else
		{	in_float_Areg(graph_2);
			graph_1 = lin_float_graph (graph_1);
			in_float_Areg(graph_1);
			instruction_void (FPREV);
		}
	}
}

static INSTRUCTION_GRAPH lin_dyadic_commutative_operator (INT16 instruction_code,INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	commutative_subgraphs (graph_1, graph_2, graph->order_left);
	
	instruction_void (instruction_code);
	pop_register();
	
	return graph;
}

static INSTRUCTION_GRAPH lin_dyadic_commutative_float_operator (INT16 instruction_code, register INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	commutative_float_subgraphs (graph_1, graph_2, graph->order_left);
	
	instruction_void (instruction_code);
	++n_free_float_registers;
	
	return graph;
}

static INSTRUCTION_GRAPH lin_dyadic_non_commutative_operator (INT16 instruction_code,INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	non_commutative_subgraphs (graph_1, graph_2, graph->order_left);
	
	instruction_void (instruction_code);
	pop_register();

	return graph;
}

static INSTRUCTION_GRAPH lin_dyadic_non_commutative_float_operator (INT16 instruction_code, register INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	non_commutative_float_subgraphs (graph_1, graph_2, graph->order_left);
	
	instruction_void (instruction_code);
	++n_free_float_registers;
	
	return graph;
}

static INSTRUCTION_GRAPH lin_float_equals_operator (register INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	commutative_float_subgraphs (graph_1, graph_2, graph->order_left);

	instruction_void (FPEQ);
	n_free_float_registers += 2;
	--n_free_registers;
	
	return graph;
}

static INSTRUCTION_GRAPH lin_float_greater_operator (register INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	non_commutative_float_subgraphs (graph_1, graph_2, graph->order_left);
	
	instruction_void (FPGT);
	n_free_float_registers += 2;
	--n_free_registers;
	
	return graph;
}

static INSTRUCTION_GRAPH lin_arith_shift_operator (register INSTRUCTION_GRAPH graph,BOOL shift_left)
{
	INSTRUCTION_GRAPH graph_1, graph_2;
		
	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	if (graph->order_left)
	{	graph_1 = lin_graph (graph_1);
		if (graph_2->uregs >= N_REGS-1)
		{	in_local(graph_1);
			graph_2 = lin_graph (graph_2);
			in_local(graph_2);
			in_Areg(graph_1);
			instruction_void(XDBLE);
			in_Areg(graph_2);
		} else
		{
			/* optimised for this case: happens quite often, e.g constant shift */
			
			in_Areg(graph_1);
			instruction_void(XDBLE);
			graph_2 = lin_graph (graph_2);
			in_Areg(graph_2);
		}
	} else
	{	graph_2 = lin_graph (graph_2);
		in_local(graph_2);
		graph_1 = lin_graph (graph_1);
		in_Areg(graph_1);
		instruction_void(XDBLE);
		in_Areg(graph_2);
	}
	if (shift_left)
	{	instruction_void(LSHL);
		instruction_void(CSNGL);
	} else
		instruction_void(LSHR);
	
	return graph;
}

/********************************************************************************
 *																				*
 *		create operations														*
 *																				*
 ********************************************************************************/

static INSTRUCTION_GRAPH heap_address (	INSTRUCTION_GRAPH graph,
										INSTRUCTION_GRAPH heap_pointer,
										INT16 offset)
{
	if (offset==0)
	{
		/* return heap pointer */

		if (heap_pointer->count>=0)
			heap_pointer->count+=(graph->count - 1);
		return change_to_indirection_node (graph, heap_pointer);
	}
	
	/* heap pointer doesn't contain correct address, so return address in Areg */
	
	instruction_short (LDL, heap_pointer->parameters[0].int16);
	instruction_short (LDNLP,offset);
	
	--n_free_registers;
	--heap_pointer->count;
	
	return graph;
}

static INSTRUCTION_GRAPH lin_create_operator (INSTRUCTION_GRAPH graph)

/* creates a node in the heap */

{
	INT16 *args;
	INT16 n_arguments, old_heap_offset;
	register INT16 i;
	INSTRUCTION_GRAPH heap_ptr;
	
	n_arguments=graph->arity;

	old_heap_offset=current_block->heap_use;
	current_block->heap_use+=n_arguments;
	
	heap_ptr=skip_indirection_nodes(curr_block_graph->heap_ptr);
	args=alloc_block_mem(n_arguments*sizeof(INT16));
	n_arguments=sort_and_calc_block_graphs (curr_block_graph,(INSTRUCTION_GRAPH *)&graph->parameters[0],0, n_arguments, args);
	
	for (i=0; i<n_arguments; i++)
	{	INSTRUCTION_GRAPH argument;
		
		argument=lin_graph (graph->parameters[args[i]].pointer);
		in_Areg(argument);
		instruction_short (LDL, heap_ptr->parameters[0].int16);
		instruction_short (STNL,old_heap_offset+args[i]);
		pop_register();
	}
	
		
	if (graph->arity>3 && graph->info->node_arity>1 && graph->parameters[0].pointer!=NULL)
	{
		/* This is a large node that does not have arity 0 or 1, so intialise empty fields
		   with the address of the node itself to prevent the garbage collector
		   from crashing. First count the amount of empty fields. */
		
		INT16 empty_count;
		
		empty_count=0;
		for (i=2; i<graph->arity; ++i)
			if (graph->parameters[i].pointer==NULL)
				++empty_count;
		
		if (empty_count > 0)
		{
			/* some empty fields have to be filled, so get address of node */
			
			instruction_short(LDL, heap_ptr->parameters[0].int16);
			
			if (old_heap_offset != 0)
				instruction_short(LDNLP, old_heap_offset);
			
			/* now the address of the node is in Areg */
			
			for (i=2; i<graph->arity; ++i)
			{
				if (graph->parameters[i].pointer==NULL)
				{
					--empty_count;
					if (empty_count>0)
						instruction_void (DUP);		/* duplicate for next one */
					
					instruction_void (DUP);
					instruction_short(STNL,i);
				}
			}
		}
	}

	return heap_address(graph,heap_ptr,old_heap_offset);
}

static INSTRUCTION_GRAPH lin_create_r_operator (INSTRUCTION_GRAPH graph)

	/* creates a real node in the heap */
{
	INSTRUCTION_GRAPH desc_graph, float_graph;
	INT16 old_heap_offset;
	INSTRUCTION_GRAPH heap_ptr;
	
	old_heap_offset=current_block->heap_use;
	current_block->heap_use+=3;
	
	heap_ptr	= skip_indirection_nodes(curr_block_graph->heap_ptr);
	desc_graph	= graph->parameters[0].pointer;
	float_graph	= graph->parameters[1].pointer;
		
	/* first fill real descriptor */

	desc_graph=lin_graph (desc_graph);
	in_Areg (desc_graph);
	instruction_short (LDL, heap_ptr->parameters[0].int16);
	instruction_short (STNL, old_heap_offset);
	pop_register();
	
	/* now fill float */
	
	float_graph = lin_float_graph (float_graph);
	
	switch(float_graph->code)
	{	case GFLOAT_LOCAL:
		{
			/* don't use fAreg to store float. Otherwise the following code would be
			   needed:
			   
			   ldlp		f
			   fpldnldb
			   ldl		heap
			   ldnlp	offset
			   fpstnldb
			   
			   this would take 14 ticks, opposed to 12 below
			*/
			
			--float_graph->count;
	
			instruction_short(LDL,float_graph->parameters[0].int16);
			instruction_short(LDL,heap_ptr->parameters[0].int16);
			instruction_short(STNL,old_heap_offset+ARG1_OFFSET);
			group_if_free(N_REGS);
			
			instruction_short(LDL,float_graph->parameters[0].int16+1);
			instruction_short(LDL,heap_ptr->parameters[0].int16);
			instruction_short(STNL,old_heap_offset+ARG1_OFFSET+1);
			group_if_free(N_REGS);
			
			if(float_graph->count==0)
			{	free_local(float_graph->parameters[0].int16);
				free_local(float_graph->parameters[0].int16+1);
			}
			break;
		}
		default:
		{
			in_float_Areg(float_graph);
			
			instruction_short(LDL,heap_ptr->parameters[0].int16);
			instruction_short(LDNLP,old_heap_offset+ARG1_OFFSET);
			instruction_void (FPSTNLDB);
			group_if_free(N_REGS);
			
			++n_free_float_registers;
		}	
	}
	return heap_address(graph,heap_ptr,old_heap_offset);
}

static INSTRUCTION_GRAPH lin_create_s_operator (INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH string_graph;
	INT32 length, words_needed;
	INT16 old_heap_offset;
	INSTRUCTION_GRAPH heap_ptr;

	heap_ptr		=skip_indirection_nodes(curr_block_graph->heap_ptr);
	string_graph	=graph->parameters[0].pointer;
	length			=graph->parameters[1].int16;
	
	words_needed=(length+7)>>2;						/* ensure word alignment */
	
	old_heap_offset=current_block->heap_use;
	current_block->heap_use+=words_needed;
	
	/* compute address of string */
	
	string_graph=lin_graph(string_graph);
	in_Areg (string_graph);
	
	/* compute heap address to put string at */
	
	instruction_short (LDL, heap_ptr->parameters[0].int16);
	if (old_heap_offset > 0)
		instruction_short (LDNLP, old_heap_offset);
	
	/* and move string in position */
	
	instruction_integer (LDC, (INT32)(length+WORD_SIZE));
	instruction_void (MOVE);
	
	empty_registers();
	
	return heap_address(graph,heap_ptr,old_heap_offset);
}

/********************************************************************************
 *																				*
 *		fill operations															*
 *																				*
 ********************************************************************************/

static VOID save_waiting_list(register INSTRUCTION_GRAPH wait_graph)
{	
	if (wait_graph->count != 0 &&
		wait_graph->code != GLOCAL &&
		wait_graph->code != GLOAD_L &&
		wait_graph->code != GEMPTY)
	{
		/* the waiting list is still needed by someone, hasn't been
		   saved yet and is about to be overwritten. Note that in_local
		   will not decrease the count of the waiting list */

		wait_graph=lin_graph(wait_graph);
		in_local(wait_graph);
	}
}

static INSTRUCTION_GRAPH lin_fill_operator (register INSTRUCTION_GRAPH graph)
{
	INT16 *args;
	INT16 n_arguments, address_offset;
	register INT16 i;
	INSTRUCTION_GRAPH address,wait_graph;
		
	n_arguments=graph->arity;
	
	args = alloc_block_mem(n_arguments*sizeof(INT16));
	n_arguments = sort_and_calc_block_graphs (curr_block_graph,(INSTRUCTION_GRAPH *)&graph->parameters[0],2, n_arguments, args);
	
	address			= graph->parameters[0].pointer;
	address			= lin_graph (address);
	address_offset	= in_local (address);
	
	wait_graph		= graph->parameters[1].pointer;
	
	for (i=2; i<n_arguments; i++)
	{	INSTRUCTION_GRAPH argument;
		INT16 argument_n;
		
		argument_n=args[i];
		argument=graph->parameters[argument_n].pointer;
				
		if (argument_n==3 && wait_graph!=NULL)
			save_waiting_list(wait_graph);
		
		argument_n-=2;
		
		if (argument->code==GLOAD_NL &&
			skip_indirection_nodes(argument->parameters[0].pointer)==address &&
			argument->parameters[1].int16==argument_n)
		{
			/* replace arg by itself, so do nothing */
			
			continue;
		}
		
		argument=lin_graph(argument);
		in_Areg(argument);
		instruction_short(LDL, address_offset);
		instruction_short(STNL, argument_n);
		pop_register();
	}
	
	if (address->count>=0)
		address->count += (graph->count - 1);
	
	return change_to_indirection_node (graph, address);
}

static INSTRUCTION_GRAPH lin_fill_r_operator (register INSTRUCTION_GRAPH graph)
{
	INT16 address_offset;
	INSTRUCTION_GRAPH address, wait_graph, desc_graph, float_graph;
		
	address		= graph->parameters[0].pointer;
	wait_graph	= graph->parameters[1].pointer;
	desc_graph	= graph->parameters[2].pointer;
	float_graph	= graph->parameters[3].pointer;
	
	address			= lin_graph (address);
	address_offset	= in_local (address);
	
	desc_graph = lin_graph (desc_graph);
	in_Areg(desc_graph);
	instruction_short (LDL, address_offset);
	instruction_short (STNL, DESCRIPTOR_OFFSET);
	pop_register();
	
	if (wait_graph!=NULL)
		save_waiting_list(wait_graph);
		
	float_graph = lin_float_graph (float_graph);

	switch(float_graph->code)
	{	case GFLOAT_LOCAL:
		{
			/* don't use fAreg to store float. Otherwise the following code would be
			   needed:
			   
			   ldlp		f
			   fpldnldb
			   ldl		address
			   ldnlp	offset
			   fpstnldb
			   
			   this would take 14 ticks, opposed to 12 below
			*/
			
			--float_graph->count;
	
			instruction_short(LDL,float_graph->parameters[0].int16);
			instruction_short(LDL,address_offset);
			instruction_short(STNL,ARG1_OFFSET);
			group_if_free(N_REGS);
			
			instruction_short(LDL,float_graph->parameters[0].int16+1);
			instruction_short(LDL,address_offset);
			instruction_short(STNL,ARG1_OFFSET+1);
			group_if_free(N_REGS);
			
			if(float_graph->count==0)
			{	free_local(float_graph->parameters[0].int16);
				free_local(float_graph->parameters[0].int16+1);
			}
			break;
		}
		default:
		{
			in_float_Areg(float_graph);
			
			instruction_short(LDL,address_offset);
			instruction_short(LDNLP,ARG1_OFFSET);
			instruction_void (FPSTNLDB);
			group_if_free(N_REGS);

			++n_free_float_registers;
		}	
	}
	if (address->count >= 0)
		address->count += (graph->count - 1);
	
	return change_to_indirection_node (graph, address);
}
 
static INSTRUCTION_GRAPH lin_fill_copy (INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH source, dest, wait_graph;
	INT16 n_bytes;
	
	source		= skip_indirection_nodes(graph->parameters[0].pointer);
	dest		= skip_indirection_nodes(graph->parameters[1].pointer);
	wait_graph	= graph->parameters[2].pointer;
	n_bytes		= graph->parameters[3].int16;
	
	if (wait_graph!=NULL)
		save_waiting_list(wait_graph);
	
	if (dest->count >= 0)
		dest->count += graph->count;	/* this will be the result of the graph */
	
	non_commutative_subgraphs (source, dest, graph->order_left);
	
	instruction_integer (LDC, (INT32)n_bytes);
	instruction_void (MOVE);
	
	empty_registers();

	return change_to_indirection_node(graph,skip_indirection_nodes(dest)); /* destination is returned */
}

/********************************************************************************
 *																				*
 *		floating point routines													*
 *																				*
 ********************************************************************************/
 
static INSTRUCTION_GRAPH lin_flow_operator (register INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH float_graph;
	
	float_graph = graph->parameters[0].pointer;
	float_graph = lin_float_graph (float_graph);
	in_float_local (float_graph);
	
	return graph;
}

static INSTRUCTION_GRAPH lin_fhigh_operator (register INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH float_graph;
		
	float_graph = graph->parameters[0].pointer;
	float_graph = lin_float_graph (float_graph);
	in_float_local (float_graph);
	
	return graph;
}

static INSTRUCTION_GRAPH lin_fjoin_operator (INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH high_graph,low_graph;
	INT16 local;
	
	/* the join operator is regarded as a non-commutative dyadic operation:
	   first the two words are pushed on the evaluation stack in the right order
	   and then they are saved in a floating point local. This is not very good
	   if both values already reside in two neighbouring locals, but the chance this
	   occurs is VERY small (as the code generator has already filtered out cases
	   where low and high nodes point to the same float graph or to neighbouring
	   values on the stack) */
	
	low_graph	=graph->parameters[0].pointer;
	high_graph	=graph->parameters[1].pointer;
	
	non_commutative_subgraphs (low_graph, high_graph, graph->order_left);
	
	/* high word is on top */
	
	local = alloc_pref_free_float_local(graph->pref_local);
	
	instruction_short (STL, local+1);
	instruction_short (STL, local);

	pop_registers(2);

	return change_to_float_local_node(graph, local);
}

static INSTRUCTION_GRAPH lin_itor_operator (register INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH graph_1;
	INT16 local;
	
	/* get integer in some local */
	
	graph_1 = graph->parameters[0].pointer;
	graph_1 = lin_graph (graph_1);
	local = in_local (graph_1);
	
	/* and get converted integer in fAreg */
	
	instruction_short (LDLP, local);
	instruction_void (FPI32TOR64);
	group_if_free(N_REGS);
	
	n_free_float_registers--;
	
	decrement_ref_count(graph_1);
		
	return graph;
}

#define TRUNCATE false

static INSTRUCTION_GRAPH lin_rtoi_operator (INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH graph_1;
	INT16 local;
	
	/* get float in fAreg */
	
	graph_1 = graph->parameters[0].pointer;
	graph_1 = lin_float_graph (graph_1);
	in_float_Areg (graph_1);

	/* and put converted float into some local */
	
	local = alloc_pref_free_local(graph->pref_local);
	
	if (! TRUNCATE)
		instruction_void (FPINT);		/* round float */
	instruction_short (LDLP,local);
	instruction_void (FPSTNLI32);
	group_if_free(N_REGS);
	
	++n_free_float_registers;
	change_to_local_node (graph, local);
	return graph;
}

static INSTRUCTION_GRAPH lin_entier_operator (INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH graph_1;
	INT16 local;
	
	/* first get float in fAreg */
	
	graph_1 = graph->parameters[0].pointer;
	graph_1 = lin_float_graph (graph_1);
	in_float_Areg (graph_1);

	/* and put entier of float in a local */
	
	local = alloc_pref_free_local(graph->pref_local);
	
	instruction_void (FPURM);		/* rounding mode to minus-infinity */
	instruction_void (FPINT);		/* round to integer */
	instruction_short (LDLP, local);
	instruction_void (FPSTNLI32);
	group_if_free(N_REGS);
	
	++n_free_float_registers;
	change_to_local_node (graph, local);
	return graph;
}

static INSTRUCTION_GRAPH lin_float_square_root (register INSTRUCTION_GRAPH graph)
{
	register INSTRUCTION_GRAPH argument;
	
	argument=graph->parameters[0].pointer;
	
	argument=lin_float_graph(argument);
	in_float_Areg(argument);
	
	instruction_void(FPUSQRTFIRST);
	instruction_void(FPUSQRTSTEP);
	instruction_void(FPUSQRTSTEP);
	instruction_void(FPUSQRTSTEP);
	instruction_void(FPUSQRTSTEP);
	instruction_void(FPUSQRTSTEP);
	instruction_void(FPUSQRTLAST);
	
	/* this instruction sequence uses Areg and frees it afterwards */
	
	group_if_free(N_REGS);

	return graph;
}

/****************************************************************************
 *																			*
 *		call routines														*
 *																			*
 ****************************************************************************/
 
static INSTRUCTION_GRAPH lin_lib_call (register INSTRUCTION_GRAPH graph, LABEL *routine)
{	
	instruction_label(LIBCALL, routine);	/* result left in Areg */
	--n_free_registers;
	
	return graph;
}

static INSTRUCTION_GRAPH lin_dyadic_comm_lib_call (register INSTRUCTION_GRAPH graph, LABEL *routine)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;

	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	commutative_subgraphs (graph_1, graph_2, graph->order_left);
	instruction_label (LIBCALL, routine);
	pop_register();
		
	return graph;
}

static INSTRUCTION_GRAPH lin_dyadic_non_comm_lib_call (register INSTRUCTION_GRAPH graph, LABEL *routine)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;

	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	non_commutative_subgraphs (graph_1, graph_2, graph->order_left);
	instruction_label (LIBCALL, routine);
	pop_register();
		
	return graph;
}

static INSTRUCTION_GRAPH lin_monadic_float_lib_call (register INSTRUCTION_GRAPH graph, LABEL *routine)
{
	register INSTRUCTION_GRAPH argument;
	
	argument=graph->parameters[0].pointer;
	
	argument=lin_float_graph(argument);
	in_float_Areg(argument);
	
	instruction_label(LIBCALL, routine);
	
	return graph;
}

static INSTRUCTION_GRAPH lin_dyadic_non_comm_float_lib_call (register INSTRUCTION_GRAPH graph, LABEL *routine)
{
	register INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=graph->parameters[0].pointer;
	graph_2=graph->parameters[1].pointer;
	
	non_commutative_float_subgraphs (graph_1, graph_2, graph->order_left);
	
	instruction_label(LIBCALL, routine);
	n_free_float_registers += 1;
	
	return graph;
}

/********************************************************************************
 *																				*
 *		store routines (for stack)												*
 *																				*
 ********************************************************************************/

static VOID save_non_local_load_graph (register INSTRUCTION_GRAPH load_graph)

/* this is used to save the value that resides in a certain non-local
   memory location into a local memory location. */
   
{
	/* the load graph may point to some node that has been put elsewhere */
		
	if (load_graph!=NULL && load_graph->count>0)
	{
		load_graph=skip_indirection_nodes(load_graph);
		switch (load_graph->code)
		{	case GLOAD_NL:
			{
				INSTRUCTION_GRAPH address;
				INT16 address_offset, offset, local;
				
				address = load_graph->parameters[0].pointer;
				
				if (address->code != GLOAD_L)
					internal_error ("save non local load graph");
				
				address_offset	= address->parameters[0].int16;
				offset			= load_graph->parameters[1].int16;

				local = alloc_pref_free_local(load_graph->pref_local);
			   	
			   	if (n_free_registers > 0)
				{
					instruction_short (LDL,address_offset);
					instruction_short (LDNL,offset);
					instruction_short (STL,local);
					group_if_free(N_REGS);
					
				} else
				{	instruction_short (STL,SAVE1);
					instruction_short (LDL,address_offset);
					instruction_short (LDNL,offset);
					instruction_short (STL,local);
					instruction_short (LDL,SAVE1);
				}
				change_to_local_node(load_graph, local);
				return;
			}
			case GFLOAD_NL:
			{
				INSTRUCTION_GRAPH address;
				INT16 address_offset, offset, local;

				address = load_graph->parameters[0].pointer;
				
				if (address->code != GLOAD_L)
					internal_error ("save_load_graph");
				
				address_offset	=address->parameters[0].int16;
				offset			=load_graph->parameters[1].int16;
				
				local = alloc_pref_free_float_local(load_graph->pref_local);
				
				if (n_free_registers > 0)
				{
					instruction_short (LDL,address_offset);
					instruction_short (LDNL,offset);
					instruction_short (STL,local);
					group_if_free(N_REGS);
				
					instruction_short (LDL,address_offset);
					instruction_short (LDNL,offset+1);
					instruction_short (STL,local+1);
					group_if_free(N_REGS);
					
				} else
				{	instruction_short (STL,SAVE1);
				
					instruction_short (LDL,address_offset);
					instruction_short (LDNL,offset);
					instruction_short (STL,local);
				
					instruction_short (LDL,address_offset);
					instruction_short (LDNL,offset+1);
					instruction_short (STL,local+1);
					
					instruction_short (LDL,SAVE1);
				}
				change_to_float_local_node(load_graph, local);
				break;
			}
			case GLOCAL:
			case GFLOAT_LOCAL:
			
				/* value has been saved earlier */
				
				return;

			default:
				internal_error ("save_non_local_load_graph");
		}
	}
}

static INSTRUCTION_GRAPH lin_store_non_local_operator (register INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH graph_1, address;
	INT16 offset;
	
	if (graph->count!=1)
		internal_error("lin_store_non_local: ref count != 1");
	
	graph_1		= graph->parameters[0].pointer;
	address		= graph->parameters[1].pointer;
	offset		= graph->parameters[2].int16;
	
	/* get address and result graph in registers */
	
	non_commutative_subgraphs (graph_1, address, graph->order_left);
	
	/* save stack values that are to be overwritten. usually only two regs are
	   used, so this isn't very troublesome */
	
	save_non_local_load_graph (graph->parameters[3].pointer);
	
	/* now store value on stack */
	
	instruction_short (STNL, offset);
	
	pop_registers(2);

	graph->code=GEMPTY;
	return graph;
}

static INSTRUCTION_GRAPH lin_float_store_non_local_operator (register INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH float_graph, address;
	INT16 offset;
	
	if (graph->count != 1)
		internal_error("lin_float_store_non_local: ref count != 1");
	
	float_graph	= graph->parameters[0].pointer;
	address		= graph->parameters[1].pointer;
	offset		= graph->parameters[2].int16;
	
	if (graph->order_left)
	{	float_graph = lin_float_graph (float_graph);
		if (address->ufloat_regs >= N_FLOAT_REGS)
		{	in_float_local(float_graph);
			address = lin_graph (address);
			in_Areg(address);
			in_float_Areg(float_graph);
		} else
		{	in_float_Areg(float_graph);
			address = lin_graph (address);
			in_Areg(address);
		}
	} else
	{	address = lin_graph (address);
		if (float_graph->uregs >= N_REGS)
		{	in_local(address);
			float_graph = lin_float_graph (float_graph);
			in_float_Areg(float_graph);
			in_Areg(address);
		} else
		{	in_Areg(address);
			float_graph = lin_float_graph (float_graph);
			in_float_Areg(float_graph);
		}
	}
	
	/* now save stack values that are about to be overwritten */
	
	save_non_local_load_graph (graph->parameters[3].pointer);
	save_non_local_load_graph (graph->parameters[4].pointer);
	
	if (offset != 0)
		instruction_short (LDNLP, offset);
	instruction_void (FPSTNLDB);
	
	pop_register();
	++n_free_float_registers;

	graph->code=GEMPTY;
	return graph;
}

/********************************************************************************
 *																				*
 *		store routines (for locals)												*
 *																				*
 ********************************************************************************/

static VOID save_parameter_local(INT16 offset)
{
	register INSTRUCTION_GRAPH load_graph;
	
	/* saves the contents of a single local and allocates it for caller */
	
	load_graph=curr_block_graph->local_parameter_node[offset];
	
	if (load_graph != NULL &&
		load_graph->count!=0)
	{
		/* the value in the local will be needed lateron, there are only 2 cases */
		
		switch(load_graph->code)
		{	case GLOCAL:
			{	register INT16 new_local;
			
				/* save single local */
				
				new_local = alloc_pref_free_local(load_graph->pref_local);
				if (n_free_registers > 0)
				{
					/* normally this is the case */
					
					instruction_short (LDL, load_graph->parameters[0].int16);
					instruction_short (STL, new_local);
					group_if_free(N_REGS);
				
				} else
				{	instruction_short (STL, SAVE1);
					instruction_short (LDL, load_graph->parameters[0].int16);
					instruction_short (STL, new_local);
					instruction_short (LDL, SAVE1);
				}
				free_local(load_graph->parameters[0].int16);
				change_to_local_node(load_graph, new_local);
				
				break;
			}
			case GFLOAT_LOCAL:
			{	register INT16 new_local;
			
				/* save floating point value in 2 locals */
				
				new_local = alloc_pref_free_float_local(load_graph->pref_local);
				
				if (n_free_registers > 0)
				{
					/* is normally the case */
					
					instruction_short (LDL, load_graph->parameters[0].int16);
					instruction_short (STL, new_local);
					group_if_free(N_REGS);
			
					instruction_short (LDL, load_graph->parameters[0].int16+1);
					instruction_short (STL, new_local+1);
					group_if_free(N_REGS);
					
				} else
				{	instruction_short (STL, SAVE1);
				
					instruction_short (LDL, load_graph->parameters[0].int16);
					instruction_short (STL, new_local);
					instruction_short (LDL, load_graph->parameters[0].int16+1);
					instruction_short (STL, new_local+1);
					
					instruction_short (LDL, SAVE1);
				}				
				free_local(load_graph->parameters[0].int16);
				free_local(load_graph->parameters[0].int16+1);
				change_to_float_local_node(load_graph, new_local);
				break;
			}
			case GINDIRECT:			
				internal_error ("save parameter local: indirect nodes must not be used as load graphs for locals");
				
			default:
				internal_error ("save parameter local: invalid node as local parameter node");
		}
	}
	alloc_parameter_local(offset);
}

static VOID save_low_parameter_local(INT16 offset)
{
	register INSTRUCTION_GRAPH load_graph;
	
	/* saves the contents of a local, but not in local+1 (used for floats) */
	
	load_graph=curr_block_graph->local_parameter_node[offset];
	
	if (load_graph != NULL &&
		load_graph->count!=0)
	{
		/* the value in the local will be needed lateron */
			
		switch(load_graph->code)
		{	case GLOCAL:
			{	register INT16 new_local;
			
				/* save single local */
				
				new_local=alloc_pref_free_local(load_graph->pref_local);
				
				if(new_local==offset+1)
				{	INT16 new_local2;
				
					new_local2=alloc_free_local();
					free_local(new_local);
					new_local=new_local2;
				}
				
				if (n_free_registers > 0)
				{
					/* normally this is the case */
					
					instruction_short (LDL, load_graph->parameters[0].int16);
					instruction_short (STL, new_local);
					group_if_free(N_REGS);
				
				} else
				{	instruction_short (STL, SAVE1);
					instruction_short (LDL, load_graph->parameters[0].int16);
					instruction_short (STL, new_local);
					instruction_short (LDL, SAVE1);
				}
				free_local(load_graph->parameters[0].int16);
				change_to_local_node(load_graph, new_local);
								
				break;
			}
			case GFLOAT_LOCAL:
			{	register INT16 new_local;
			
				new_local = alloc_pref_free_float_local(load_graph->pref_local);
				
				if(new_local==offset+1)
				{	INT16 new_local2;
				
					free_local(new_local+1);
					new_local2 = alloc_free_float_local();
					free_local(new_local);
					new_local=new_local2;
				}
				
				if (n_free_registers > 0)
				{
					/* is normally the case */
					
					instruction_short (LDL, load_graph->parameters[0].int16);
					instruction_short (STL, new_local);
					group_if_free(N_REGS);
			
					instruction_short (LDL, load_graph->parameters[0].int16+1);
					instruction_short (STL, new_local+1);
					group_if_free(N_REGS);
					
				} else
				{	instruction_short (STL, SAVE1);
				
					instruction_short (LDL, load_graph->parameters[0].int16);
					instruction_short (STL, new_local);
					instruction_short (LDL, load_graph->parameters[0].int16+1);
					instruction_short (STL, new_local+1);
					
					instruction_short (LDL, SAVE1);
				}
				free_local(load_graph->parameters[0].int16);
				free_local(load_graph->parameters[0].int16+1);
				change_to_float_local_node(load_graph, new_local);
				break;
			}
			case GINDIRECT:
				internal_error ("save parameter low local: indirect nodes must not be used as load graphs for locals");
			default:
				internal_error ("save parameter low local: invalid node as local parameter node");
		}
	}
	alloc_parameter_local(offset);
}

static INSTRUCTION_GRAPH lin_store_local_node (INSTRUCTION_GRAPH store_graph)
{
	INSTRUCTION_GRAPH result_graph;
	INT16 store_local;
	
	if(store_graph->count!=1)
		internal_error("lin_store_local: ref count != 1");
	
	store_local=store_graph->parameters[1].int16;
	if (store_local >= N_LOCALS)
		internal_error("lin store local: local offset is too big");

	/* indicate this local is a parameter for the next block */
	
	set_end_local(store_local);
	
	/* evaluate the result graph */

	result_graph=store_graph->parameters[0].pointer;	
	result_graph=lin_graph (result_graph);
	
	if (result_graph->code == GLOCAL &&
		result_graph->parameters[0].int16==store_local)
	{
		/* result is already in correct local */
		
		/* store_graph->code=GEMPTY;*/		/* forget store node */
		result_graph->count=-1;				/* fix result local */
		return change_to_indirection_node(store_graph,result_graph);
	}
	
	/* get result in Areg to be stored in correct local. Don't use the routine
	   in_Areg, as this could allocate an extra local for the result if it's shared,
	   which is not necessary, as the result will be saved anyway. For the rest the
	   lines below act more or less the same as in_Areg */
	
	--result_graph->count;
	switch (result_graph->code)
	{
		case GLOCAL:
		{	
			instruction_short (LDL, result_graph->parameters[0].int16);
			--n_free_registers;
			
			/* if count greater than zero this local may be moved */
			
			if (result_graph->count>=0)
			{	free_local(result_graph->parameters[0].int16);					
				if (result_graph->count==0)
					result_graph->code=GEMPTY;
				else
					change_to_indirection_node(result_graph,store_graph);
			}
			break;
		}
		case GFLOAT_LOCAL:
			internal_error ("lin_store_local: a float (in locals) cannot be stored in local");
		
		case GFLOW:
		{	INSTRUCTION_GRAPH float_graph;
		
			/* get low word from float local */
			
			float_graph=result_graph->parameters[0].pointer;
			if (float_graph->code!=GFLOAT_LOCAL)
				internal_error("in Areg: a float local node was expected");

			instruction_short (LDL, float_graph->parameters[0].int16);
			--n_free_registers;
			
			if (float_graph->count>=0)				/* float graph isn't fixed */
			{	--float_graph->count;				/* because result redirected */
				if(float_graph->count==0)
				{	free_local(float_graph->parameters[0].int16);
					free_local(float_graph->parameters[0].int16+1);
				}
				if (result_graph->count==0)
					result_graph->code=GEMPTY;
				else
					change_to_indirection_node(result_graph,store_graph);
			}
			break;
		}
		case GFHIGH:
		{	INSTRUCTION_GRAPH float_graph;
		
			/* get high word from float local */
			
			float_graph=result_graph->parameters[0].pointer;
			if (float_graph->code!=GFLOAT_LOCAL)
				internal_error("lin store local: a float local node was expected");

			instruction_short (LDL, float_graph->parameters[0].int16+1);
			--n_free_registers;
			
			if (result_graph->count>=0)
			{	--float_graph->count;
				if(float_graph->count==0)
				{	free_local(float_graph->parameters[0].int16);
					free_local(float_graph->parameters[0].int16+1);
				}
				if (result_graph->count==0)
					result_graph->code=GEMPTY;
				else					
					change_to_indirection_node(result_graph,store_graph);
			}
			break;
		}
		case GEMPTY:
			internal_error("lin store local: no result available to store in Areg");
		
		case GINDIRECT:
			internal_error("lin store local indirect node should not be returned for result");
		
		case GLOAD_L:

			/* result in fixed node, so get it out */
			
			instruction_short (LDL, result_graph->parameters[0].int16);
			--n_free_registers;
			
			break;
		
		default:
		{
			/* result is in Areg, overwrite result node with indirection */
						
			if (result_graph->count==0)
			{	result_graph->code=GEMPTY;
				break;
			}
			if (is_small_comp(result_graph))
				break;
			
			change_to_indirection_node(result_graph,store_graph);
			break;
		}
	}

	/* now result of computation is in Areg.
	   the old value in the local is about to be overwritten. Save it if necessary
	   (thus freeing the result local) */
	
	save_parameter_local(store_local);
	
	/* now the result is in Areg and the result local is free to store value in */
	
	instruction_short(STL,store_local);
	pop_register();
	
	/* and overwrite the store node with a local node */
	
	store_graph->count=-1;
	change_to_local_node(store_graph,store_local);		
	return store_graph;
}

static INSTRUCTION_GRAPH lin_store_float_local_node (INSTRUCTION_GRAPH store_graph)
{
	INSTRUCTION_GRAPH result_graph;
	INT16 store_local;
		
	if(store_graph->count!=1)
		internal_error("lin_store_float_local: ref count != 1");
	
	store_local	= store_graph->parameters[1].int16;
		
	if (store_local >= N_LOCALS-1)
		internal_error("lin store float local: local offset is too big");

	set_end_local(store_local);
	set_end_local(store_local+1);

	/* evaluate the result graph */

	result_graph=store_graph->parameters[0].pointer;
	result_graph = lin_float_graph (result_graph);
	
	if (result_graph->code == GFLOAT_LOCAL &&
		result_graph->parameters[0].int16==store_local)
	{
		/* result is already in correct local */
		
		store_graph->code=GEMPTY;		/* forget store node */
		result_graph->count=-1;			/* fix result local */
		return result_graph;
	}
	
	--result_graph->count;

	if (result_graph->code==GLOCAL ||
		result_graph->code==GFLOW ||
		result_graph->code==GFHIGH ||
		result_graph->code==GEMPTY ||
		result_graph->code==GINDIRECT ||
		result_graph->code==GLOAD_L)
		internal_error("lin store float: invalid result graph");
	
	if (result_graph->code==GFLOAT_LOCAL)
	{	
		/* result is in another float local, move it into the right one */
		
		if (n_free_registers>1)
		{
			n_free_registers-=2;

			instruction_short (LDL, result_graph->parameters[0].int16);
			instruction_short (LDL, result_graph->parameters[0].int16+1);
			
			if (result_graph->count>=0)
			{
				/* result isn't fixed */
				
				free_local(result_graph->parameters[0].int16);
				free_local(result_graph->parameters[0].int16+1);
				
				if(result_graph->count==0)
					result_graph->code=GEMPTY;
				else
					change_to_indirection_node(result_graph,store_graph);
			}
			
			save_low_parameter_local	(store_local);
			save_parameter_local		(store_local+1);
			
			instruction_short(STL,store_local+1);
			instruction_short(STL,store_local);
			
			pop_registers(2);
		
		} else if (n_free_registers==1)
		{
			--n_free_registers;
			instruction_short (LDL, result_graph->parameters[0].int16);
			
			if (result_graph->count>=0)
				free_local(result_graph->parameters[0].int16);
			
			save_low_parameter_local (store_local);
			instruction_short (STL, store_local);
			pop_register();
			
			--n_free_registers;
			instruction_short (LDL, result_graph->parameters[0].int16+1);
			
			if  (result_graph->count>=0)
			{
				free_local(result_graph->parameters[0].int16+1);
				if(result_graph->count==0)
					result_graph->code=GEMPTY;
				else
					change_to_indirection_node(result_graph,store_graph);
			}
			
			save_parameter_local (store_local+1);
			instruction_short (STL, store_local+1);
			pop_register();
		
		} else
		{
			/* no registers free, use SAVE2 to temporary store Areg, as SAVE1
			   could be used by  'save local' This doesn't happen normally.
			   Note that this can be improved (this could result in some
			   needless stores/loads). Grouping is useless here: not enough
			   registers are free */
			
			instruction_short (STL,SAVE2);
			
			instruction_short (LDL, result_graph->parameters[0].int16);
			if (result_graph->count>=0)
				free_local(result_graph->parameters[0].int16);
			
			save_low_parameter_local(store_local);
			instruction_short(STL,store_local);
			
			instruction_short (LDL, result_graph->parameters[0].int16+1);
			if (result_graph->count>=0)
			{	free_local(result_graph->parameters[0].int16+1);
				if(result_graph->count==0)
					result_graph->code=GEMPTY;
				else
					change_to_indirection_node(result_graph,store_graph);
			}
			
			save_parameter_local (store_local+1);
			instruction_short (STL, store_local+1);
			
			instruction_short (LDL,SAVE2);
		}
	} else
	{
		/* the result is supposed to be in fAreg */
		
		if(result_graph->count==0)
			result_graph->code=GEMPTY;
		else
			change_to_indirection_node(result_graph,store_graph);
		
		save_low_parameter_local(store_local);
		save_parameter_local(store_local+1);
			
		if (n_free_registers>0)
		{
			instruction_short(LDLP,store_local);
			instruction_void(FPSTNLDB);
			group_if_free(N_REGS);
			
		} else
		{	instruction_short(STL,SAVE1);
			instruction_short(LDLP,store_local);
			instruction_void(FPSTNLDB);
			instruction_short(LDL,SAVE1);
		}
		n_free_float_registers++;
	}
	/* and overwrite the store node with a float local node */
	
	store_graph->count=-1;
	change_to_float_local_node(store_graph,store_local);
	return store_graph;
}

/********************************************************************************
 *																				*
 *		main linearisation routines												*
 *																				*
 ********************************************************************************/

static INSTRUCTION_GRAPH lin_graph (register INSTRUCTION_GRAPH graph)
{	
	switch (graph->code)
	{	case GINDIRECT:
			return lin_graph(graph->parameters[0].pointer);
					
		case GLOCAL:
			return graph;
		
		case GLOAD_CONST:
		{		
			instruction_integer (LDC,graph->parameters[0].int32);
			
			n_free_registers--;
			return graph;
		}
		case GLOAD_ADDR:
		{			
			instruction_label_short (LDADDR,graph->parameters[0].label,graph->parameters[1].int16);
			
			n_free_registers--;
			return graph;
		}
		case GLOAD_DESC:
		{
			instruction_label_short (LDDESC,graph->parameters[0].label,graph->parameters[1].int16);
			
			n_free_registers--;
			return graph;
		}
		case GLOAD_L:
		{
			/* only used for 'global' vars, otherwise GLOCAL should be used, this
			   routine does nothing: it just returns the GLOAD_L node so others
			   know where to get the result from */
						
			return graph;
		}
		case GPROC_INFO:
		{
			instruction_void(MINT);
			instruction_short(LDNL,HARD_ROOT);
			instruction_short(LDNL,COMM_INFO);
			
			n_free_registers--;
			return graph;
		}
		case GDESC_TABLE:
		{
			instruction_void(MINT);
			instruction_short(LDNL,HARD_DESCRIPTORS);
			
			n_free_registers--;
			return graph;
		}
		case GID:
		{
			instruction_short(LDL,HEAP_START);
			instruction_short(LDNL,ENV_PID_OFFSET);
			
			n_free_registers--;
			return graph;

		}
		case GMINT:
		{
			instruction_void(MINT);
			
			n_free_registers--;
			return graph;
		}
		case GLOAD_NL:
		{
			INSTRUCTION_GRAPH address;
			INT16 offset;
			
			address	=graph->parameters[0].pointer;
			offset	=graph->parameters[1].int16;
						
			address=lin_graph (address);
			in_Areg (address);
						
			instruction_short (LDNL,offset);
			
			return graph;
		}
		case GLOAD_NLP:
		{
			INSTRUCTION_GRAPH address;
			INT16 offset;
			
			address	=graph->parameters[0].pointer;
			offset	=graph->parameters[1].int16;
			
			address		=lin_graph (address);
			in_Areg (address);
			
			if (offset != 0)
				instruction_short (LDNLP, offset);
			
			return graph;
		}
		case GLOAD_B:
		{
			INSTRUCTION_GRAPH address;
			
			address=graph->parameters[0].pointer;
			address = lin_graph (address);
			in_Areg (address);
			
			instruction_void (LBYTE);
			
			return graph;
		}

		case GKEEP:
		{
			register INSTRUCTION_GRAPH graph_1,graph_2;
			
			graph_1=graph->parameters[0].pointer;
			graph_2=graph->parameters[1].pointer;

			/*	GKEEP:
				the first graph just has to be linearised and the result can be
				discarded (for this node). If the keep graph performs a fill, this
				fill will be executed, but the resulting address will be discarded
				if no-one else needs it. If the count of the keep graph is larger
				than 1, the result will be stored in a temporary. In contrast to a
				GSAVE instruction, the keep graph will be evaluated ALWAYS (as the
				count of the keep graph will be larger than zero), so that
				it is possible to evaluate graphs that are not on the stack
				anymore
			*/
			
			if (graph_1->code != GLOCAL && graph_1->count > 0)
			{
				graph_1=lin_graph (graph_1);
				discard_result(graph_1);
			}
	
			if (graph_2->count>=0)
				graph_2->count+=(graph->count-1);
			
			graph_2=lin_graph(graph_2);
			
			return change_to_indirection_node (graph,graph_2);
		}
		case GSAVE:
		{
			register INSTRUCTION_GRAPH graph_1,graph_2;
			
			graph_1=graph->parameters[0].pointer;
			graph_2=graph->parameters[1].pointer;

			/*	GSAVE:
				the first graph just has to be linearised and the result can be
				discarded (for this node). This construct is used if evaluation
				of the second graph is 'dangerous' for the result of the first
				one, e.g. if it overwrites a field in a node that needs to be
				read by the first graph. The first graph will only be linearised
				if someone still needs it, which is the most important difference
				with GKEEP. This is automatically achieved by the difference in
				computing the count for save/keep graphs
			*/

			if (graph_1->code != GLOCAL && graph_1->count > 0)
			{
				/*	count more than zero: somebody else needs this graph,
					the count must be increased here to avoid zeroing it
					too soon
				*/
				
				++graph_1->count;
				graph_1=lin_graph (graph_1);
				discard_result(graph_1);
			}
	
			if (graph_2->count>=0)
				graph_2->count+=(graph->count-1);
			
			graph_2=lin_graph(graph_2);
			
			return change_to_indirection_node (graph,graph_2);
		}

		case GADD:
			return lin_dyadic_commutative_operator (ADD,graph);
			
		case GSUB:
			return lin_dyadic_non_commutative_operator (SUB,graph);
			
		case GDIFF:
			return lin_dyadic_non_commutative_operator (DIFF,graph);
			
		case GMUL:
			return lin_dyadic_commutative_operator (MUL,graph);
		
		case GPROD:
			return lin_dyadic_commutative_operator (PROD,graph);
			
		case GDIV:
			return lin_dyadic_non_commutative_operator (DIV,graph);
			
		case GMOD:
			return lin_dyadic_non_commutative_operator (REM,graph);
		
		case GBSUB:
			return lin_dyadic_commutative_operator (BSUB, graph);
			
		case GWSUB:
			return lin_dyadic_non_commutative_operator (WSUB, graph);
			
		case GAND:
			return lin_dyadic_commutative_operator (AND,graph);
			
		case GOR:
			return lin_dyadic_commutative_operator (OR,graph);
			
		case GXOR:
			return lin_dyadic_commutative_operator (XOR,graph);
			
		case GLSR:
			return lin_dyadic_non_commutative_operator (SHR,graph);
			
		case GLSL:
			return lin_dyadic_non_commutative_operator (SHL, graph);
			
		case GASL:
			return lin_arith_shift_operator(graph, true);
			
		case GASR:
			return lin_arith_shift_operator(graph, false);
			
		case GNOT:
		{
			register INSTRUCTION_GRAPH bool_graph;
	
			bool_graph = graph->parameters[0].pointer;
			bool_graph = lin_graph (bool_graph);
			in_Areg (bool_graph);
	
			instruction_integer (EQC, (INT32)0);
	
			return graph;
		}
		case GABS:
		{
			register INSTRUCTION_GRAPH integer;
	
			integer = graph->parameters[0].pointer;
			integer = lin_graph (integer);
			in_Areg (integer);

			instruction_void(ABS);			
	
			return graph;
		}
		case GEQ:
		{	INSTRUCTION_GRAPH result;
		
			result = lin_dyadic_commutative_operator (DIFF, graph);
			instruction_integer (EQC, (INT32)0);
			
			return graph;
		}
		case GEQC:
		{
			register INSTRUCTION_GRAPH graph_1;
			INT32 constant;
	
			graph_1 = graph->parameters[0].pointer;
			constant = graph->parameters[1].int32;
	
			graph_1 = lin_graph (graph_1);
			in_Areg (graph_1);
			instruction_integer (EQC,constant);
	
			return graph;
		}
		case GEQD:
		{
			register INSTRUCTION_GRAPH graph_1;
	
			graph_1 = graph->parameters[0].pointer;
	
			graph_1 = lin_graph (graph_1);
			in_Areg (graph_1);
			instruction_label_short (EQD,graph->parameters[1].label,graph->parameters[2].int16);
	
			return graph;
		}

		case GADC:
		{
			register INSTRUCTION_GRAPH graph_1;
			INT32 constant;
	
			graph_1 = graph->parameters[0].pointer;
			constant = graph->parameters[1].int32;
	
			graph_1 = lin_graph (graph_1);
			in_Areg (graph_1);
			instruction_integer (ADC, constant);
	
			return graph;
		}
		case G_BCOUNT:
		{
			register INSTRUCTION_GRAPH graph_1;
	
			graph_1 = graph->parameters[0].pointer;
			graph_1 = lin_graph (graph_1);
			in_Areg (graph_1);
			instruction_void (BCNT);
	
			return graph;
		}
				
		case GGT:
			return lin_dyadic_non_commutative_operator (GT, graph);
		
		/*	the GSTORE... instruction represent the top of each graph,
			so the register stack can be emptied. GFSTORE_L and GSTORE_L
			set the parameter registers for the next block. */
			
		case GSTORE_NL:
			empty_registers();
			n_free_float_registers=N_FLOAT_REGS;
			return lin_store_non_local_operator (graph);
		
		case GSTORE_L:
			empty_registers();
			n_free_float_registers=N_FLOAT_REGS;
			return lin_store_local_node (graph);
		
		case GFSTORE_NL:
			empty_registers();
			n_free_float_registers=N_FLOAT_REGS;
			return lin_float_store_non_local_operator (graph);
		
		case GFSTORE_L:
			empty_registers();
			n_free_float_registers=N_FLOAT_REGS;
			return lin_store_float_local_node(graph);
			
		case GCREATE:
			return lin_create_operator (graph);
			
		case GFILL:
			return lin_fill_operator (graph);
			
		case GCREATE_R:
			return lin_create_r_operator (graph);
		
		case GFILL_R:
			return lin_fill_r_operator (graph);
		
		case GFILL_COPY:
			return lin_fill_copy (graph);

		case GCREATE_S:
			return lin_create_s_operator (graph);
			
		case GFHIGH:
			return lin_fhigh_operator (graph);
			
		case GFLOW:
			return lin_flow_operator (graph);
			
		case GFEQ:
			return lin_float_equals_operator (graph);
			
		case GFGT:
			return lin_float_greater_operator (graph);
			
		case GFRTOI:
			return lin_rtoi_operator (graph);
			
		case GFPUENTIER:
			return lin_entier_operator (graph);
			
		case GRANDOM:
			return lin_lib_call (graph, random_label);
			
		case GEQ_SYMBOL:
			return lin_dyadic_comm_lib_call (graph, eq_symbol_label);
			
		case GSTR_EQ:
			return lin_dyadic_comm_lib_call (graph, equal_string_label);
			
		case GSTR_CMP:
			return lin_dyadic_non_comm_lib_call (graph, cmp_string_label);
		
		case GEMPTY:
			internal_error ("lin_graph: an empty node cannot be evaluated");
		
		default:
			internal_error ("lin_graph: unknown node: %d",graph->code);
	}
	return graph; /* reached when error */
}

static INSTRUCTION_GRAPH lin_float_graph (register INSTRUCTION_GRAPH graph)
{
	switch (graph->code)
	{	case GINDIRECT:
			return lin_float_graph (graph->parameters[0].pointer);
			
		case GFLOAT_LOCAL:
			return graph;
		
		case GFLOAD_NL:
		{
			INSTRUCTION_GRAPH address;
			INT16 offset;
			
			address	=graph->parameters[0].pointer;
			offset	=graph->parameters[1].int16;
			
			address = lin_graph (address);
			in_Areg (address);
			
			if (offset != 0)
				instruction_short (LDNLP, offset);
			instruction_void (FPLDNLDB);
			
			pop_register();
			--n_free_float_registers;
			
			return graph;
		}
		case GFLOAD_CONST:
		{
			FLOAT	f;
			
			/*	if float zero: fpldzerodb, else put a floating point number
				in the code, load its address and load the float itself */
			
			f.join.high	=graph->parameters[0].int32;
			f.join.low	=graph->parameters[1].int32;
			
			if (f.f == 0.0)
			{
				instruction_void(FPLDZERODB);
				n_free_float_registers--;
				
				return graph;
				
			} else
			{
				LABEL	*flabel;

				flabel=define_float(&f);
			
				instruction_label_short (LDADDR,flabel,0);
				instruction_void (FPLDNLDB);
				
				group_if_free(N_REGS);
				n_free_float_registers--;
				
				return graph;
			}
		}
		case GFADD:
			return lin_dyadic_commutative_float_operator (FPADD,graph);
			
		case GFSUB:
			return lin_dyadic_non_commutative_float_operator (FPSUB,graph);
			
		case GFMUL:
			return lin_dyadic_commutative_float_operator (FPMUL,graph);
			
		case GFDIV:
			return lin_dyadic_non_commutative_float_operator (FPDIV,graph);
			
		case GFREM:
			return lin_dyadic_non_commutative_float_operator (FPREM,graph);
						
		case GFJOIN:
			return lin_fjoin_operator (graph);
			
		case GFITOR:
			return lin_itor_operator (graph);
			
		case GFPUSQRT:
			return lin_float_square_root(graph);

		case GFPUMULBY2:
		
			lin_monadic_float_operator(FPUMULBY2,graph);
			group_if_free(N_REGS);
						
			return graph;
			
		case GFPUDIVBY2:

			lin_monadic_float_operator(FPUDIVBY2,graph);
			group_if_free(N_REGS);
			
			return graph;
			
		case GFPUEXPINC32:

			lin_monadic_float_operator(FPUEXPINC32,graph);
			group_if_free(N_REGS);
			
			return graph;
			
		case GFPUEXPDEC32:

			lin_monadic_float_operator(FPUEXPDEC32,graph);
			group_if_free(N_REGS);
			
			return graph;
			
		case GFSIN:
			return lin_monadic_float_lib_call(graph, sin_label);
		case GFCOS:
			return lin_monadic_float_lib_call(graph, cos_label);
		case GFTAN:
			return lin_monadic_float_lib_call(graph, tan_label);
			
		case GFASIN:
			return lin_monadic_float_lib_call(graph, asin_label);
		case GFACOS:
			return lin_monadic_float_lib_call(graph, acos_label);
		case GFATAN:
			return lin_monadic_float_lib_call(graph, atan_label);
			
		case GFLN:
			return lin_monadic_float_lib_call(graph, ln_label);
		case GFLOG10:
			return lin_monadic_float_lib_call(graph, log10_label);
		case GFEXP:
			return lin_monadic_float_lib_call(graph, exp_label);
		
		case GFPOWER:
			return lin_dyadic_non_comm_float_lib_call(graph, power_label);
		
		case GEMPTY:
			internal_error("lin float graph: empty nodes cannot be evaluated");
			
		default:
			internal_error ("lin float graph: unknown node");
	}
	return graph; /* reached when error */
}

INSTRUCTION_BLOCK *lin_block (BLOCK_GRAPH *block_graph,INT16 n_elements,INSTRUCTION_GRAPH graphs[])
{
	INT16 n;
	INT16 *args;

	handle_compute_event();
	
	curr_block_graph=block_graph;
	
	/* end previous block and allocate a new instruction block */
	
	new_instruction_block();
	
	current_block->start_label			=block_graph->start_label;
	current_block->node_descriptor		=block_graph->start_descriptor;
	current_block->node_arity			=block_graph->start_arity;

/*	heap use is zero and will be increased during linearisation
	check_size is increased during local allocation if more stack space is needed
	for virtual locals */

	current_block->a_increase			=block_graph->a_increase;
	current_block->b_increase			=block_graph->b_increase;
	current_block->begin_n_a_locals		=block_graph->begin_n_a_locals;
	current_block->begin_n_b_locals		=block_graph->begin_n_b_locals;
	
	args = alloc_block_mem(n_elements*sizeof(INT16));
	n_elements = sort_and_calc_block_graphs (curr_block_graph,graphs, 0, n_elements, args);
	
	free_all_locals();
	
	/* allocate used parameter locals */
	
	for (n=0; n<N_LOCALS; ++n)
	{	register INSTRUCTION_GRAPH local_graph;
	
		local_graph=block_graph->local_parameter_node[n];
		if (local_graph !=NULL && local_graph->count!=0)
			alloc_parameter_local(n);
	}

	/* linearise all graphs */
	
	for (n=0; n<n_elements; n++)
		lin_graph (graphs[args[n]]);
				
	return current_block;
}

INSTRUCTION_BLOCK *lin_graph_Areg (BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph)
{
	INSTRUCTION_GRAPH result;
	
	if (block_graph!=curr_block_graph)
		internal_error("lin graph Areg: wrong basic block passed");
	
	calc_block_graph (block_graph,graph);

	result=lin_graph (graph);
	in_Areg (result);
	
	/* indicate that 1 reg is in use */
	
	group(N_REGS-1);
	
	return current_block;
}

VOID block_ready(BLOCK_GRAPH *block_graph, BOOL combine_checks)
{
	INSTRUCTION_BLOCK *block;
	
	block=block_graph->instruction_block;
		
	if (block->ready)
		internal_error("block_ready called twice for the same block");
		
	block->ready=true;

	/* write intermediate code to file if required */
	
	if (options.output_intermediate_code1)
		write_intermediate_block(block);

	if (n_virtual_locals>N_LOCALS)
	{
		schedule_locals(block,n_virtual_locals);
		ungroup_block(block);
		
	} else
		ungroup_block(block);
		
	if (combine_checks)
		combine_stack_checks(block_graph);

	/* release block memory */
	
	free_block_mem();
}

INSTRUCTION_BLOCK *get_current_block(VOID)
{
	return current_block;
}

INSTRUCTION_BLOCK *initialize_linearization (VOID)
{
	number_of_blocks=0;
	
	n_free_registers = N_REGS;
	n_free_float_registers = N_FLOAT_REGS;
	
	free_locals_list.next=NULL;
	free_all_locals();

	uncoded_floats				= NULL;
	uncoded_strings				= NULL;
		
	first_uncoded_symbol		= NULL;
	last_uncoded_symbol			= NULL;
	first_uncoded_descriptor	= NULL;
	last_uncoded_descriptor		= NULL;
	
	first_block=NULL;
	current_block=NULL;
	first_group=NULL;
	last_group=NULL;
	
	new_instruction_block();
	
	return current_block;
}
