#ifndef _LIN_
#define _LIN_

#ifndef _TYPES_
#include "cctypes.h"
#endif

extern INSTRUCTION_BLOCK *initialize_linearization(VOID);

/************************************************************************************
 *																					*
 *		routines to linearise the graphs of a certain block graph					*
 *																					*
 ************************************************************************************/

extern INSTRUCTION_BLOCK *lin_block			(BLOCK_GRAPH *block_graph,INT16 n_elements,INSTRUCTION_GRAPH *graphs);
extern INSTRUCTION_BLOCK *lin_graph_Areg	(BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph);

/************************************************************************************
 *																					*
 *		routine get the last linearised block										*
 *																					*
 ************************************************************************************/
 
extern INSTRUCTION_BLOCK *get_current_block(VOID);

/************************************************************************************
 *																					*
 *		routine to add declared data to a certain block								*
 *																					*
 ************************************************************************************/

extern VOID lin_cdata (INSTRUCTION_BLOCK *block);

/************************************************************************************
 *																					*
 *		routines to define data														*
 *																					*
 ************************************************************************************/

extern VOID		define_symbol			(LABEL *label);
extern LABEL	*define_string			(STRING *string);

extern VOID		define_label_string		(LABEL *label, STRING *string);
extern VOID 	define_label_descriptor	(LABEL *label, LABEL *node_entry_label, LABEL *lazy_entry_label, INT16 arity,STRING *descriptor_name);

/************************************************************************************
 *																					*
 *		routines to append instructions to a block									*
 *																					*
 ************************************************************************************/

extern INSTRUCTION *append_instr_void			(INSTRUCTION_BLOCK *block,INT16 i_code);
extern INSTRUCTION *append_instr_integer		(INSTRUCTION_BLOCK *block,INT16 i_code, INT32 constant);
extern INSTRUCTION *append_instr_short			(INSTRUCTION_BLOCK *block,INT16 i_code, INT16 constant);
extern INSTRUCTION *append_instr_short2			(INSTRUCTION_BLOCK *block,INT16 instruction_code, INT16 n1,INT16 n2);
extern INSTRUCTION *append_instr_short4			(INSTRUCTION_BLOCK *block,INT16 i_code, INT16 c1,INT16 c2,INT16 c3,INT16 c4);
extern INSTRUCTION *append_instr_label			(INSTRUCTION_BLOCK *block,INT16 i_code, LABEL *label);
extern INSTRUCTION *append_instr_label_int16	(INSTRUCTION_BLOCK *block,INT16 i_code, LABEL *label,INT16 n);
extern INSTRUCTION *append_instr_int16_int16	(INSTRUCTION_BLOCK *block,INT16 i_code, INT16 n1,INT16 n2);
extern INSTRUCTION *append_instr_float			(INSTRUCTION_BLOCK *block,INT16 i_code, DOUBLE r);

/*	this routine should be called after all intermediate instructions for a
	certain block have been generated. This should be done in the right order
	and exactly once for each block. This routine will invoke the local scheduler
	and evt. the intermediate code printer.
*/

extern VOID block_ready(BLOCK_GRAPH *block_graph, BOOL combine_checks);

extern TINSTRUCTION **append_inline_block		(INSTRUCTION_BLOCK *block);

#endif