#include <string.h>

#include "ifinterface.h"

#include "cgoutputabc.h"
#include "cglabels.h"
#include "cgmemory.h"
#include "cgoutputs.h"
#include "cgwrites.h"
#include "cgscanner.h"

#include "ccoffsets.h"
#include "ccmacros.h"
#include "ccintermediate.h"

/* from cgcode.c */

extern DDESCRIPTOR		*first_uncoded_descriptor;
extern INT32			number_of_blocks;
extern INT32			number_of_inline_instructions;
extern INT32			inline_instruction_number;

static INT32			block_number;

/********************************************************************
 *																	*
 *		global data													*
 *																	*
 ********************************************************************/

static INSTRUCTION_BLOCK *current_block;

static BOOL		current_ctop_stack;
static INT16	current_ctop_offset;	/* should be positive */

static BOOL reachable;

/********************************************************************
 *																	*
 *		auxiliary routines											*
 *																	*
 ********************************************************************/

#define increase_pointer(p,n)			{o_ldl16(p);o_ldnlp16(n);o_stl16(p);}
#define prepare_context_switch(a_pars)	{o_ldc16(a_pars);o_stl16(EXTRA_LOCAL);}
#define o_desc_table()					{o_mint();o_ldnl16(HARD_DESCRIPTORS);}

/* note: the following macro's leave Areg intact */

#define put_ctop_stack(n)	{o_ldl16(RETURN_ADDRESS);o_ldl16(B_STACK_POINTER);o_stnl16(n);}
#define put_ctop_local(n)	{o_ldl16(B_STACK_POINTER);o_ldnl16(n);o_stl16(RETURN_ADDRESS);}

static VOID adjust_stack_pointers_begin(VOID)
{	
	/*	stack ptr is kept as low as possible during basic block,
		so only adjust at beginning of block if a shrinks and/or b grows */
	
	if (current_block->a_increase < 0)
		increase_pointer(A_STACK_POINTER,current_block->a_increase);
			
	if (current_block->b_increase > 0)
	{	increase_pointer(B_STACK_POINTER,-current_block->b_increase);
		current_ctop_offset+=current_block->b_increase;
	}
}

static VOID adjust_stack_pointers_end(VOID)
{	
	if (current_block->a_increase > 0)
		increase_pointer(A_STACK_POINTER,current_block->a_increase);
			
	if (current_block->b_increase < 0)
	{	increase_pointer(B_STACK_POINTER,-current_block->b_increase);
		current_ctop_offset+=current_block->b_increase;
	}
	
	current_block->a_increase=0;
	current_block->b_increase=0;
}

static BOOL is_label_ctop_stack (LABEL *label)
{	
	if (label->flags & FIXED_CTOP)
	{	if (label->flags & CTOP_STACK)
			return true;
		return false;
	}
	/* not yet defined, take current situation */

	label->flags|=FIXED_CTOP;

	if(current_ctop_stack)
	{	label->flags|=CTOP_STACK;
		return true;
	}
	label->flags&=~CTOP_STACK;
	return false;
}

static VOID adjust_label_ctop(LABEL *label)
{
	if (label->flags & FIXED_CTOP)
	{	if(label->flags & CTOP_STACK)
		{	if(!current_ctop_stack)
			{	put_ctop_stack(current_ctop_offset);
				current_ctop_stack=true;
			}
		} else
		{	if(current_ctop_stack)
			{	put_ctop_local(current_ctop_offset);
				current_ctop_stack=false;
			}
		}
	}else
	{	label->flags |= FIXED_CTOP;
		if (current_ctop_stack)
			label->flags |= CTOP_STACK;
		else
			label->flags &=~ CTOP_STACK;
	}
}

/****************************************************************************
 *																			*
 *		routines to save/restore locals on/from stack when performing		*
 *		a conditional call													*
 *																			*
 ****************************************************************************/

static VOID eval_save_stack (INT16 n_a_locals,INT16 n_b_locals, INT16 reserve_return,INT16 a_param_local)
{
	/*	save all b-locals and keep top free for return address, if there
		are more than 2 locals to be saved it is better to use a single
		move than separate loads/stores. Note that not only the instruction
		times as given in the compiler writers guide are important, but
		the instruction fetch time must be considered as well: the variant
		with the move uses less words (for 3 words: 6 bytes vs. 9 bytes).
		For restoration of locals the situation is slightly more in favour
		of the separate loads/stores as a single stl costs only 1 cycle. */
	
	/*	make room on b-stack for locals and possibly a return address. Note that
		the field reserve_return must be 0 or 1 */

	if (!current_ctop_stack && current_ctop_offset == 0 && n_b_locals > 1)
	{
		/*	the ctop must be stored on the stack. A word has been reserved
			for it already and this word is on top of the stack. This means
			that all locals+return address can be stored in one go. Note that
			the return address is actually not placed in the word that has been
			reserved for it, but this should be no problem. */
		
		increase_pointer(B_STACK_POINTER,-(n_b_locals+reserve_return));

		o_ldlp16(RETURN_ADDRESS);
		o_ldl16(B_STACK_POINTER);
		if (reserve_return)
			o_ldnlp16(1);
		o_ldc16((n_b_locals+1) * WORD_SIZE);
		o_move();
		
	} else
	{
		if (!current_ctop_stack)
			put_ctop_stack(current_ctop_offset);
			
		if (n_b_locals+reserve_return > 0)
			increase_pointer(B_STACK_POINTER,-(n_b_locals+reserve_return));
	
		if (n_b_locals > 2)
		{
			o_ldlp16(B_PAR_FIRST_LOCAL);
			o_ldl16(B_STACK_POINTER);
			if (reserve_return)
				o_ldnlp16(1);
			o_ldc16(n_b_locals * WORD_SIZE);
			o_move();
		
		} else
		{
			INT16 i,b_offset;
			
			b_offset=reserve_return+n_b_locals;
			for (i=B_PAR_FIRST_LOCAL+n_b_locals-1; i>=B_PAR_FIRST_LOCAL;--i)
			{
				o_ldl16(i);
				o_ldl16(B_STACK_POINTER);
				o_stnl16(--b_offset);
			}
		}
	}

	if (n_a_locals > 3 && (a_param_local==A_PAR_FIRST_LOCAL || a_param_local==A_PAR_FIRST_LOCAL+n_a_locals-1))
	{
		/*	more than 2 locals need to be saved and the parameter local is not
			between other locals
		*/
		
		if (a_param_local == A_PAR_FIRST_LOCAL)
			o_ldlp16(A_PAR_FIRST_LOCAL+1);
		else
			o_ldlp16(A_PAR_FIRST_LOCAL);
		o_ldl16(A_STACK_POINTER);
		o_ldc16((n_a_locals-1) * WORD_SIZE);
		o_move();
		
	} else
	{
		INT16 i,a_offset;
		
		a_offset=n_a_locals-1;
		for (i=A_PAR_FIRST_LOCAL+a_offset; i>=A_PAR_FIRST_LOCAL;--i)
		{	if(i!=a_param_local)
			{
				o_ldl16(i);
				o_ldl16(A_STACK_POINTER);
				o_stnl16(--a_offset);
			}
		}
	}

	/* make room for locals-parameter */
	
	if (n_a_locals > 1)
		increase_pointer(A_STACK_POINTER,n_a_locals-1);
}

static VOID eval_restore_stack (INT16 n_a_locals,INT16 n_b_locals,INT16 free_return,INT16 a_param_local)
{
	if (n_a_locals > 1)
		increase_pointer(A_STACK_POINTER,-(n_a_locals-1));

	if (n_a_locals > 3 && (a_param_local==A_PAR_FIRST_LOCAL || a_param_local==A_PAR_FIRST_LOCAL+n_a_locals-1))
	{
		/*	more than 2 locals need to be restored and the parameter local is not
			between other locals
		*/

		o_ldl16(A_STACK_POINTER);		
		if (a_param_local == A_PAR_FIRST_LOCAL)
			o_ldlp16(A_PAR_FIRST_LOCAL+1);
		else
			o_ldlp16(A_PAR_FIRST_LOCAL);
		o_ldc16((n_a_locals-1) * WORD_SIZE);
		o_move();
		
	} else
	{
		INT16 i,a_offset;
		
		a_offset=n_a_locals-1;
		for (i=A_PAR_FIRST_LOCAL+a_offset; i>=A_PAR_FIRST_LOCAL;--i)
		{	if (i!=a_param_local)
			{
				o_ldl16(A_STACK_POINTER);
				o_ldnl16(--a_offset);
				o_stl16(i);
			}
		}
	}

	if (!current_ctop_stack && current_ctop_offset == 0 && n_b_locals > 1)
	{
		/* inverse of the routine above */

		o_ldl16(B_STACK_POINTER);
		if (free_return)
			o_ldnlp16(1);
		o_ldlp16(RETURN_ADDRESS);
		o_ldc16((n_b_locals+1) * WORD_SIZE);
		o_move();
		
		increase_pointer(B_STACK_POINTER,n_b_locals+free_return);
	
	} else
	{
		if (n_b_locals > 2)
		{
			o_ldl16(B_STACK_POINTER);
			if (free_return)
				o_ldnlp16(1);
			o_ldlp16(B_PAR_FIRST_LOCAL);
			o_ldc16(n_b_locals * WORD_SIZE);
			o_move();
			
		} else
		{
			INT16 i,b_offset;
			
			b_offset=free_return+n_b_locals;
			for (i=B_PAR_FIRST_LOCAL+n_b_locals-1; i>=B_PAR_FIRST_LOCAL;--i)
			{
				o_ldl16(B_STACK_POINTER);
				o_ldnl16(--b_offset);
				o_stl16(i);
			}
		}
		
		/* remove space from b-stack */
		
		if (n_b_locals+free_return > 0)
			increase_pointer(B_STACK_POINTER,n_b_locals+free_return);
	
		/* restore return address */
		
		if (!current_ctop_stack)
			put_ctop_local(current_ctop_offset);
	}
}

/************************************************************************************
 *																					*
 *		simple non-standard transputer instructions									*
 *																					*
 ************************************************************************************/

static VOID out_eqd(INSTRUCTION *instruction)
{
	o_eqc();
	o_desc_id(instruction->parameters[0].label,instruction->parameters[1].int16,false);
}

static VOID out_minus(INSTRUCTION *instruction)
{
	/* use 1 reg, integer in Areg */
	
	o_not();
	o_adc32(1);
}

static VOID out_abs(INSTRUCTION *instruction)
{
	/* integer in Areg, 3 locals may be used... */

	INT16 positive;
	
	positive=new_label_id();

	o_dup();						/* remember integer */
	
	o_mint();
	o_and();
	o_cj_local(positive);			/* integer is positive */

	/* integer is negative, get unary minus */
	
	out_minus(instruction);
	
	o_ldc32(0);
	o_ldef_label(positive);
	o_diff();						/* remove 0 from Areg */	
}

static VOID out_lea(INSTRUCTION *instruction)
{
	o_address(instruction->parameters[0].label,instruction->parameters[1].int16);
}

static VOID out_ldesc(INSTRUCTION *instruction)
{
	o_ldc();
	o_desc_id(instruction->parameters[0].label,instruction->parameters[1].int16,false);
}

static VOID out_fprem(INSTRUCTION *instruction)
{
	INT16 loop,next;
	
	loop=new_label_id();
	next=new_label_id();
	
	/*	fpremfirst
		eqc			0
		cj			next
	loop
		fpremstep
		cj			loop
	next
	
	*/
	
		o_fpremfirst();
		o_eqc16(0);
		o_cj_local(next);
		
	o_ldef_label(loop);
	
		o_fpremstep();
		o_cj_local(loop);
		
	o_ldef_label(next);
}

/************************************************************************************
 *																					*
 *		flow of control routines													*
 *																					*
 ************************************************************************************/

static VOID out_j(INSTRUCTION *instruction)
{
	register LABEL *label;
	
	label=instruction->parameters[0].label;

	adjust_label_ctop(label);	
	adjust_stack_pointers_end();	

	if (label->flags & FIXED_CTOP_OFFSET)
	{	if (label->b.c_offset != current_ctop_offset)
			fatal_error("jump (%s) ctop offset ambiguous",label->name);
	}else
	{	label->flags|=FIXED_CTOP_OFFSET;
		label->b.c_offset = current_ctop_offset;
	}
	
	/* output jump instruction */

	if (options.parallel_code)
	{
		prepare_context_switch(label->a.n_locals);

		o_j();
		
	} else
	{
		o_ldc16(0);
		o_cj();
	}
	
	o_jlabel(label);
	o_newline();
	
	current_ctop_offset=0;
	current_ctop_stack=true;
	reachable=false;
}

static VOID out_cj(INSTRUCTION *instruction)
{
	register LABEL *label;
	
	label=instruction->parameters[0].label;
	
	adjust_label_ctop(label);
	adjust_stack_pointers_end();		/* keep Areg intact */

	if (label->flags & FIXED_CTOP_OFFSET)
	{	if (label->b.c_offset != current_ctop_offset)
			fatal_error("conditional jump ctop offset ambiguous");
	}else
	{	label->flags |= FIXED_CTOP_OFFSET;
		label->b.c_offset = current_ctop_offset;
	}

	/* output conditional jump instruction */
	
	o_cj();
	o_jlabel(label);
	o_newline();
}

static VOID out_gcall(INSTRUCTION *instruction)
{
	LABEL *gcall_label;

	gcall_label=instruction->parameters[0].label;

	if (!(gcall_label->flags & FIXED_CTOP_OFFSET))
		internal_error("out_gcall");

	if (!current_ctop_stack)
	{	put_ctop_stack(current_ctop_offset);
		current_ctop_stack=true;
	}
	
	adjust_stack_pointers_end();
	
	o_address(gcall_label,0);
	o_gcall();
	
	/* the amount of parameters on the b-stack may have been increased/decreased */
	
	current_ctop_offset+=instruction->parameters[1].int16;
}

static VOID out_jsr(INSTRUCTION *instruction)
{
	LABEL *jsr_label;
	
	jsr_label=instruction->parameters[0].label;

	if (!(jsr_label->flags & FIXED_CTOP_OFFSET))
		internal_error("out_jsr");
	
	/* if c top cached in local, store it on stack */
	
	if (!current_ctop_stack)
		put_ctop_stack(current_ctop_offset);

	adjust_stack_pointers_end();

	/* for parallel code the jsr is implemented by means of a jump instruction
	   this inserts a descheduling point, which is slightly less efficient, but fair.
	   for sequential code a gcall is used for local functions, which also implies
	   the callee should store the return address (for imported functions a
	   jump is used, as the gcall-label may not be defined in the other module). */
	
	if (options.parallel_code || !(jsr_label->flags & DEF_LABEL))
	{
		INT16 return_label;
		
		prepare_context_switch(jsr_label->a.n_locals);

		return_label=new_label_id ();
		o_local_address(return_label,0);
		
		current_ctop_stack=true;				/* preference for the routine that is called (and after return) */
		if (is_label_ctop_stack(jsr_label))
		{
			o_ldl16(B_STACK_POINTER);
			o_stnl16(jsr_label->b.c_offset);		
		} else
		{
			o_stl16(RETURN_ADDRESS);
		}
		
		o_j();	
		o_jlabel(jsr_label);
		o_newline();
		
		o_align();
		o_ldef_label(return_label);
		
	} else
	{
		current_ctop_stack=true;
		
		o_address(jsr_label,0);
		o_gcall();
	}
	
	/* the amount of parameters on the b-stack may have been increased/decreased */

	current_ctop_offset+=instruction->parameters[1].int16;
}

static VOID out_rtn(INSTRUCTION *instruction)
{
	INT16 rtn_ctop_offset;
	
	adjust_stack_pointers_end();

	rtn_ctop_offset=instruction->parameters[0].int16;
	if(rtn_ctop_offset != current_ctop_offset)
		fatal_error("rtn ambiguous ctop offset (rtn %d current %d)",(int)rtn_ctop_offset,(int)current_ctop_offset);
	
	if (current_ctop_stack)
	{
		o_ldl16(B_STACK_POINTER);
		o_ldnl16(rtn_ctop_offset);
	} else
	{
		o_ldl16(RETURN_ADDRESS);
	}
	
	o_gcall();
	
	current_ctop_stack=true;
	current_ctop_offset=0;
	reachable=false;
}

static VOID out_jsr_ap(INSTRUCTION *instruction)
{
	/* local 6 contains a pointer to the node of which the
	   lazy entry has to be called (i.e. the top of the a-stack) */
			
	if (!current_ctop_stack)
	{	put_ctop_stack(current_ctop_offset);
		current_ctop_stack=true;
	}
	
	adjust_stack_pointers_end();
	
	/* now get descriptor of node */
	
	o_ldl16(A2);
	o_ldnl16(DESCRIPTOR_OFFSET);		/* get descriptor offset */
	o_desc_table();						/* get address of desc table */
	o_add();							/* get address of descriptor */
	
	/* and call lazy entry */
	
	o_dup();
	o_ldnl16(1);
	o_add();
	o_gcall();
		
	/* the ctop offset should remain the same */
}

static VOID out_jmp_ap(INSTRUCTION *instruction)
{
	/* local 6 contains a pointer to the node of which the
	   lazy entry has to be called (i.e. the top of the a-stack) */
	
	adjust_stack_pointers_end();
	if (current_ctop_offset != 0)
		internal_error("out_jmp_ap");

	/* load current return address */
	
	if (current_ctop_stack)
	{
		o_ldl16(B_STACK_POINTER);
		o_ldnl16(current_ctop_offset);		
	} else
	{
		o_ldl16(RETURN_ADDRESS);
	}
	
	/* get descriptor */

	o_ldl16(A2);
	o_ldnl16(DESCRIPTOR_OFFSET);		/* get desc offset */
	o_desc_table();								/* get desc table */
	o_add();							/* get desc address */

	/* and call entry -1 (to keep old return address) */
	
	o_dup();
	o_ldnl16(1);
	o_add();
	o_adc16(-1);
	o_gcall();
	
	current_ctop_stack=true;
	current_ctop_offset=0;
	reachable=false;
}

static VOID out_jsr_eval(INSTRUCTION *instruction)
{
	INT16 end_label;
	
	INT16 n_a_locals, n_b_locals;
	INT16 check_size;
	INT16 param_local;
	
	end_label=new_label_id ();

	check_size		=instruction->parameters[1].int16;
	n_a_locals		=instruction->parameters[2].int16;
	n_b_locals		=instruction->parameters[3].int16;
	
	param_local		=A_PAR_FIRST_LOCAL+n_a_locals-instruction->parameters[0].int16-1;
	
	adjust_stack_pointers_end();

	/* HNF TEST: get evaluation code in Areg */
	
	o_ldl16(param_local);
	o_ldnl16(CODE_OFFSET);
	
	/* jump around if positive (a descriptor) */
	
	o_mint();
	o_and();
	o_cj_local(end_label);	
	
	
	/* EVALUATION CODE: check if enough heap is available to save locals */
	
	if (options.check_stacks && check_size > 0)
	{
		INT16 ready_label;
	
		ready_label=new_label_id ();
		
		/* test stack */
		
		o_ldl16(A_STACK_POINTER);
		o_ldnlp16(check_size);
		o_ldl16(B_STACK_POINTER);
		o_gt();
		o_cj_local(ready_label);
		
		/* call reallocation routine */
		
		o_ldc16(n_a_locals);
		o_stl16(EXTRA_LOCAL);
		o_ldc16(check_size);
		
		o_call();
		o_inline_label(realloc_label);
		o_newline();
		
		o_ldef_label(ready_label);
	}
		
	/* save locals,return address, and leave room for return address on b-stack */
		
	eval_save_stack(n_a_locals,n_b_locals,1,param_local);
	
	/* put param in correct local and call eval code */
	
	o_ldl16(param_local);
	if (param_local!=A_PAR_FIRST_LOCAL)
	{
		o_dup();
		o_stl16(A_PAR_FIRST_LOCAL);
	}
	
	o_ldnl16(CODE_OFFSET);
	o_gcall();

	/* node evaluated: restore locals */
	
	if (param_local!=A_PAR_FIRST_LOCAL)
	{
		o_ldl16(A_PAR_FIRST_LOCAL);
		o_stl16(param_local);
	}
	
	eval_restore_stack(n_a_locals,n_b_locals,1,param_local);
	
	/* HNF CODE: output overide label */
	
	o_ldef_label(end_label);	
}


static VOID out_jmp_eval(INSTRUCTION *instruction)
{
	INT16 return_label;
	
	/* top of a-stack is in A_PAR_FIRST_LOCAL */
		
	return_label=new_label_id();
	adjust_stack_pointers_end();
	if (current_ctop_offset!=0)
		fatal_error("jmp_eval ambiguous ctop offset");

	/* get current return address in Areg */
	
	if (current_ctop_stack)
	{
		o_ldl16(B_STACK_POINTER);
		o_ldnl16(current_ctop_offset);
	
	} else
	{
		o_ldl16(RETURN_ADDRESS);
	}

	/* HNF TEST: get eval code and test for HNF */
	
	o_ldl16(A_PAR_FIRST_LOCAL);
	o_ldnl16(CODE_OFFSET);
	
	/* jump around if positive (a descriptor) */
	
	o_mint();
	o_and();
	o_cj_local(return_label);	
	
	/* EVAL CODE: return address is in Areg, call eval code -1 */
	
	o_ldl16(A_PAR_FIRST_LOCAL);
	o_ldnl16(CODE_OFFSET);
	o_adc16(-1);
	o_gcall();
		
	/* HNF CODE: remove zero in Areg and return (to address on eval stack) */
	
	o_ldef_label(return_label);
	o_diff();
	o_gcall();
		
	current_ctop_offset=0;
	current_ctop_stack=true;
	reachable=false;
}

static VOID out_cs(INSTRUCTION *instruction)
{
	INT16 n_a_locals;
	INSTRUCTION_BLOCK *next_block;
	
	next_block=current_block->next;
	if(next_block==NULL)
		internal_error("out_cs");

	n_a_locals=next_block->begin_n_a_locals;
	
	adjust_stack_pointers_end();
	
	prepare_context_switch(n_a_locals);

	o_ldlp16(0);
	o_adc16(1);
	o_runp();
	o_stopp();
}

static VOID out_enable_cs(INSTRUCTION *instruction)
{
	INT16 n_a_locals;
	INSTRUCTION_BLOCK *next_block;
	
	next_block=current_block->next;
	if(next_block==NULL)
		internal_error("out_enable_cs");

	n_a_locals=next_block->begin_n_a_locals;
	
	adjust_stack_pointers_end();
	
	prepare_context_switch(n_a_locals);

	/* code to decrease priority */

	o_ldlp16(0);
	o_adc16(1);
	o_runp();
	o_stopp();
}

static VOID out_disable_cs(INSTRUCTION *instruction)
{
	INT16 end_label;
	
	end_label=new_label_id();
	
	adjust_stack_pointers_end();
	
	/*	code to increase priority, one does not have to store
		the amount of pointers */

	o_local_address(end_label,0);
	o_stl16(-1);
	o_mint();
	o_ldnlp16(HARD_STOP_WSP);
	o_gajw();
	o_runp();
	o_stopp();
	o_ldef_label(end_label);
}

static VOID out_suspend(INSTRUCTION *instruction)
{
	INT16 n_a_locals;
	INSTRUCTION_BLOCK *next_block;

	next_block=current_block->next;
	if(next_block==NULL)
		internal_error("out_suspend");

	n_a_locals=next_block->begin_n_a_locals;
	
	adjust_stack_pointers_end();
	prepare_context_switch(n_a_locals);
	
	/*	if polling enabled, this instruction should not be generated */
	
	if (options.poll_eval)
		internal_error("out_suspend (2)");
	
	/* set highest bit in pid of this process */

	o_ldl16(HEAP_START);
	o_ldnl16(ENV_PID_OFFSET);
	o_mint();
	o_or();
	o_ldl16(HEAP_START);
	o_stnl16(ENV_PID_OFFSET);
	
	o_stopp();
}

static VOID out_suspend_uneval(INSTRUCTION *instruction)
{
	INT16 start_label, end_label;
	INT16 param_local,n_a_locals;
	INSTRUCTION_BLOCK *next_block;
	
	next_block=current_block->next;
	if(next_block==NULL)
		internal_error("out_suspend_uneval");

	n_a_locals=next_block->begin_n_a_locals;
	param_local=A_PAR_FIRST_LOCAL+n_a_locals-instruction->parameters[0].int16-1;
	
	adjust_stack_pointers_end();

	/*	if polling enabled, the reducer must not become a waiting one, but must stay
		active, else it will no be rescheduled. */
	
	if (options.poll_eval)
	{		
		/* generate a loop with a context switch */
		
		start_label	=new_label_id();
		end_label	=new_label_id();
		
		prepare_context_switch(n_a_locals);

		o_ldef_label(start_label);

		/* insert low prio context switch at start of loop */

		o_ldlp16(0);
		o_adc16(1);
		o_runp();
		o_stopp();

		/* test if parameter in HNF after context switch */
	
		o_ldl16(param_local);
		o_ldnl16(CODE_OFFSET);
		o_mint();
		o_and();
		o_cj_local(end_label);
		
		/* not evaluated: jump to start of loop */
		
		o_j_local(start_label);
		
		o_ldef_label(end_label);
		
	} else
	{
		end_label=new_label_id ();

		/* test if parameter in HNF */
		
		o_ldl16(param_local);
		o_ldnl16(CODE_OFFSET);
		o_mint();
		o_and();
		o_cj_local(end_label);
			
		/* not evaluated: suspend */
		
		prepare_context_switch(n_a_locals);
		
		/* set highest bit in pid */

		o_ldl16(HEAP_START);
		o_ldnl16(ENV_PID_OFFSET);
		o_mint();
		o_or();
		o_ldl16(HEAP_START);
		o_stnl16(ENV_PID_OFFSET);
	
		o_stopp();
		
		o_ldef_label(end_label);
	}
}

static VOID out_release(INSTRUCTION *instruction)
{
	/* the release instruction works the same as a jsr_eval: it's supposed not to
	   be needed very often. The parameter is the waiting list. There's no
	   need to save an entry on the b-stack for the return address however,
	   as the release routine doesn't require this */

	INT16 end_label;
	
	INT16 n_a_locals, n_b_locals;
	INT16 check_size;
	INT16 param_local;
	
	end_label=new_label_id ();

	check_size		=instruction->parameters[1].int16;
	n_a_locals		=instruction->parameters[2].int16;
	n_b_locals		=instruction->parameters[3].int16;
	
	param_local		=A_PAR_FIRST_LOCAL+n_a_locals-instruction->parameters[0].int16-1;
	
	adjust_stack_pointers_end();

	/* WAITLIST TEST: jump around if no one is waiting */
	
	o_ldl16(param_local);
	o_cj_local(end_label);
	
	/* RELEASE CODE: check if enough heap is available to save locals */
	
	if (options.check_stacks && check_size > 0)
	{	
		INT16 ready_label;
	
		ready_label=new_label_id ();
		
		/* test stack */
		
		o_ldl16(A_STACK_POINTER);
		o_ldnlp16(check_size);
		o_ldl16(B_STACK_POINTER);
		o_gt();
		o_cj_local(ready_label);
		
		/* call reallocation routine */
		
		o_ldc16(n_a_locals);
		o_stl16(EXTRA_LOCAL);
		o_ldc16(check_size);
		
		o_call();
		o_inline_label(realloc_label);
		o_newline();
		
		o_ldef_label(ready_label);
	}
		
	/* save locals and return address */
	
	eval_save_stack(n_a_locals,n_b_locals,0,param_local);
		
	/*	put param in correct local and release the waiting list by evaluating
		the first node */
	
	o_ldl16(param_local);
	
	if (param_local!=A_PAR_FIRST_LOCAL)
	{
		o_dup();
		o_stl16(A_PAR_FIRST_LOCAL);		
	}

	o_ldnl16(CODE_OFFSET);
	o_gcall();
		
	/* waitlist released: restore locals */
	
	if (param_local!=A_PAR_FIRST_LOCAL)
	{
		o_ldl16(A_PAR_FIRST_LOCAL);
		o_stl16(param_local);
	}
	
	eval_restore_stack(n_a_locals,n_b_locals,0,param_local);

	/* NORMAL CODE: output overide label */
	
	o_ldef_label(end_label);	
}

static VOID out_call(INSTRUCTION *instruction)
{
	/* not used at the moment, could be used to call subroutines by calling their
	   jump label (call with wsp+4: return address is automatically stored in local */
	
	error("WARNING: call used\n");
	
	o_call();
	o_jlabel(instruction->parameters[0].label);
	o_newline();
}

static VOID out_libcall(INSTRUCTION *instruction)
{
	/* used for library routines that don't use stack space nor locals
	   just call appropriate routine. context switches are not to be feared */
	
	o_call();
	o_jlabel(instruction->parameters[0].label);
	o_newline();
}

static VOID out_signal(INSTRUCTION *instruction)
{
	/*	the return address may stay whereever it is, no context switching
		possible at this point */

	adjust_stack_pointers_end();
	
	o_ldc16(instruction->parameters[0].int16);
	
	o_call();
	o_inline_label(signal_label);
	o_newline();
	
	/*	the amount of parameters on the b-stack may have been increased/decreased by
		the runtime system */

	current_ctop_offset+=instruction->parameters[1].int16;
}

static VOID out_inline(INSTRUCTION *instruction)
{
	/* used to output a block of inline instructions */

	out_inline_instructions(instruction->parameters[0].inline_block);	
}

static void out_group(INSTRUCTION *instruction)
{
	internal_error("group instructions not alllowed in final code");
}

/********************************************************************
 *																	*
 *		start of block												*
 *																	*
 ********************************************************************/

static VOID out_block_entry(INSTRUCTION_BLOCK *block)
{
	register LABEL *block_label;
	BOOL label_ctop_stack;
	INT16 label_ctop_offset;
	
	block_label=block->start_label;
	
	if (block_label==NULL)
		return;												/* no entry needed */
	
	if (!reachable && !(block_label->flags & USE_LABEL))
		return;												/* entry not reachable */
	
	/* determine offset to ctop */
	
	if (block_label->flags & FIXED_CTOP_OFFSET)
	{
		/*	stack layout of this entry is fixed, if this block is reachable the
			stack layout must be checked against the previous block */
		
		label_ctop_offset=block_label->b.c_offset;
		
		if ((block_label->flags & REACHABLE) && label_ctop_offset != current_ctop_offset)			
			fatal_error("block entry (%s) ambiguous ctop offset",block_label->name);
	
	} else
	{
		/*	Stack layout for this label is not yet known. If this label is
			reachable from the previous block, AND the previous block itself
			is reachable, the label should get the same stack layout as the
			previous block. Else the stack layout of this label is unknown
			and it is considered unreachable. Note that this implies that
			local labels (without stack info) must be placed after some
			label with stack layout, i.e. they must not be encountered first.
		*/
		
		if ((block_label->flags & REACHABLE) && reachable)
		{
			/* label reachable from reachable block */
			
			label_ctop_offset			=current_ctop_offset;
			block_label->b.c_offset		=current_ctop_offset;
			block_label->flags |= FIXED_CTOP_OFFSET;
		
		} else
		{
			/*	label is not reachable from previous block and/or the previous
				block is not reachable itself. */
			
			error("WARNING: label %s assumed unreachable (cannot determine c top offset)\n",block_label->name);
			
			block_label->flags&=~DEF_LABEL;			/* no label definition			*/
			reachable=false;						/* this block is unreachable	*/
			return;
		}
	}
	reachable=true;
	current_ctop_offset=label_ctop_offset;
	
	/* return for readability */

	o_newline();

	/* determine location of ctop and relocate if needed */
	
	label_ctop_stack = is_label_ctop_stack(block_label);
	
	if (block_label->flags & REACHABLE)
	{	if (label_ctop_stack != current_ctop_stack)
		{	if (label_ctop_stack)
			{	put_ctop_stack(current_ctop_offset);
			} else
			{	put_ctop_local(current_ctop_offset);
			}
		}
	}
	current_ctop_stack = label_ctop_stack;
	
	if(block_label->flags & GCALL_LABEL)		
	{
		INT16 node_arity;
		LABEL *descriptor;

		/* code must be inserted to store return address and perhaps this is a node
		   entry */

		descriptor = block->node_descriptor;
		node_arity = block->node_arity;
		
		if (block_label->flags & REACHABLE)
		{
			/* some jumping around is necessary */
			
			block_label->flags |= JUMP_LABEL;
			
			if (options.parallel_code)
			{
				prepare_context_switch(block->begin_n_a_locals);
			
				o_j();
			
			} else
			{
				o_ldc16(0);
				o_cj();
			}
			o_jlabel(block_label);
			o_newline();
		}
		
		if(descriptor)
		{
			/* this is a node entry: put arity and descriptor before code entry */
			
			if (node_arity > MAX_CLOSURE_ARITY)
			{	error("node has to much arguments");
				abort_code_generation();
			}

			/* entry MUST be aligned! */
			
			o_align();
			o_xwords16(0);
			
			/* fill field with marked offset to correct descriptor entry */
			
			o_desc_id(descriptor,node_arity,true);
			
			/* and insert size, entry type, and arity */
			
			if (node_arity <= 2)
			{
				/*	this is a node entry for a small node. For small nodes
					the size is always 3 except for a few special ones
				*/
				
				if (block_label->entry_type & CHANNEL_NODE)
				{
					/* channel nodes or request nodes have special sizes */
					
					if (strcmp(block_label->name,channel_label->name)==0)
						o_ibytes16(3,CHANNEL_NODE_SIZE,block_label->entry_type,node_arity);
					else
						o_ibytes16(3,REQUEST_NODE_SIZE,block_label->entry_type,node_arity);
				} else
					o_ibytes16(3,3,block_label->entry_type,node_arity);
			} else
				o_ibytes16(3,node_arity+1,block_label->entry_type,node_arity);
						
		} else
		{
			/* just align (not demanded) */
			
			o_align();
			o_nop(3);
		}
		
		/* now insert code to save return address */
		
		o_rev();					/* for jmp's to entries */
		o_glabel(block_label);
		o_newline();
		
		if (current_ctop_stack)
		{
			o_ldl16(B_STACK_POINTER);
			o_stnl16(current_ctop_offset);
		
		} else
		{
			o_stl16(RETURN_ADDRESS);
		}
		
		if (block_label->flags & JUMP_LABEL)
		{	o_jlabel(block_label);
			o_newline();
		}
		
	} else
	{	if (block_label->flags & JUMP_LABEL)
		{
			if (!(block_label->flags & REACHABLE))
				o_align();
			
			o_jlabel(block_label);
			o_newline();
		}
	}
}

/************************************************************************************
 *																					*
 *		stack checks																*
 *																					*
 ************************************************************************************/

static VOID out_stack_check(INSTRUCTION_BLOCK *block)
{	
	if (block->check_size > 0)
	{
		INT16 ready_label;
	
		ready_label=new_label_id ();
		
		/*	test stack, note that ldnlp should not result in an overflow: all
			addresses are negative... */
		
		o_ldl16(A_STACK_POINTER);
		o_ldnlp16(block->check_size);
		o_ldl16(B_STACK_POINTER);
		o_gt();
		o_cj_local(ready_label);
				
		/* call reallocation routine */
		
		o_ldc16(block->begin_n_a_locals);
		o_stl16(EXTRA_LOCAL);
		o_ldc16(block->check_size);
		o_call();
		
		o_inline_label(realloc_label);
		o_newline();
		
		o_align();
		o_ldef_label(ready_label);
	}
}

/************************************************************************************
 *																					*
 *		heap allocation																*
 *																					*
 ************************************************************************************/

static VOID out_heap_allocation(INSTRUCTION_BLOCK *block)
{	
	/*	less than 4K can be allocated quite safely: the heap is not in
		internal memory and subtracting less than 4K from the end of the local
		heap cannot result in an overflow. This is not true when more has to
		be allocated: if this is the case the error flag should be checked as
		well, to catch overflows, or one should not use a subtraction... note
		that adding a large value to an address cannot result in an overflow,
		as all valid addresses are (supposed to be) negative.
	*/
	
	if (block->heap_use > 0)
	{
		if (block->heap_use < (4*KILOBYTE)/WORD_SIZE)
		{
			INT16 ready_label;
	
			ready_label	=new_label_id ();
			
			o_ldl16(HEAP_START);
			o_ldl16(HEAP_END);					/* get current end of heap */
			o_ldnlp16(-block->heap_use);		/* subtract amount needed */
			o_dup();
			o_stl16(HEAP_END);					/* update current end of heap */
			o_gt();								/* test if start > end */
			o_cj_local(ready_label);			/* no: start <= end, so everything is allright */
			
			/* call allocate routine to get proper amount of heap */
					
			o_ldc16(block->begin_n_a_locals);
			o_stl16(EXTRA_LOCAL);
			o_ldc16(block->heap_use);
			o_call();
			
			o_inline_label(extend_label);
			o_newline();
					
			o_align();
			o_ldef_label(ready_label);
		} else
		{
			INT16 ready_label, error_label;
			
			/*	a large amount of memory will probably not fit in the local heap
				anyway, so one could consider to call the heap extension routine
				directly to get the right amount of memory.
			*/
			
			ready_label	=new_label_id();
			error_label	=new_label_id();
			
			o_ldl16(HEAP_END);
			o_ldl16(HEAP_START);
			o_ldnlp16(block->heap_use-1);
			o_gt();								/* new start-1 < end ? */
			o_cj_local(error_label);			/* new start-1 >= end, extend heap */
			
			/*	enough local heap available, allocate it (subtracting cannot result
				overflow here */
			
			o_ldl16(HEAP_END);
			o_ldnlp16(-block->heap_use);
			o_stl16(HEAP_END);
			
			/*	jump around heap extension, use a conditional jump to avoid
				time slice here: you don't want the garbage collector to
				throw away the allocated heap! */
			
			o_ldc16(0);
			o_cj_local(ready_label);
			
			o_ldef_label(error_label);
						
			/* call allocate routine to get proper amount of heap */
					
			o_ldc16(block->begin_n_a_locals);
			o_stl16(EXTRA_LOCAL);
			o_ldc16(block->heap_use);
			o_call();
			
			o_inline_label(extend_label);
			o_newline();
					
			o_align();
			o_ldef_label(ready_label);
		}
	}
}

/************************************************************************************
 *																					*
 *		data and symbols															*
 *																					*
 ************************************************************************************/

static VOID out_symbols(DSYMBOL *symbol,INT16 start_label)
{
	register LABEL *label;

	for(;symbol!=NULL;symbol=symbol->next)
	{		
		label = symbol->label;
		
		if ((label->flags & GCALL_LABEL) && (label->flags & DEF_LABEL))
		{
			o_ldc();
			o_glabel(label);
			o_char('-');
			o_luse_label(start_label);
			o_newline();
			
			o_ibytes16(0);
			o_string(label->name);
			o_newline();					
		}
		
		if ((label->flags & JUMP_LABEL) && (label->flags & DEF_LABEL))
		{
			o_ldc();
			o_jlabel(label);
			o_char('-');
			o_luse_label(start_label);
			o_newline();
			
			o_ibytes16(0);
			o_string(label->name);
			o_newline();
		}
	}
}

static VOID out_desc_entries (register DDESCRIPTOR *descriptor)
{	
	register INT16 n,arity;
	
	arity=descriptor->d.arity;
	
	for (n=0; n<=arity; ++n)
	{
		fprintf(codefile,"%s%d,\t",dsymbol[WORD_DIR],(int)n);
		
		if (n < arity-1)
		{	if (n > MAX_YET_ARGS_NEEDED_ARITY)
				o_glabel(init_yet_args_needed_label);
			else
				o_glabel(init_yet_args_needed_labels[n]);
		} else
		{	if (n==arity-1)
				o_glabel(descriptor->d.lazy_entry_label);
			else
				o_glabel(descriptor->d.node_entry_label);
		}
		
		/*	Get offset relative to start of entry. Subtract (descriptor label
			+ (node arity) * (word length)). For some assemblers it may be sufficient
			to add (word length).
		*/
		
		switch (options.assembler)
		{
			case PACT_ASSEMBLER:
			case PARC_ASSEMBLER:
			{
				o_char('-');
				o_dlabel(descriptor->label);
				
				if(n>0)
					fprintf(codefile,"-%d\n",(int)n<<3);
				else
					o_newline();
				break;
			}
			/*
			{
				fprintf(codefile,"+%d\n",(int)WORD_SIZE);
				break;
			}*/

			default:
				internal_error("out_desc_entries");
		}
	}
}

static VOID out_descriptors (DDESCRIPTOR *descriptor)
{
	for (;descriptor!=NULL; descriptor=descriptor->next)
	{
		o_newline();
		o_align();
		
		/*	-descriptor arity
			-descriptor label
			-descriptor arity times:
				-node arity
				-offset to lazy entry
			-string label (optional)
			-string length
			-string
		*/
				
		o_iwords16((INT16)1,(INT16)descriptor->d.arity);
		
		o_dlabel(descriptor->label);
		o_newline();
		
		out_desc_entries(descriptor);
		
		if (descriptor->string_label->flags & USE_LABEL)
		{	o_dlabel(descriptor->string_label);
			o_newline();
		}
		
		o_iwords16((INT16)1,(INT16)descriptor->d.string_length);
		o_ibytes16(0);				
		o_string(descriptor->d.string);
		o_newline();
	}
}

static VOID out_strings (register DSTRING *string)
{
	for (;string!=NULL; string=string->right)
	{
		o_align();
		
		o_dlabel(string->label);
		o_newline();
		
		o_iwords16(1,string->s.length);
		o_ibytes16(0);
		o_string(string->s.string);
		o_newline();
	}
}

static VOID out_floats(register DFLOAT *f)
{
	for(;f!=NULL;f=f->right)
	{
		o_align();
		
		o_dlabel(f->label);
		o_newline();				
		o_float(&f->f);
	}
}

static VOID out_code_data(INSTRUCTION_BLOCK *block)
{
	BLOCK_DATA *data;
	
	data=block->data;
	
	if (data==NULL)
		return;
	
	if (options.output_debug_symbols)
	{
		INT16 symbol_label,data_label,end_label;
		
		symbol_label	=new_label_id();
		data_label		=new_label_id();
		end_label		=new_label_id();

		/* put mark in code and label at start of symbols */
		
		o_ldnlp16(0);
		o_ldef_label(symbol_label);

		out_symbols(data->first_symbol,symbol_label);
		
		/* put offset to end in code */
		
		o_ldc();
		o_luse_label(end_label);
		o_char('-');
		o_luse_label(data_label);
		o_newline();
		
		o_ldef_label(data_label);
		
		out_strings	(data->first_string);
		out_floats	(data->first_float);
		
		o_ldef_label(end_label);
		o_ldnlp16(0);
		o_newline();
				
	} else
	{
		out_strings	(data->first_string);
		out_floats	(data->first_float);
	}
}

static VOID data_area(VOID)
{
	if (first_uncoded_descriptor==NULL)
		return;
	
	fputs("\n; data area\n\n",codefile);
	
	out_descriptors(first_uncoded_descriptor);
}

/************************************************************************************
 *																					*
 *		imports and exports															*
 *																					*
 ************************************************************************************/

VOID import_inline_labels (register LABEL_NODE *label)
{
	fputs("\n; import inline labels\n\n",codefile);
	
	for(;label!=NULL;label=label->right)
	{
		fputs(dsymbol[EXTERN_DIR],codefile);
		o_label(&label->label);
		o_newline();
	}
}

VOID import_data_labels (register LABEL_NODE *label)
{
	fputs("\n; import data\n\n",codefile);
	
	for(;label!=NULL;label=label->right)
	{
		fputs(dsymbol[EXTERN_DIR],codefile);
		o_dlabel(&label->label);
		o_newline();
	}
}

VOID import_code_labels (register LABEL_NODE *label)
{
	fputs("\n; import code\n\n",codefile);
	
	for(;label!=NULL;label=label->right)
	{
		if (label->label.flags & GCALL_LABEL)
		{
			fputs(dsymbol[EXTERN_DIR],codefile);
			o_glabel(&label->label);
			o_newline();
		}
		if (label->label.flags & JUMP_LABEL)
		{
			fputs(dsymbol[EXTERN_DIR],codefile);
			o_jlabel(&label->label);
			o_newline();
		}
	}
}

VOID export_inline_labels (register LABEL_NODE *label)
{
	fputs("\n; export inline labels\n\n",codefile);
	
	for(;label!=NULL;label=label->right)
	{
		fputs(dsymbol[GLOBAL_DIR],codefile);
		o_label(&label->label);
		o_newline();
	}
}

VOID export_data_labels (register LABEL_NODE *label)
{
	fputs("\n; export data\n\n",codefile);
	
	for(;label!=NULL;label=label->right)
	{
		fputs(dsymbol[GLOBAL_DIR],codefile);
		o_dlabel(&label->label);
		o_newline();
	}
}

VOID export_code_labels (register LABEL_NODE *label)
{
	fputs("\n; export code\n\n",codefile);
	
	for(;label!=NULL;label=label->right)
	{	if (label->label.flags & GCALL_LABEL)
		{
			fputs(dsymbol[GLOBAL_DIR],codefile);
			o_glabel(&label->label);
			o_newline();
		}
		if (label->label.flags & JUMP_LABEL)
		{
			fputs(dsymbol[GLOBAL_DIR],codefile);
			o_jlabel(&label->label);
			o_newline();
		}
	}
}

/************************************************************************************
 *																					*
 *		runtime linking of external descriptor entries								*
 *																					*
 ************************************************************************************/

static VOID desc_init_code(VOID)
{
	INT16 n;
	
	/* at the moment this is only needed for Par.C assembly files */
	
	switch (options.assembler)
	{
		case PARC_ASSEMBLER:
	
			error("runtime link routines use old descriptor representation");
			
			/* output link routine for nodes with arity > 2 */
			
			if (yet_args_needed_label != NULL)
			{
				fputs("\n; runtime link routine for lazy entries with arity > 2\n\n",codefile);
				
				/* The lazy entry is called with 'x' in local 5 and 'f+arguments' in local 6.
				   If this routine is called not enough parameters are available. The lazy
				   entry is, however an external routine and the ParC linker cannot fill in
				   this address in the descriptor of the node (for some obscure reason). The
				   code below performs some runtime linking: the first time the lazy entry
				   is called the correct address will be filled in the descriptor entry. Be
				   carefull not to loose the return address when changing the code below */
				
				o_rev();
				o_glabel(init_yet_args_needed_label);
				o_newline();
				
				/* get descriptor address from local 6 and use local 7 to store it */
				
				o_ldc16(0);
				o_ldl16(A2);
				o_ldnl16(DESCRIPTOR_OFFSET);
				o_diff();
				o_stl16(A3);
						
				o_address(yet_args_needed_label, 0);
				
				/* put offset to real entry in decriptor */
				
				o_ldl16(A3);
				o_diff();
				o_ldl16(A3);
				o_stnl16(1);
							
				/* and use cond jump to goto right entry (-1 to get return address right) */
				
				o_ldc16(0);
				o_cj();
				o_glabel(yet_args_needed_label);
				fputs("-1\n",codefile);
			}
			
			/* output link routines for curried nodes with arity <= 2 */
			
			for (n=0; n <= MAX_YET_ARGS_NEEDED_ARITY; n++)
			{	if (yet_args_needed_labels[n] != NULL)
				{
					fprintf(codefile,"\n; runtime link routine for lazy entries with arity %d\n\n",(int)n);
		
					o_rev();
					o_glabel(init_yet_args_needed_labels[n]);
					o_newline();
					
					/* get descriptor address from local 6 and use local 7 to store it */
				
					o_ldc16(0);
					o_ldl16(A2);
					o_ldnl16(DESCRIPTOR_OFFSET);
					o_diff();
					o_stl16(A3);
		
					o_address(yet_args_needed_labels[n], 0);
					
					/* put offset to real entry in decriptor */
				
					o_ldl16(A3);
					o_diff();
					o_ldl16(A3);
					o_stnl16(1);
					
					/* and use cond jump to goto right entry (-1 to get return address right) */
				
					o_ldc16(0);
					o_cj();
					o_glabel(yet_args_needed_labels[n]);
					fputs("-1\n",codefile);
				}
			}
			
			/* output link routine for hnf entry */
			
			if (init_hnf_label != NULL)
			{
				fputs("\n; runtime link routine for hnf entry\n\n",codefile);
				
				/* in fact this routine will not be called. It's here for completeness, and
				   it doesn't link to the actual routine, but just implements the _hnf
				   code, which is simply a return instruction */
				
				o_rev();
				o_glabel(init_hnf_label);
				o_newline();
				
				o_gcall();		
			}
			break;
		
		case PACT_ASSEMBLER:
			break;
		
		default:
			internal_error("output_desc_init_code: unknown assembler type");
	}
}

/************************************************************************************
 *																					*
 *		output of alternative entries for closures									*
 *																					*
 ************************************************************************************/

static VOID output_alt_entry(ENTRY_LABEL *entry_label)
{
	/* just output node entries as specified, the offset to the ctop is not
	   important as this will not be needed in these entries. reachability is
	   of no concern and no jumping around is required, as these entries will
	   be placed a non-reachable area.
	*/
		
	if (entry_label->left)
		output_alt_entry(entry_label->left);
	if (entry_label->right)
		output_alt_entry(entry_label->right);

	if (entry_label->arity > MAX_CLOSURE_ARITY)
	{
		error("alternative entry has too much arguments");
		abort_code_generation();
	}
	
	/* entry MUST be aligned! */
	
	o_newline();
	o_align();

	if (options.assembler == PACT_ASSEMBLER)
	{
		/* pact is able to handle external refs, so write descriptor... */
		
		o_xwords16(0);
		
		/* fill field with marked offset (from node entry) to correct descriptor entry */
		
		if (entry_label->descriptor==NULL)
			internal_error("desc of alt entry is NULL");
		
		o_desc_id(entry_label->descriptor,entry_label->arity,true);
	}
		
	/* insert node type and arity. This should be available for all entries */
	
	o_ibytes16(3,entry_label->size,entry_label->label.entry_type,entry_label->arity);	
	o_rev();
	
	o_glabel(&entry_label->label);
	o_newline();
	
	/* and write code to jump to correct entry in an undivisable way */
	
	o_ldc16(0);
	o_cj();
	o_glabel(entry_label->entry);
	fputs("-1\n",codefile);
}

VOID output_alt_entries(ENTRY_LABEL *entry_label, CHAR *comment)
{
	if (entry_label)
	{	fprintf(codefile,"\n;alternative node entries (%s)\n",comment);
	
		output_alt_entry(entry_label);
	}
}

/************************************************************************************
 *																					*
 *		output of prelude															*
 *																					*
 ************************************************************************************/

static VOID prelude (VOID)
{
	switch (options.assembler)
	{
		case PACT_ASSEMBLER:
		{
			/*	this part of code is called by the runtime system with PACT
				calling conventions. At least, this part of code should tell
				the runtime system about the start address. Note that this
				routine should not use any stack, as this could break the
				runtime system when it calls this routine. It may overwrite
				the hidden arguments, as these are not needed.
			*/

			fputs("\tglobal\tprelude\n",codefile);
			
			o_align();
			fputs("prelude\n",codefile);				/* label to be used by runtime sys */
			
			o_stl16(PRELUDE_RETURN);					/* store return address */
			
			/* pass ptr to desc area */
			
			o_address(nil_label,0);
			o_mint();
			o_stnl16(HARD_DESCRIPTORS);
			
			/* important code addresses */
			
			o_mint();
			o_ldnl16(HARD_ROOT);						/* get ptr to root */
			o_ldnl16(CODE_INFO);						/* get ptr to code info structure */
			o_stl16(PRELUDE_SCRATCH1);					/* store in temp */
			
			o_address(start_label,0);					/* address of start code */
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(START_CODE_OFFSET);				/* store in start field */
			
			o_address(reserve_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(RESERVE_CODE_OFFSET);
			o_address(reserve1_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(RESERVE1_CODE_OFFSET);
			
			o_address(request_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(REQUEST_CODE_OFFSET);
			o_address(request1_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(REQUEST1_CODE_OFFSET);
			
			o_address(release_proc_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(RELEASE_PROC_OFFSET);
			o_address(release_proc1_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(RELEASE_PROC1_OFFSET);
			
			o_address(release_request_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(RELEASE_REQUEST_OFFSET);
			o_address(release_request1_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(RELEASE_REQUEST1_OFFSET);

			o_address(channel_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(CHANNEL_CODE_OFFSET);

			o_address(ea_process_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(EA_PROCESS_CODE_OFFSET);
			
			o_address(ea_reduce_label,0);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(EA_REDUCE_CODE_OFFSET);
									
			/* important descriptors */
			
			o_mint();
			o_ldnl16(HARD_ROOT);						/* get ptr to root */
			o_ldnl16(DESC_INFO);						/* get ptr to desc info structure */
			o_stl16(PRELUDE_SCRATCH1);					/* store in temp */

			o_ldc();
			o_desc_id(nil_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(NIL_OFFSET);
			o_ldc();
			o_desc_id(cons_label,2,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(CONS_OFFSET);
			
			o_ldc();
			o_desc_id(tuple_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(TUPLE_OFFSET);

			o_ldc();
			o_desc_id(INT_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(INT_OFFSET);
			o_ldc();
			o_desc_id(BOOL_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(BOOL_OFFSET);
			o_ldc();
			o_desc_id(CHAR_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(CHAR_OFFSET);
			o_ldc();
			o_desc_id(REAL_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(REAL_OFFSET);
			o_ldc();
			o_desc_id(FILE_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(FILE_OFFSET);
			o_ldc();
			o_desc_id(STRING1_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(STRING1_OFFSET);
			o_ldc();
			o_desc_id(STRING2_label,0,false);
			o_ldl16(PRELUDE_SCRATCH1);
			o_stnl16(STRING2_OFFSET);
			
			o_ret();						/* and return */
			
			break;
		}		
		case PARC_ASSEMBLER:
			error("out_prelude: Par.C assembler prelude not supported\n");
			break;
			
		default:
			internal_error("out_prelude: unknown assembler type");
	}
}

/************************************************************************************
 *																					*
 *		main output routine															*
 *																					*
 ************************************************************************************/

typedef VOID (*out_instr_function) (INSTRUCTION *instruction);

static out_instr_function out_instr_table[] =
{
	/* general instructions */
	
	out_ldl,
	out_stl,
	out_ldnl,
	out_stnl,
	out_lb,
	out_ldc,
	out_adc,
	out_eqc,
	out_eqd,
	out_ldlp,
	out_ldnlp,
	out_ajw,
	out_j,
	out_cj,	
	out_ret,
	out_lea,
	out_ldesc,
	out_ldpi,
	out_gt,
	out_rev,
	out_dup,
	
	out_add,
	out_sub,
	out_mul,
	out_div,
	out_rem,
	out_sum,
	out_diff,
	out_prod,
	out_and,
	out_or,
	out_xor,
	out_shl,
	out_shr,
	out_not,
	out_abs,
		
	out_lshl,
	out_lshr,
	out_xdble,
	out_csngl,
	
	out_bsub,
	out_wsub,
	
	out_move,
	out_bcnt,
	out_wcnt,
	out_mint,
	
	/* floating point instructions */
	
	out_fpdup,
	out_fprev,
	out_fpldnldb,
	out_fpstnldb,
	out_fpadd,
	out_fpsub,
	out_fpmul,
	out_fpdiv,
	out_fprem,
	out_fpgt,
	out_fpeq,
	
	out_fpldzerodb,
	
	out_fpremfirst,
	out_fpremstep,
	out_fpusqrtfirst,
	out_fpusqrtstep,
	out_fpusqrtlast,
	
	out_fpumulby2,
	out_fpudivby2,
	out_fpuexpinc32,
	out_fpuexpdec32,
	
	out_fpint,
	out_fpi32tor64,
	out_fpstnli32,
	out_fpurm,
	
	/* flow of control */
	
	out_rtn,
	out_jmp_ap,
	out_jmp_eval,
	
	out_stopp,
	out_runp,
	out_call,
	out_gajw,

	out_libcall,
	out_inline,
	out_group,

	out_jsr_eval,
	out_release,
	
	out_suspend_uneval,
	out_suspend,
	
	out_cs,
	out_enable_cs,
	out_disable_cs,
	
	out_jsr,
	out_gcall,
	out_jsr_ap,
	out_signal,
	
	out_error,
	out_error
};

static VOID out_switch (BOOL b)
{
	if (b)
		fputs("on\n",codefile);
	else
		fputs("off\n",codefile);
}

VOID out_code (CHAR *filename,INSTRUCTION_BLOCK *first_block)
{
	if ((codefile = open_assembly_file (filename))!=NULL)
	{
		set_file_stage("writing \'%s\'",filename);

		block_number=0;
		inline_instruction_number=0;

		current_ctop_stack=true;
		current_ctop_offset=0;
				
		switch (options.assembler)
		{	case PARC_ASSEMBLER:
				fputs("; Par.C assembly file\n",codefile);
				
				break;
			case PACT_ASSEMBLER:
				fputs("; PACT assembly file\n",codefile);
				
				break;
			default:
				internal_error("out_code: unknown assembler type");
		}

				
		if (options.parallel_code)
			fputs("; parallel code ",codefile);
		else
			fputs("; sequential code ",codefile);
		
		fputs("generated by version 0.8 of transputer code generator\n\n",codefile);
		
		
		if (options.output_settings)
		{
			fputs("; compiler settings:\n\n",codefile);
			
			fputs("; parallel_code:         ",codefile); out_switch(options.parallel_code);
			fputs("; inline_instructions:   ",codefile); out_switch(options.inline_instructions);
			fputs("; implicit_imports:      ",codefile); out_switch(options.implicit_imports);
			fputs("; ignore_depend:         ",codefile); out_switch(options.ignore_depend);
			fputs("; ignore_defer:          ",codefile), out_switch(options.ignore_defer);
		
			fputs("; full_label_names:      ",codefile); out_switch(options.output_full_label_names);
			fputs("; debug_symbols:         ",codefile); out_switch(options.output_debug_symbols);
			fputs("; default_attributes:    ",codefile); out_switch(options.output_default_attributes);
			
			fputs("; optimise_arith:        ",codefile); out_switch(options.optimise_arith);
			fputs("; optimise_jumps:        ",codefile); out_switch(options.optimise_jumps);
			fputs("; optimise_a_stack:      ",codefile); out_switch(options.optimise_a_stack);
			fputs("; optimise_b_stack:      ",codefile); out_switch(options.optimise_b_stack);
			fputs("; optimise_a_parameter:  ",codefile); out_switch(options.optimise_a_parameter);
			fputs("; optimise_stack_checks: ",codefile); out_switch(options.optimise_stack_checks);
		
			fputs("; any_proc_eval:         ",codefile); out_switch(options.any_proc_eval);
			fputs("; poll_eval:             ",codefile); out_switch(options.poll_eval);
	
			fputs("; check_stacks:          ",codefile); out_switch(options.check_stacks);
			fputs("; check_arithmetic:      ",codefile); out_switch(options.check_arithmetic);
		}

		if (options.assembler==PACT_ASSEMBLER && options.output_default_attributes)
		{
			o_newline();
			
			fputs("\tattrib\terrmode=dont_care\n",codefile);
			fputs("\tattrib\tiomode=instr\n",codefile);
			fputs("\tattrib\tfpmode=non_fpu_calling\n",codefile);
			fputs("\tattrib\tmemstart=dont_care\n",codefile);
			fputs("\tattrib\twordsize=4\n",codefile);
	
		}

		port_labels();					/* imports and exports */

		if (start_label != NULL && start_label->flags & IMPORT_LABEL)
		{
			fputs("\n; prelude (system file only)\n",codefile);
			
			prelude();
		}

		desc_init_code();				/* runtime linking of lazy entries (Par.C only) */
		data_area();

		define_entry_labels();			/* extra entries for defered (etc.) nodes */
						
		fputs("\n\n; start of code\n\n",codefile);
				
		reachable=false;
		for(current_block=first_block;current_block!=NULL;current_block=current_block->next,++block_number)
		{
			register INSTRUCTION *current_instruction;
		
			handle_write_code();
						
			out_block_entry(current_block);
			
			if(reachable)
			{
				/*	Check and reallocate stack if necessary. Do this
					before allocating heap space: if a stack reallocation
					requires garbage collection, any reserved heap may
					get destroyed!!
				*/
				
				if (options.check_stacks)
					out_stack_check(current_block);
				
				out_heap_allocation(current_block);
								
				adjust_stack_pointers_begin();
	
				for(current_instruction=current_block->instructions;
					current_instruction!=NULL;
					current_instruction=current_instruction->next)
				{
					(*out_instr_table[current_instruction->code])(current_instruction);
				}
	
				adjust_stack_pointers_end();
				
			}else
			{
				/* scan for inline instructions and output only these */
				
				for(current_instruction=current_block->instructions;
					current_instruction!=NULL;
					current_instruction=current_instruction->next)
				{
					if (current_instruction->code==INLINE)
					{
						reachable=true;
						(*out_instr_table[INLINE])(current_instruction);
					}
				}
			}
						
			out_code_data(current_block);	
		}
		
		fputs("\n; end of code\n",codefile);

		first_block=NULL;
		
		close_output_file();

	}else
		fatal_error("output file \'%s\' couldn't be opened\n",filename);	
}

void get_code_position(long *position,long *max)
{
	*position	=block_number+inline_instruction_number;
	*max		=number_of_blocks+number_of_inline_instructions;
}
