#include "ifinterface.h"

#include "cgoutputinter.h"
#include "cgwrites.h"

static INT16	block_number;

/********************************************************************************
 *																				*
 *		output non-standard intermediate instructions							*
 *																				*
 ********************************************************************************/

static VOID out_eqd(INSTRUCTION *instruction)
{
	fputs("\teqd\t",codefile);
	o_label_name(instruction->parameters[0].label);
	o_newline();
}

#if _COMPILE_UNUSED_

static VOID out_minus(INSTRUCTION *instruction)
{
	fputs("\tminus\n",codefile);
}

#endif

static VOID out_abs(INSTRUCTION *instruction)
{
	fputs("\tabs\n",codefile);
}

static VOID out_lea(INSTRUCTION *instruction)
{
	INT16 offset;
	
	fputs("\tlea\t",codefile);
	o_label_name(instruction->parameters[0].label);
	
	offset=instruction->parameters[1].int16;
	if (offset != 0)
		fprintf(codefile,"+%d\n",offset);
	else
		o_newline();
}

static VOID out_ldesc(INSTRUCTION *instruction)
{
	fputs("\tldesc\t",codefile);
		o_label_name(instruction->parameters[0].label);
	o_newline();
}

static VOID out_fprem(INSTRUCTION *instruction)
{
	fputs("\tfprem\n",codefile);
}

static VOID out_j(INSTRUCTION *instruction)
{
	o_j();
	o_label_name(instruction->parameters[0].label);
	o_newline();
}

static VOID out_cj(INSTRUCTION *instruction)
{
	o_cj();
	o_label_name(instruction->parameters[0].label);
	o_newline();
}

static VOID out_gcall(INSTRUCTION *instruction)
{
	fputs("\tlea\t",codefile);
	o_label_name(instruction->parameters[0].label);
	o_newline();
	o_gcall();
}

static VOID out_jsr(INSTRUCTION *instruction)
{
	fputs("\tjsr\t",codefile);
	o_label_name(instruction->parameters[0].label);
	o_newline();
}

static VOID out_rtn(INSTRUCTION *instruction)
{
	fputs("\trtn\n",codefile);
}

static VOID out_jsr_ap(INSTRUCTION *instruction)
{
	fputs("\tjsr_ap\n",codefile);
}

static VOID out_jmp_ap(INSTRUCTION *instruction)
{
	fputs("\tjmp_ap\n",codefile);
}

static VOID out_jsr_eval(INSTRUCTION *instruction)
{
	fputs("\tjsr_eval\n",codefile);
}


static VOID out_jmp_eval(INSTRUCTION *instruction)
{
	fputs("\tjmp_eval\n",codefile);
}

static VOID out_cs(INSTRUCTION *instruction)
{
	fputs("\tcs\n",codefile);
}

static VOID out_enable_cs(INSTRUCTION *instruction)
{
	fputs("\tecs\n",codefile);
}

static VOID out_disable_cs(INSTRUCTION *instruction)
{
	fputs("\tdcs\n",codefile);
}

static VOID out_suspend(INSTRUCTION *instruction)
{
	fputs("\tsuspend\n",codefile);
}

static VOID out_suspend_uneval(INSTRUCTION *instruction)
{
	fputs("\tsuspend_uneval\n",codefile);
}

static VOID out_release(INSTRUCTION *instruction)
{
	fputs("\trelease\n",codefile);
}

static VOID out_call(INSTRUCTION *instruction)
{
	/* not used at the moment, could be used to call subroutines by calling their
	   jump label (call with wsp+4: return address is automatically stored in local */
	
	error("WARNING: call used\n");
}

static VOID out_libcall(INSTRUCTION *instruction)
{
	fputs("\tlib_call\t",codefile);
	o_label_name(instruction->parameters[0].label);
	o_newline();
}

static VOID out_signal(INSTRUCTION *instruction)
{
	o_signal16(instruction->parameters[0].int16);
}

static VOID out_inline(INSTRUCTION *instruction)
{
	fputs("\n\t...inline instructions...\n\n",codefile);
}

static void out_group(INSTRUCTION *instruction)
{
	INSTRUCTION_GROUP *group;
	
	group=(INSTRUCTION_GROUP *)instruction;
	
	o_group16(group->free);
}

/********************************************************************************
 *																				*
 *		data																	*
 *																				*
 ********************************************************************************/

#if _COMPILE_UNUSED

static VOID out_data(VOID)
{
}

#endif

/************************************************************************************
 *																					*
 *		main output routine															*
 *																					*
 ************************************************************************************/

typedef VOID (*out_instr_function) (INSTRUCTION *instruction);

static out_instr_function out_instr_table[] =
{
	/* general instructions */
	
	out_ldl,
	out_stl,
	out_ldnl,
	out_stnl,
	out_lb,
	out_ldc,
	out_adc,
	out_eqc,
	out_eqd,
	out_ldlp,
	out_ldnlp,
	out_ajw,
	out_j,
	out_cj,	
	out_ret,
	out_lea,
	out_ldesc,
	out_ldpi,
	out_gt,
	out_rev,
	out_dup,
	
	out_add,
	out_sub,
	out_mul,
	out_div,
	out_rem,
	out_sum,
	out_diff,
	out_prod,
	out_and,
	out_or,
	out_xor,
	out_shl,
	out_shr,
	out_not,
	out_abs,
		
	out_lshl,
	out_lshr,
	out_xdble,
	out_csngl,
	
	out_bsub,
	out_wsub,
	
	out_move,
	out_bcnt,
	out_wcnt,
	out_mint,
	
	/* floating point instructions */
	
	out_fpdup,
	out_fprev,
	out_fpldnldb,
	out_fpstnldb,
	out_fpadd,
	out_fpsub,
	out_fpmul,
	out_fpdiv,
	out_fprem,
	out_fpgt,
	out_fpeq,
	
	out_fpldzerodb,
	
	out_fpremfirst,
	out_fpremstep,
	out_fpusqrtfirst,
	out_fpusqrtstep,
	out_fpusqrtlast,
	
	out_fpumulby2,
	out_fpudivby2,
	out_fpuexpinc32,
	out_fpuexpdec32,
	
	out_fpint,
	out_fpi32tor64,
	out_fpstnli32,
	out_fpurm,
	
	/* flow of control */
	
	out_rtn,
	out_jmp_ap,
	out_jmp_eval,
	
	out_stopp,
	out_runp,
	out_call,
	out_gajw,
	
	out_libcall,
	out_inline,
	out_group,
	
	out_jsr_eval,
	out_release,
	
	out_suspend_uneval,
	out_suspend,
	
	out_cs,
	out_enable_cs,
	out_disable_cs,
	
	out_jsr,
	out_gcall,
	out_jsr_ap,
	out_signal,
	
	out_error,
	out_error,
};

VOID init_write_intermediate(VOID)
{
	block_number=0;
}

VOID write_intermediate_block(INSTRUCTION_BLOCK *block)
{
	register INSTRUCTION *current_instruction;
	
	fprintf(codefile,"\n;block %d\n\n",block_number++);

	if (block->start_label!=NULL)
	{
		o_label_name(block->start_label);
		o_newline();
	}
		
	for(current_instruction=block->instructions;
		current_instruction!=NULL;
		current_instruction=current_instruction->next)
	{
		(*out_instr_table[current_instruction->code])(current_instruction);
	}
}

VOID write_intermediate_code(INSTRUCTION_BLOCK *code)
{
	INSTRUCTION_BLOCK *current_block;
	
	init_write_intermediate();
	
	for(current_block=code; current_block != NULL; current_block=current_block->next)
		write_intermediate_block(current_block);
}
