#ifndef _SCANNER_
#define _SCANNER_

#ifndef _TYPES_
#include "cctypes.h"
#endif

typedef enum
{
	END_OF_FILE			= EOF,
	PROCEED				= 0,
	END_OF_INLINE		= 1,
	END_OF_MACRO		= 2
	
} FILE_STATUS;

typedef struct FILE_DESCRIPTOR
{
	struct FILE_DESCRIPTOR	*prev;
	
	FILE					*file;
	CHAR					*filename;
	U_INT16					line_nr;
	INT16					curchar;
	
} FILE_DESCRIPTOR;

VOID scan_error					(CHAR *str, CHAR *mess, ...);
VOID scan_warning				(CHAR *str, CHAR *mess, ...);

/* skipping */

VOID	SkipToChar				(CHAR c);
VOID	skip_whites				(VOID);
INT16	SkipRest				(VOID);
INT16	SkipComment				(VOID);
INT16	skip_assembler_comment	(VOID);
INT16	skip_assembler			(VOID);

/* tests */

BOOL	is_directive					(VOID);
BOOL	is_char							(VOID);
BOOL	is_string						(VOID);
BOOL	is_number						(VOID);

/* basic terms */

BOOL	scan_ident				(STRING *s);
BOOL	scan_assembler_ident 	(STRING *s);

BOOL	scan_number32			(INT32 *value);
BOOL	scan_unumber32			(U_INT32 *value);
BOOL	scan_number16			(INT16 *value);
BOOL	scan_unumber16			(U_INT16 *value);

BOOL	scan_real				(FLOAT *value);
BOOL	scan_string				(STRING *string);
BOOL	scan_char				(INT32 *c);
BOOL	scan_bool				(INT32 *b);

BOOL	ScanBVector				(INT16 *size, register U_INT32 *vector);
BOOL	scan_parameters			(INT16 *n_a_params, INT16 *n_b_params, U_INT32 *vector, STRING *params);

/* constructors and operators */

BOOL	scan_inline_start				(VOID);

BOOL	scan_monadic_assembler_operator	(INT16 *operator);
BOOL	scan_prio2_assembler_operator	(INT16 *operator);
BOOL	scan_prio3_assembler_operator	(INT16 *operator);
BOOL	scan_prio4_assembler_operator	(INT16 *operator);

BOOL	scan_assembler_separator		(CHAR *separator);
BOOL	scan_equal						(CHAR *equal);
BOOL	scan_colon						(CHAR *colon);
BOOL	scan_open						(CHAR *open);
BOOL	scan_close						(CHAR *close);
BOOL	scan_begin_macro				(CHAR *open);
BOOL	scan_end_macro					(CHAR *close);

/* set scanning for certain file */

FILE *open_source_file			(CHAR *filename);
FILE *open_depend_file			(STRING *filename);
FILE *open_assembly_file		(char *name);
FILE *open_intermediate_file	(char *name,char *extension);

void get_source_position(long *position,long *max);
char *get_curr_file_name(short *main);

VOID close_input_file(VOID);
VOID close_output_file(VOID);
VOID close_all_files(VOID);

/* initialisation */

VOID initialise_scanner(VOID);

#endif