#include "ifinterface.h"

#include "cgstacks.h"
#include "cggraphs.h"
#include "cgmemory.h"
#include "cglin.h"
#include "cgcalc.h"
#include "cglabels.h"
#include "cgscanner.h"

#include "ccmacros.h"
#include "ccnodes.h"
#include "ccintermediate.h"

#define	ELEMENT_MAY_BE_REMOVED	1	/* both load graph and result graph are unused
									   this is a dummy stack element */
#define	ELEMENT_USED_BEFORE_JSR	2

extern BOOL code_reachable;

/* global variables for block graphs */

static BLOCK_GRAPH *first_block_graph,*last_block_graph;

/* exported standard vectors */

U_INT32 e_vector[] = { 0 };
U_INT32 i_vector[] = { 0 };
U_INT32 i_i_vector[] = { 0 };
U_INT32 i_i_i_vector[]= { 0 };

/****************************************************************************************
 *                                                                                      *
 *									stack simulation									*
 *																						*
 ****************************************************************************************/

INSTRUCTION_GRAPH s_get_a (BLOCK_GRAPH *block_graph, INT16 offset)
{
	register A_STACK **element_p,*new_element;
	register INT16 required_offset;
	INSTRUCTION_GRAPH graph;
	
	required_offset=block_graph->a_stack_top_offset+offset;
	
	element_p=&block_graph->a_stack;
	while (*element_p!=NULL)
	{	register INT16 element_offset=(*element_p)->offset;
		
		if (element_offset==required_offset)
			return (*element_p)->graph;
			
		if (element_offset>required_offset)
			break;
			
		element_p=&(*element_p)->next;
	}
	
	graph=g_load_as (block_graph,-(required_offset+block_graph->a_stack_begin_displacement+1));

	new_element=alloc_graph_mem(sizeof (A_STACK));
	
	new_element->offset		=required_offset;
	new_element->graph		=graph;
	new_element->load_graph	=graph;
	new_element->flags		=0;
	
	/* link new elem in a stack */
	
	new_element->next=*element_p;
	*element_p=new_element;
	
	return graph;
}

INSTRUCTION_GRAPH s_get_b (BLOCK_GRAPH *block_graph,INT16 offset)
{
	register B_STACK **element_p,*new_element;
	register INT16 required_offset;
	INSTRUCTION_GRAPH graph;
	
	required_offset=block_graph->b_stack_top_offset+offset;
	
	element_p=&block_graph->b_stack;
	while (*element_p!=NULL)
	{	INT16 element_offset=(*element_p)->offset;
		
		if (element_offset==required_offset)
			return (*element_p)->graph;
			
		if (element_offset>required_offset)
			break;
			
		element_p=&(*element_p)->next;
	}

	graph=g_load_bs (block_graph,required_offset+block_graph->b_stack_begin_displacement);
	
	new_element=alloc_graph_mem(sizeof (B_STACK));
	
	new_element->offset		=required_offset;
	new_element->graph		=graph;
	new_element->load_graph	=graph;
	new_element->flags		=0;

	/* link new elem in a stack */
	
	new_element->next=*element_p;
	*element_p=new_element;
	
	return graph;
}

static void s_put_a_l (BLOCK_GRAPH *block_graph,INT16 offset,INSTRUCTION_GRAPH graph,INSTRUCTION_GRAPH new_load_graph)
{
	register A_STACK **element_p,*new_element;
	register INT16 required_offset;
	
	required_offset=block_graph->a_stack_top_offset+offset;
	
	element_p=&block_graph->a_stack;
	while (*element_p!=NULL)
	{	INT16 element_offset=(*element_p)->offset;
		
		if (element_offset==required_offset)
		{	(*element_p)->graph=graph;
			return;
		}
			
		if (element_offset>required_offset)
			break;
			
		element_p=&(*element_p)->next;
	}
	
	new_element=alloc_graph_mem(sizeof (A_STACK));
	
	new_element->offset		=required_offset;
	new_element->graph		=graph;
	new_element->load_graph	=new_load_graph;
	new_element->flags		=0;

	/* link new elem in a stack */
	
	new_element->next=*element_p;
	*element_p=new_element;
}

static VOID s_put_b_l (BLOCK_GRAPH *block_graph,INT16 offset,INSTRUCTION_GRAPH graph,INSTRUCTION_GRAPH new_load_graph)
{
	register B_STACK **element_p,*new_element;
	register INT16 required_offset;
		
	required_offset=block_graph->b_stack_top_offset+offset;
	
	element_p=&block_graph->b_stack;
	while (*element_p!=NULL)
	{	INT16 element_offset=(*element_p)->offset;
		
		if (element_offset==required_offset)
		{	(*element_p)->graph=graph;
			return;
		}
			
		if (element_offset>required_offset)
			break;
			
		element_p=&(*element_p)->next;
	}
	
	new_element=alloc_graph_mem(sizeof (B_STACK));
	
	new_element->offset		=required_offset;
	new_element->graph		=graph;
	new_element->load_graph	=new_load_graph;
	new_element->flags		=0;

	/* link new elem in a stack */
	
	new_element->next=*element_p;
	*element_p=new_element;
}

VOID s_put_a (BLOCK_GRAPH *block_graph,INT16 offset,INSTRUCTION_GRAPH graph)
{
	s_put_a_l (block_graph,offset,graph,NULL);
}

VOID s_put_b (BLOCK_GRAPH *block_graph,INT16 offset,INSTRUCTION_GRAPH graph)
{
	s_put_b_l (block_graph,offset,graph,NULL);
}

VOID s_push_a (BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph)
{
	--block_graph->a_stack_top_offset;
	s_put_a (block_graph,0,graph);
}

VOID s_push_b (BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph)
{
	--block_graph->b_stack_top_offset;
	s_put_b (block_graph,0,graph);
}

INSTRUCTION_GRAPH s_get_pop_a (BLOCK_GRAPH *block_graph)
{
	INSTRUCTION_GRAPH graph;
	
	graph=s_get_a (block_graph,0);
	++block_graph->a_stack_top_offset;
	
	return graph;
}

INSTRUCTION_GRAPH s_get_pop_b (BLOCK_GRAPH *block_graph)
{
	INSTRUCTION_GRAPH graph;
	
	graph=s_get_b (block_graph,0);
	++block_graph->b_stack_top_offset;
	
	return graph;
}

VOID s_pop_a (BLOCK_GRAPH *block_graph)
{
	++block_graph->a_stack_top_offset;
}

VOID s_pop_b (BLOCK_GRAPH *block_graph)
{
	++block_graph->b_stack_top_offset;
}

VOID s_pop_a_n (BLOCK_GRAPH *block_graph,INT16 n)
{
	block_graph->a_stack_top_offset+=n;
}

VOID s_pop_b_n (BLOCK_GRAPH *block_graph,INT16 n)
{
	block_graph->b_stack_top_offset+=n;
}

/********************************************************************************
 *																				*
 *		floating point stack simulation											*
 *																				*
 ********************************************************************************/

INSTRUCTION_GRAPH s_get_pop_float (BLOCK_GRAPH *block_graph)
{
	INSTRUCTION_GRAPH result;
	
	result=g_fjoin(s_get_b(block_graph,0),s_get_b(block_graph,1));

	block_graph->b_stack_top_offset+=2;

	return result;
}

VOID s_push_float (BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH f)
{
	block_graph->b_stack_top_offset-=2;

	s_put_b(block_graph,1,g_fhigh(f));
	s_put_b(block_graph,0,g_flow(f));
}

VOID s_put_float (BLOCK_GRAPH *block_graph,INT16 offset,INSTRUCTION_GRAPH f)
{
	s_put_b(block_graph,offset+1,g_fhigh(f));
	s_put_b(block_graph,offset,g_flow(f));
}

/************************************************************************************
 *																					*
 *																					*
 *																					*
 ************************************************************************************/

static VOID empty_stacks (BLOCK_GRAPH *block_graph)
{
	register INT16 n;
	
	block_graph->a_stack=NULL;
	block_graph->b_stack=NULL;
	
	for (n=0; n<N_LOCALS; ++n)
		block_graph->local_parameter_node[n]=NULL;

	block_graph->begin_a_stack_size=0;
	block_graph->begin_b_stack_size=0;
	block_graph->end_a_stack_size=0;
	block_graph->end_b_stack_size=0;

	block_graph->end_stack_vector=&block_graph->small_end_stack_vector;
	block_graph->small_end_stack_vector=*e_vector;
	block_graph->begin_stack_vector=&block_graph->small_begin_stack_vector;
	block_graph->small_begin_stack_vector=*e_vector;

	block_graph->a_stack_top_offset=0;
	block_graph->a_stack_begin_displacement=0;
	block_graph->a_stack_end_displacement=0;
	block_graph->used_a_stack_elements=0;
	
	block_graph->b_stack_top_offset=0;
	block_graph->b_stack_begin_displacement=0;
	block_graph->b_stack_end_displacement=0;
	block_graph->used_b_stack_elements=0;

	block_graph->begin_n_a_locals=0;
	block_graph->begin_n_b_locals=0;
	block_graph->a_increase=0;
	block_graph->b_increase=0;
	
	block_graph->stack_use=0;
}

static INT16 count_a_params (BLOCK_GRAPH *block_graph)
{
	register A_STACK *a_element;
	register INT16 a_stack_top_offset;
	register INT16 a_stack_size,offset;

	a_element=block_graph->a_stack;
	a_stack_top_offset=block_graph->a_stack_top_offset;
	
	/* counts the size of the stack structure until a hole or a dummy element
	   is encountered */
	
	a_stack_size=0;
		
	while (a_element!=NULL && a_element->offset<a_stack_top_offset)
		a_element=a_element->next;
	
	offset=a_stack_top_offset;

	while (a_element!=NULL && a_element->offset==offset
		&& !((a_element->flags & ELEMENT_MAY_BE_REMOVED) 
			  && a_element->graph==NULL))
	{
		++a_stack_size;	
		a_element=a_element->next;
		++offset;
	}
	block_graph->end_a_stack_size=a_stack_size;
	return a_stack_size;
}

static INT16 count_b_params (BLOCK_GRAPH *block_graph)
{
	U_INT32 **vector_p;
	INT16 b_stack_top_offset;
	
	register B_STACK *first_b_element,*b_element;
	register INT16 b_stack_size,offset,i;
	U_INT32 *vector;

	/* counts the size of the stack structure until a hole or a dummy element
	   is encountered */
	
	block_graph->end_stack_vector=&block_graph->small_end_stack_vector;
	vector_p=&block_graph->end_stack_vector;
	b_stack_top_offset=block_graph->b_stack_top_offset;

	b_stack_size=0;
	first_b_element=block_graph->b_stack;
	
	while (first_b_element!=NULL && first_b_element->offset<b_stack_top_offset)
		first_b_element=first_b_element->next;
		
	offset=b_stack_top_offset;
	b_element=first_b_element;
	
	while (b_element!=NULL && b_element->offset==offset
		&& !((b_element->flags & ELEMENT_MAY_BE_REMOVED) 
			 && b_element->graph==NULL))
	{
		++b_stack_size;
		b_element=b_element->next;
		++offset;
	}
	
	/* stack size determined, try find floats on stack */
	
	if (b_stack_size<=VECTOR_ELEMENT_SIZE)
		vector=*vector_p;
	else
	{	vector=alloc_graph_mem
			(vector_elem(b_stack_size+VECTOR_ELEMENT_SIZE-1)*sizeof(U_INT32));
		*vector_p=vector;
	}
	
	b_element=first_b_element;
	
	for (i=0; i<b_stack_size; )
	{	register B_STACK *next_b_element;
		INSTRUCTION_GRAPH graph,next_graph;
			
		if (i<b_stack_size-1 
			&& (graph=b_element->graph)->code==GFLOW
			&& (next_b_element=b_element->next)!=NULL
			&& (next_graph=next_b_element->graph)->code==GFHIGH
			&& next_graph->parameters[0].pointer==graph->parameters[0].pointer)
		{
			set_vector_bit(vector,i);
			++i;
			set_vector_bit(vector,i);
			++i;
			b_element=next_b_element->next;
		} else
		{	clear_vector_bit(vector,i);
			++i;
			b_element=b_element->next;
		}
	}
	block_graph->end_b_stack_size=b_stack_size;
	return b_stack_size;
}

static INT16 count_b_params_2 (BLOCK_GRAPH *block_graph)
{
	U_INT32 **vector_p;
	INT16 b_stack_top_offset;
		
	/* the same a count_b_stack_size, except for the fact that it stops somewhat
	   sooner (used to determine stack parameter count before jsr_eval) */
	
	register B_STACK *first_b_element,*b_element;
	register INT16 b_stack_size,offset,i;
	U_INT32 *vector;
	
	block_graph->end_stack_vector=&block_graph->small_end_stack_vector;
	vector_p=&block_graph->end_stack_vector;
	b_stack_top_offset=block_graph->b_stack_top_offset;

	b_stack_size=0;
	first_b_element=block_graph->b_stack;
	
	while (first_b_element!=NULL && first_b_element->offset<b_stack_top_offset)
		first_b_element=first_b_element->next;
		
	offset=b_stack_top_offset;
	b_element=first_b_element;
	
	while (b_element!=NULL && b_element->offset==offset
		&& b_element->graph!=NULL
		&& b_element->flags & ELEMENT_USED_BEFORE_JSR
		&& !((b_element->flags & ELEMENT_MAY_BE_REMOVED)
			 && b_element->graph==NULL))
	{
		++b_stack_size;
		b_element=b_element->next;
		++offset;
	}

	/* stack size determined, try find floats on stack */
	
	if (b_stack_size<=VECTOR_ELEMENT_SIZE)
		vector=*vector_p;
	else
	{	vector=alloc_graph_mem
			(vector_elem(b_stack_size+VECTOR_ELEMENT_SIZE-1)*sizeof(U_INT32));
		*vector_p=vector;
	}
	
	b_element=first_b_element;

	for (i=0; i<b_stack_size; )
	{	register B_STACK *next_b_element;
		INSTRUCTION_GRAPH graph,next_graph;
			
		if (i<b_stack_size-1 
			&& (graph=b_element->graph)->code==GFLOW
			&& (next_b_element=b_element->next)!=NULL
			&& (next_graph=next_b_element->graph)->code==GFHIGH
			&& next_graph->parameters[0].pointer==graph->parameters[0].pointer)
		{
			set_vector_bit(vector,i);
			++i;
			set_vector_bit(vector,i);
			++i;
			b_element=next_b_element->next;
		} else
		{	clear_vector_bit(vector,i);
			++i;
			b_element=b_element->next;
		}
	}
	block_graph->end_b_stack_size=b_stack_size;
	return b_stack_size;
}

static VOID copy_vector (U_INT32 **dest_vector_p, U_INT32 source_vector[], INT16 source_vector_size)
{
	/* Copies the source vector into the dest vector. It is assumed the dest vector
	   can hold a single vector element */
	
	U_INT32 *vector;
	INT16 index;

	if (source_vector_size<=VECTOR_ELEMENT_SIZE)
		*dest_vector_p[0] = source_vector[0];
	else
	{	vector=alloc_heap_mem(vector_elem(source_vector_size+VECTOR_ELEMENT_SIZE-1)*sizeof(U_INT32));
		*dest_vector_p=vector;

		for (index=vector_elem(source_vector_size); index >= 0; index--)
			vector[index]=source_vector[index];
	}
}

static VOID default_block_label_layout (BLOCK_GRAPH *block_graph, LABEL *label)
{
	if (label->flags & FIXED_LOCALS)
	{	block_graph->end_a_stack_size=label->a.stack_size;
		block_graph->end_b_stack_size=label->b.stack_size;
		block_graph->end_stack_vector=&block_graph->small_end_stack_vector;
		copy_vector(&block_graph->end_stack_vector,label->vector,label->b.stack_size);
	} else
	{	label->a.stack_size=count_a_params(block_graph);
		label->b.stack_size=count_b_params(block_graph);
		label->vector=&label->small_vector;
		copy_vector(&label->vector,block_graph->end_stack_vector,block_graph->end_b_stack_size);
		label->flags |= FIXED_LOCALS;
	}
}

VOID set_label_layout (	LABEL *label,
						INT16 a_stack_size,
						INT16 b_stack_size,
						U_INT32 vector[])
{
	if (label->flags & FIXED_LOCALS)
	{
		if (label->a.stack_size!=a_stack_size ||
			label->b.stack_size!=b_stack_size)
			label_error(label,"label layout not consistent");
		
		return;
	}
	
	label->a.stack_size=a_stack_size;
	label->b.stack_size=b_stack_size;
	label->vector=&label->small_vector;
	copy_vector(&label->vector,vector,b_stack_size);
	label->flags |= FIXED_LOCALS;
}

static VOID set_block_end_layout (	BLOCK_GRAPH *block_graph,
									INT16 a_stack_size,
									INT16 b_stack_size,
									U_INT32 vector[])
{
	block_graph->end_a_stack_size=a_stack_size;
	block_graph->end_b_stack_size=b_stack_size;
	block_graph->end_stack_vector=&block_graph->small_end_stack_vector;
	copy_vector(&block_graph->end_stack_vector,vector,b_stack_size);
}

static VOID set_block_begin_layout (	BLOCK_GRAPH *block_graph,
										INT16 a_stack_size,
										INT16 b_stack_size,
										U_INT32 vector[])
{
	block_graph->begin_a_stack_size=a_stack_size;
	block_graph->begin_b_stack_size=b_stack_size;
	block_graph->begin_stack_vector=&block_graph->small_begin_stack_vector;
	copy_vector(&block_graph->begin_stack_vector,vector,b_stack_size);
}

/********************************************************************************
 *																				*
 *		generation of store nodes for stack elements							*
 *																				*
 ********************************************************************************/
 
static VOID store_a_stack_values (BLOCK_GRAPH *block_graph)
{
	/* generates store nodes for each stack element that hasn't been popped */
	
	register A_STACK *a_element;
	INT16 n_parameters;
	
	n_parameters=block_graph->end_a_stack_size;
	
	if (n_parameters>N_A_PARAMETER_LOCALS)
		n_parameters=N_A_PARAMETER_LOCALS;
	
	/* remove popped values */
	
	for (a_element=block_graph->a_stack;
		a_element!=NULL && a_element->offset<block_graph->a_stack_top_offset;
		a_element=a_element->next)
	{	
		a_element->graph=NULL;
	}
	
	/* generate store local nodes */
	
	for(;a_element!=NULL && a_element->offset-block_graph->a_stack_top_offset<n_parameters;
			a_element=a_element->next)
	{
		if (a_element->graph!=NULL)
		{	register INT16 local_number;
			
			local_number=A_PAR_FIRST_LOCAL+n_parameters-1-(a_element->offset-block_graph->a_stack_top_offset);
			a_element->graph=g_store_l(a_element->graph,local_number);
		}
	}
	
	/* generate store non-local nodes for the rest */

	for(;a_element!=NULL;a_element=a_element->next)
	{
		if(a_element->graph!=NULL)
		{	if (a_element->graph==a_element->load_graph
				&& a_element->load_graph->code!=GLOCAL)
			{
				/* value on stack hasn't changed and isn't in local now, so
				   nothing has to be stored on stack */
				
				a_element->graph=NULL;
			} else
			{	INSTRUCTION_GRAPH address_graph;
		
				address_graph=a_element->graph;
				while (address_graph->code==GFILL)
					address_graph=address_graph->parameters[0].pointer;
				
				if (address_graph!=a_element->load_graph ||
					a_element->load_graph->code==GLOCAL)
				{	a_element->graph=g_store_as (
						block_graph,
						a_element->graph,
						-(1+a_element->offset+block_graph->a_stack_begin_displacement),
						a_element->load_graph);
				}
			}
		}
	}
	block_graph->a_stack_end_displacement=-n_parameters;
}

static void store_b_stack_values (BLOCK_GRAPH *block_graph)
{
	/* generates store nodes for each stack element that hasn't been popped */
	
	register B_STACK *b_element;
	register INT16 stack_displacement,offset,all_parameters_in_locals;
	INT16 number_of_b_parameters;
	
	INT16 n_parameters;
	U_INT32 *vector;
	
	n_parameters=block_graph->end_b_stack_size;
	vector=block_graph->end_stack_vector;
	
	/* determine amount of locals to store */
	
	number_of_b_parameters=0;
	for (offset=0; offset<n_parameters; ++offset)
	{	if (is_vector_bit_set (vector,offset))
		{	if (number_of_b_parameters<N_B_PARAMETER_LOCALS-1)
				number_of_b_parameters+=2;
			++offset;
		} else
		{	if (number_of_b_parameters<N_B_PARAMETER_LOCALS)
				++number_of_b_parameters;
		}
	}
	
	/* remove popped values */
	
	for (b_element=block_graph->b_stack;
		b_element!=NULL && b_element->offset<block_graph->b_stack_top_offset;
		b_element=b_element->next)
	{
		b_element->graph=NULL;
	}
	
	stack_displacement=0;
	all_parameters_in_locals=true;
	
	for(;b_element!=NULL; b_element=b_element->next)
	{	INSTRUCTION_GRAPH graph;
		INT16 parameter_number;

		graph=b_element->graph;
		parameter_number=b_element->offset-block_graph->b_stack_top_offset;
		
		if (parameter_number<n_parameters)
		{
			/* if possible result should be left in local */
		
			if (is_vector_bit_set (vector,parameter_number))
			{
				/* a floating point number */
				
				if (number_of_b_parameters>1)
				{
					/* float fits in locals */
					
					B_STACK *next_element;
					INSTRUCTION_GRAPH next_graph, float_graph;
					
					next_element=b_element->next;
					
					if (graph==NULL || next_element==NULL || (next_graph=next_element->graph)==NULL)
					{	internal_error("b_stack_stores: a floating point number was expected on stack");
						return;
					}

					number_of_b_parameters-=2;

					float_graph=g_fjoin(graph,next_graph);
					
					b_element->graph=g_fstore_l (float_graph,B_PAR_FIRST_LOCAL+number_of_b_parameters);
					next_element->graph=NULL;
					
					b_element=next_element;
					
					if (all_parameters_in_locals)
						stack_displacement-=2;

					continue;

				} else
					all_parameters_in_locals=false;
			} else
			{
				/* graph is a basic value */
				
				if (number_of_b_parameters>0)
				{
					/* basic value fits in local */
					
					--number_of_b_parameters;
					
					b_element->graph=g_store_l (graph,B_PAR_FIRST_LOCAL+number_of_b_parameters);
										
					if(all_parameters_in_locals)
						--stack_displacement;
					
					continue;
				} else
					all_parameters_in_locals=false;
			}
		}
		
		/* graph (if it exists) must be stored on stack (if it exists) */
		
		if (graph!=NULL)
		{	B_STACK *next_element;
			INSTRUCTION_GRAPH next_graph;

			if (graph->code==GFLOW
				&& (next_element=b_element->next)!=NULL
				&& (next_graph=next_element->graph)!=NULL
				&& next_graph->code==GFHIGH
				&& next_element->offset==b_element->offset+1
				&& next_graph->parameters[0].pointer==graph->parameters[0].pointer)
			{
				/* the graph is a floating point number that can't be stored in local */
				
				if (graph==b_element->load_graph &&
					next_graph==next_element->load_graph &&
					graph->parameters[0].pointer->code!=GFLOAT_LOCAL)
				{
					/* the floating point number hasn't been changed and doesn't reside
					   in locals, so nothing has to be stored on stack */
					
					b_element->graph=NULL;
					next_element->graph=NULL;
				} else
				{	INSTRUCTION_GRAPH float_graph;
				
					/* store floating point number on stack */

					float_graph=graph->parameters[0].pointer;
					b_element->graph=
						g_fstore_bs (
							block_graph,
							float_graph,
							b_element->offset+block_graph->b_stack_begin_displacement,
							b_element->load_graph,
							next_element->load_graph);
					
					next_element->graph=NULL;
				}
				b_element=next_element;
				continue;
			}
			
			/* the graph is a basic value that couldn't be stored in local */
			
			if (graph==b_element->load_graph &&
				graph->code!=GLOCAL &&
				!(	(graph->code==GFLOW || graph->code==GFHIGH) &&
					((INSTRUCTION_GRAPH)graph->parameters[0].pointer)->code==GFLOAT_LOCAL))
			{
				/* if graph points to same load value, this value doesn't have to be
				   saved on the stack, unless it is in a local */
				
				/* the basic value isn't changed and is on stack already */
				
				b_element->graph=NULL;
			} else
			{
				b_element->graph=
					g_store_bs (block_graph,
							graph,
							b_element->offset+block_graph->b_stack_begin_displacement,
							b_element->load_graph);
			}
		}
	}	
	block_graph->b_stack_end_displacement=stack_displacement;
}

/****************************************************************************
 *																			*
 *		generation of local nodes for parameters							*
 *																			*
 ****************************************************************************/

static VOID load_a_param_locals (BLOCK_GRAPH *block_graph)
{
	/* generates load node for each a-stack element (to be used at start of new block
	   when no graphs have been built yet and when it's clear that these parameters
	   will be placed in locals indeed) */
	
	INT16 n;
	INT16 a_stack_size;
	
	a_stack_size=block_graph->begin_a_stack_size;
		
	if (a_stack_size>N_A_PARAMETER_LOCALS)
		a_stack_size=N_A_PARAMETER_LOCALS;
	
	block_graph->begin_n_a_locals=a_stack_size;
	
	for (n=0; n<a_stack_size; ++n)
	{	INSTRUCTION_GRAPH local_graph;
				
		local_graph=g_local (A_PAR_FIRST_LOCAL+a_stack_size-1-n,block_graph->local_parameter_node);
		s_put_a_l (block_graph,n,local_graph,local_graph);
	}
	block_graph->a_stack_begin_displacement=-a_stack_size;
}

static VOID load_b_param_locals (BLOCK_GRAPH *block_graph)
{
	/* generates load node for each b-stack element (to be used at start of new
	   block when no graphs have been built yet and when it is clear that these
	   parameters will be placed in locals indeed) */

	INT16 offset,stack_displacement,all_parameters_in_locals;
	INT16 number_of_b_parameters;
	
	INT16 b_stack_size;
	U_INT32 *vector;
	
	b_stack_size=block_graph->begin_b_stack_size;
	vector=block_graph->begin_stack_vector;
	
	/* the difference between the virtual stack top and the real stack top isn't
	   always exactly the amount of locals that is used to cache the top of the
	   stack. From the point where the locals don't represent a contiguous piece of
	   stack space the corresponding stack entries are kept free, so that it stays
	   possible to swap entries between stack and locals without affecting the
	   position of other entries on the stack */
	
	/* first determine number of locals needed */
	
	number_of_b_parameters=0;
	for (offset=0; offset<b_stack_size; ++offset)
	{	if (is_vector_bit_set (vector,offset))
		{	if (number_of_b_parameters<N_B_PARAMETER_LOCALS-1)
				number_of_b_parameters+=2;
			++offset;
		} else
		{	if (number_of_b_parameters<N_B_PARAMETER_LOCALS)
				++number_of_b_parameters;
		}
	}
	block_graph->begin_n_b_locals=number_of_b_parameters;

	stack_displacement=0;
	all_parameters_in_locals=true;
	
	for (offset=0; offset<b_stack_size;++offset)
	{	if (is_vector_bit_set (vector,offset))
		{	
			INSTRUCTION_GRAPH f_graph,l_graph,h_graph;
			
			/* a floating point value has to be loaded */
			
			if (number_of_b_parameters>1)
			{	
				/* floating point number fits in locals, so build a float local node */
	
				number_of_b_parameters-=2;
				
				f_graph=g_flocal(B_PAR_FIRST_LOCAL+number_of_b_parameters,block_graph->local_parameter_node);
				h_graph=g_fhigh (f_graph);
				l_graph=g_flow (f_graph);
				
				s_put_b_l (block_graph,offset,l_graph,l_graph);
				s_put_b_l (block_graph,offset+1,h_graph,h_graph);
				
				if (all_parameters_in_locals)
					stack_displacement-=2;
			} else
			{
				/* float doesn't fit in locals */
				
				f_graph=g_fload_bs (block_graph,offset+stack_displacement); /*stack displacement fixed now */
				h_graph=g_fhigh (f_graph);
				l_graph=g_flow (f_graph);
		
				s_put_b_l (block_graph,offset,l_graph,l_graph);
				s_put_b_l (block_graph,offset+1,h_graph,h_graph);
	
				all_parameters_in_locals=false;
			}
			++offset;
		} else
		{	INSTRUCTION_GRAPH b_graph;
			
			/* a basic value has to be loaded */
			
			if (number_of_b_parameters>0)
			{
				/* basic fits in local */
				
				--number_of_b_parameters;
				
				b_graph=g_local (B_PAR_FIRST_LOCAL+number_of_b_parameters,block_graph->local_parameter_node);
				s_put_b_l(block_graph,offset,b_graph,b_graph);
				
				if (all_parameters_in_locals)
					--stack_displacement;
			} else
				all_parameters_in_locals=false;
		}
	}
	block_graph->b_stack_begin_displacement=stack_displacement;
}

static VOID load_a_param_locals_2 (BLOCK_GRAPH *block_graph)
{
	register INT16 offset;
	register A_STACK **element_p;
	INT16 n_a_locals;
	INSTRUCTION_GRAPH *local_parameter_node;
	
	element_p=&block_graph->a_stack;
	n_a_locals=block_graph->begin_a_stack_size;
	local_parameter_node=block_graph->local_parameter_node;

	if (n_a_locals>N_A_PARAMETER_LOCALS)
		n_a_locals=N_A_PARAMETER_LOCALS;

	block_graph->begin_n_a_locals=n_a_locals;
	
	/* generates proper load nodes for each parameter for a certain block. This
	   routine should be called after graphs have been built for this block and
	   when it becomes clear that these parameters will be placed in locals
	   indeed (in fact only needed for JSR_EVAL) */
	
	while (*element_p!=NULL && (*element_p)->offset<0)
		element_p=&(*element_p)->next;
	
	for (offset=0; offset<n_a_locals; ++offset)
	{	register INSTRUCTION_GRAPH graph;
		
		if (*element_p!=NULL && (*element_p)->offset==offset)
		{	register A_STACK *element;
			
			/* the stack element for this parameter already exists */
			
			element=*element_p;
			
			if (element->flags & ELEMENT_MAY_BE_REMOVED)
			{	/* this is a dummy stack element: the load graph isn't used and no
				   result is stored. Build a local node to fool the local allocator
				   so memory is 'reserved' for this parameter at the start of the block.
				   To ensure this local contains valid information (it may get pushed
				   on the stack...look out for garbage collections) the the local node
				   is put in the stack element */
				
				graph=g_local(A_PAR_FIRST_LOCAL+n_a_locals-1-offset,local_parameter_node);
							
				element->graph=graph;
				element->load_graph=graph;
			} else
			{	/* replace the load graph by a local node, if there's no load graph,
				   a load local node is created. This ensures a local is reserved at
				   the start of a block, so nothing (the heap pointer in particular)
				   will corrupt it. */
				
				graph=element->load_graph;
				if (graph!=NULL)
				{	graph->code=GLOCAL;
					graph->arity=1;
					graph->parameters[0].int16=(A_PAR_FIRST_LOCAL+n_a_locals-1-offset);
					local_parameter_node[A_PAR_FIRST_LOCAL+n_a_locals-1-offset]=graph;
				} else
				{	graph=g_local(A_PAR_FIRST_LOCAL+n_a_locals-1-offset,local_parameter_node);
					element->load_graph=graph;
				}
			}
			element_p=&(*element_p)->next;
		} else
		{	register A_STACK *new_element;
			
			/* the stack element for this parameter doesn't exist, but
			   it may yet be used much later (a store_nl node might need access to it) */
			
			graph=g_local(A_PAR_FIRST_LOCAL+n_a_locals-1-offset,local_parameter_node);
			
			new_element=alloc_graph_mem (sizeof (A_STACK));
			
			new_element->offset		=offset;
			new_element->graph		=graph;
			new_element->load_graph	=graph;
			new_element->flags		=0;
			
			new_element->next=*element_p;
			*element_p=new_element;
			element_p=&new_element->next;
		}
	}
	block_graph->a_stack_begin_displacement=-n_a_locals;
}

static VOID load_b_param_locals_2 (BLOCK_GRAPH *block_graph)
{
	register INT16 offset,stack_displacement;
	register INT16 n_b_locals,all_parameters_in_locals;

	register B_STACK **element_p;
	INT16 b_stack_size;
	U_INT32 *vector;
	INSTRUCTION_GRAPH *local_parameter_node;
	
	element_p=&block_graph->b_stack;
	b_stack_size=block_graph->begin_b_stack_size;
	vector=block_graph->begin_stack_vector;
	local_parameter_node=block_graph->local_parameter_node;

	/* generates proper load nodes for each parameter for a certain block. This
	   routine should be called after graphs have been built for this block and
	   when it becomes clear that these locals will be placed in locals indeed
	   (in fact only needed for JSR_EVAL) */

	/* determine amount of locals that are needed for parameters */
	
	n_b_locals=0;
	for (offset=0; offset<b_stack_size; ++offset)
	{	if (is_vector_bit_set (vector,offset))
		{	if (n_b_locals<N_B_PARAMETER_LOCALS-1)
				n_b_locals+=2;
			++offset;
		} else
		{	if (n_b_locals<N_B_PARAMETER_LOCALS)
				++n_b_locals;
		}
	}
	block_graph->begin_n_b_locals=n_b_locals;

	stack_displacement=0;
	all_parameters_in_locals=true;
	
	for (offset=0; offset<b_stack_size; ++offset)
	{	register INSTRUCTION_GRAPH graph;
		INSTRUCTION_GRAPH f_graph;
		INT16 f_local_not_used_flag;
		register B_STACK *element;

		if (!is_vector_bit_set (vector,offset))
		{
			/* the parameter is a basic value */
			
			if (n_b_locals<=0)
			{	all_parameters_in_locals=false;
				continue;
			}
			--n_b_locals;

			if (all_parameters_in_locals)
				--stack_displacement;
			
			while (*element_p!=NULL && (*element_p)->offset<offset)
				element_p=&(*element_p)->next;
			
			if (*element_p!=NULL && (*element_p)->offset==offset)
			{	
				element=*element_p;
				
				if (element->flags & ELEMENT_MAY_BE_REMOVED)
				{
					/* this is a dummy graph. Fool the local allocator by generating
					   a local node, so memory is reserved for this local and other
					   ones are allocated at the correct places. The graphs of the
					   stack element may be empty however, because the local doesn't
					   have to contain valid information (garbage collector doesn't
					   require valid information on b-stack) */
					
					graph=g_local(B_PAR_FIRST_LOCAL+n_b_locals,local_parameter_node);
					
					element->graph=NULL;
					element->load_graph=NULL;
				} else
				{					
					graph=element->load_graph;					
					if (graph!=NULL)
					{
						/* replace graph by proper local node */

						if (graph->code==GFLOW &&
							graph->parameters[0].pointer->code==GFLOAD_NL)
						{
							/*
								this turned out to be a floating point number, that
								is (partially) placed in locals.
							*/
							
							B_STACK *next_element;
							INSTRUCTION_GRAPH fload_graph,high_graph;
							
							fload_graph=graph->parameters[0].pointer;
							
							next_element=element->next;
							if (next_element==NULL || next_element->offset!=offset+1)
								internal_error ("load_b_param_locals_2");
							
							high_graph=next_element->load_graph;
							if (high_graph==NULL || high_graph->code!=GFHIGH
								|| high_graph->parameters[0].pointer!=fload_graph)
								internal_error ("load_b_param_locals_2");
							
							/*
								the floating point number can (partially) be found
								in locals (at least the low word will be placed
								there). Replace the original fload node by a
								join node to the two words on the stack. NOTE THAT
								A CYCLE IS CREATED FOR THE HIGH GRAPH IF THIS
								DOES NOT FIT IN A LOCAL, this must be taken care of
								later!.
							*/
							
							fload_graph->code=GFJOIN;
							fload_graph->parameters[0].pointer=graph;
							fload_graph->parameters[1].pointer=high_graph;
						}

						graph->code=GLOCAL;
						graph->arity=1;
						graph->parameters[0].int16=(B_PAR_FIRST_LOCAL+n_b_locals);

					}
					local_parameter_node[B_PAR_FIRST_LOCAL+n_b_locals]=graph;
				}
				element_p=&(*element_p)->next;
			} else
			{
				/* the stack element for this parameter doesn't exist,
				   but it may yet be used lateron */
			
				graph=g_local(B_PAR_FIRST_LOCAL+n_b_locals,local_parameter_node);
			
				element=alloc_graph_mem (sizeof (B_STACK));
				
				element->offset		=offset;
				element->graph		=graph;
				element->load_graph	=graph;
				element->flags		=0;
								
				element->next=*element_p;
				*element_p=element;
				element_p=&element->next;
			}
			continue;
		}
				
		/* the parameter is a float */
		
		if (n_b_locals<=1)
		{	all_parameters_in_locals=false;
			continue;
		}
		
		/* the float fits in locals */
		
		n_b_locals-=2;
		if (all_parameters_in_locals)
			stack_displacement-=2;

		while (*element_p!=NULL && (*element_p)->offset<offset)
			element_p=&(*element_p)->next;
		
		f_graph=NULL;
		
		/* if LOW and/or HIGH node available on stack: overwrite original float node by
		   FLOAT_LOCAL node, else create new FLOAT_LOCAL node and make stack entries
		   point to it by means of LOW/HIGH nodes */
		
		if ((element=*element_p)!=NULL)
		{	if (element->offset==offset)
			{	if ((graph=element->load_graph)!=NULL &&
					graph->code==GFLOW)
				{
					f_graph=graph->parameters[0].pointer;
					f_graph->code=GFLOAT_LOCAL;
					f_graph->arity=0;
					f_graph->parameters[0].int16=(B_PAR_FIRST_LOCAL+n_b_locals);					
				}
				element=element->next;
			}
			if (element!=NULL && element->offset==offset+1 &&
				(graph=element->load_graph)!=NULL &&
				graph->code==GFHIGH)
			{
				f_graph=graph->parameters[0].pointer;
				f_graph->code=GFLOAT_LOCAL;
				f_graph->arity=0;
				f_graph->parameters[0].int16=(B_PAR_FIRST_LOCAL+n_b_locals);
			}
		}
		
		if (f_graph==NULL)
		{	f_graph=g_new_node (GFLOAT_LOCAL,0,sizeof (PARAMETER));
			f_graph->parameters[0].int16=(B_PAR_FIRST_LOCAL+n_b_locals);
		}
				
		/* now f_graph consists of a float_local node */
		
		f_local_not_used_flag=0;	/* both words supposed needed */
		
   		if ((element=*element_p)!=NULL && element->offset==offset)
		{	if (element->flags & ELEMENT_MAY_BE_REMOVED)
			{	/* the first word of the float is a dummy graph */
				
				/* LOW GRAPH NOT NEEDED
				
				graph=g_new_node (GFLOW,0,sizeof (PARAMETER));
				graph->parameters[0].pointer=f_graph;
				*/
										
				element->graph=NULL;
				element->load_graph=NULL;
			} else
			{
				/* replace load graph by a LOW->FLOAT_LOCAL graph */
				
				graph=element->load_graph;
				if (graph!=NULL)
				{	graph->code=GFLOW;
					graph->arity=0;
					graph->parameters[0].pointer=f_graph;
					element->load_graph=f_graph;
				} else
					++f_local_not_used_flag;	/* low load graph not needed */
			}
			element_p=&(*element_p)->next;
		} else
		{
			/* stack elem isn't available yet, build one */
			
			graph=g_new_node (GFLOW,0,sizeof (PARAMETER));
			graph->parameters[0].pointer=f_graph;
			
			element=alloc_graph_mem (sizeof (B_STACK));
			
			element->offset		=offset;
			element->flags		=0;
			element->graph		=graph;
			element->load_graph	=f_graph;
			
			element->next=*element_p;
			*element_p=element;
			element_p=&element->next;
		}
		
		++offset;
		
		/* low word has been built on stack */
		
		if ((element=*element_p)!=NULL && (*element_p)->offset==offset)
		{	if (element->flags & ELEMENT_MAY_BE_REMOVED)
			{
				/* HIGH GRAPH NOT NEEDED
				
				graph=g_new_node (GFHIGH,0,sizeof (PARAMETER));
				graph->parameters[0].pointer=f_graph;
				*/
				
				element->graph=NULL;
				element->load_graph=NULL;
			} else
			{	graph=element->load_graph;
				if (graph!=NULL)
				{	graph->code=GFHIGH;
					graph->arity=0;
					graph->parameters[0].pointer=f_graph;
					element->load_graph=f_graph;
				} else
					++f_local_not_used_flag;	/* high load graph not needed */
			}
			element_p=&(*element_p)->next;
		} else
		{
			graph=g_new_node (GFHIGH,0,sizeof (PARAMETER));
			graph->parameters[0].pointer=f_graph;
			
			element=alloc_graph_mem (sizeof (B_STACK));
			
			element->offset		=offset;
			element->flags		=0;
			element->graph		=graph;
			element->load_graph	=f_graph;
				
			element->next=*element_p;
			*element_p=element;
			element_p=&element->next;
		}
		
		/* high word has been built on the stack */
		
		if(f_local_not_used_flag==2)
		{	local_parameter_node[B_PAR_FIRST_LOCAL+n_b_locals]=NULL;
			local_parameter_node[B_PAR_FIRST_LOCAL+n_b_locals+1]=NULL;
		} else
		{	local_parameter_node[B_PAR_FIRST_LOCAL+n_b_locals]=f_graph;
			local_parameter_node[B_PAR_FIRST_LOCAL+n_b_locals+1]=f_graph;
		}
	}
	block_graph->b_stack_begin_displacement=stack_displacement;
}

static VOID load_nodes_for_end_a_params (BLOCK_GRAPH *block_graph)
{
	/* builds default load graphs for END parameters, so that the store local
	   nodes that will be build for them refer to valid information */
	
	register A_STACK **element_p;
	register INT16 parameter_n;
	INT16 n_parameters;
	
	n_parameters=block_graph->end_a_stack_size;
	
	if (n_parameters>N_A_PARAMETER_LOCALS)
		n_parameters=N_A_PARAMETER_LOCALS;
	
	element_p=&block_graph->a_stack;
	for (parameter_n=0; parameter_n<n_parameters; ++parameter_n)
	{	register A_STACK *new_element;
		register INT16 required_offset;
		INSTRUCTION_GRAPH graph;
		
		required_offset=block_graph->a_stack_top_offset+parameter_n;
		
		while (*element_p!=NULL && (*element_p)->offset<required_offset)
			element_p=&(*element_p)->next;
		
		if (*element_p!=NULL && (*element_p)->offset==required_offset)
		{	register A_STACK *element;
			
			element=*element_p;
			if (element->flags & ELEMENT_MAY_BE_REMOVED
				&& element->graph==NULL)
			{
				graph=g_load_as(block_graph,-(1+required_offset+block_graph->a_stack_begin_displacement));
				element->graph=graph;
				element->load_graph=graph;
			}
			element_p=&(*element_p)->next;
		} else {
			graph=g_load_as(block_graph,-(1+required_offset+block_graph->a_stack_begin_displacement));

			new_element=alloc_graph_mem (sizeof (A_STACK));
			
			new_element->offset		=required_offset;
			new_element->graph		=graph;
			new_element->load_graph	=graph;
			new_element->flags		=0;
			
			new_element->next=*element_p;
			*element_p=new_element;
			element_p=&new_element->next;
		}
	}
}

static VOID load_nodes_for_end_b_params (BLOCK_GRAPH *block_graph)
{
	/* builds default load graphs for END parameters, so that the store local
	   nodes that will be build for them refer to valid information */

	register B_STACK **element_p;
	register INT16 parameter_n;
	INT16 number_of_b_parameters;
	
	INT16 n_parameters;
	U_INT32 *vector;
	
	n_parameters=block_graph->end_b_stack_size;
	vector=block_graph->end_stack_vector;
	
	number_of_b_parameters=0;
	
	element_p=&block_graph->b_stack;
	for (parameter_n=0; parameter_n<n_parameters; ++parameter_n)
	{	register B_STACK *new_element;
		register INT16 required_offset;
		INSTRUCTION_GRAPH graph;
		
		required_offset=block_graph->b_stack_top_offset+parameter_n;
		
		while (*element_p!=NULL && (*element_p)->offset<required_offset)
			element_p=&(*element_p)->next;
		
		if (is_vector_bit_set(vector,parameter_n))
		{	if (!(number_of_b_parameters < N_B_PARAMETER_LOCALS-1))
			{
				/* result float doesn't fit, skip this parameter */
				
				++parameter_n;
				continue;
			}
			/* result float fits in local, handle first element of it */
			
			++number_of_b_parameters;
			
			if (*element_p!=NULL && (*element_p)->offset==required_offset)
			{	register B_STACK *element;
				
				element=*element_p;
				if (element->flags & ELEMENT_MAY_BE_REMOVED
					&& element->graph==NULL)
				{
					graph=g_load_bs(block_graph,required_offset+block_graph->b_stack_begin_displacement);
					element->graph=graph;
					element->load_graph=graph;
				}
				element_p=&(*element_p)->next;
			} else
			{	graph=g_load_bs(block_graph,required_offset+block_graph->b_stack_begin_displacement);

				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=required_offset;
				new_element->graph		=graph;
				new_element->load_graph	=graph;
				new_element->flags		=0;
			
				new_element->next=*element_p;
				*element_p=new_element;
				element_p=&new_element->next;
			}
			++parameter_n;
			required_offset=block_graph->b_stack_top_offset+parameter_n;
		
			while (*element_p!=NULL && (*element_p)->offset<required_offset)
				element_p=&(*element_p)->next;
				
			/* and fall through to normal case for next word */
		}
		if (number_of_b_parameters < N_B_PARAMETER_LOCALS)
		{
			/* this result parameter will placed in local */
			
			number_of_b_parameters++;
			
			if (*element_p!=NULL && (*element_p)->offset==required_offset)
			{	register B_STACK *element;
				
				element=*element_p;
				if (element->flags & ELEMENT_MAY_BE_REMOVED
					&& element->graph==NULL)
				{
					graph=g_load_bs(block_graph,required_offset+block_graph->b_stack_begin_displacement);
					element->graph=graph;
					element->load_graph=graph;
				}
				element_p=&(*element_p)->next;
			} else
			{	graph=g_load_bs(block_graph,required_offset+block_graph->b_stack_begin_displacement);

				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=required_offset;
				new_element->graph		=graph;
				new_element->load_graph	=graph;
				new_element->flags		=0;
			
				new_element->next=*element_p;
				*element_p=new_element;
				element_p=&new_element->next;
			}
		}
	}
}

/********************************************************************************
 *																				*
 *		insertion/removal of stack elements										*
 *																				*
 ********************************************************************************/

/* the insert and remove routines are implemented in such a way that the load
   graph of a stack element sticks to its stack element. In this way it stays
   possible to adjust the stack offsets properly */

VOID insert_graph_in_b_stack (BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph,INT16 b_stack_size,U_INT32 *vector)
{
	register INT16 offset,required_offset;
	register B_STACK **element_p,*new_element;
	INSTRUCTION_GRAPH *graph_p;
	
	/* get offset to top of stack */
	
	required_offset=block_graph->b_stack_top_offset;

	/* find top stack element */
	
	element_p=&block_graph->b_stack;
	while (*element_p!=NULL && (*element_p)->offset<required_offset)
		element_p=&(*element_p)->next;
	
	for (offset=0; offset<b_stack_size; ++offset,++required_offset)
	{
		/* for each offset within stack size, make sure stack elems are initialised */
		
		if (is_vector_bit_set(vector,offset))
		{
			/* if a float word resides in this element */
			
			if (*element_p==NULL || (*element_p)->offset>required_offset+1)
			{	INSTRUCTION_GRAPH f_graph,l_graph,h_graph;
				
				/* both stack elements haven't been inialised, so create graphs */
				
				f_graph=g_fload_bs(block_graph,required_offset+block_graph->b_stack_begin_displacement);
				
				h_graph=g_fhigh(f_graph);
				l_graph=g_flow(f_graph);

				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=required_offset;
				new_element->graph		=l_graph;
				new_element->load_graph	=l_graph;
				new_element->flags		=0;
		
				new_element->next=*element_p;
				*element_p=new_element;

				element_p=&new_element->next;
				++offset;
				++required_offset;
				
				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=required_offset;
				new_element->graph		=h_graph;
				new_element->load_graph	=h_graph;
				new_element->flags		=0;
		
				new_element->next=*element_p;
				*element_p=new_element;

				element_p=&new_element->next;
				continue;
			} else
			{	if (*element_p==NULL || (*element_p)->offset!=required_offset)
				{	INSTRUCTION_GRAPH load_graph;
					
					/* first element hasn't been initialised, build normal load node */
					
					load_graph=g_load_bs(block_graph,required_offset+block_graph->b_stack_begin_displacement);
					
					new_element=alloc_graph_mem (sizeof (B_STACK));
					
					new_element->offset		=required_offset;
					new_element->graph		=load_graph;
					new_element->load_graph	=load_graph;
					new_element->flags		=0;
		
					new_element->next=*element_p;
					*element_p=new_element;
				}
				element_p=&(*element_p)->next;

				++offset;
				++required_offset;
				
				/* and fall through to normal case */
			}
		}
		
		if (*element_p==NULL || (*element_p)->offset!=required_offset)
		{	INSTRUCTION_GRAPH load_graph;
			
			load_graph=g_load_bs(block_graph,required_offset+block_graph->b_stack_begin_displacement);
			
			new_element=alloc_graph_mem (sizeof (B_STACK));
			
			new_element->offset		=required_offset;
			new_element->graph		=load_graph;
			new_element->load_graph	=load_graph;
			new_element->flags		=0;

			new_element->next=*element_p;
			*element_p=new_element;
		}
		element_p=&(*element_p)->next;
	}
	
	/* now all stack elements within size have been initialised for sure, compute
	   offset needed to push element */
	
	required_offset=block_graph->b_stack_top_offset-1;

	/* pop elements before that point and allocate elem for the required offset */
	
	element_p=&block_graph->b_stack;
	while (*element_p!=NULL && (*element_p)->offset<required_offset)
	{
		(*element_p)->graph=NULL;
	
	/* CHANGED 14-09-92
	
		(*element_p)->load_graph=NULL;
	*/
	
		element_p=&(*element_p)->next;
	}

	if (*element_p==NULL || (*element_p)->offset!=required_offset)
	{	new_element=alloc_graph_mem (sizeof (B_STACK));
	
		new_element->offset		=required_offset;
		new_element->load_graph	=NULL;
		new_element->flags		=0;
		
		/* graph will be filled in below */
		
		new_element->next=*element_p;
		*element_p=new_element;
	}

	/* shift all result graphs within size one place forward (reverse cleanup)
	   and insert new graph at the end */
	
	graph_p=&(*element_p)->graph;
	++required_offset;
	element_p=&(*element_p)->next;
	
	for (offset=0; offset<b_stack_size; ++offset)
	{	*graph_p=(*element_p)->graph;
		graph_p=&(*element_p)->graph;
		element_p=&(*element_p)->next;
	}
	*graph_p=graph;

	--block_graph->b_stack_top_offset; /* one additional element on stack */
}


static VOID remove_a_stack_element (BLOCK_GRAPH *block_graph,register INT16 offset)
{
	/* removes a stack element (graph and load graph) at a certain offset by taking
	   the stack element out of the stack structure and decreasing the stack offset
	   of following elements */
	
	register A_STACK **element_p;
	register A_STACK *element;
	
	element_p=&block_graph->a_stack;
	
	element=*element_p;
	while (element!=NULL && element->offset<offset)
	{	element_p=&element->next;
		element=*element_p;
	}
	
	element=element->next;
	*element_p=element;
	
	while (element!=NULL)
	{	--element->offset;
		element=element->next;
	}
	
	/* top offset doesn't have to be adjusted */
}

static VOID remove_b_stack_element (BLOCK_GRAPH *block_graph,register INT16 offset)
{
	/* removes a stack element (graph and load graph) at a certain offset by taking
	   the stack element out of the stack structure and decreasing the stack offset
	   of following elements */
	
	register B_STACK *element;
	register B_STACK **element_p;
	
	element_p=&block_graph->b_stack;
	
	element=*element_p;
	while (element!=NULL && element->offset<offset)
	{	element_p=&element->next;
		element=*element_p;
	}
	
	element=element->next;
	*element_p=element;
	
	while (element!=NULL)
	{	--element->offset;
		element=element->next;
	}
	
	/* top offset doesn't have to be adjusted */
}

static VOID remove_result_b_stack_element(BLOCK_GRAPH *block_graph,register INT16 remove_offset)
{
	/* removes the result graph at a certain offset by shifting all previous result
	   graphs one element further (see also ABC cleanup) */
	
	register B_STACK **element_p;
	register B_STACK *element;
	register INSTRUCTION_GRAPH curr_graph;
	register INT16 offset,curr_flags,begin_offset;
	
	INT16 stack_begin_displacement;
	INT16 stack_top_offset;
		
	element_p=&block_graph->b_stack;
	stack_begin_displacement=block_graph->b_stack_begin_displacement;
	stack_top_offset=block_graph->b_stack_top_offset;
	
	element=*element_p;
	
	/* skip insignificant stack element */
	
	begin_offset=stack_top_offset<=0 ? stack_top_offset : 0;
	while (element!=NULL && element->offset<begin_offset)
	{	element_p=&element->next;
		element=*element_p;
	}
	
	curr_graph=NULL;
	curr_flags=0;

	for (offset=begin_offset; offset<remove_offset; ++offset)
	{	register INSTRUCTION_GRAPH prev_graph;
		register INT16 prev_flags;	
		
		if (element==NULL || element->offset!=offset)
		{	register B_STACK *new_element;
			INSTRUCTION_GRAPH graph;

			graph=g_load_bs(block_graph,offset+stack_begin_displacement);

			graph->used_mark=offset<stack_top_offset ? 0 : 2;
			
			new_element=alloc_graph_mem (sizeof (B_STACK));
			
			new_element->offset		=offset;
			new_element->graph		=graph;
			new_element->load_graph	=graph;
			new_element->flags		=0;
	
			new_element->next=element;
			*element_p=new_element;
			element=new_element;
		}
				
		prev_graph=curr_graph;
		prev_flags=curr_flags;
		
		curr_graph=element->graph;
		curr_flags=element->flags;
		
		element->graph=prev_graph;
		element->flags=prev_flags;
		
		element_p=&element->next;
		element=*element_p;
	}
	
	if (element->offset!=remove_offset)
		internal_error("remove_end_stack_element");
		
	element->graph=curr_graph;
	element->flags=curr_flags;
	
	++block_graph->b_stack_top_offset;
}

static VOID remove_result_a_stack_element(BLOCK_GRAPH *block_graph,register INT16 remove_offset)
{
	/* removes the result graph at a certain offset by shifting all previous result
	   graphs one element further (see also ABC cleanup) */
	
	register A_STACK **element_p;
	register A_STACK *element;
	register INSTRUCTION_GRAPH curr_graph;
	register INT16 offset,curr_flags,begin_offset;
	
	INT16 stack_begin_displacement;
	INT16 stack_top_offset;
	
	element_p=&block_graph->a_stack;
	stack_begin_displacement=block_graph->a_stack_begin_displacement;
	stack_top_offset=block_graph->a_stack_top_offset;
	
	element=*element_p;
	
	/* skip insignificant stack element */
	
	begin_offset=stack_top_offset<=0 ? stack_top_offset : 0;
	while (element!=NULL && element->offset<begin_offset)
	{	element_p=&element->next;
		element=*element_p;
	}
	
	curr_graph=NULL;
	curr_flags=0;

	for (offset=begin_offset; offset<remove_offset; ++offset)
	{	register INSTRUCTION_GRAPH prev_graph;
		register INT16 prev_flags;	
		
		if (element==NULL || element->offset!=offset)
		{	register A_STACK *new_element;
			INSTRUCTION_GRAPH graph;

			graph=g_load_as(block_graph,-(offset+stack_begin_displacement+1));

			graph->used_mark=offset<stack_top_offset ? 0 : 2;
			
			new_element=alloc_graph_mem (sizeof (A_STACK));
			
			new_element->offset		=offset;
			new_element->graph		=graph;
			new_element->load_graph	=graph;
			new_element->flags		=0;
	
			new_element->next=element;
			*element_p=new_element;
			element=new_element;
		}
				
		prev_graph=curr_graph;
		prev_flags=curr_flags;
		
		curr_graph=element->graph;
		curr_flags=element->flags;
		
		element->graph=prev_graph;
		element->flags=prev_flags;
		
		element_p=&element->next;
		element=*element_p;
	}
	
	if (element->offset!=remove_offset)
		internal_error("remove_end_stack_element");
		
	element->graph=curr_graph;
	element->flags=curr_flags;
	
	++block_graph->a_stack_top_offset;
}

static VOID remove_load_a_stack_element(BLOCK_GRAPH *block_graph,register INT16 remove_offset)
{
	/* removes the load graph at a certain offset by shifting all previous result
	   graphs one element back, removing the stack element at the given location
	   and adjusting the offsets at the following stack elements */
	
	register A_STACK *element,*previous_element;
	register INT16 offset,begin_offset;
	register A_STACK **element_p;
	INT16 stack_begin_displacement;
	INT16 stack_top_offset;
			
	element_p=&block_graph->a_stack;
	stack_begin_displacement=block_graph->a_stack_begin_displacement;
	stack_top_offset=block_graph->a_stack_top_offset;
	
	/* skip insignificant stack elements */
	
	begin_offset=stack_top_offset<=0 ? stack_top_offset : 0;

	previous_element=NULL;	
	element=*element_p;
	
	while (element!=NULL && element->offset<begin_offset){
		previous_element=element;
		element_p=&element->next;
		element=*element_p;
	}
	
	for (offset=begin_offset; offset<=remove_offset; ++offset){
		if (element==NULL || element->offset!=offset){
			register A_STACK *new_element;
			INSTRUCTION_GRAPH graph;
			
			graph=g_load_as(block_graph,-(offset+stack_begin_displacement+1));
			
			graph->used_mark=offset<stack_top_offset ? 0 : 2;
			
			new_element=alloc_graph_mem (sizeof (A_STACK));
			
			new_element->offset		=offset;
			new_element->graph		=graph;
			new_element->load_graph	=graph;
			new_element->flags		=0;
	
			new_element->next=element;
			*element_p=new_element;
			element=new_element;
		}
		
		if (previous_element==NULL || previous_element->offset!=offset-1)
		{	register A_STACK *new_element;
			
			new_element=alloc_graph_mem (sizeof (A_STACK));
			
			new_element->offset		=offset-1;
			new_element->graph		=NULL;
			new_element->load_graph	=NULL;
			new_element->flags		=0;
			
			new_element->next=element;
			*element_p=new_element;
			element_p=&new_element->next;
			previous_element=new_element;
		}
		
		previous_element->graph=element->graph;
		previous_element->flags=element->flags;
		
		element->graph=NULL;
		element->flags=0;

		if (offset<remove_offset)
		{	previous_element=element;
			element_p=&element->next;
			element=*element_p;
		}
	}
	
	/* take stack elem with unused load graph out of list */
	
	if (element!=NULL)
		*element_p=element->next;
	
	for (element=*element_p; element!=NULL; element=element->next)
		--element->offset;

	--block_graph->a_stack_top_offset;
}

static void remove_load_b_stack_element(BLOCK_GRAPH *block_graph,register INT16 remove_offset)
{
	/* removes the load graph at a certain offset by shifting all previous result
	   graphs one element back, removing the stack element at the given location
	   and adjusting the offsets at the following stack elements */
	
	register B_STACK *element,*previous_element;
	register INT16 offset,begin_offset;
	register B_STACK **element_p;
	INT16 stack_begin_displacement;
	INT16 stack_top_offset;
	
	element_p=&block_graph->b_stack;
	stack_begin_displacement=block_graph->b_stack_begin_displacement;
	stack_top_offset=block_graph->b_stack_top_offset;
	
	/* skip insignificant stack elements */
	
	begin_offset=stack_top_offset<=0 ? stack_top_offset : 0;

	previous_element=NULL;	
	element=*element_p;
	
	while (element!=NULL && element->offset<begin_offset)
	{	previous_element=element;
		element_p=&element->next;
		element=*element_p;
	}

	for (offset=begin_offset; offset<=remove_offset; ++offset)
	{	if (element==NULL || element->offset!=offset)
		{	register B_STACK *new_element;
			INSTRUCTION_GRAPH graph;

			graph=g_load_bs(block_graph,offset+stack_begin_displacement);
			
			graph->used_mark=offset<stack_top_offset ? 0 : 2;
			
			new_element=alloc_graph_mem (sizeof (B_STACK));
			
			new_element->offset		=offset;
			new_element->graph		=graph;
			new_element->load_graph	=graph;
			new_element->flags		=0;
	
			new_element->next=element;
			*element_p=new_element;
			element=new_element;
		}
		
		if (previous_element==NULL || previous_element->offset!=offset-1)
		{	register B_STACK *new_element;
			
			new_element=alloc_graph_mem (sizeof (B_STACK));
			
			new_element->offset		=offset-1;
			new_element->graph		=NULL;
			new_element->load_graph	=NULL;
			new_element->flags		=0;
			
			new_element->next=element;
			*element_p=new_element;
			element_p=&new_element->next;
			previous_element=new_element;
		}
		
		previous_element->graph=element->graph;
		previous_element->flags=element->flags;
		
		element->graph=NULL;
		element->flags=0;

		if (offset<remove_offset)
		{	previous_element=element;
			element_p=&element->next;
			element=*element_p;
		}
	}
	
	/* take stack elem with unused load graph out of list */
	
	if (element!=NULL)
		*element_p=element->next;
	
	for (element=*element_p; element!=NULL; element=element->next)
		--element->offset;

	--block_graph->b_stack_top_offset;
}

static VOID remove_unused_a_stack_elements (register BLOCK_GRAPH *block_graph)
{
	/* determines which elements aren't used and removes them with the help
	   of the routines above */
	
	register A_STACK *curr_element,**curr_element_p;
	register A_STACK *next_element,**next_element_p;
	register INT16 stack_offset_difference,first_non_parameter_offset,offset;
	BLOCK_GRAPH *next_block_graph;
	
	next_block_graph=block_graph->next;
	
	/* relative offset of elements of next block: first of
	   all corresponding elements of the next block have an offset that is
	   'top_offset' (at the end of the current block) less, and furthermore
	   one has to take into account the difference between the amount of parameters
	   between two basic blocks */
	
	stack_offset_difference=	block_graph->a_stack_top_offset
								+block_graph->end_a_stack_size
								-next_block_graph->begin_a_stack_size;
	
	/* offset to first non-parameter at end of current block */
	
	first_non_parameter_offset=	block_graph->a_stack_top_offset+
								block_graph->used_a_stack_elements;
	
	/* parameters may not be removed, so first find the first non-parameter
	   at end of current block */
	
	curr_element_p=&block_graph->a_stack;
	curr_element=*curr_element_p;
	
	offset=block_graph->a_stack_top_offset;
	while (curr_element!=NULL && curr_element->offset<offset)
	{	curr_element_p=&curr_element->next;
		curr_element=*curr_element_p;
	}
		
	for (; offset<first_non_parameter_offset; ++offset)
	{	if (curr_element==NULL || curr_element->offset!=offset)
			return;
		
		curr_element_p=&curr_element->next;
		curr_element=*curr_element_p;
	}
	
	/* for each stack element at end of current block, find coresponding element of
	   next block and determine whether it should be removed or not. Note that only
	   elements are removed that form a contiguous area of stack elements from the
	   top. Otherwise a lot of stack elements might be moved on the stacks itself
	   which is not very profitable. */
	
	next_element_p=&next_block_graph->a_stack;
	next_element=*next_element_p;
	
	while (curr_element!=NULL && curr_element->offset==offset)
	{	register INT16 next_offset;
		
		next_offset=offset-stack_offset_difference;
		
		while (next_element!=NULL && next_element->offset<next_offset)
		{	next_element_p=&next_element->next;
			next_element=*next_element_p;
		}
		
		if (	(next_element!=NULL && next_element->offset==next_offset)
			?	(next_element->load_graph==NULL
				|| !next_element->load_graph->used_mark)
			:	next_offset<next_block_graph->a_stack_top_offset	)
		{
			/* in the next block, the element is:
				- unused and overwritten, or
				- not marked (but used by unused result graph), or
				- popped
			*/
							
			remove_result_a_stack_element(block_graph,offset);
			curr_element=*curr_element_p;
						
			if (next_element!=NULL && next_element->offset==next_offset
				&& next_element->flags & ELEMENT_MAY_BE_REMOVED)
			{
				/* in the next block, the load graph, as well as the result
				   graph of this stack element aren't used: remove both (and
				   repeat for next blocks). */
				
				A_STACK *element_to_be_removed;
				BLOCK_GRAPH *block_graph_to_remove_from,*previous_block_graph_to_remove_from;
				INT16 remove_offset;
				
				block_graph_to_remove_from=block_graph->next;
				element_to_be_removed=next_element;
				
				do {
					remove_offset=element_to_be_removed->offset;
					remove_a_stack_element(block_graph_to_remove_from,remove_offset);

					previous_block_graph_to_remove_from=block_graph_to_remove_from;
					block_graph_to_remove_from=block_graph_to_remove_from->next;
					element_to_be_removed=element_to_be_removed->next_block_element;
				} while (element_to_be_removed!=NULL);
								
				remove_offset-=
					previous_block_graph_to_remove_from->a_stack_top_offset
					+previous_block_graph_to_remove_from->end_a_stack_size
					-block_graph_to_remove_from->begin_a_stack_size;

				remove_load_a_stack_element(block_graph_to_remove_from,remove_offset);

				next_element=*next_element_p;
				++stack_offset_difference;	/* set relation to offset-1 */
			} else
			{
				remove_load_a_stack_element(next_block_graph,next_offset);
				next_element=*next_element_p;
				
				++stack_offset_difference;	/* set relation to offset-1 */
			}
		}
		
		++offset;
		curr_element_p=&curr_element->next;
		curr_element=*curr_element_p;
	}
}

static VOID remove_unused_b_stack_elements (register BLOCK_GRAPH *block_graph)
{
	/* determines which elements aren't used and removes them with the help
	   of the routines above (also see previous routine) */

	register B_STACK *curr_element,**curr_element_p;
	register B_STACK *next_element,**next_element_p;
	register INT16 stack_offset_difference,first_non_parameter_offset,curr_offset;
	BLOCK_GRAPH *next_block_graph;

	next_block_graph=block_graph->next;
		
	stack_offset_difference=
		block_graph->b_stack_top_offset
		+block_graph->end_b_stack_size
		-next_block_graph->begin_b_stack_size;
	
	first_non_parameter_offset=
		block_graph->used_b_stack_elements
		+block_graph->b_stack_top_offset;
	
	/* find first non-parameter of current block */
	
	curr_element_p=&block_graph->b_stack;
	curr_element=*curr_element_p;
	
	curr_offset=block_graph->b_stack_top_offset;
	while (curr_element!=NULL && curr_element->offset<curr_offset){
		curr_element_p=&curr_element->next;
		curr_element=*curr_element_p;
	}
	
	for (; curr_offset<first_non_parameter_offset; ++curr_offset)
	{	if (curr_element==NULL || curr_element->offset!=curr_offset)
			return;
		
		curr_element_p=&curr_element->next;
		curr_element=*curr_element_p;
	}
	
	next_element_p=&next_block_graph->b_stack;
	next_element=*next_element_p;
	
	while (curr_element!=NULL && curr_element->offset==curr_offset)
	{	INSTRUCTION_GRAPH load_graph;
		
		register INT16 next_offset;
		
		next_offset=curr_offset-stack_offset_difference;
		
		while (next_element!=NULL && next_element->offset<next_offset)
		{	next_element_p=&next_element->next;
			next_element=*next_element_p;
		}
		
		if ((next_element!=NULL && next_element->offset==next_offset)
			?  ((load_graph=next_element->load_graph)==NULL || 
				(!load_graph->used_mark && 
				 !((load_graph->code==GFHIGH ||
				   load_graph->code==GFLOW)
				  && load_graph->parameters[0].pointer->used_mark)))
			:  next_offset<next_block_graph->b_stack_top_offset)
		{
			/* in the next block, the element is:
				- unused and overwritten, or
				- not marked (but used by unused result graph), or
				- popped
			*/

			remove_result_b_stack_element(block_graph,curr_offset);
			curr_element=*curr_element_p;
						
			if (next_element!=NULL && next_element->offset==next_offset
				&& next_element->flags & ELEMENT_MAY_BE_REMOVED)
			{
				B_STACK *b_element_to_be_removed;
				BLOCK_GRAPH *block_graph_to_remove_from,*previous_block_graph_to_remove_from;
				INT16 remove_offset;
				
				block_graph_to_remove_from=block_graph->next;
				b_element_to_be_removed=next_element;
				
				do {
					remove_offset=b_element_to_be_removed->offset;
					remove_b_stack_element(block_graph_to_remove_from,remove_offset);
					
					previous_block_graph_to_remove_from=block_graph_to_remove_from;
					block_graph_to_remove_from=block_graph_to_remove_from->next;
					b_element_to_be_removed=b_element_to_be_removed->next_block_element;

				} while (b_element_to_be_removed!=NULL);
				
				remove_offset-=
					previous_block_graph_to_remove_from->b_stack_top_offset
					+previous_block_graph_to_remove_from->end_b_stack_size
					-block_graph_to_remove_from->begin_b_stack_size;

				remove_load_b_stack_element(block_graph_to_remove_from,remove_offset);
				
				next_element=*next_element_p;
				++stack_offset_difference;
			} else
			{	remove_load_b_stack_element(next_block_graph,next_offset);
				next_element=*next_element_p;
				
				++stack_offset_difference;
			}
		}
		
		++curr_offset;
		curr_element_p=&curr_element->next;
		curr_element=*curr_element_p;
	}
}

static INT16 try_remove_a_top_param (BLOCK_GRAPH *block_graph,INT16 a_stack_size)
{
	/* tries to remove top element of a-stack if possible */
	
	register A_STACK *a_element,*a_element_0;
	register A_STACK *begin_a_element;
	register INT16 offset,n;
	INSTRUCTION_GRAPH graph;
	BLOCK_GRAPH *next_block_graph;
		
	next_block_graph=block_graph->next;
	
	offset=block_graph->a_stack_top_offset;
	a_element=block_graph->a_stack;
	while (a_element!=NULL && a_element->offset<offset)
		a_element=a_element->next;
	
	if (a_element==NULL || a_element->offset!=offset)
		return a_stack_size;
	
	a_element_0=a_element;
	graph=a_element->graph;
	
	/* graph contains the top graph at end of block */
		
	begin_a_element=next_block_graph->a_stack;
	while (begin_a_element!=NULL && begin_a_element->offset<0)
		begin_a_element=begin_a_element->next;
	
	if (begin_a_element!=NULL && begin_a_element->offset==0)
	{	if (begin_a_element->load_graph!=NULL &&
			begin_a_element->load_graph->used_mark)
			return a_stack_size;
	} else
	{	if (next_block_graph->a_stack_top_offset<=0)
			return a_stack_size;
	}
	
	/* the top graph will be popped in next block */
		
	for (n=1; n<a_stack_size; ++n)
	{	a_element=a_element->next;
		if (a_element==NULL || a_element->offset!=offset+n)
			return a_stack_size;
		if (a_element->graph==graph)
			break;
	}
	
	if (n>=a_stack_size)
		return a_stack_size;

	/* the top graph has been found somewhat deeper on the stack (but within amount
	   of parameters). Remove the top element at the beginning of the next block and
	   at the end of the current block and set the offset in the jsr_eval instruction
	   accordingly. Note that the parameter will certainly be placed in a local (as
	   this routine is called with a low enough value for a_stack_size) */
		
	remove_load_a_stack_element(next_block_graph,0);
	s_pop_a (block_graph);
	
	a_element_0->graph=NULL;
 	
	block_graph->param.offset=n-1;
	
	return a_stack_size-1;
}

/********************************************************************************
 *																				*
 *		marking (un)used stack elements											*
 *																				*
 ********************************************************************************/

static VOID link_unused_a_stack_elements (register BLOCK_GRAPH *block_graph)
{
	register A_STACK *curr_element,**curr_element_p;
	register A_STACK *next_element;
	register INT16 stack_offset_difference,curr_first_non_parameter,next_first_non_parameter;
	register INT16 curr_offset,next_offset;
	BLOCK_GRAPH *next_block_graph;
	
	next_block_graph=block_graph->next;
	
	/*	insert dummy graphs in the current block for stack elements that aren't
		used anymore in following blocks */

	stack_offset_difference=	block_graph->a_stack_top_offset
								+block_graph->end_a_stack_size
								-next_block_graph->begin_a_stack_size;
	
	curr_first_non_parameter=	block_graph->a_stack_top_offset
								+block_graph->used_a_stack_elements;
	
	curr_element_p=&block_graph->a_stack;
	curr_element=*curr_element_p;
	
	next_element=next_block_graph->a_stack;
	next_first_non_parameter=curr_first_non_parameter-stack_offset_difference;
	while (next_element!=NULL && next_element->offset<next_first_non_parameter)
		next_element=next_element->next;

	/* for all non-parameters that are referenced by a stack element in the next block */
	
	for (;next_element!=NULL; next_element=next_element->next)
	{
		INSTRUCTION_GRAPH load_graph;
		
		if (next_element->flags & ELEMENT_MAY_BE_REMOVED
			|| (load_graph=next_element->load_graph)==NULL
			|| !load_graph->used_mark)
		{
			/* the element isn't used in the next block */
			
			curr_offset=next_element->offset+stack_offset_difference;
		
			/* now find corresponding element in the current block */
			
			while (curr_element!=NULL && curr_element->offset<curr_offset)
			{	curr_element_p=&curr_element->next;
				curr_element=*curr_element_p;
			}
			
			/* if the element exists nothing has to be done, else
			   a link must be made */

			if (curr_element==NULL || curr_element->offset!=curr_offset)
			{	register A_STACK *new_element;
								
				new_element=alloc_graph_mem (sizeof (A_STACK));
				
				new_element->offset		=curr_offset;
				new_element->graph		=NULL;
				new_element->load_graph	=NULL;
				new_element->flags		=ELEMENT_MAY_BE_REMOVED;

				if (next_element->flags & ELEMENT_MAY_BE_REMOVED)
					new_element->next_block_element=next_element;
				else
					new_element->next_block_element=NULL;
				
				new_element->next=curr_element;
				*curr_element_p=new_element;
				curr_element_p=&new_element->next;
			}
		}
	}
		
	curr_element_p=&block_graph->a_stack;
	curr_element=*curr_element_p;
	
	next_element=next_block_graph->a_stack;
	
	next_offset=next_first_non_parameter;
	if (next_offset<0)
		next_offset=0;
	
	/* the same for all non-parameters that are popped in the next block */
	
	for (; next_offset<next_block_graph->a_stack_top_offset; ++next_offset)
	{
		while (next_element!=NULL && next_element->offset<next_offset)
			next_element=next_element->next;
		
		if (next_element==NULL || next_element->offset!=next_offset)
		{	
			/* the element will be popped in the next block, and as no
			   stack element has been allocated to it, this means nothing
			   will be done with it (and it cannot have been encountered in the
			   round above, so...) */
			   
			/* find corresponding element in current block */
			
			while (curr_element!=NULL && curr_element->offset<next_offset+stack_offset_difference)
			{	curr_element_p=&curr_element->next;
				curr_element=*curr_element_p;
			}
			
			if (curr_element==NULL || curr_element->offset!=next_offset+stack_offset_difference){
				register A_STACK *new_element;
				
				new_element=alloc_graph_mem (sizeof (A_STACK));
				
				new_element->offset		=next_offset+stack_offset_difference;
				new_element->graph		=NULL;
				new_element->load_graph	=NULL;
				new_element->flags		=ELEMENT_MAY_BE_REMOVED;
				
				new_element->next_block_element=NULL; /* will be popped anyway */
	
				new_element->next=curr_element;
				*curr_element_p=new_element;
				curr_element_p=&new_element->next;
			} 
		}
	}
}

static VOID link_unused_b_stack_elements (register BLOCK_GRAPH *block_graph)
{
	register B_STACK *curr_element,**curr_element_p,*next_element;
	register INT16 stack_offset_difference,curr_offset,next_offset;
	register INT16 curr_first_non_parameter,next_first_non_parameter;
	BLOCK_GRAPH *next_block_graph;
	
	next_block_graph=block_graph->next;
	
	/*	insert dummy graphs in the current block for stack elements that aren't
		used anymore in following blocks (see also routine above) */

	stack_offset_difference=	block_graph->b_stack_top_offset
								+block_graph->end_b_stack_size
								-next_block_graph->begin_b_stack_size;
	
	curr_first_non_parameter=	block_graph->used_b_stack_elements
								+block_graph->b_stack_top_offset;
	
	curr_element_p=&block_graph->b_stack;
	curr_element=*curr_element_p;
	
	next_element=next_block_graph->b_stack;

	next_first_non_parameter=curr_first_non_parameter-stack_offset_difference;
	while (next_element!=NULL && next_element->offset<next_first_non_parameter)
		next_element=next_element->next;

	for (; next_element!=NULL; next_element=next_element->next)
	{	INSTRUCTION_GRAPH load_graph;
		
		if (next_element->flags & ELEMENT_MAY_BE_REMOVED || 
			((load_graph=next_element->load_graph)==NULL ||
			 (!load_graph->used_mark && 
			  !((load_graph->code==GFHIGH || 
				 load_graph->code==GFLOW) && 
				load_graph->parameters[0].pointer->used_mark))))
		{
			curr_offset=next_element->offset+stack_offset_difference;
		
			while (curr_element!=NULL && curr_element->offset<curr_offset){
				curr_element_p=&curr_element->next;
				curr_element=*curr_element_p;
			}
	
			if (curr_element==NULL || curr_element->offset!=curr_offset){
				register B_STACK *new_element;
				
				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=curr_offset;
				new_element->graph		=NULL;
				new_element->load_graph	=NULL;
				new_element->flags		=ELEMENT_MAY_BE_REMOVED;
				
				if (next_element->flags & ELEMENT_MAY_BE_REMOVED)
					new_element->next_block_element=next_element;
				else
					new_element->next_block_element=NULL;
				
				new_element->next=curr_element;
				*curr_element_p=new_element;
				curr_element_p=&new_element->next;
			}
		}
	}
	
	curr_element_p=&block_graph->b_stack;
	curr_element=*curr_element_p;
	
	next_element=next_block_graph->b_stack;
	
	next_offset=next_first_non_parameter;
	if (next_offset<0)
		next_offset=0;
	
	for (; next_offset<next_block_graph->b_stack_top_offset; ++next_offset)
	{
		while (next_element!=NULL && next_element->offset<next_offset)
			next_element=next_element->next;
			
		if (next_element==NULL || next_element->offset!=next_offset)
		{	curr_offset=next_offset+stack_offset_difference;
		
			while (curr_element!=NULL && curr_element->offset<curr_offset)
			{	curr_element_p=&curr_element->next;
				curr_element=*curr_element_p;
			}
			
			if (curr_element==NULL || curr_element->offset!=curr_offset){
				register B_STACK *new_element;
				
				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=curr_offset;
				new_element->graph		=NULL;
				new_element->load_graph	=NULL;
				new_element->flags		=ELEMENT_MAY_BE_REMOVED;
				
				new_element->next_block_element=NULL;
	
				new_element->next=curr_element;
				*curr_element_p=new_element;
				curr_element_p=&new_element->next;
			}
		}
	}
}

static VOID mark_stack_graphs_1 (BLOCK_GRAPH *block_graph,BOOL consider_next_block)
{
	INT16 stack_offset_difference,first_non_parameter_offset,offset;
	register B_STACK *b_element_1,*b_element_2,**b_element_1_p;
	register A_STACK *a_element;
	BLOCK_GRAPH *next_block_graph;
	
	next_block_graph=block_graph->next;
	
	/* mark all graphs on a-stack */
	
	for (a_element=block_graph->a_stack; 
		a_element!=NULL; a_element=a_element->next)
	{
		if (a_element->graph!=NULL)
			mark_block_graph_2 (block_graph,a_element->graph);
	}

	stack_offset_difference=	block_graph->b_stack_top_offset
								+block_graph->end_b_stack_size
								-next_block_graph->begin_b_stack_size;
	
	first_non_parameter_offset=	block_graph->b_stack_top_offset
								+block_graph->used_b_stack_elements;
	
	b_element_2=next_block_graph->b_stack;
	
	/* now mark all graphs that are reachable from the result graphs on the b-stack */
	
	for (b_element_1=block_graph->b_stack;b_element_1!=NULL; b_element_1=b_element_1->next)
	{	if (b_element_1->graph!=NULL)
		{
			if (b_element_1->offset<first_non_parameter_offset ||
				!consider_next_block)
			{
				/* this stack element is considered necessary for jsr */
				
				b_element_1->flags|=ELEMENT_USED_BEFORE_JSR;
				mark_block_graph_2 (block_graph,b_element_1->graph);
			} else
			{
				/* depending on kind of block, mark with 1 or 2 */
				
				switch(block_graph->kind)
				{	case	NORMAL_BLOCK:
					case	INCOMPLETE_BLOCK:
						internal_error("mark stack graphs 1");
					
					case	JSR:
					case 	GCALL:
					case	JSR_AP:
					case	SIGNAL:
					
					/* this block will certainly break execution sequence, next block
					   unimportant */

						b_element_1->flags|=ELEMENT_USED_BEFORE_JSR;
						mark_block_graph_2 (block_graph,b_element_1->graph);
						break;
						
					default:
					
					/* one of the blocks for wich stack layout has not yet been
					   established, so consider next block */

					{	INT16 required_offset;
						
						required_offset=b_element_1->offset-stack_offset_difference;
						
						while (b_element_2!=NULL && b_element_2->offset<required_offset)
							b_element_2=b_element_2->next;
						
						if (b_element_2!=NULL && b_element_2->offset==required_offset
							  && b_element_2->load_graph!=NULL 
							  && (b_element_2->load_graph->used_mark==2 || 
								  ((b_element_2->load_graph->code==GFHIGH
									|| b_element_2->load_graph->code==GFLOW)
								   && b_element_2->load_graph->parameters[0].pointer->used_mark==2)))
						{
							/* the element will certainly be used in the next block */
							
							b_element_1->flags|=ELEMENT_USED_BEFORE_JSR;
							mark_block_graph_2 (block_graph,b_element_1->graph);
						
						} else
						{	/* this block ends with a jsr_eval. The stack element isn't
							   used (directly) in the next block. This means that it's probably
							   better not to load this element in a local before the jsr_eval.
							   To indicate this the flag 'USED_BEFORE_JSR' isn't set. The top
							   node of the result graph is marked with '1' instead of '2' to
							   tell previous blocks that the result might not be stored in a
							   local (the rest of the graph still gets marked with 2!). Note that
							   the mark '1' only persists in the current block if the load graph
							   equals the result graph and no one else uses this result.
							*/
							   
							mark_block_graph_1 (block_graph,b_element_1->graph);
						}
						break;
					}
				}
			}
		}
	}

	/* now the result graphs of both stacks have been marked. To allow for some
	   optimisations for floats the information obtained in later blocks will be
	   passed forward. Also mark 2's of the next block will be passed forward
	   for non-existent stack elements of the current block */
	
	/* first do this for non-parameters of a block that may fall through to next
	   block */
	
	switch(block_graph->kind)
	{
		case	NORMAL_BLOCK:
		case	INCOMPLETE_BLOCK:

			internal_error("mark stack graphs 1");
					
		case	JSR:
		case 	GCALL:
		case	SIGNAL:
		case	JSR_AP:
		
			break;

		default:
		{
			b_element_1_p=&block_graph->b_stack;
			b_element_1=*b_element_1_p;
		
			for (b_element_2=next_block_graph->b_stack;
				b_element_2!=NULL; b_element_2=b_element_2->next)
			{
				INSTRUCTION_GRAPH load_graph,next_load_graph,graph_1,next_graph_1;
				B_STACK *next_b_element_2,*next_b_element_1;
	
				load_graph=b_element_2->load_graph;
				if (load_graph!=NULL &&
					(load_graph->used_mark==2 || 
						((load_graph->code==GFHIGH || load_graph->code==GFLOW)
						 && load_graph->parameters[0].pointer->used_mark==2)) && 
					(offset=b_element_2->offset+stack_offset_difference)>=first_non_parameter_offset)
				{	
					/* the next element is not a parameter and certainly used */
					
					while (b_element_1!=NULL && b_element_1->offset<offset)
					{	b_element_1_p=&b_element_1->next;
						b_element_1=*b_element_1_p;
					}
					
					if (load_graph->code==GFLOW && 
						(next_b_element_2=b_element_2->next)!=NULL && 
						next_b_element_2->offset==b_element_2->offset+1 && 
						(next_load_graph=next_b_element_2->load_graph)!=NULL && 
						next_load_graph->code==GFHIGH && 
						next_load_graph->parameters[0].pointer==load_graph->parameters[0].pointer &&
						b_element_1!=NULL && b_element_1->offset==offset &&
						(graph_1=b_element_1->graph)!=NULL && 
						graph_1->code==GLOAD_NL && 
						(next_b_element_1=b_element_1->next)!=NULL && 
						next_b_element_1->offset==offset+1 && 
						(next_graph_1=next_b_element_1->graph)!=NULL && 
						next_graph_1->code==GLOAD_NL && 
						graph_1->parameters[0].pointer==next_graph_1->parameters[0].pointer && 
						graph_1->parameters[1].int16+1==next_graph_1->parameters[1].int16)
					{
						INSTRUCTION_GRAPH f_graph;
												
						/* in the next block the stack contains LOW and HIGH load graphs
						   for the same float and the current block contains two LOAD_NL
						   result graphs for neighbouring memory locations on the stack
						*/

						f_graph=g_fload_nl(graph_1->parameters[0].pointer,graph_1->parameters[1].int16);
						f_graph->used_mark=2;
						
						/* change the load_nl nodes into LOW and HIGH nodes */
						
						graph_1->code=GFLOW;
						graph_1->parameters[0].pointer=f_graph;
						graph_1->used_mark=2;
						
						next_graph_1->code=GFHIGH;
						next_graph_1->parameters[0].pointer=f_graph;
						next_graph_1->used_mark=2;
					
					} else if (b_element_1==NULL || b_element_1->offset!=offset)
					{
						/* a corresponding element couldn't be found in current block */
						
						if (load_graph->code==GFLOW &&
							(next_b_element_2=b_element_2->next)!=NULL &&
							next_b_element_2->offset==b_element_2->offset+1 && 
							(next_load_graph=next_b_element_2->load_graph)!=NULL && 
							next_load_graph->code==GFHIGH && 
							next_load_graph->parameters[0].pointer==load_graph->parameters[0].pointer && 
							(b_element_1==NULL || b_element_1->offset!=offset+1))
						{
							/* a LOW and HIGH element on next stack, but both corresponding
							   elements not on current stack, so build them */
							
							register B_STACK *new_element;
							INSTRUCTION_GRAPH f_graph,l_graph,h_graph;
										
							f_graph=g_fload_bs(block_graph,offset+block_graph->b_stack_begin_displacement);
							f_graph->used_mark=2;
							
							h_graph=g_fhigh (f_graph);
							h_graph->used_mark=2;
							l_graph=g_flow (f_graph);
							l_graph->used_mark=2;
							
							new_element=alloc_graph_mem (sizeof (B_STACK));
							
							new_element->offset		=offset;
							new_element->graph		=l_graph;
							new_element->load_graph	=l_graph;
							new_element->flags		=ELEMENT_USED_BEFORE_JSR;
						
							new_element->next=b_element_1;
							*b_element_1_p=new_element;
							b_element_1_p=&new_element->next;
	
							new_element=alloc_graph_mem (sizeof (B_STACK));
							
							new_element->offset		=offset+1;
							new_element->graph		=h_graph;
							new_element->load_graph	=h_graph;
							new_element->flags		=ELEMENT_USED_BEFORE_JSR;
						
							new_element->next=b_element_1;
							*b_element_1_p=new_element;
							b_element_1_p=&new_element->next;
							
							b_element_2=next_b_element_2;
						} else
						{
							register B_STACK *new_element;
							INSTRUCTION_GRAPH graph;
					
							/* no corresponding element in current block */
							
							graph=g_load_bs(block_graph,offset+block_graph->b_stack_begin_displacement);
							graph->used_mark=2;
							
							new_element=alloc_graph_mem (sizeof (B_STACK));
							
							new_element->offset		=offset;
							new_element->graph		=graph;
							new_element->load_graph	=graph;
							new_element->flags		=ELEMENT_USED_BEFORE_JSR;
						
							new_element->next=b_element_1;
							*b_element_1_p=new_element;
							b_element_1_p=&new_element->next;
						}
					}
				}
			}
		}
	}
	
	b_element_1_p=&block_graph->b_stack;
	b_element_1=*b_element_1_p;
	
	/* now do the same as above, but now for parameters of any block */
	
	for (offset=block_graph->b_stack_top_offset; offset<first_non_parameter_offset; ++offset)
	{
		while (b_element_1!=NULL && b_element_1->offset<offset)
		{	b_element_1_p=&b_element_1->next;
			b_element_1=*b_element_1_p;
		}
			
		if (b_element_1==NULL || b_element_1->offset!=offset)
		{	INT16 n;
			
			n=offset-block_graph->b_stack_top_offset;
			
			if ((unsigned)n < (unsigned)block_graph->end_b_stack_size
				&& is_vector_bit_set (block_graph->end_stack_vector,n)
				&& (b_element_1==NULL || b_element_1->offset!=offset+1))
			{
				/* the parameter should be a float, but both words don't exist yet */
				
				register B_STACK *new_element;
				INSTRUCTION_GRAPH f_graph,l_graph,h_graph;
		
				f_graph=g_fload_bs(block_graph,offset+block_graph->b_stack_begin_displacement);
				f_graph->used_mark=2;
				
				h_graph=g_fhigh (f_graph);
				h_graph->used_mark=2;
				l_graph=g_flow (f_graph);
				l_graph->used_mark=2;
				
				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=offset;
				new_element->graph		=l_graph;
				new_element->load_graph	=l_graph;
				new_element->flags		=ELEMENT_USED_BEFORE_JSR;
			
				new_element->next=b_element_1;
				*b_element_1_p=new_element;
				b_element_1_p=&new_element->next;

				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=offset+1;
				new_element->graph		=h_graph;
				new_element->load_graph	=h_graph;
				new_element->flags		=ELEMENT_USED_BEFORE_JSR;
			
				new_element->next=b_element_1;
				*b_element_1_p=new_element;
				b_element_1_p=&new_element->next;
				
				++offset;
			} else
			{	/* just insert a stack element */
			
				register B_STACK *new_element;
				INSTRUCTION_GRAPH graph;
		
				graph=g_load_bs(block_graph,offset+block_graph->b_stack_begin_displacement);
				graph->used_mark=2;
				
				new_element=alloc_graph_mem (sizeof (B_STACK));
				
				new_element->offset		=offset;
				new_element->graph		=graph;
				new_element->load_graph	=graph;
				new_element->flags		=ELEMENT_USED_BEFORE_JSR;
				
				new_element->next=b_element_1;
				*b_element_1_p=new_element;
				b_element_1_p=&new_element->next;
			}
		}
	}
}

static VOID mark_stack_graphs_2(BLOCK_GRAPH *block_graph)
{
	register A_STACK *a_element;
	register B_STACK *b_element;
		
	/* mark the results graphs of both stacks with 2 */
	
	for (a_element=block_graph->a_stack; a_element!=NULL; a_element=a_element->next)
	{	if (a_element->graph!=NULL)
			mark_block_graph_2 (block_graph,a_element->graph);
	}

	for (b_element=block_graph->b_stack; b_element!=NULL; b_element=b_element->next)
	{	if (b_element->graph!=NULL)
		{	b_element->flags|=ELEMENT_USED_BEFORE_JSR;
			mark_block_graph_2 (block_graph,b_element->graph);
		}
	}
}

/********************************************************************************
 *																				*
 *		adjustment of load and store nodes with respect to the stack ptr		*
 *																				*
 ********************************************************************************/

static VOID adjust_a_stack_offsets (BLOCK_GRAPH *block_graph,INT16 stack_offset)
{
	register A_STACK *a_element;
	INT16 begin_displacement;
	
	begin_displacement=block_graph->a_stack_begin_displacement;
	
	for (a_element=block_graph->a_stack; a_element!=NULL; a_element=a_element->next)
	{	register INSTRUCTION_GRAPH load_graph, result_graph;
				
		load_graph=a_element->load_graph;
		if (load_graph!=NULL)
		{	switch (load_graph->code)
			{	case GLOAD_NL:
					load_graph->parameters[1].int16=-(1+a_element->offset+begin_displacement)+stack_offset;
					break;
				case GLOCAL:
					break;
				default:
					internal_error("adjust_a_stack_offsets (load)");
			}
		}
		result_graph=a_element->graph;
		if (result_graph!=NULL)
		{	switch (result_graph->code)
			{	case GSTORE_NL:
					result_graph->parameters[2].int16+=stack_offset;
				case GSTORE_L:
					break;
				case GFILL:
					break;

					/* apparently the address of the node doesn't have to be stored */
										
				default:
					internal_error("adjust_a_stack_offsets (store) invalid node: %d",result_graph->code);
			}
		}
	}
}

static VOID adjust_b_stack_offsets (BLOCK_GRAPH *block_graph,INT16 stack_offset)
{
	register B_STACK *b_element;
	INT16 begin_displacement;
	
	begin_displacement=block_graph->b_stack_begin_displacement;
	
	for (b_element=block_graph->b_stack; b_element!=NULL; b_element=b_element->next)
	{	register INSTRUCTION_GRAPH load_graph, result_graph;

		load_graph=b_element->load_graph;
		if (load_graph!=NULL)
		{	switch (load_graph->code)
			{	case GLOAD_NL:
					load_graph->parameters[1].int16=b_element->offset+begin_displacement+stack_offset;
					break;
					
				case GFLOW:
				{
					INSTRUCTION_GRAPH float_graph;
					
					float_graph=load_graph->parameters[0].pointer;
					if (float_graph->code==GFLOAD_NL)
						float_graph->parameters[1].int16=b_element->offset+begin_displacement+stack_offset;
					else if (float_graph->code!=GFLOAT_LOCAL)
						internal_error("adjust_b_stack_offsets (load LOW)");
								
					break;
				}
				case GFHIGH:
				{
					INSTRUCTION_GRAPH float_graph;
					
					float_graph=load_graph->parameters[0].pointer;
					if (float_graph->code==GFLOAD_NL)
						float_graph->parameters[1].int16=b_element->offset-1+begin_displacement+stack_offset;
					else if (float_graph->code!=GFLOAT_LOCAL)
						internal_error("adjust_b_stack_offsets (load HIGH)");
					
					break;
				}

				case GLOCAL:
				case GFLOAT_LOCAL:
					break;

				default:
					internal_error("adjust_b_stack_offsets: invalid load graph: %d",load_graph->code);
			}
		}
		result_graph=b_element->graph;			
		if (result_graph!=NULL)
		{	switch (result_graph->code)
			{	case GSTORE_NL:
				case GFSTORE_NL:
					result_graph->parameters[2].int16+=stack_offset;
					break;
				case GSTORE_L:
				case GFSTORE_L:
					break;
				
				default:
					internal_error("adjust_b_stack_offsets");
			}
		}
	}
}

static VOID adjust_stack_offsets(BLOCK_GRAPH *block_graph)
{
	INT16 a_increase,b_increase;
	
	/* stack_top_offset starts at 0, this means that this value, combined with the
	   difference in start and end displacement indicates the increase/decrease that
	   is needed for the stack pointer. Note that this routine can only be called
	   after store nodes have been generated, as only then it's known what the
	   stack displacement at the end will be
	*/
	
	a_increase=-(	block_graph->a_stack_top_offset+
					block_graph->a_stack_begin_displacement-
					block_graph->a_stack_end_displacement);
	b_increase=-(	block_graph->b_stack_top_offset+
					block_graph->b_stack_begin_displacement-
					block_graph->b_stack_end_displacement);
	
	block_graph->a_increase=a_increase;
	block_graph->b_increase=b_increase;
	
	/* If one assumes the stack pointer is only adjusted at the end of each block,
	   the following holds:
	   -	offsets within store_nl nodes of result graphs are correct with
	   		respect to the stack pointer.
	   -	offsets within load_nl nodes of load graphs are incorrect with respect
	   		to the stack pointer (due to removal of unused elements). The offset
	   		of the stack element they are referenced by is correct however (see
	   		comment above remove/insert routines).

	   BUT: the stack pointer will not always be adjusted at the end of each block.
	   Sometimes it is better to do this at the beginning to keep offsets positive
	   as much as possible. In fact the stack top pointer should always point to
	   the lowest address possible. This means that it is adjusted at the start of
	   a block when:
	   -	the a-stack shrinks	(a-stack grows from low to high)
	   -	the b-stack grows	(b-stack grows grom high to low)
	   in these cases the offsets of the affected store_nl/load_nl nodes should
	   be increased by the amount the stack pointer moves.
	*/
	
	if (a_increase >= 0)
		adjust_a_stack_offsets(block_graph,0);
	else
		adjust_a_stack_offsets(block_graph,-a_increase);
	
	if (b_increase > 0)
		adjust_b_stack_offsets(block_graph, b_increase);
	else
		adjust_b_stack_offsets(block_graph, 0);
}

/************************************************************************************
 *																					*
 *		finish some block after linearisation										*
 *																					*
 ************************************************************************************/

static VOID set_stack_check_and_end_instruction(BLOCK_GRAPH *block_graph)
{
	/* determines (after linearisation) the amount of stack space to check per
	   block, the stack use and the instruction that must be appended at inserted
	   blocks */
	
	INSTRUCTION_BLOCK *instruction_block;
	INT16 a_increase,b_increase;
	
	instruction_block=block_graph->instruction_block;
	
	a_increase=block_graph->a_increase;
	b_increase=block_graph->b_increase;

	/*	compute basic stack use of block, i.e. if the block increase the total
		amount of stack space, or decreases it */
	
	block_graph->stack_use=a_increase+b_increase;

	/* compute least amount of stack space to check for this block */
	
	instruction_block->check_size+=(b_increase>0 ? b_increase : 0)+(a_increase>0 ? a_increase : 0);
	
	switch(block_graph->kind)
	{
		/*	for all instructions that may increase/decrease the amount of parameters
			on the stack: increase/decrease stack use accordingly. This is only
			necessary for jsr, call, and signal instructions The check size doesn't have
			to be changed as this block won't push/pop these values; it's only
			important for following blocks if checks are combined */
	
		case JSR:
		case GCALL:
		case SIGNAL:
			
		{	BLOCK_GRAPH *next_block_graph;
	
			next_block_graph=block_graph->next;
			if(next_block_graph==NULL)
				internal_error("determine_stack_check_size");

			/* increase of parameters not in locals before and after counts */
									
			a_increase=
				(next_block_graph->begin_a_stack_size+
				 next_block_graph->a_stack_begin_displacement)-
				(block_graph->end_a_stack_size+
				 block_graph->a_stack_end_displacement);
			
			b_increase=
				(next_block_graph->begin_b_stack_size+
				 next_block_graph->b_stack_begin_displacement)-
				(block_graph->end_b_stack_size+
				 block_graph->b_stack_end_displacement);

			block_graph->stack_use+=(a_increase+b_increase);

			/* the amount of stack space that is certainly free after shrinking a
			   stack is the value left free by the runtime system minus the parameters
			   that have been pushed on the stack by the called routine */
			
			next_block_graph->max_check_size=MAX_CHECK_SIZE
				-(next_block_graph->begin_a_stack_size+
				next_block_graph->a_stack_begin_displacement)
				-(next_block_graph->begin_b_stack_size+
				next_block_graph->b_stack_begin_displacement);

			/*	append correct instruction, note that the increase of the b-stack must
				be appended as well, so that the new c-top offset can be computed */
			
			switch (block_graph->kind)
			{
				case JSR:
				case GCALL:
					append_instr_label_int16(instruction_block,block_graph->kind,block_graph->param.label,b_increase);
					break;
				case SIGNAL:
					append_instr_short2(instruction_block,block_graph->kind,block_graph->param.number,b_increase);
					break;
				default:
					internal_error("set_stack_check_and_end_instruction");
			}				
			break;
		}
		case JSR_AP:
		{	
			BLOCK_GRAPH *next_block_graph;

			/* doesn't increase parameters on stack, no params left on stack */
			
			next_block_graph=block_graph->next;
			if(next_block_graph==NULL)
				internal_error("determine_stack_check_size");

			next_block_graph->max_check_size=MAX_CHECK_SIZE;
			
			append_instr_label(instruction_block,block_graph->kind,block_graph->param.label);
			
			break;
		}
		
		/* for these 3 instructions no locals have to be saved */

		case SUSPEND:
		case SUSPEND_UNEVAL:
		
		case CS:
		case ENABLE_CS:
		case DISABLE_CS:
		
		{
			BLOCK_GRAPH *next_block_graph;

			/* doesn't increase parameters on stack, no params left on stack */
			
			next_block_graph=block_graph->next;
			if(next_block_graph==NULL)
				internal_error("determine_stack_check_size");

			next_block_graph->max_check_size=MAX_CHECK_SIZE;

			switch (block_graph->kind)
			{
				case SUSPEND:
				
				case CS:
				case ENABLE_CS:
				case DISABLE_CS:

					append_instr_void(instruction_block,block_graph->kind);
					break;

				case SUSPEND_UNEVAL:
					append_instr_short(instruction_block,block_graph->kind,block_graph->param.offset);
					break;					
			}
			break;
		}
		
		/* for the following blocks some locals have to be saved if the evaluation
		   routine is called */
		
		case JSR_EVAL:
		case RELEASE:
		{
			BLOCK_GRAPH *next_block_graph;
			register INT16 b_stack_size, n_b_locals, n;
			INT16 n_a_locals;
			U_INT32 *vector;
			INT16 check_size;
			
			next_block_graph=block_graph->next;
			if(next_block_graph==NULL)
				internal_error("determine_stack_check_size");
		
			vector=block_graph->end_stack_vector;
			b_stack_size=block_graph->end_b_stack_size;
			
			n_b_locals=0;
			for (n=0; n<b_stack_size; ++n)
			{	if (is_vector_bit_set(vector,n))
				{	if (n_b_locals<N_B_PARAMETER_LOCALS-1)
						n_b_locals+=2;
					++n;
				} else
				{	if (n_b_locals<N_B_PARAMETER_LOCALS)
						++n_b_locals;
				}
			}
			n_a_locals=block_graph->end_a_stack_size;
			if(n_a_locals>N_A_PARAMETER_LOCALS)
				n_a_locals=N_A_PARAMETER_LOCALS;
			
			/* amount to check: a_size-1 + b_size +1 (-param + return address) */
			
			check_size=n_a_locals+n_b_locals;
			
			/*	test if this amount must really be checked: if the block does not increase the
				stack as much as the amount tested for earlier, the space left on the stack
				may be suficient to store the parameters and no check is needed. Note that
				if there's not enough space left the check size should remain the same */
			
			if (options.optimise_stack_checks &&
				block_graph->stack_use < block_graph->instruction_block->check_size &&
				check_size <= (block_graph->instruction_block->check_size-block_graph->stack_use))
			{
				check_size=0;
			}
			
			/* it's possible to consider somewhat more stack available: if no eval is
			   performed the stack will not be shrunk and otherwise the space that
			   was used to save the locals will become available too (after return) */
			
			next_block_graph->max_check_size=MAX_CHECK_SIZE+n_a_locals+n_b_locals;

			block_graph->param.instruction=append_instr_short4(instruction_block,block_graph->kind,block_graph->param.offset,check_size,n_a_locals,n_b_locals);

			break;
		}
		default:
			/* incomplete or normal block: do nothing, it may not be ready yet */
			
			break;
	}
}

/****************************************************************************************
 *                                                                                      *
 *								basic blocks											*
 *																						*
 ****************************************************************************************/

static INSTRUCTION_BLOCK *evaluate_stacks (BLOCK_GRAPH *block_graph)
{	
	register INT16 element_n,n_a_elements,n_b_elements;
	register A_STACK *a_element;
	register B_STACK *b_element;
	register INSTRUCTION_GRAPH *graphs;
	INT16 n_elements;

	/* takes all elements of the stack structures and puts these in an array to
	   generate code for these elements */
	   
	n_a_elements=0;
	for (a_element=block_graph->a_stack; a_element!=NULL; a_element=a_element->next)
		if (a_element->graph!=NULL)
			++n_a_elements;
	
	n_b_elements=0;
	for (b_element=block_graph->b_stack; b_element!=NULL; b_element=b_element->next)
		if (b_element->graph!=NULL)
			++n_b_elements;
	
	graphs=alloc_block_mem((n_a_elements+n_b_elements)*sizeof (INSTRUCTION_GRAPH));
	
	/* for the M68000 it matters in which sequence stack elems are evaluated, as
	   the stack pointer is adjusted 'along' the way. For the tramsputer this is
	   irrelevant */
	
	element_n=n_b_elements;
	for (a_element=block_graph->a_stack; a_element!=NULL; a_element=a_element->next)
		if (a_element->graph!=NULL)
			graphs[element_n++]=a_element->graph;
			
	element_n=0;
	for (b_element=block_graph->b_stack; b_element!=NULL; b_element=b_element->next)
		if (b_element->graph!=NULL)
			graphs[element_n++]=b_element->graph;
	
	n_elements=n_a_elements+n_b_elements;
		
	for (element_n=0; element_n<n_elements; ++element_n)
	{	register INSTRUCTION_GRAPH graph;
		
		graph=graphs[element_n];
		
		if (graph->code!=GFILL)
			count_block_graph (block_graph,graph);
		else
		{	register INT16 n;
						
			/* node address doesn't have to be stored anywhere */
			
			if (graph->parameters[0].pointer!=NULL)
				count_block_graph(block_graph,graph->parameters[0].pointer);
				
				/* skip wait_graph */
			
			for (n=2; n<graph->arity; ++n)
				if (graph->parameters[n].pointer!=NULL)
					count_block_graph (block_graph,graph->parameters[n].pointer);
		}
	}
	
	return lin_block (block_graph,n_elements,graphs);
}

static VOID evaluate_block (register BLOCK_GRAPH *block_graph)
{
	load_nodes_for_end_a_params (block_graph);
	load_nodes_for_end_b_params (block_graph);
		
	store_a_stack_values (block_graph);
	store_b_stack_values (block_graph);
	
	adjust_stack_offsets(block_graph);
	
	block_graph->instruction_block=evaluate_stacks(block_graph);
	
	set_stack_check_and_end_instruction(block_graph);
}

static VOID evaluate_previous_blocks (BLOCK_GRAPH *last_block_graph)
{
	BLOCK_GRAPH *curr_block_graph;
	BOOL consider_next_block;
	
	/* for the current block no code will be generated yet. It will only be used
	   to optimise stack access. */
	
	/* default end layout for last block */

	set_block_end_layout(last_block_graph,0,0,e_vector);
	mark_stack_graphs_2 (last_block_graph);

	if (last_block_graph->previous==NULL)
	{
		/* no previous blocks */
		
		return;
	}
	
	/* mark all previous blocks from the last-1 downto the first */
	
	consider_next_block=last_block_graph->kind!=INCOMPLETE_BLOCK;
	for (curr_block_graph=last_block_graph->previous;curr_block_graph!=NULL;curr_block_graph=curr_block_graph->previous)
	{	
		link_unused_a_stack_elements (curr_block_graph);
		link_unused_b_stack_elements (curr_block_graph);

		mark_stack_graphs_1 (curr_block_graph,consider_next_block);
		
		if (curr_block_graph->kind==JSR ||
			curr_block_graph->kind==GCALL ||
			curr_block_graph->kind==JSR_AP ||
			curr_block_graph->kind==SIGNAL)
			consider_next_block=true;
	}
	
	/* remove unused stack elements from the first block until last block (thus
	   removing an element as soon as it becomes unused) */
	
	for (curr_block_graph=first_block_graph;curr_block_graph!=last_block_graph;curr_block_graph=curr_block_graph->next)
	{	
		if (options.optimise_a_stack)
			remove_unused_a_stack_elements (curr_block_graph);
		if (options.optimise_b_stack)
			remove_unused_b_stack_elements (curr_block_graph);
	}
	
	/* determine the amount of parameters for blocks for which this has not been
	   established yet */
	
	for (curr_block_graph=first_block_graph;curr_block_graph!=last_block_graph;curr_block_graph=curr_block_graph->next)
	{
		switch(curr_block_graph->kind)
		{
			case	JSR_EVAL:
			case	RELEASE:
			case	SUSPEND_UNEVAL:
			{
				/* these blocks use a parameter on top of the a-stack (and leave it
				   there) The stack layout has been delayed */
				
				BLOCK_GRAPH *next_block_graph;
				INT16 a_stack_size,b_stack_size;
				U_INT32 *vector;
				
				next_block_graph=curr_block_graph->next;
				if (next_block_graph==NULL)
					internal_error("evaluate_previous_blocks");
	
				a_stack_size=count_a_params (curr_block_graph);
				
				if (a_stack_size<1)
					a_stack_size=1;
				else
					if (a_stack_size>N_A_PARAMETER_LOCALS+1)
						a_stack_size=N_A_PARAMETER_LOCALS+1;
				
				b_stack_size=count_b_params_2(curr_block_graph);
				vector=curr_block_graph->end_stack_vector;
	
				/* determine local that will hold parameter in the end */
				
				curr_block_graph->param.offset=0;
				if (options.optimise_a_parameter)
					a_stack_size=try_remove_a_top_param(curr_block_graph,a_stack_size);
	
				if (a_stack_size>N_A_PARAMETER_LOCALS)
					a_stack_size=N_A_PARAMETER_LOCALS;
				
				/* allocate minimum amount of parameters */
				
				curr_block_graph->end_a_stack_size=a_stack_size;
				curr_block_graph->end_b_stack_size=b_stack_size;
							
				set_block_begin_layout(next_block_graph,a_stack_size,b_stack_size,vector);
				
				load_a_param_locals_2(next_block_graph);
				load_b_param_locals_2(next_block_graph);
				
				break;
			}
						
			case SUSPEND:
			case CS:
			case ENABLE_CS:
			case DISABLE_CS:	
			{
				/* these blocks don't use a parameter, but stack layout has been
				   delayed */
				
				BLOCK_GRAPH *next_block_graph;
				INT16 a_stack_size,b_stack_size;
				U_INT32 *vector;
				
				next_block_graph=curr_block_graph->next;
				if (next_block_graph==NULL)
					internal_error("evaluate_previous_blocks");
	
				a_stack_size=count_a_params (curr_block_graph);
				if (a_stack_size>N_A_PARAMETER_LOCALS)
					a_stack_size=N_A_PARAMETER_LOCALS;
				
				b_stack_size=count_b_params_2(curr_block_graph);
				vector=curr_block_graph->end_stack_vector;
					
				/* allocate minimum amount of parameters */
				
				curr_block_graph->end_a_stack_size=a_stack_size;
				curr_block_graph->end_b_stack_size=b_stack_size;
							
				set_block_begin_layout(next_block_graph,a_stack_size,b_stack_size,vector);
				
				load_a_param_locals_2(next_block_graph);
				load_b_param_locals_2(next_block_graph);
				
				break;
			}
			default:
				break;
		}
	}

	/* generate code for each block from low to high */
		
	for (curr_block_graph=first_block_graph;curr_block_graph!=last_block_graph;curr_block_graph=curr_block_graph->next)
	{	evaluate_block(curr_block_graph);
		block_ready(curr_block_graph,false);
	}
}

/************************************************************************************
 *																					*
 *		allocation/deallocation of block graphs										*
 *																					*
 ************************************************************************************/

static VOID free_graph_blocks(VOID)
{
	/* code has been generated for all previous block graphs. This means that
	   block graphs and nodes can be freed. Note that this routine should be
	   called after local allocation (to be specific after combination of stack
	   checks) unless no code will be generated. It is called by the combine_stack_checks
	   routine and when no code is generated (when code is not reachable) by the
	   end_block/insert_block routines below. */
	
	free_graph_mem();
	
	first_block_graph=NULL;
	last_block_graph=NULL;
}

static BLOCK_GRAPH *allocate_block(VOID)
{
	register BLOCK_GRAPH *new_block;
	
	new_block=alloc_graph_mem (sizeof (BLOCK_GRAPH));
	
	empty_stacks(new_block);

	init_heap_pointer(new_block);
	init_a_stack_pointer(new_block);
	init_b_stack_pointer(new_block);
	
	new_block->start_label=NULL;
	new_block->start_descriptor=NULL;
	new_block->start_arity=-1;
	
	if (last_block_graph==NULL)
		first_block_graph=new_block;
	else
		last_block_graph->next=new_block;
	
	new_block->previous=last_block_graph;
	new_block->next=NULL;
	last_block_graph=new_block;

	return new_block;
}

/************************************************************************************
 *																					*
 *		combination of (conditional) stack checks									*
 *		this routine is called after local allocation								*
 *																					*
 ************************************************************************************/

static BOOL unconditional_check_needed(register BLOCK_GRAPH *block_graph)
{
	register INT16 check_size;
	
	/*	tells if there's need for an unconditional stack check before the
		given block. Note that we cannot stop as soon as the check size
		becomes positive, as it may decrease as well (when the first block
		pops many stack elements for instance) */
	
	if(block_graph==NULL)
		return false;
	
	check_size=0;
	while(block_graph!=first_block_graph)
	{
		/* checks are shifted forward as much as possible */
		
		if(block_graph->instruction_block->check_size > check_size)
			check_size=block_graph->instruction_block->check_size;
		
		/*	if this amount exceeds the maximum check size, there is certainly
			a check needed here. */
		
		if(check_size > block_graph->max_check_size)
			return true;
		
		block_graph=block_graph->previous;
		
		check_size+=block_graph->stack_use;
	}
	if(check_size>0)
		return true;
	
	return block_graph->instruction_block->check_size > 0;
}

static BLOCK_GRAPH *insert_unconditional_check(register BLOCK_GRAPH *block_graph)
{
	register INT16 check_size;
	register INT16 block_check_size;
	
	/* inserts a stack check before the current block. it is assumed that
	   checks for following blocks have been inserted already. As an unconditional
	   check is needed anyway, conditional checks are combined with this
	   block. the block at which the check is inserted is returned */
	
	check_size=0;
	while(block_graph!=first_block_graph)
	{	
		if (block_graph->kind==JSR_EVAL || block_graph->kind==RELEASE)
		{
			/*	possibly a conditional stack check needed here,
				determine amount of stack needed for block alone */
			
			block_check_size=block_graph->stack_use+block_graph->param.instruction->parameters[1].int16;
						
			if (block_graph->instruction_block->check_size>block_check_size)
				block_check_size=block_graph->instruction_block->check_size;
			
			block_graph->param.instruction->parameters[1].int16=0;
			
		} else
			block_check_size=block_graph->instruction_block->check_size;

		if(block_check_size>check_size)
			check_size=block_check_size;
		
		if(check_size > block_graph->max_check_size)
		{
			/* stack check can't be forwarded any more, so insert it here and
			   return previous block */
			
			block_graph->instruction_block->check_size=check_size;
			return block_graph;
		}
		
		/* forward stack check and go on */
		
		block_graph->instruction_block->check_size=0;
		block_graph=block_graph->previous;
		check_size+=block_graph->stack_use;
	}
	
	/* we've arrived at the first block: insert a check here */
		
	if (block_graph->kind==JSR_EVAL || block_graph->kind==RELEASE)
	{
		/* test for condional stack check */
		
		block_check_size=block_graph->stack_use+block_graph->param.instruction->parameters[1].int16;
		if (block_graph->instruction_block->check_size>block_check_size)
			block_check_size=block_graph->instruction_block->check_size;
		
		block_graph->param.instruction->parameters[1].int16=0;
		
	} else
		block_check_size=block_graph->instruction_block->check_size;

	block_graph->instruction_block->check_size=max(check_size,block_check_size);
	
	return block_graph;
}

VOID combine_stack_checks(BLOCK_GRAPH *last_block_graph)
{
	/* stack stacks are combined from back to front. Within a single block there's
	   no limit on the amount of stack that may be checked at once. If several blocks
	   are combined there is. This has to do with shrinking stacks: when a stack is
	   made smaller, a certain amount (say 1K) is kept free. This means that one
	   has to make sure that blocks that follow a block with a stack check do not
	   use more memory than this amount alltogether. This routine uses two passes:
	   the first one determines if unconditional stack stacks are needed or not
	   (normally this boils down to the question whether the stack is checked right at
	   the beginning of a sequence of blocks or not). If only conditional stack checks
	   are used, these are left as they are. If an unconditional stack check is needed
	   the conditional stack checks are combined with these checks */
	
	if (!options.optimise_stack_checks)
	{
		free_graph_blocks();
		return;
	}
	
	while(unconditional_check_needed(last_block_graph))
	{
		last_block_graph=insert_unconditional_check(last_block_graph);
		
		if(last_block_graph==first_block_graph)
		{
			free_graph_blocks();
			return;
		}
		
		last_block_graph=last_block_graph->previous;
	}
	
	/* leave all (conditional) checks as they are */
	
	free_graph_blocks();
}

/********************************************************************************
 *																				*
 *		begin_block/insert_block/end_block routines								*
 *																				*
 ********************************************************************************/

/* Begin block routines link new block graphs in the list end sets their start
   layout (i.e. parameters that are in locals for sure are initialised). Each
   routine returns the new block graph */

BLOCK_GRAPH *begin_layout_block (INT16 a_stack_size, INT16 b_stack_size, U_INT32 vector [])
{
	register BLOCK_GRAPH *new_block;
	
	new_block=allocate_block();
	
	set_block_begin_layout(new_block,a_stack_size,b_stack_size,vector);
	
	load_a_param_locals (new_block);
	load_b_param_locals (new_block);
	
	return new_block;
}

BLOCK_GRAPH *begin_label_layout_block (LABEL *label)
{
	register BLOCK_GRAPH *new_block;
	
	new_block=allocate_block();
	
	if (label->flags & FIXED_LOCALS)
	{	set_block_begin_layout(new_block,label->a.stack_size,label->b.stack_size,label->vector);
		load_a_param_locals (new_block);
		load_b_param_locals (new_block);
			
		return new_block;
	}
	internal_error("begin label layout: layout of label unknown");
	return NULL;
}

/* End block routines generate code for the current block graph. For insert blocks this
   is delayed (jsr...).  */

VOID insert_label_layout_block(BLOCK_GRAPH *block_graph,INT16 kind,LABEL *label)
{	
	if (!code_reachable)
	{	free_graph_blocks();
		return;
	}
	
	block_graph->kind=kind;
	block_graph->used_a_stack_elements=label->a.stack_size;
	block_graph->used_b_stack_elements=label->b.stack_size;
	
	block_graph->param.label=label;

	set_block_end_layout(block_graph,label->a.stack_size,label->b.stack_size,label->vector);
}

VOID insert_layout_block(BLOCK_GRAPH *block_graph,INT16 kind,INT16 n,INT16 n_a_parameters,INT16 n_b_parameters,U_INT32 vector[])
{
	if (!code_reachable)
	{	free_graph_blocks();
		return;
	}
	
	block_graph->kind=kind;
	block_graph->used_a_stack_elements=n_a_parameters;
	block_graph->used_b_stack_elements=n_b_parameters;
	
	block_graph->param.number=n;
	
	set_block_end_layout(block_graph,n_a_parameters,n_b_parameters,vector);
}

/* blocks that end with using a parameter from the a-stack */

VOID insert_delay_layout_a_param_block(BLOCK_GRAPH *block_graph,INT16 kind)
{	
	if (!code_reachable)
	{	free_graph_blocks();
		return;
	}
	
	block_graph->kind=kind;
	block_graph->used_a_stack_elements=1;
	block_graph->used_b_stack_elements=0;

	set_block_end_layout(block_graph,0,0,e_vector);
}

VOID insert_delay_layout_block(BLOCK_GRAPH *block_graph,INT16 kind)
{	
	if (!code_reachable)
	{	free_graph_blocks();
		return;
	}
	
	block_graph->kind=kind;
	block_graph->used_a_stack_elements=0;
	block_graph->used_b_stack_elements=0;

	set_block_end_layout(block_graph,0,0,e_vector);
}

/*	blocks that are the last in a sequence of blocks. If the block is
	reachable, the previous block are as well and code must be generated for
	them. If this block is not reachable, no code should be generated. One
	should not throw away the block yet. It will get the previous instruction
	block in one of its fields. This can be used to append some instructions
	if one likes (inline instuctions for example). The block administration does
	have to be reset however. */

VOID end_layout_block(BLOCK_GRAPH *block_graph,INT16 n_a_parameters,INT16 n_b_parameters,U_INT32 vector[])
{
	if (code_reachable)
	{	
		block_graph->kind=NORMAL_BLOCK;
		
		evaluate_previous_blocks(block_graph);
		
		set_block_end_layout(block_graph,n_a_parameters,n_b_parameters,vector);
		
		evaluate_block(block_graph);
				
	} else
	{
		block_graph->instruction_block=get_current_block();
		first_block_graph=NULL;
		last_block_graph=NULL;
	}
}

VOID end_label_layout_block(BLOCK_GRAPH *block_graph,LABEL *label)
{
	if (code_reachable)
	{
		block_graph->kind=NORMAL_BLOCK;
		
		evaluate_previous_blocks(block_graph);
		
		default_block_label_layout(block_graph,label);
		
		evaluate_block(block_graph);
		
	} else
	{
		block_graph->instruction_block=get_current_block();
		first_block_graph=NULL;
		last_block_graph=NULL;
	}
}

VOID end_graph_label_layout_block (BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph,LABEL *label)
{
	if (code_reachable)
	{
		block_graph->kind=INCOMPLETE_BLOCK;
			
		mark_and_count_block_graph(block_graph,graph);
		evaluate_previous_blocks(block_graph);
	
		default_block_label_layout(block_graph,label);
	
		load_nodes_for_end_a_params (block_graph);
		load_nodes_for_end_b_params (block_graph);
		store_a_stack_values (block_graph);
		store_b_stack_values (block_graph);
		
		adjust_stack_offsets(block_graph);
		
		evaluate_stacks(block_graph);
		block_graph->instruction_block=lin_graph_Areg(block_graph,graph);
		
		set_stack_check_and_end_instruction(block_graph);
		
	} else
	{
		block_graph->instruction_block=get_current_block();
		first_block_graph=NULL;
		last_block_graph=NULL;
	}
}

VOID end_dummy_block (BLOCK_GRAPH *block_graph)
{
	/* no code will be generated for this block, so graphs can be freed */
	
	free_graph_blocks();
}

VOID end_dummy_label_block (BLOCK_GRAPH *block_graph, LABEL *label)
{
	/* see routine above */
	
	default_block_label_layout(block_graph,label);
	free_graph_blocks();
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

VOID initialise_stacks()
{
	first_block_graph	=NULL;
	last_block_graph	=NULL;
}

/************************************************************************************
 *																					*
 *		checking of parameters														*
 *																					*
 ************************************************************************************/

BOOL check_parameters (INT16 n_a_parameters, INT16 n_b_parameters, U_INT32 vector[], STRING *params)
{
	register CHAR *p, *end;
	INT16	a_stack_offset, b_stack_offset, n_parameters, i;
	BOOL result;

	result=true;
	p=params->string;
	n_parameters=n_a_parameters+n_b_parameters;
	
	if (n_b_parameters <= N_B_PARAMETER_LOCALS)
		b_stack_offset=0;
	else
		b_stack_offset=(n_b_parameters-N_B_PARAMETER_LOCALS);
	
	if (n_a_parameters <= N_A_PARAMETER_LOCALS)
		a_stack_offset=0;
	else
		a_stack_offset=(n_a_parameters-N_A_PARAMETER_LOCALS);


	/*	check if parameter description is valid. There may have been less parameters
		described than the n_parameters, but the description must conform to certain
		conventions */

	for (i=1, end=p+params->length ; p<end; ++i)
	{
		INT16 stack, stack_offset, place, place_offset;
		
		stack=*p++;
		get_int16(p,stack_offset);
		place=*p++;
		get_int16(p,place_offset);
		
		switch (stack)
		{
			case 'A':
			{
				switch (place)
				{
					case 'L':
					case 'l':
					
						/* force in local */
						
						if (stack_offset >= N_A_PARAMETER_LOCALS)
						{	scan_error(NULL,"parameter %d should be placed on a-stack\n",i);
							result=false;
							break;
						}

						if (place_offset < A_PAR_FIRST_LOCAL || place_offset > A_PAR_LAST_LOCAL)
						{	scan_error(NULL,"parameter %d has invalid a-local offset %d\n",i,place_offset);
							result=false;
							break;
						}
						
						if (place_offset != A_PAR_FIRST_LOCAL+n_a_parameters-1-stack_offset)
						{	scan_error(NULL,"parameter %d has wrong a-local offset %d\n",i,place_offset);
							result=false;
						}
						
						break;
					
					case 'A':
					case 'a':

						/* could be in local, or on stack */
					
						if (stack_offset < N_A_PARAMETER_LOCALS)
						{
							/*	the parameter should be in a local, but note
								counting starts at 1 (L7==A1) */

							if (place_offset < 1 || place_offset > N_A_PARAMETER_LOCALS)
							{	scan_error(NULL,"parameter %d has invalid a-local offset %d\n",i,place_offset);
								result=false;
								break;
							}
							
							if (place_offset != 1+n_a_parameters-1-stack_offset)
							{	scan_error(NULL,"parameter %d has wrong a-local offset %d\n",i,place_offset);
								result=false;
							}
							
							/* all right */
							
							break;
						}
						
						if (place_offset != stack_offset-a_stack_offset)
						{	scan_error(NULL,"parameter %d has incorrect a-stack offset\n",i);
							result=false;
						}
						
						break;
						
					case 'B':
					case 'b':
						scan_error(NULL,"parameter %d cannot be placed on b-stack\n",i);
						result=false;
						break;
						
					default:
						scan_error(NULL,"parameter %d has unknown parameter place: %c\n",i,place);
						result=false;
				}
				break;
			}	
			case 'B':
			{
				switch (place)
				{
					case 'L':
					case 'l':
					
						if (stack_offset >= N_B_PARAMETER_LOCALS)
						{	scan_error(NULL,"parameter %d should be placed on b-stack\n",i);
							result=false;
							break;
						}
						
						if (place_offset < B_PAR_FIRST_LOCAL || place_offset > B_PAR_LAST_LOCAL)
						{	scan_error(NULL,"parameter %d has invalid b-local offset %d\n",i,place_offset);
							result=false;
							break;
						}
						
						/*	If the vector bit is set for this parameter, it is a
							real. */
						
						if (is_vector_bit_set(vector,stack_offset))
						{
							if (place_offset != B_PAR_FIRST_LOCAL+n_b_parameters-2-stack_offset)
							{	scan_error(NULL,"real parameter %d has wrong b-local offset %d\n",i,place_offset);
								result=false;
							}
						} else
						{
							if (place_offset != B_PAR_FIRST_LOCAL+n_b_parameters-1-stack_offset)
							{	scan_error(NULL,"parameter %d has wrong b-local offset %d\n",i,place_offset);
								result=false;
							}
						}
						
						break;
						
					case 'B':
					case 'b':
					
						if (stack_offset < N_B_PARAMETER_LOCALS)
						{
							/*	the parameter should be in a local, but note
								counting starts at 1 (L1==B1), which is the same by coincidence */

							if (place_offset < 1 || place_offset > N_B_PARAMETER_LOCALS)
							{	scan_error(NULL,"parameter %d has invalid b-local offset %d\n",i,place_offset);
								result=false;
								break;
							}
							
							/*	If the vector bit is set for this parameter, it is a
								real. */
							
							if (is_vector_bit_set(vector,stack_offset))
							{
								if (place_offset != 1+n_b_parameters-2-stack_offset)
								{	scan_error(NULL,"real parameter %d has wrong b-local offset %d\n",i,place_offset);
									result=false;
								}
							} else
							{
								if (place_offset != 1+n_b_parameters-1-stack_offset)
								{	scan_error(NULL,"parameter %d has wrong b-local offset %d\n",i,place_offset);
									result=false;
								}
							}
							
							/* all right */
							
							break;
						}
						
						if (place_offset != stack_offset-b_stack_offset)
						{	scan_error(NULL,"parameter %d has incorrect b-stack offset\n",i);
							result=false;
						}

						break;
						
					case 'A':
					case 'a':
					
						scan_error(NULL,"parameter %d cannot be placed on a-stack\n",i);
						result=false;
						break;
						
					default:
						scan_error(NULL,"parameter %d has unknown parameter place: %c\n",i,place);
						result=false;
				}
				break;
			}	
			default:
				internal_error("check_parameters");
		}
	}
	return result;
}