#include "ifinterface.h"

#include "cgwrites.h"
#include "cglabels.h"
#include "cgmemory.h"

#include "ccmacros.h"

extern LABEL *nil_label;

/********************************************************************
 *																	*
 *		instruction symbols											*
 *																	*
 ********************************************************************/

FILE *codefile;

CHAR *isymbol[] = {

	"\tldl\t",
	"\tstl\t",
	"\tldnl\t",
	"\tstnl\t",
	"\tlb\n",
	"\tldc\t",
	"\tadc\t",
	"\teqc\t",
	"\teqd\n",
	"\tldlp\t",
	"\tldnlp\t",
	"\tajw\t",
	"\tj\t",
	"\tcj\t",
	"\tret\n",
	"\tlea\n",
	"\tldesc\n",
	"\tldpi\n",
	"\tgt\n",
	"\trev\n",
	"\tdup\n",
	
	"\tadd\n",
	"\tsub\n",
	"\tmul\n",
	"\tdiv\n",
	"\trem\n",
	"\tsum\n",
	"\tdiff\n",
	"\tprod\n",
	"\tand\n",
	"\tor\n",
	"\txor\n",
	"\tshl\n",
	"\tshr\n",
	"\tnot\n",
	"\tabs\n",
	
	"\tlshl\n",
	"\tlshr\n",
	"\txdble\n",
	"\tcsngl\n",
	
	"\tbsub\n",
	"\twsub\n",
	
	"\tmove\n",
	"\tbcnt\n",
	"\twcnt\n",
	"\tmint\n",
		
	"\tfpdup\n",
	"\tfprev\n",
	"\tfpldnldb\n",
	"\tfpstnldb\n",
	"\tfpadd\n",
	"\tfpsub\n",
	"\tfpmul\n",
	"\tfpdiv\n",
	"\tfprem\n",
	"\tfpgt\n",
	"\tfpeq\n",
	
	"\tfpldzerodb\n",
	
	"\tfpremfirst\n",
	"\tfpremstep\n",
	"\tfpusqrtfirst\n",
	"\tfpusqrtstep\n",
	"\tfpusqrtlast\n",

	"\tfpumulby2\n",
	"\tfpudivby2\n",
	"\tfpuexpinc32\n",
	"\tfpuexpdec32\n",
	
	"\tfpint\n",
	"\tfpi32tor64\n",
	"\tfpstnli32\n",
	"\tfpurm\n",

	"\trtn\n",
	"\tjmp_ap\n",
	"\tjmp_eval\n",

	"\tstopp\n",
	"\trunp\n",
	"\tcall\t",
	"\tgajw\n",

	"\tlibcall\n",
	"\tinline\n",
	"\tgroup\t",

	"\tjsr_eval\n",
	"\trelease\n",
		
	"\tsuspend_uneval\n",
	"\tsuspend\n",
	
	"\tcs\n",
	"\tenable_cs\n",
	"\tdisable_cs\n",
	
	"\tjsr\n",
	"\tgcall\n",
	"\tjsr_ap\n",
	"\tsignal\t",
	
	"\terror\n",
	"\terror\n",
};

static CHAR *parc_dsymbol[] = {

	"\t.word\t",
	"\t.byte\t",
	"\t.align\n",	
	"\t.equ\t",
	
	"\t.public\t",
	"\t.public\t",
	
	"\terror\n",
	"\t.cseg\n",
	"\t.dseg\n"	
};

static CHAR *pact_dsymbol[] = {

	"\tword\t",
	"\tbyte\t",
	"\talign\n",
	"\tequ\t",
	
	"\textern\t",
	"\tglobal\t",
	
	"\tattrib\t",
	"\tcsec\n",
	"\tdsec\n"
};

CHAR **dsymbol;

static BOOL label_table[257];		/* table with true for label chars, false otherwise */

/****************************************************************************
 *																			*
 *		basic output routines												*
 *																			*
 ****************************************************************************/

static VOID initialise_label_table(VOID)
{
	INT16 i;
	
	U_CHAR non_label_chars[]="@#$%^&*-+/=<>\\_~?\'\"";

	for (i=0; i<257; ++i)
		label_table[i]=true;
	
	label_table[0]=false;
	for (i=0; i<sizeof (non_label_chars); ++i)
		label_table[1+non_label_chars[i]]=false;
}

VOID o_string (register CHAR *Str)

{   register CHAR *p;

	p = Str;

	switch (options.assembler)
	{	case PARC_ASSEMBLER:
		{	
			if (is_print_char(*p))
		    {   o_char(dquote);
		        while (is_print_char(*p)) o_char((int)*p++);
		        fprintf(codefile, "\",$%02x", (int)*p);
		    }
		    else
		        fprintf(codefile, "$%02x", (int)*p);
		
		    while (*p)
		    {   o_char(comma);
		        p++;
		        if (is_print_char(*p))
		        {   o_char(dquote);
		            while (is_print_char(*p)) o_char((int)*p++);
		            fprintf(codefile, "\",$%02x", (int)*p);
		        }
		        else
		            fprintf(codefile, "$%02x", (int)*p);
		    }
			break;
		}
		case PACT_ASSEMBLER:
		{
			o_char(dquote);
			
			while(*p)
			{
				while (is_print_char(*p))
					o_char((int)*p++);
				
				if (*p)
				{	switch(*p)
					{	case '\n':
							fputs("\\n",codefile);
							break;
						case '\b':
							fputs("\\b",codefile);
							break;
						case '\f':
							fputs("\\f",codefile);
							break;
						case '\r':
							fputs("\\r",codefile);
							break;
						case '\t':
							fputs("\\t",codefile);
							break;
	
						default:
							fprintf(codefile,"\\x%x", (int)*p);
					}
					++p;
				}
			}
			fputs("\\0\"",codefile);
			
			break;
		}
		default:
			internal_error("o_string: unknown assembler option");
	}
}

VOID o_label_name (LABEL *label)
{
	/*	This routine can be used to simply print label names without
		serious conversions: useful for intermediate code.
	*/

	if (label->name != NULL)
		fputs(label->name,codefile);
	else
		fprintf(codefile,"lab%d",(int)label->number);
}

VOID o_label (LABEL *label)
{
	register CHAR *name;

	name=label->name;
	
	if (name!=NULL)
	{
		for (;;)
		{
			while(label_table[(*name)+1])
				o_char(*name++);
			
			switch (*name++)
			{
				case '\0':
					return;
					
				case '_':
					o_char('_');
					o_char('_');
					break;
				case '*':
					o_char('_');
					o_char('M');
					break;
				case '-':
					o_char('_');
					o_char('S');
					break;
				case '+':
					o_char('_');
					o_char('A');
					break;
				case '=':
					o_char('_');
					o_char('E');
					break;
				case '~':
					o_char('_');
					o_char('T');
					break;
				case '<':
					o_char('_');
					o_char('L');
					break;
				case '>':
					o_char('_');
					o_char('G');
					break;
				case '/':
					o_char('_');
					o_char('D');
					break;
				case '?':
					o_char('_');
					o_char('Q');
					break;
				case '#':
					o_char('_');
					o_char('H');
					break;
				case '$':
					o_char('_');
					o_char('N');
					o_char('D');
					break;
				case '^':
					o_char('_');
					o_char('N');
					o_char('C');
					break;
				case '@':
					o_char('_');
					o_char('N');
					o_char('T');
					break;
				case '&':
					o_char('_');
					o_char('N');
					o_char('A');
					break;
				case '%':
					o_char('_');
					o_char('N');
					o_char('P');
					break;
				case '\'':
					o_char('_');
					o_char('N');
					o_char('S');
					break;
				case '\"':
					o_char('_');
					o_char('N');
					o_char('Q');
					break;
				default:
					internal_error("o_label");
			}
		}
	} else
		fprintf(codefile,"%d",(int)label->number);
}

VOID o_dlabel(LABEL *label)
{
	if (options.output_full_label_names ||
		(label->flags & EXPORT_LABEL) ||
		(label->flags & IMPORT_LABEL))
	{
		o_char('.');
		o_char('D');
		o_label(label);
		return;
	}

	fprintf(codefile,".D%d",(int)label->number);
}

VOID o_glabel(LABEL *label)
{
	if (options.output_full_label_names ||
		(label->flags & EXPORT_LABEL) ||
		(label->flags & IMPORT_LABEL))
	{
		o_char('.');
		o_char('G');
		o_label(label);
		return;
	}
	
	fprintf(codefile,".G%d",(int)label->number);
}

VOID o_jlabel(LABEL *label)
{
	if (options.output_full_label_names ||
		(label->flags & EXPORT_LABEL) ||
		(label->flags & IMPORT_LABEL))
	{	o_char('.');
		o_char('J');
		o_label(label);
		return;
	}
	
	fprintf(codefile,".J%d",(int)label->number);
}

VOID o_inline_label(LABEL *label)
{
	if (options.output_full_label_names ||
		(label->flags & EXPORT_LABEL) ||
		(label->flags & IMPORT_LABEL))
	{	o_label(label);
		return;
	}
	fprintf(codefile,".L%d",(int)label->number);
}

VOID o_iwords16 (INT16 n, ...)
{
	va_list argp;
	INT16 word;
	
    va_start(argp, n);
    
    switch (n)
    {	case 0:
    		fputs(dsymbol[WORD_DIR],codefile);
    		break;
    	
    	case 1:
    		word=next_short(argp);
    		fprintf(codefile,"%s%d\n",dsymbol[WORD_DIR],(int)word);
    		break;

		default:
		{
			INT16 i;
			
			word=next_short(argp);
			fprintf(codefile,"%s%d",dsymbol[WORD_DIR],(int)word);
			
			for (i=1; i<n; ++i)
			{
				word=next_short(argp);
				fprintf(codefile,",%d",(int)word);
			}
			o_newline();
		}
	}
    va_end(argp);
}

VOID o_xwords16 (INT16 n, ...)
{
	va_list argp;
	INT16 word;
	
    va_start(argp, n);
    
    switch (options.assembler)
    {	case PARC_ASSEMBLER:
	    {	switch (n)
		    {	case 0:
		    		fputs(dsymbol[WORD_DIR],codefile);
		    		break;
		    	
		    	case 1:
		    		word=next_short(argp);
		    		fprintf(codefile,"%s$%08x\n",dsymbol[WORD_DIR],(int)word);
		    		break;
		
				default:
				{
					INT16 i;
					
					word=next_short(argp);
					fprintf(codefile,"%s$%08x",dsymbol[WORD_DIR],(int)word);
					
					for (i=1; i<n; ++i)
					{
						word=next_short(argp);
						fprintf(codefile,",$%08x",(int)word);
					}
					o_newline();
				}
			}
			break;
		}
		case PACT_ASSEMBLER:
		{
		    switch (n)
		    {	case 0:
		    		fputs(dsymbol[WORD_DIR],codefile);
		    		break;
		    	
		    	case 1:
		    		word=next_short(argp);
		    		fprintf(codefile,"%s0x%08x\n",dsymbol[WORD_DIR],(int)word);
		    		break;
		
				default:
				{
					INT16 i;
					
					word=next_short(argp);
					fprintf(codefile,"%s0x%08x",dsymbol[WORD_DIR],(int)word);
					
					for (i=1; i<n; ++i)
					{
						word=next_short(argp);
						fprintf(codefile,",0x%08x",(int)word);
					}
					o_newline();
				}
			}
			break;
		}
		default:
			internal_error("xwords16: unknown assembler type");
	}
    va_end(argp);
}

VOID o_ibytes16 (INT16 n, ...)
{
	va_list argp;
	INT16 word;
	
    va_start(argp, n);
    
	switch (n)
    {	case 0:
    		fputs(dsymbol[BYTE_DIR],codefile);
    		break;
    	
    	case 1:
    		word=next_short(argp);
    		fprintf(codefile,"%s%d\n",dsymbol[BYTE_DIR],(int)word);
    		break;

		default:
		{
			INT16 i;
			
			word=next_short(argp);
			fprintf(codefile,"%s%d",dsymbol[BYTE_DIR],(int)word);
			
			for (i=1; i<n; ++i)
			{
				word=next_short(argp);
				fprintf(codefile,",%d",(int)word);
			}
			o_newline();
		}
	}
	va_end(argp);
}

VOID o_xbytes16 (INT16 n, ...)
{
	va_list argp;
	INT16 word;
	
    va_start(argp, n);
    
    switch (options.assembler)
    {	case PARC_ASSEMBLER:
	    {	switch (n)
		    {	case 0:
		    		fputs(dsymbol[BYTE_DIR],codefile);
		    		break;
		    	
		    	case 1:
		    		word=next_short(argp);
		    		fprintf(codefile,"%s$%02x\n",dsymbol[BYTE_DIR],(int)word);
		    		break;
		
				default:
				{
					INT16 i;
					
					word=next_short(argp);
					fprintf(codefile,"%s$%02x",dsymbol[BYTE_DIR],(int)word);
					
					for (i=1; i<n; ++i)
					{
						word=next_short(argp);
						fprintf(codefile,",$%02x",(int)word);
					}
					o_newline();
				}
			}
			break;
		}
		case PACT_ASSEMBLER:
		{
		    switch (n)
		    {	case 0:
		    		fputs(dsymbol[BYTE_DIR],codefile);
		    		break;
		    	
		    	case 1:
		    		word=next_short(argp);
		    		fprintf(codefile,"%s0x%02x\n",dsymbol[BYTE_DIR],(int)word);
		    		break;
		
				default:
				{
					INT16 i;
					
					word=next_short(argp);
					fprintf(codefile,"%s0x%02x",dsymbol[BYTE_DIR],(int)word);
					
					for (i=1; i<n; ++i)
					{
						word=next_short(argp);
						fprintf(codefile,",0x%02x",(int)word);
					}
					o_newline();
				}
			}
			break;
		}
		default:
			internal_error("xbytes16: unknown assembler type");
	}
    va_end(argp);
}

VOID o_float (FLOAT *r)
{
	switch (options.assembler)
	{
		case PARC_ASSEMBLER:
			fprintf(codefile,"%s$%08lx,$%08lx\n", dsymbol[WORD_DIR], (long)r->join.low, (long)r->join.high);
			break;
		
		case PACT_ASSEMBLER:
			fprintf(codefile,"%s0x%08lx,0x%08lx\n", dsymbol[WORD_DIR], (long)r->join.low, (long)r->join.high);
			break;
		
		default:
			internal_error("o_float: unknown assembler type");
	}
}

VOID o_address(LABEL *label, INT16 offset)
{	
	o_ldc();
	
	/* output destination label */
	
	if (label->flags & DATA_LABEL)
		o_dlabel(label);
	else if (label->flags & GCALL_LABEL)
		o_glabel(label);
	else if (label->flags & INLINE_LABEL)
		o_inline_label(label);
	else
		o_jlabel(label);
	
	/*	Create an offset relative to the following ldpi instruction. For some
		assemblers this can simply be done by subtracting the size of the ldpi
		instruction from the label (evaluating to an offset). Others require a
		local label.
	*/

	switch (options.assembler)
	{		
		case PACT_ASSEMBLER:
		{
			offset=WORD_SIZE*offset-LDPI_SIZE;
						
			if (offset < 0)
				fprintf(codefile,"-%d\n",(int)-offset);
			else if (offset > 0)
				fprintf(codefile,"+%d\n",(int)offset);

			o_ldpi();
		
			break;
		}
		case PARC_ASSEMBLER:
		{
			INT16 local_id;

			local_id=new_label_id();

			o_char('-');
			o_luse_label(local_id);
		
			if (offset==0)
			{
				o_newline();
			} else
			{		
				fprintf(codefile,"+%d\n",(int)WORD_SIZE*offset);
			}
			o_ldpi();
			o_ldef_label(local_id);
		
			break;
		}	
		default:
			internal_error("o_address: unknown assembler type");
	}
}

VOID o_local_address(INT16 id, INT16 offset)
{	
	o_ldc();
	o_luse_label(id);
	
	/* see also routine above */

	switch (options.assembler)
	{		
		case PACT_ASSEMBLER:
		{
			offset=WORD_SIZE*offset-LDPI_SIZE;
			
			if (offset < 0)
				fprintf(codefile,"-%d\n",(int)-offset);
			else if (offset > 0)
				fprintf(codefile,"+%d\n",(int)offset);

			o_ldpi();
		
			break;
		}
		case PARC_ASSEMBLER:
		{
			INT16 local_id;
			
			local_id=new_label_id();
			
			o_char('-');
			o_luse_label(local_id);
		
			if (offset==0)
			{
				o_newline();
			} else
			{
				fprintf(codefile,"+%d\n",(int)WORD_SIZE*offset);
			}
			o_ldpi();
			o_ldef_label(local_id);
			
			break;
		}
		default:
			internal_error("o_local_address: unknown assembler type");
	}
}

VOID o_desc_id(LABEL *label, INT16 arity, BOOL mark)
{
	/* a marked descriptor is used for descriptors of closures, so that
	   these can be distinguished from hnf nodes when copied... */
			
	if (!(label->flags & DATA_LABEL))
	{	label_error (label,"descriptor label is not a data label");
		return;
	}
	
	if (label==nil_label)
	{
		/* offset to nil descriptor is always 0, forget arity... */
		
		if (mark)
			o_char('2');
		else
			o_char('0');
		
		o_newline();
		
		return;
	}
	
	o_dlabel(label);
	o_char('-');
	o_dlabel(nil_label);
	
	if (mark)
	{
		if (arity==0)
			fprintf(codefile,"+2\n");
		else
			fprintf(codefile,"+%d\n",((int)2*WORD_SIZE*arity)+2);
		
	} else
	{
		if (arity==0)
			o_newline();
		else
			fprintf(codefile,"+%d\n",(int)2*WORD_SIZE*arity);
	}
}

VOID o_nop (INT16 n)
{
	switch(n)
	{	case 0:
			break;
		case 1:
			o_xbytes16(1,(INT16)0x20);
			break;
		case 2:
			o_xbytes16(2,(INT16)0x20,(INT16)0x20);
			break;
		case 3:
			o_xbytes16(3,(INT16)0x20,(INT16)0x20,(INT16)0x20);
			break;
		default:
			internal_error("output_nop: too many nop operations");
	}
}

/********************************************************************************
 *																				*
 *		routines to output standard transputer instructions: all these			*
 *		routines are of the form o_instruction(INSTRUCTION *instruction)		*
 *		and are normally placed in a table.										*
 *																				*
 ********************************************************************************/

/************************************************************************************
 *																					*
 *		load / store instructions													*
 *																					*
 ************************************************************************************/

VOID out_ldl(INSTRUCTION *instruction)
{
	o_ldl16(instruction->parameters[0].int16);
}

VOID out_stl(INSTRUCTION *instruction)
{
	o_stl16(instruction->parameters[0].int16);
}

VOID out_ldnl(INSTRUCTION *instruction)
{
	o_ldnl16(instruction->parameters[0].int16);
}

VOID out_stnl(INSTRUCTION *instruction)
{
	o_stnl16(instruction->parameters[0].int16);
}

VOID out_lb(INSTRUCTION *instruction)
{
	o_lb();
}

VOID out_ldlp(INSTRUCTION *instruction)
{
	o_ldlp16(instruction->parameters[0].int16);
}

VOID out_ldnlp(INSTRUCTION *instruction)
{
	o_ldnlp16(instruction->parameters[0].int16);
}

VOID out_ajw(INSTRUCTION *instruction)
{
	o_ldnlp16(instruction->parameters[0].int16);
}

/************************************************************************************
 *																					*
 *		arithmetic instructions														*
 *																					*
 ************************************************************************************/

VOID out_ldc(INSTRUCTION *instruction)
{
	o_ldc32(instruction->parameters[0].int32);
}

VOID out_adc(INSTRUCTION *instruction)
{
	o_adc32(instruction->parameters[0].int32);
}

VOID out_eqc(INSTRUCTION *instruction)
{
	o_eqc32(instruction->parameters[0].int32);
}

VOID out_gt(INSTRUCTION *instruction)
{
	o_gt();
}

VOID out_dup(INSTRUCTION *instruction)
{
	o_dup();
}

VOID out_rev(INSTRUCTION *instruction)
{
	o_rev();
}

VOID out_add(INSTRUCTION *instruction)
{
	o_add();
}

VOID out_sub(INSTRUCTION *instruction)
{
	o_sub();
}

VOID out_mul(INSTRUCTION *instruction)
{
	o_mul();
}

VOID out_div(INSTRUCTION *instruction)
{
	o_div();
}

VOID out_rem(INSTRUCTION *instruction)
{
	o_rem();
}

VOID out_sum(INSTRUCTION *instruction)
{
	o_sum();
}

VOID out_diff(INSTRUCTION *instruction)
{
	o_diff();
}

VOID out_prod(INSTRUCTION *instruction)
{
	o_prod();
}

VOID out_and(INSTRUCTION *instruction)
{
	o_and();
}

VOID out_or(INSTRUCTION *instruction)
{
	o_or();
}

VOID out_xor(INSTRUCTION *instruction)
{
	o_xor();
}

VOID out_shl(INSTRUCTION *instruction)
{
	o_shl();
}

VOID out_shr(INSTRUCTION *instruction)
{
	o_shr();
}

VOID out_not(INSTRUCTION *instruction)
{
	o_not();
}

VOID out_lshr(INSTRUCTION *instruction)
{
	o_lshr();
}

VOID out_lshl(INSTRUCTION *instruction)
{
	o_lshl();
}

VOID out_xdble(INSTRUCTION *instruction)
{
	o_xdble();
}

VOID out_csngl(INSTRUCTION *instruction)
{
	o_csngl();
}

VOID out_bsub(INSTRUCTION *instruction)
{
	o_bsub();
}

VOID out_wsub(INSTRUCTION *instruction)
{
	o_wsub();
}

VOID out_move(INSTRUCTION *instruction)
{
	o_move();
}

VOID out_bcnt(INSTRUCTION *instruction)
{
	o_bcnt();
}

VOID out_wcnt(INSTRUCTION *instruction)
{
	o_wcnt();
}

VOID out_mint(INSTRUCTION *instruction)
{
	o_mint();
}

/************************************************************************************
 *																					*
 *		address instructions														*
 *																					*
 ************************************************************************************/

VOID out_ldpi(INSTRUCTION *instruction)
{
	o_ldpi();
}

/************************************************************************************
 *																					*
 *		floating point instructions													*
 *																					*
 ************************************************************************************/

VOID out_fpdup(INSTRUCTION *instruction)
{
	o_fpdup();
}

VOID out_fprev(INSTRUCTION *instruction)
{
	o_fprev();
}

VOID out_fpldnldb(INSTRUCTION *instruction)
{
	o_fpldnldb();
}

VOID out_fpstnldb(INSTRUCTION *instruction)
{
	o_fpstnldb();
}

VOID out_fpadd(INSTRUCTION *instruction)
{
	o_fpadd();
}

VOID out_fpsub(INSTRUCTION *instruction)
{
	o_fpsub();
}

VOID out_fpmul(INSTRUCTION *instruction)
{
	o_fpmul();
}

VOID out_fpdiv(INSTRUCTION *instruction)
{
	o_fpdiv();
}

VOID out_fpgt(INSTRUCTION *instruction)
{
	o_fpgt();
}

VOID out_fpeq(INSTRUCTION *instruction)
{
	o_fpeq();
}

VOID out_fpldzerodb(INSTRUCTION *instruction)
{
	o_fpldzerodb();
}

VOID out_fpremfirst(INSTRUCTION *instruction)
{
	o_fpremfirst();
}

VOID out_fpremstep(INSTRUCTION *instruction)
{
	o_fpremstep();
}

VOID out_fpusqrtfirst(INSTRUCTION *instruction)
{
	o_fpusqrtfirst();
}

VOID out_fpusqrtstep(INSTRUCTION *instruction)
{
	o_fpusqrtstep();
}

VOID out_fpusqrtlast(INSTRUCTION *instruction)
{
	o_fpusqrtlast();
}

VOID out_fpumulby2(INSTRUCTION *instruction)
{
	o_fpumulby2();
}

VOID out_fpudivby2(INSTRUCTION *instruction)
{
	o_fpudivby2();
}

VOID out_fpuexpinc32(INSTRUCTION *instruction)
{
	o_fpuexpinc32();
}

VOID out_fpuexpdec32(INSTRUCTION *instruction)
{
	o_fpuexpdec32();
}

VOID out_fpint(INSTRUCTION *instruction)
{
	o_fpint();
}

VOID out_fpi32tor64(INSTRUCTION *instruction)
{
	o_fpi32tor64();
}

VOID out_fpstnli32(INSTRUCTION *instruction)
{
	o_fpstnli32();
}

VOID out_fpurm(INSTRUCTION *instruction)
{
	o_fpurm();
}

/********************************************************************************
 *																				*
 *		flow of control instructions											*
 *																				*
 ********************************************************************************/

VOID out_ret (INSTRUCTION *instruction)
{
	o_ret();
}

VOID out_stopp(INSTRUCTION *instruction)
{
	o_stopp();
}

VOID out_runp(INSTRUCTION *instruction)
{
	o_runp();
}

VOID out_gajw(INSTRUCTION *instruction)
{
	o_gajw();
}

VOID out_error(INSTRUCTION *instruction)
{
	o_error();
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

VOID init_writes(VOID)
{
	initialise_label_table();
	
	switch (options.assembler)
	{	case PARC_ASSEMBLER:
			dsymbol=parc_dsymbol;
			break;
		case PACT_ASSEMBLER:
			dsymbol=pact_dsymbol;
			break;
		default:
			internal_error("init_writes: unknown assembler type");
	}
}
