
#include "ifaux.h"
#include "ifkeys.h"
#include "ifinterface.h"
#include "ifmacinterface.h"

#define	BlinkDelay		8

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

static	Str255	text;
static	Str255	trunc_text;

/********************************************************************************
 *																				*
 *		miscelaneous															*
 *																				*
 ********************************************************************************/

void frame_item(DialogPtr dialog, int item)
{
	short type;
	Handle handle;
	Rect box;
	
	GetDItem(dialog,item,&type,&handle,&box);
	
	SetPort(dialog);
	PenSize(1,1);
	PenPat(&qd.gray);
	InsetRect(&box,-4,-4);
	FrameRect(&box);
}

void set_hilite_state(DialogPtr dialog, int item, Boolean active)
{
	short type;
	Handle handle;
	Rect box;
	
	GetDItem(dialog,item,&type,&handle,&box);
	HiliteControl((ControlHandle)handle,active ? 0 : 255);
}

void set_check_box(DialogPtr dialog, int item, Boolean active)
{
	short type;
	Handle handle;
	Rect box;
	
	GetDItem(dialog,item,&type,&handle,&box);
	SetCtlValue((ControlHandle)handle,active ? 1 : 0);
}

/********************************************************************************
 *																				*
 *		text and items															*
 *																				*
 ********************************************************************************/

void set_item_text(DialogPtr dialog, int item, Str255 text)
{
	short	type;
	Handle	item_handle;
	Rect	box;

	GetDItem(dialog,item,&type,&item_handle,&box);
	SetIText(item_handle,(ConstStr255Param)text);
}

void set_item_ttext(DialogPtr dialog, int item, Str255 text)
{
	short	type;
	Handle	item_handle;
	Rect	box;
	short	result, width;

	GetDItem(dialog,item,&type,&item_handle,&box);
	
	TextFont(systemFont);
	TextSize(12);
	width=box.right-box.left-1;
	result=TruncString(width,text,truncMiddle);

	SetIText(item_handle,(ConstStr255Param)text);
}

void get_item_text(DialogPtr dialog, int item, Str255 text)
{
	short	type;
	Handle	item_handle;
	Rect	box;

	GetDItem(dialog,item,&type,&item_handle,&box);
	GetIText(item_handle,text);
}

void set_item_chars(DialogPtr dialog, int item,char *s)
{
	Str255 text;
	
	set_str255(text,s)
	set_item_text(dialog,item,text);
}

void set_item_tchars(DialogPtr dialog, int item,char *s)
{
	Str255 text;
	
	set_str255(text,s)
	set_item_ttext(dialog,item,text);
}

void get_item_chars(DialogPtr dialog, int item,char *s)
{
	Str255 text;
	
	get_item_text(dialog,item,text);
	get_str255(s,text);
}

void vset_item_chars(DialogPtr dialog, int item,char *s,va_list argp)
{
	Str255 text;
	
	vset_str255(text,s,argp)
	set_item_text(dialog,item,text);
}

/********************************************************************************
 *																				*
 *		operations on buttons													*
 *																				*
 ********************************************************************************/

void set_default_button(DialogPtr dialog, int item,Boolean active)
{
	short type;
	Handle handle;
	Rect box;
	
	GetDItem(dialog,item,&type,&handle,&box);
	
	SetPort(dialog);
	PenSize(3,3);
	InsetRect(&box,-4,-4);
	PenPat(active ? &qd.black : &qd.gray);
	FrameRoundRect(&box,16,16);
}

void blink_button(DialogPtr dialog, int item)
{
	short	type;
	Handle	item_handle;
	Rect	box;
	long	finalTicks;

	GetDItem(dialog,item,&type,&item_handle,&box);
	HiliteControl((ControlHandle)item_handle,inButton);
	Delay(BlinkDelay,&finalTicks);
	HiliteControl((ControlHandle)item_handle,0);
}

void set_button_title(DialogPtr dialog, int item, Str255 title)
{
	short	type;
	Handle	item_handle;
	Rect	box;
	short	result, width;
	
	GetDItem(dialog,item,&type,&item_handle,&box);

	TextFont(systemFont);
	TextSize(12);
	
	set_str255(text,"Select \"");
	width=(box.right-box.left)-(StringWidth(text)+CharWidth(' ')+CharWidth('"'));
	
	cpy_str255(trunc_text,title);
	result=TruncString(width,trunc_text,truncMiddle);
	
	str255_append_str255(text,trunc_text);
	str255_append_char(text,'"');
	
	SetCTitle((ControlHandle)item_handle,text);
	ValidRect(&box);
}

/********************************************************************************
 *																				*
 *		error window															*
 *																				*
 ********************************************************************************/

void verror (char *mess,va_list argp)
{
	int i;
	
	Str255	string;
	int		length;
	
    length=vsprintf((char *)string,mess, argp);
    
    for(i=0;i<length;i++)
	{	if(string[i]=='\n')
			string[i]=returnKey;
	}
	
	if (windows.messages_window.window != FrontWindow())
		SelectWindow(windows.messages_window.window);

	ShowWindow(windows.messages_window.window);
		
	TEInsert(string,length,windows.messages_window.text);
}

void error (char *mess, ...)
{
	va_list argp;
	
    va_start(argp, mess);
    verror(mess, argp);
    va_end(argp);
}

void internal_error(char *mess, ...)
{
	DialogPtr	fatal_dialog;
	short		item;
	
	short	type;
	Handle	handle;
	Rect	box;
	char	text[257];
	
	va_list argp;
	
    va_start(argp, mess);
    text[0]=vsprintf(text+1,mess, argp);
    va_end(argp);
    
	fatal_dialog=GetNewDialog(dialog_id_internal_error,NULL,(WindowPtr)NULL);

	GetDItem(fatal_dialog,5,&type,&handle,&box);
	SetIText(handle,(ConstStr255Param)text);
	
	SysBeep(1);
	
	SelectWindow(fatal_dialog);
	ShowWindow(fatal_dialog);
	
	set_default_button(fatal_dialog,1,true);

	ModalDialog(NULL,&item);
	DisposDialog(fatal_dialog);
	
	if (item==1)
		quit_application(1);			/* note that cleaning up might fail... */
	
	abort_code_generation();
}
