#include "ifaux.h"
#include "ifkeys.h"
#include "ifinterface.h"
#include "ifmacinterface.h"
#include "ifdialogs.h"

#define FolderBit			4
#define SelectFolderButton	10

/********************************************************************************
 *																				*
 *																				*
 *																				*
 ********************************************************************************/

typedef struct SelectFolderReply
{
	StandardFileReply 		standard;		/* used by standard functions			*/
	FSSpec					selected;		/* the selected folder					*/
	Boolean					good;			/* true if selection has been made		*/
	
} SelectFolderReply;

/********************************************************************************
 *																				*
 *		low level file access functions											*
 *																				*
 ********************************************************************************/

static void GetFullPath(FSSpec *folder,Str255 path)
{
	CInfoPBRec	pb;
	Str255		directory;
	
	set_str255(path,"");
	
	pb.dirInfo.ioCompletion		=NULL;
	pb.dirInfo.ioNamePtr		=directory;
	pb.dirInfo.ioVRefNum		=folder->vRefNum;
	pb.dirInfo.ioDrParID		=folder->parID;
	pb.dirInfo.ioFDirIndex		=-1;
	
	for (;;)
	{
		/* get info about parent upto root */
		
		pb.dirInfo.ioDrDirID = pb.dirInfo.ioDrParID;
		
		if (PBGetCatInfo(&pb,false) != noErr)
			return;
					
		str255_append_char(directory,':');
		str255_append_str255(directory,path);
		cpy_str255(path,directory);
		
	 	if (pb.dirInfo.ioDrDirID == fsRtDirID)
	 		break;									/* arrived at root */
	}
	
	/* path to folder found, now append folder name */
	
	str255_append_str255(path,folder->name);
	str255_append_char(path,':');
}

/********************************************************************************
 *																				*
 *		a special dialog filters to avoid update events from clogging the		*
 *		event queue																*
 *																				*
 ********************************************************************************/

static pascal Boolean ModalFilter (DialogPtr dialog,EventRecord *event,short *itemhit)
{
	switch(event->what)
	{
		case updateEvt:
		{
			if ((DialogPtr)event->message != dialog)
			{
				/* handle event and clear it */
				
				do_update(event);
				event->what=nullEvent;
			}
			
			return false;
		}

		case keyDown:
		case autoKey:
		{
			char c;
	
			c=event->message & charCodeMask;
	
			if (c==enterKey || c==returnKey)
			{
				blink_button(dialog,1);
				*itemhit=1;
				return true;
			}
			if (c==escKey || ((event->modifiers & cmdKey) && c=='.'))
			{
				blink_button(dialog,2);
				*itemhit=2;
				return true;
			}
			
			return false;
		}
		default:
			return false;
	}
}

static pascal Boolean ModalFilterAbout (DialogPtr dialog,EventRecord *event,short *itemhit)
{
	switch(event->what)
	{
		case updateEvt:
		{
			if ((DialogPtr)event->message != dialog)
			{
				/* handle event and clear it */
				
				do_update(event);
				event->what=nullEvent;
			}
			
			return false;
		}

		case keyDown:
		case autoKey:
		{
			/*	No matter what key is pressed, just remove the dialog */
		
			*itemhit=0;					/* dummy item */
			return true;
		}

		default:
			return false;
	}
}

/********************************************************************************
 *																				*
 *		Extra File Selector Routines											*
 *																				*
 ********************************************************************************/

static pascal Boolean SelectFolderFileFilter (CInfoPBPtr info, SelectFolderReply *reply)
{
	return !(info->hFileInfo.ioFlAttrib & (1<<FolderBit));
}

static pascal Boolean SelectFolderHandleEvent (DialogPtr dialog,EventRecord *event,short *itemhit, SelectFolderReply *reply)
{
	/*	first perform normal dialog filtering, if this returns true, return
		immediately, else perform some extra filtering */
	
	if (ModalFilter(dialog,event,itemhit))
		return true;
	
	switch(event->what)
	{
		case keyDown:
		case autoKey:
		{
			char c;
	
			c=event->message & charCodeMask;
	
			if ((event->modifiers & cmdKey) && c=='s')
			{
				/* act if select button has been pushed */
				
				blink_button(dialog,SelectFolderButton);
				*itemhit=SelectFolderButton;
				return true;
			}
			return false;
		}
		default:
			return false;
	}
}

static pascal short SelectFolderItemHit (short item, DialogPtr dialog, SelectFolderReply *reply)
{	
	/*	this function is only needed for the main dialog box, so if this
		is not the case, just return the item  */
	
	if (GetWRefCon((WindowPtr)dialog)!=(long)(sfMainDialogRefCon))
		return item;
	
	switch (item)
	{
		case sfHookFirstCall:
		{
			if (reply->good)
			{
				/*	current selected folder is valid				
					set reply record to the selected folder */
								
				set_button_title(dialog,SelectFolderButton,reply->selected.name);
				
				reply->standard.sfFile		=reply->selected;
				reply->standard.sfScript	=smSystemScript;
				reply->standard.sfFlags		=0;
				reply->standard.sfIsFolder	=true;
				reply->standard.sfIsVolume	=false;
	
				return sfHookChangeSelection;
			}
			reply->good=true;
			
			return item;
		}	
		case sfItemCancelButton:
		
			reply->good=false;
		
		case SelectFolderButton:
		
			return sfItemCancelButton;
				
		default:
		
			/*	check if selection has changed and correct select button
				title if so
			*/
			
			if (reply->standard.sfIsFolder || reply->standard.sfIsVolume)
			{
				if (reply->standard.sfFile.vRefNum	!= reply->selected.vRefNum ||
					reply->standard.sfFile.parID	!= reply->selected.parID ||
					!EqualString(reply->standard.sfFile.name,reply->selected.name,true,true))
				{
					reply->selected=reply->standard.sfFile;
					set_button_title(dialog,SelectFolderButton,reply->selected.name);
				}
			}
	}
		
	/* For all other buttons do nothing special. */
	
	return item;
}

static void GetPath(char *pathname)
{
	static FileFilterYDUPP	file_filter	=NULL;
	static DlgHookYDUPP		hook		=NULL;
	static ModalFilterYDUPP	filter		=NULL;

	SelectFolderReply		reply;	
	Str255					path;
	Point					where={-1,-1};
	
	if (file_filter == NULL)
		file_filter = NewFileFilterYDProc(SelectFolderFileFilter);
	if (hook == NULL)
		hook = NewDlgHookYDProc(SelectFolderItemHit);
	if (filter == NULL)
		filter = NewModalFilterYDProc(SelectFolderHandleEvent);

	set_str255(path,pathname);
	
	reply.good=(FSMakeFSSpec(0,0,path,&reply.selected)==noErr);

	CustomGetFile(	file_filter,		/* filter out files */
					-1, NULL,			/* pass folders only */
					&reply.standard,
					dialog_id_select,
					where,				/* centered on screen */
					hook,				/* handle item hit: adjusts title of select button */
					filter,				/* handle events: react to command keys */
					NULL,				/* no active list */
					NULL,				/* no activation procedure */
					&reply				/* pass reply record */
				);
	
	if (reply.good)
	{
		/*	user did not cancel, retrieve full pathname */
		
		GetFullPath(&reply.selected,path);
		get_str255(pathname,path);
	}
}

/********************************************************************************
 *																				*
 *		about box																*
 *																				*
 ********************************************************************************/

void show_about_box(void)
{
	static ModalFilterUPP filter=NULL;

	DialogPtr about_box;
	short item;
	
	if (filter == NULL)
		filter = NewModalFilterProc(ModalFilterAbout);

	about_box=GetNewDialog(dialog_id_about,NULL,(WindowPtr)NULL);
	
	SelectWindow(about_box);
	ShowWindow(about_box);
	
	ModalDialog(filter,&item);
	
	DisposDialog(about_box);
}

/********************************************************************************
 *																				*
 *		processes dialog														*
 *																				*
 ********************************************************************************/

Boolean processes_dialog(void)
{
	static		ModalFilterUPP filter=NULL;

	OPTIONS		new_options;
	DialogPtr	dialog;
	short		item;

	if (filter == NULL)
		filter = NewModalFilterProc(ModalFilter);

	new_options=options;
	
	dialog=GetNewDialog(dialog_id_procesess,NULL,(WindowPtr)NULL);
	set_check_box(dialog,3,new_options.any_proc_eval);
	set_check_box(dialog,4,new_options.poll_eval);
		
	SelectWindow(dialog);
	ShowWindow(dialog);

	set_default_button(dialog,1,true);
	
	for(;;)
	{
		ModalDialog(filter,&item);
		switch(item)
		{
			case 3:
				new_options.any_proc_eval=!new_options.any_proc_eval;
				set_check_box(dialog,item,new_options.any_proc_eval);
				break;
			case 4:
				new_options.poll_eval=!new_options.poll_eval;
				set_check_box(dialog,item,new_options.poll_eval);
				break;
			case 1:
				options=new_options;
				DisposDialog(dialog);
				return true;
			case 2:
				DisposDialog(dialog);
				return false;
		}
	}
}

/********************************************************************************
 *																				*
 *		optimisation dialog														*
 *																				*
 ********************************************************************************/

Boolean optimisations_dialog(void)
{
	static		ModalFilterUPP filter=NULL;

	OPTIONS		new_options;
	DialogPtr	dialog;
	short		item;
	
	if (filter == NULL)
		filter = NewModalFilterProc(ModalFilter);

	new_options=options;
	
	dialog=GetNewDialog(dialog_id_optimisations,NULL,(WindowPtr)NULL);
	set_check_box(dialog,3,new_options.optimise_a_stack);
	set_check_box(dialog,4,new_options.optimise_b_stack);
	set_check_box(dialog,5,new_options.optimise_a_parameter);
	set_check_box(dialog,6,new_options.optimise_stack_checks);
	set_check_box(dialog,7,new_options.optimise_jumps);
	set_check_box(dialog,10,new_options.optimise_arith);
	
	SelectWindow(dialog);
	ShowWindow(dialog);
	
	set_default_button(dialog,1,true);
	
	for(;;)
	{
		ModalDialog(filter,&item);
		switch(item)
		{
			case 3:
				new_options.optimise_a_stack=!new_options.optimise_a_stack;
				set_check_box(dialog,item,new_options.optimise_a_stack);
				break;
			case 4:
				new_options.optimise_b_stack=!new_options.optimise_b_stack;
				set_check_box(dialog,item,new_options.optimise_b_stack);
				break;
			case 5:
				new_options.optimise_a_parameter=!new_options.optimise_a_parameter;
				set_check_box(dialog,item,new_options.optimise_a_parameter);
				break;
			case 6:
				new_options.optimise_stack_checks=!new_options.optimise_stack_checks;
				set_check_box(dialog,item,new_options.optimise_stack_checks);
				break;
			case 7:
				new_options.optimise_jumps=!new_options.optimise_jumps;
				set_check_box(dialog,item,new_options.optimise_jumps);
				break;
			case 10:
				new_options.optimise_arith=!new_options.optimise_arith;
				set_check_box(dialog,item,new_options.optimise_arith);
				break;
			case 1:
				options=new_options;
				DisposDialog(dialog);
				return true;
			case 2:
				DisposDialog(dialog);
				return false;
		}
	}
}

/********************************************************************************
 *																				*
 *		checks dialog															*
 *																				*
 ********************************************************************************/
 
Boolean checks_dialog(void)
{
	static		ModalFilterUPP filter=NULL;

	OPTIONS		new_options;
	DialogPtr	dialog;
	short		item;
	
	if (filter == NULL)
		filter = NewModalFilterProc(ModalFilter);

	new_options=options;
	
	dialog=GetNewDialog(dialog_id_checks,NULL,(WindowPtr)NULL);
	set_check_box(dialog,3,new_options.check_stacks);
	set_check_box(dialog,4,new_options.check_arithmetic);
	set_hilite_state(dialog,5,false);

	SelectWindow(dialog);
	ShowWindow(dialog);
		
	set_default_button(dialog,1,true);

	for(;;)
	{
		ModalDialog(filter,&item);
		switch(item)
		{
			case 3:
				new_options.check_stacks=!new_options.check_stacks;
				set_check_box(dialog,item,new_options.check_stacks);
				break;
			case 4:
				new_options.check_arithmetic=!new_options.check_arithmetic;
				set_check_box(dialog,item,new_options.check_arithmetic);
				break;				
			case 1:
				options=new_options;
				DisposDialog(dialog);
				return true;
			case 2:
				DisposDialog(dialog);
				return false;
		}
	}
}

/************************************************************************************
 *																					*
 *		output dialog																*
 *																					*
 ************************************************************************************/
 
Boolean output_dialog(void)
{
	static		ModalFilterUPP filter=NULL;

	OPTIONS		new_options;
	DialogPtr	dialog;
	short		item;
	
	if (filter == NULL)
		filter = NewModalFilterProc(ModalFilter);

	new_options=options;

	dialog=GetNewDialog(dialog_id_assembly_format,NULL,(WindowPtr)NULL);
	set_check_box(dialog,3,new_options.output_full_label_names);
	set_check_box(dialog,4,new_options.output_debug_symbols);
	set_check_box(dialog,9,new_options.output_intermediate_code1);
	set_check_box(dialog,10,new_options.output_intermediate_code2);
	set_check_box(dialog,11,new_options.output_intermediate_code3);
	set_check_box(dialog,12,new_options.output_settings);

	SelectWindow(dialog);
	ShowWindow(dialog);
	
	switch (options.assembler)
	{
		case PACT_ASSEMBLER:
			set_radio_button(dialog,6,true);
			set_radio_button(dialog,7,false);
			break;
		
		case PARC_ASSEMBLER:
			set_radio_button(dialog,6,false);
			set_radio_button(dialog,7,true);
			break;
		
		default:
			internal_error("unknown assembler type");
	}
	
	set_default_button(dialog,1,true);

	for (;;)
	{
		ModalDialog(filter,&item);
		switch(item)
		{
			case 3:
				new_options.output_full_label_names=!new_options.output_full_label_names;
				set_check_box(dialog,item,new_options.output_full_label_names);
				break;
			case 4:
				new_options.output_debug_symbols=!new_options.output_debug_symbols;
				set_check_box(dialog,item,new_options.output_debug_symbols);
				break;
			case 9:
				new_options.output_intermediate_code1=!new_options.output_intermediate_code1;
				set_check_box(dialog,item,new_options.output_intermediate_code1);
				break;
			case 10:
				new_options.output_intermediate_code2=!new_options.output_intermediate_code2;
				set_check_box(dialog,item,new_options.output_intermediate_code2);
				break;
			case 11:
				new_options.output_intermediate_code3=!new_options.output_intermediate_code3;
				set_check_box(dialog,item,new_options.output_intermediate_code3);
				break;
			case 12:
				new_options.output_settings=!new_options.output_settings;
				set_check_box(dialog,item,new_options.output_settings);
				break;
			case 6:
				if (new_options.assembler!=PACT_ASSEMBLER)
				{
					set_radio_button(dialog,7,false);
					set_radio_button(dialog,6,true);
					new_options.assembler=PACT_ASSEMBLER;
				}
				break;
			case 7:
				if (new_options.assembler!=PARC_ASSEMBLER)
				{
					set_radio_button(dialog,6,false);
					set_radio_button(dialog,7,true);
					new_options.assembler=PARC_ASSEMBLER;
				}
				break;
			case 1:
				options=new_options;
				DisposDialog(dialog);
				return true;
			case 2:
				DisposDialog(dialog);
				return false;
		}
	}	
}

/************************************************************************************
 *																					*
 *		input dialog																*
 *																					*
 ************************************************************************************/
 
Boolean input_dialog(void)
{
	static		ModalFilterUPP filter=NULL;

	OPTIONS		new_options;
	DialogPtr	dialog;
	short		item;
	
	if (filter == NULL)
		filter = NewModalFilterProc(ModalFilter);

	new_options=options;

	dialog=GetNewDialog(dialog_id_abc_format,NULL,(WindowPtr)NULL);
	set_check_box(dialog,3,new_options.inline_instructions);
	set_check_box(dialog,4,new_options.parallel_code);
	set_check_box(dialog,6,new_options.implicit_imports);
	set_check_box(dialog,7,new_options.ignore_depend);
	set_check_box(dialog,8,new_options.ignore_defer);

	SelectWindow(dialog);
	ShowWindow(dialog);
		
	set_default_button(dialog,1,true);

	for (;;)
	{
		ModalDialog(filter,&item);
		switch(item)
		{
			case 3:
				new_options.inline_instructions=!new_options.inline_instructions;
				set_check_box(dialog,item,new_options.inline_instructions);
				break;
			case 4:
				new_options.parallel_code=!new_options.parallel_code;
				set_check_box(dialog,item,new_options.parallel_code);
				break;
			case 6:
				new_options.implicit_imports=!new_options.implicit_imports;
				set_check_box(dialog,item,new_options.implicit_imports);
				break;
			case 7:
				new_options.ignore_depend=!new_options.ignore_depend;
				set_check_box(dialog,item,new_options.ignore_depend);
				break;
			case 8:
				new_options.ignore_defer=!new_options.ignore_defer;
				set_check_box(dialog,item,new_options.ignore_defer);
				break;
			case 1:
				options=new_options;
				DisposDialog(dialog);
				return true;
			case 2:
				DisposDialog(dialog);
				return false;
		}
	}	
}

/************************************************************************************
 *																					*
 *		messages dialog																*
 *																					*
 ************************************************************************************/
 
Boolean messages_dialog(void)
{
	static		ModalFilterUPP filter=NULL;

	OPTIONS		new_options;
	DialogPtr	dialog;
	short		item;
	
	if (filter == NULL)
		filter = NewModalFilterProc(ModalFilter);

	new_options=options;

	dialog=GetNewDialog(dialog_id_messages,NULL,(WindowPtr)NULL);
	set_check_box(dialog,3,new_options.verbose_compile);

	SelectWindow(dialog);
	ShowWindow(dialog);
		
	set_default_button(dialog,1,true);

	for (;;)
	{
		ModalDialog(filter,&item);
		switch(item)
		{
			case 3:
				new_options.verbose_compile=!new_options.verbose_compile;
				set_check_box(dialog,item,new_options.verbose_compile);
				break;
			case 1:
			{
				options=new_options;
				
				if (options.verbose_compile)
					windows.generate_dialog=GetNewDialog(Cdialog_id_generate_verbose,NULL,(WindowPtr)NULL);
				else
					windows.generate_dialog=GetNewDialog(Cdialog_id_generate,NULL,(WindowPtr)NULL);
				
				DisposDialog(dialog);
				return true;
			}
			case 2:
				DisposDialog(dialog);
				return false;
		}
	}	
}

/************************************************************************************
 *																					*
 *		paths dialog																*
 *																					*
 ************************************************************************************/

static void draw_paths_dialog(DialogPtr dialog)
{
	set_default_button(dialog,1,true);
	frame_item(dialog,5);
	frame_item(dialog,7);
	frame_item(dialog,9);
}

Boolean paths_dialog(void)
{
	static		ModalFilterUPP filter=NULL;

	OPTIONS		new_options;
	DialogPtr	dialog;
	short		item;
	
	if (filter == NULL)
		filter = NewModalFilterProc(ModalFilter);

	new_options=options;
	
	dialog=GetNewDialog(dialog_id_paths,NULL,(WindowPtr)NULL);
	set_item_tchars(dialog,5,new_options.source_directory);
	set_item_tchars(dialog,7,new_options.depend_directory);
	set_item_tchars(dialog,9,new_options.assembly_directory);

	SelectWindow(dialog);
	ShowWindow(dialog);
	
	draw_paths_dialog(dialog);
	
	for (;;)
	{
		ModalDialog(filter,&item);
		switch(item)
		{
			case 10:
				GetPath(new_options.source_directory);
				set_item_tchars(dialog,5,new_options.source_directory);
				draw_paths_dialog(dialog);
				break;

			case 11:
				GetPath(new_options.depend_directory);
				set_item_tchars(dialog,7,new_options.depend_directory);
				draw_paths_dialog(dialog);
				break;
				
			case 12:
				GetPath(new_options.assembly_directory);
				set_item_tchars(dialog,9,new_options.assembly_directory);
				draw_paths_dialog(dialog);
				break;
			
			case 1:
				options=new_options;
				DisposDialog(dialog);
				return true;
			case 2:
				DisposDialog(dialog);
				return false;
		}
	}
}