
/* standard c headers */

#include <string.h>
#include <stdarg.h>
#include <setjmp.h>

#include "ifaux.h"
#include "ifgenstate.h"
#include "ifkeys.h"
#include "ifinterface.h"
#include "ifmacinterface.h"

#include "cgmemory.h"
#include "cg.h"
#include "cgscanner.h"
#include "cgoutputabc.h"

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

static Boolean					generator_running;
static jmp_buf					abort_position;

static struct code_stage
{
	long			position;
	long			maximum;
	
	Rect			frame;
	Rect			fill;
	Rect			nofill;
	
	PixPatHandle	fill_pattern;
	PixPatHandle	nofill_pattern;

	long			max_fill;
	
} code_stage;

/********************************************************************************
 *																				*
 *		trace messages (called by code generator)								*
 *																				*
 ********************************************************************************/

void vset_file_stage(char *stage,va_list argp)
{
	if (options.verbose_compile)
		vset_item_chars(windows.generate_dialog,4,stage,argp);
}

void vset_code_stage(char *stage,va_list argp)
{
	vset_item_chars(windows.generate_dialog,2,stage,argp);
}

void set_code_stage (char *stage, ...)
{
	va_list argp;
	
    va_start(argp, stage);
    vset_code_stage(stage, argp);
    va_end(argp);
}

void set_file_stage (char *stage, ...)
{
	va_list argp;
	
    va_start(argp, stage);
    vset_file_stage(stage, argp);
    va_end(argp);
}

/********************************************************************************
 *																				*
 *		routines called by user interface										*
 *																				*
 ********************************************************************************/

static void init_code_stage(Boolean colour)
{
	short type;
	Handle item;
	
	/* get correct box of current generate dialog */
	
	GetDItem(windows.generate_dialog,3,&type,&item,&code_stage.frame);

	code_stage.position	=0L;
	code_stage.maximum	=1000L;
		
	code_stage.fill.left	=code_stage.frame.left+1;
	code_stage.fill.top		=code_stage.frame.top+1;
	code_stage.fill.bottom	=code_stage.frame.bottom-1;
	code_stage.fill.right	=code_stage.fill.left;
	
	code_stage.nofill		=code_stage.fill;
	code_stage.nofill.right	=code_stage.frame.right-1;
	
	code_stage.max_fill		=code_stage.frame.right - code_stage.frame.left -2;

	if (colour)
	{	code_stage.fill_pattern		=GetPixPat(pattern_id_generate1);
		code_stage.nofill_pattern	=GetPixPat(pattern_id_generate2);
	} else
	{	code_stage.fill_pattern		=NULL;
		code_stage.nofill_pattern	=NULL;
	}
}

static void update_code_stage(long position, long maximum)
{
	code_stage.position	=position;
	code_stage.maximum	=maximum;
	
	if (maximum > 0 && position < maximum)
		code_stage.fill.right=code_stage.fill.left+(position*code_stage.max_fill)/maximum;
	else
		code_stage.fill.right=code_stage.fill.left+code_stage.max_fill;
	
	code_stage.nofill.left=code_stage.fill.right;

	SetPort(windows.generate_dialog);
	InvalRect(&code_stage.fill);
}

Boolean init_generate_window(void)
{
	windows.generate_dialog=GetNewDialog(options.verbose_compile ? Cdialog_id_generate_verbose : Cdialog_id_generate,NULL,(WindowPtr)NULL);
	generator_running=false;
	
	return windows.generate_dialog != NULL;
}

void draw_code_stage(Boolean colour)
{
	SetPort(windows.generate_dialog);
	
	FrameRect(&code_stage.frame);

	if (colour)
	{
		FillCRect(&code_stage.nofill,code_stage.nofill_pattern);
		FillCRect(&code_stage.fill,code_stage.fill_pattern);
		
	} else
		FillRect(&code_stage.fill,&qd.gray);	
}

/********************************************************************************
 *																				*
 *		command execution														*
 *																				*
 ********************************************************************************/

void abort_code_generation(void)
{
	if (generator_running)
		longjmp(abort_position,1);
}

void execute_command(Boolean colour)
{
	char command[257];
	int	length;
	int position,i;
	char *p;
	
	if (generator_running)
		return;
		
	position	=(*windows.command_window.text)->selEnd;
	p=(char *)(*TEGetText(windows.command_window.text))+position-1;
	if (*p==returnKey || *p==enterKey)
	{	--p;
		--position;
	}
	length=position;
			
	if (length>255)
		length=255;
	
	command[length+1]='\0';
	
	for(i=length;i>0;--i)
	{	
		if (*p==' ' || *p==returnKey)
			break;
				
		command[i]=*p--;
	}
	command[i]=length-i;

	if (command[i])
	{
		SetWTitle		(windows.generate_dialog,(ConstStr255Param)&command[i]);
		SelectWindow	(windows.generate_dialog);
		ShowWindow		(windows.generate_dialog);
	
		TESetSelect(0,32767,windows.messages_window.text);
		TEDelete(windows.messages_window.text);
		
		alloc_code_memory();

		generator_running=true;

		if (setjmp(abort_position)==0)
		{
			init_code_stage(colour);
			draw_code_stage(colour);
			
			ticks_passed=TickCount();
			
			generate_code(&command[i+1]);
			
			update_code_stage(1000L,1000L);
			draw_code_stage(colour);
			
			/* if everything goes well, select the command window again */
			
			SelectWindow (windows.command_window.window);
		}
		generator_running=false;

		close_all_files();
		free_code_memory();
		
		HideWindow(windows.generate_dialog);		
	}
}

/********************************************************************************
 *																				*
 *		event handling during code generation									*
 *																				*
 ********************************************************************************/

void handle_generate_intermediate(void)
{
	/* see handle_compute_event: this routine prints some extra info */
	
	if(TickCount()-ticks_passed >= 6)
	{
		long position, maximum;
		
		get_source_position(&position,&maximum);		
		update_code_stage(position,maximum+maximum);
		
		handle_event();
		ticks_passed=TickCount();
	}
}

void handle_write_code(void)
{
	/* see handle_compute_event: this routine prints some extra info */
	
	if(TickCount()-ticks_passed >= 6)
	{
		long position, maximum;
		
		get_code_position(&position,&maximum);		
		update_code_stage(maximum+position,maximum+maximum);
		
		handle_event();
		ticks_passed=TickCount();
	}
}