
#include "ifkeys.h"
#include "ifaux.h"
#include "ifwindows.h"
#include "ifdialogs.h"
#include "ifgenstate.h"
#include "ifinterface.h"
#include "ifmacinterface.h"
#include "ifserver.h"

/* #define _ERIC_ */

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

OPTIONS				options;

static SysEnvRec	theWorld;

unsigned long		ticks_passed;

struct Strings		strings;
struct Windows		windows;
struct Menus		menus;

static Boolean		changed_prefs;

#ifdef POWERPC
	static char pref_file[]	="abc->trans Prefs (ppc)";
#else
	static char pref_file[]	="abc->trans Prefs";
#endif

/************************************************************************************
 *																					*
 *		initialisation																*
 *																					*
 ************************************************************************************/

static void DefaultPrefs (void)
{
	options.max_errors					=20;

	options.verbose_compile				=false;
	options.verbose_warnings			=true;
	options.verbose_errors				=true;
	options.verbose_messages			=true;

	options.parallel_code				=true;
	options.inline_instructions			=true;
	options.implicit_imports			=true;
	options.ignore_depend				=true;
	options.ignore_defer				=true;
	
	options.output_full_label_names		=false;
	options.output_debug_symbols		=false;
	options.output_settings				=true;

	options.assembler					=PACT_ASSEMBLER;

	options.output_intermediate_code1	=false;
	options.output_intermediate_code2	=false;
	options.output_intermediate_code3	=false;

	options.optimise_arith				=true;
	options.optimise_jumps				=true;
	options.optimise_a_stack			=true;
	options.optimise_b_stack			=true;
	options.optimise_a_parameter		=true;
	options.optimise_stack_checks		=true;

#ifdef _ERIC_
	options.any_proc_eval				=false;
#else
	options.any_proc_eval				=true;
#endif
	options.poll_eval					=false;
	
	options.check_stacks				=true;
	options.check_arithmetic			=false;

	strcpy(options.source_extension,		".abc");
	strcpy(options.assembly_extension,		".s");
	strcpy(options.object_extension,		".o");
	
	strcpy(options.intermediate_extension1,	".im1");
	strcpy(options.intermediate_extension2,	".im2");
	strcpy(options.intermediate_extension3,	".im3");
	
	strcpy(options.depend_directory,	"");
	strcpy(options.source_directory,	"");
	strcpy(options.assembly_directory,	"");
	strcpy(options.object_directory,	"");

	/* set window position to invalid ones, so prefs will be ignored */

	options.graphics_position.hsize		=0;
	options.graphics_position.vsize		=0;
	options.messages_position.hsize		=0;
	options.messages_position.vsize		=0;
	options.commands_position.hsize		=0;
	options.commands_position.vsize		=0;
}

static void InitPreferences (void)
{
	/* intialise code generation options and window positions */

	FILE *prefs;

	changed_prefs=false;

	/* if possible, read preferences from file, else keep default options */

	if ((prefs=fopen (pref_file, "rb"))!= NULL)
	{
		if (fread(&options,sizeof(OPTIONS),1,prefs)!=1)
			DefaultPrefs();
	
		fclose(prefs);
		
	} else
		DefaultPrefs();
}

static void InitManagers(void)
{
	InitGraf(&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs((long)NULL);
}

static Boolean init_menu_bar(void)
{
	Handle menu_bar;
	
	/* get variable menu strings */
	
	if ((strings.enable_string		=GetString(string_id_enable))	==NULL ||
		(strings.disable_string		=GetString(string_id_disable))	==NULL)
		return false;
	
	/* get menu bar from resource file */
	
	if ((menu_bar=GetNewMBar(128))==NULL)
		return false;
		
	SetMenuBar(menu_bar);
	
	/* get pull-down menus from resource file */
	
	if ((menus.apple_menu		=GetMHandle(menu_id_apple))		== NULL ||
		(menus.file_menu		=GetMHandle(menu_id_file))		== NULL	||
		(menus.edit_menu		=GetMHandle(menu_id_edit))		== NULL	||
		(menus.options_menu		=GetMHandle(menu_id_options))	== NULL	||
		(menus.remote_menu		=GetMHandle(menu_id_remote))	== NULL	||
		(menus.window_menu		=GetMHandle(menu_id_window))	== NULL)
		return false;
	
	/* add system stuff to apple menu */
	
	AddResMenu(menus.apple_menu,'DRVR');
					
	DrawMenuBar();
	
	return true;
}

/********************************************************************************
 *																				*
 *		cleaning up																*
 *																				*
 ********************************************************************************/

void quit_application(int n)
{
	WindowPosition pos;

	GetWindowPosition(&windows.graphics_window,&pos);
	if (!EqualWindowPosition(&pos,&options.graphics_position))
	{
		options.graphics_position=pos;
		changed_prefs=true;
	}

	GetWindowPosition(&windows.messages_window,&pos);
	if (!EqualWindowPosition(&pos,&options.messages_position))
	{
		options.messages_position=pos;
		changed_prefs=true;
	}

	GetWindowPosition(&windows.command_window,&pos);
	if (!EqualWindowPosition(&pos,&options.commands_position))
	{
		options.commands_position=pos;
		changed_prefs=true;
	}

	if (changed_prefs)
	{
		FILE *prefs;
	
		/* try to save preferences and window positions */
	
		if ((prefs = fopen (pref_file, "wb"))!=NULL)
		{
			FileParam		pb;
			Str255			filename;
	
			fwrite(&options,sizeof(OPTIONS),1,prefs);

			fclose(prefs);
	
			set_str255(filename,pref_file);
	
			pb.ioNamePtr	=filename;
			pb.ioVRefNum	=0;
			pb.ioFVersNum	=0;
			pb.ioFDirIndex	=0;
				
			if (PBGetFInfoSync((ParmBlkPtr)&pb)==noErr)
			{
				pb.ioFlFndrInfo.fdType		='PREF';
				pb.ioFlFndrInfo.fdCreator	='MKcg';
		
				PBSetFInfoSync((ParmBlkPtr)&pb);
			}
		}
	}


	if (!DisposeRemote())
		return;

	DisposeGraphWindow(&windows.graphics_window);
	DisposeGraphWindow(&windows.command_window);
	DisposeGraphWindow(&windows.messages_window);
	
	exit(n);
}

/************************************************************************************
 *																					*
 *		menu handling																*
 *																					*
 ************************************************************************************/
 
static void handle_menu_command(long menu_result)
{
	short	menu;
	short	item;
	
	item=LoWord(menu_result);
	menu=HiWord(menu_result);
	
	switch(menu)
	{
		case menu_id_apple:
		{	switch(item)
			{
				case menu_item_about:
					show_about_box();
					break;
				
				default:
				{
					Str255		name;
					GrafPtr		current_port;
										
					GetPort(&current_port);
					GetItem(menus.apple_menu,item,name);
					OpenDeskAcc(name);
					SetPort(current_port);
				}
			}
			break;
		}
		case menu_id_file:
		{	switch(item)
			{
				case menu_item_quit:
					quit_application(0);
				
				default:
					break;
			}
			break;
		}
		
		case menu_id_edit:
		{
			if (SystemEdit(item-1))
				return;
			
			/* edit yourself */
			
			break;
		}
		case menu_id_options:
		{
			switch(item)
			{
				case menu_item_processes:
					changed_prefs=processes_dialog();
					break;
				case menu_item_optimisations:
					changed_prefs=optimisations_dialog();
					break;
				case menu_item_checks:
					changed_prefs=checks_dialog();
					break;
				case menu_item_input:
					changed_prefs=input_dialog();
					break;
				case menu_item_output:
					changed_prefs=output_dialog();
					break;
				case menu_item_messages:
					changed_prefs=messages_dialog();
					break;
				case menu_item_paths:
					changed_prefs=paths_dialog();
					break;
			}
			break;
		}
		case menu_id_remote:
		{
			switch(item)
			{
				case menu_item_force_disconnect:

					DisconnectRemote();				
					break;

				case menu_item_remote_enable:
				
					FlipRemoteAble();
					break;
			}
			break;
		}
		case menu_id_window:
		{
			struct GraphWindow *window;
			
			switch(item)
			{
				case menu_item_command_window:
					window=&windows.command_window;
					break;
				case menu_item_messages_window:
					window=&windows.messages_window;
					break;
				case menu_item_graphics_window:
					window=&windows.graphics_window;
					AllocWindowOffScreen(&windows.graphics_window,8);
					break;
				default:
					internal_error("handle_menu_command: menu_id_window");
			}
				
			if (window->window != FrontWindow())
				SelectWindow(window->window);
	
			ShowWindow(window->window);
			
			break;
		}
	}
	HiliteMenu(0);
}

/********************************************************************************
 *																				*
 *		mouse handling															*
 *																				*
 ********************************************************************************/

static void close_window(WindowPtr window)
{
	if (window == windows.graphics_window.window)
		DisposWindowOffScreen(&windows.graphics_window);
		
	HideWindow(window);
}

static void do_mouse_down(EventRecord *event)
{
	WindowPtr event_window;
	
	switch(FindWindow(event->where,&event_window))
	{
		case inDesk:
			return;
			
		case inMenuBar:
			handle_menu_command(MenuSelect(event->where));
			return;
		
		case inSysWindow:
			SystemClick(event,event_window);
			return;
		
		case inContent:
		{
			if (event_window != FrontWindow())
			{	SelectWindow(event_window);
				return;
			}
			
			/* if click in text window, handle it */
			
			if (event_window == windows.command_window.window)
				GraphWindowClick(&windows.command_window,event);
			if (event_window == windows.messages_window.window)
				GraphWindowClick(&windows.messages_window,event);
						
			return;
		}
		case inDrag:
		{	
			Rect drag_rect;

			SetRect(&drag_rect,4,24,qd.screenBits.bounds.right-4,qd.screenBits.bounds.bottom-4);

			DragWindow(event_window,event->where,&drag_rect);
			return;
		}
		case inGrow:
		{
			/* if in text window, handle text grow */
			
			if (event_window == windows.command_window.window)
				GraphWindowGrow(&windows.command_window,event);
			if (event_window == windows.messages_window.window)
				GraphWindowGrow(&windows.messages_window,event);
				
			return;
		}
		case inGoAway:
		{	
			if (!TrackGoAway(event_window,event->where))
				return;
			
			close_window(event_window);			
			return;
		}
	}
}

/********************************************************************************
 *																				*
 *		keyboard handling														*
 *																				*
 ********************************************************************************/

static void do_key_down(EventRecord *event)
{
	char c;
	
	c=event->message & charCodeMask;
	
	if (event->modifiers & cmdKey)
	{
		switch (c)
		{
			case 'w':
				
				if (FrontWindow() != NULL)
					close_window(FrontWindow());
				
				break;
			
			default:
				handle_menu_command(MenuKey(c));
		}
	} else
	{
		/* depending on window, handle char */
		
		if (windows.command_window.window==FrontWindow())
		{
			if (c!=enterKey)
				GraphWindowKey(&windows.command_window,c);
			if (c==enterKey || c==returnKey)
				execute_command(theWorld.hasColorQD);
		}
		if (windows.messages_window.window==FrontWindow())
			GraphWindowKey(&windows.messages_window,c);
	}
}

/********************************************************************************
 *																				*
 *		event handling															*
 *																				*
 ********************************************************************************/

static Boolean do_dialog(EventRecord *event)
{
	DialogPtr	dialog;
	short		item;
	
	/*	Handle event for modeless dialog. If the some additional handling must be
		done this routine returns false. Otherwise it will return true.
	*/
	
	if (DialogSelect(event,&dialog,&item))
	{
		if (dialog==windows.generate_dialog && item==1)
			abort_code_generation();
			
		return true;
	}
	return false;
}

void do_activate(EventRecord *event)
{
	WindowPtr window;
	
	window=(WindowPtr)event->message;
	
	if (window==windows.command_window.window)
		GraphWindowActivate(&windows.command_window,event);
	else if (window==windows.messages_window.window)
		GraphWindowActivate(&windows.messages_window,event);
	else if (window==windows.graphics_window.window)
		GraphWindowActivate(&windows.graphics_window,event);
}

void do_update(EventRecord *event)	
{
	WindowPtr window;
	
	window=(WindowPtr)event->message;

	if (window==windows.generate_dialog)
	{
		draw_code_stage(theWorld.hasColorQD);
		return;
	}
	
	BeginUpdate(window);

	if (window == windows.command_window.window)
		GraphWindowUpdate(&windows.command_window,event);
	else if (window == windows.messages_window.window)
		GraphWindowUpdate(&windows.messages_window,event);
	else if (window == windows.graphics_window.window)
		GraphWindowUpdate(&windows.graphics_window,event);
	
	EndUpdate(window);
}

void handle_event(void)
{
	EventRecord event;
	
	/* perform system tasks */
	
	SystemTask();
	
	/* blink cursor of text windows */
	
	TEIdle(windows.command_window.text);
	TEIdle(windows.messages_window.text);
	
	/* get event and handle it */
	
	if (GetNextEvent(everyEvent,&event))
	{
		if (IsDialogEvent(&event) && do_dialog(&event))
			return;
		
		switch(event.what)
		{
			case mouseDown:
				do_mouse_down(&event);
				break;
			
			case keyDown:
			case autoKey:
				do_key_down(&event);
				break;
			
			case activateEvt:
				
				do_activate(&event);
				break;
			
			case updateEvt:
			
				do_update(&event);
				break;
			
			default:
				break;
		}
	}
}

void handle_compute_event(void)
{
	/*	this routine handles events while some computation is running. It only
		handles events once in a while, to avoid slowing down the computation.
	*/
	
	if(TickCount()-ticks_passed >= 6)
	{
		/* at least 1/10 second has passed, handle events */

		handle_event();
		ticks_passed=TickCount();
	}
}

/********************************************************************************
 *																				*
 *		main																	*
 *																				*
 ********************************************************************************/

int main(int argc, char *argv[])
{	
	/* intialise basic structures (these routines should not fail) */
		
	InitManagers();
	InitPreferences();

	FlushEvents(everyEvent,0);
	InitCursor();

	/* determine system enviroment for current version of SysEnvirons.
	   If this does not succeed: quit */
	
	if (SysEnvirons(curSysEnvVers,&theWorld) != noErr)
		return 1;
		
	InitTCP();
	
	/* draw desktop */
	
	if (!init_menu_bar() ||
		!init_generate_window())
		return 1;
	
	InitGraphWindow(&windows.messages_window,window_id_error,&options.messages_position,TRUE,TRUE,TRUE,theWorld.hasColorQD);
	InitGraphWindow(&windows.command_window,window_id_command,&options.commands_position,TRUE,TRUE,TRUE,theWorld.hasColorQD);
	InitGraphWindow(&windows.graphics_window,window_id_graphics,&options.graphics_position,FALSE,FALSE,FALSE,theWorld.hasColorQD);

	ShowWindow(windows.command_window.window);

	/* at this moment, override graphics window position */

	SizeWindow(windows.graphics_window.window,560,320,true);
	
	InitRemote();	

	EventHandler();
	
	return 0;
}