#include <string.h>

#include "ifwindows.h"
#include "ifinterface.h"
#include "ifplot.h"

void PlotPixel (long x,long y,long value,struct GraphWindow *w)
{
	unsigned char mask,shiftBits,*thePixel,pixelDepth;
	unsigned long rowBytes;
	PixMapHandle thePixMap;
	struct Rect rectangle;

	/* only plot if off-screen pixel map available */
	
	if (w->off_screen == NULL)
		return;
	
	thePixMap=((CGrafPtr)w->off_screen)->portPixMap;
	
	/* check if pixel within bounds of window pixel map... */

	if (x <	 (*thePixMap)->bounds.left	||
		x >= (*thePixMap)->bounds.right	||
		y <	 (*thePixMap)->bounds.top	||
		y >= (*thePixMap)->bounds.bottom)
	{
		error("PlotPixel: x: %ld, y: %ld\n",x,y);
		return;
	}

	LockPixels (thePixMap);
 
	rowBytes = (unsigned long) ((*thePixMap)->rowBytes & 0x3fff);	/* mask out high 3 bits */
	thePixel = (unsigned char *) (*thePixMap)->baseAddr;
    
	pixelDepth = (*thePixMap)->pixelSize;
 
	switch (pixelDepth)
	{
		case 1:
			thePixel += (rowBytes * y) + (((unsigned long) x) >> 3);
			shiftBits = 7 - (x & 7);
 
			mask = (unsigned char) 1 << shiftBits;
 			value &= 1;

			*thePixel &= ~mask;
			*thePixel |= (unsigned char) value << shiftBits;

			break;
		case 2:
			thePixel += (rowBytes * y) + (((unsigned long) 2 * x) >> 3);
 			shiftBits = 6 - ((x * 2) & 7);
 
			mask = (unsigned char) 3 << shiftBits;
 			value &= 3;

			*thePixel &= ~mask;
			*thePixel |= (unsigned char) value << shiftBits;

			break;
		case 4:
			thePixel += (rowBytes * y) + (((unsigned long) 4 * x) >> 3);
 
			shiftBits = 4 - ((x * 4) & 7);
 
			mask = (unsigned char) 15 << shiftBits;
			value &= 15;
 
			*thePixel &= ~mask;
			*thePixel |= (unsigned char) value << shiftBits;

			break;
		case 8:
			thePixel += (rowBytes * y) + (((unsigned long) 8 * x) >> 3);
   
			*thePixel = (unsigned char) value;

			break;
		case 16:
			thePixel += (rowBytes * y) + (2L * x);
 
			*((unsigned short *)thePixel) = (unsigned short) value;

			break;        
		case 24:    /* untested */
			thePixel += (rowBytes * y) + (3L * x);

			*thePixel = (unsigned char) (value >> 16);
			*((unsigned short *)(thePixel+1)) = (unsigned short) value;

			break;
		case 32:
			thePixel += (rowBytes * y) + (4L * x);
            
			*((unsigned long *)thePixel) = value;
	}
 
	UnlockPixels (thePixMap);
	
	/*
 		pixel has been set in off-chip pixel map, copy it to window
 	*/
 	
	rectangle.top		=y;
	rectangle.bottom	=y+1;
	rectangle.left		=x;
	rectangle.right		=x+1;

	SetPort(w->window);
	CopyBits ((BitMap *)*(w->off_screen->portPixMap),&(w->window->portBits),&rectangle,&rectangle,srcCopy,NULL);
}

static void PlotPixelMap8(long x,long y,long dx,long dy,long offset,long length,unsigned char *map,struct GraphWindow *w)
{
	unsigned char *BasePixel,pixelSize;
	unsigned long rowBytes;
	long skip_rows, skip_columns;
	PixMapHandle thePixMap;
		
	thePixMap=((CGrafPtr)w->off_screen)->portPixMap;
	pixelSize = (*thePixMap)->pixelSize;

	if (pixelSize!=8)
	{
		internal_error("PlotPixelMap8");
		return;
	}

	/* check if pixel map within bounds of window pixel map */

	if (x <	 (*thePixMap)->bounds.left			||
		y <	 (*thePixMap)->bounds.top			||
		x + dx > (*thePixMap)->bounds.right	||
		y + dy > (*thePixMap)->bounds.bottom)
	{
		error("PlotPixelMap8: x: %ld, y: %ld\n",x,y);
		return;
	}

	LockPixels (thePixMap);
 
	rowBytes = (unsigned long) ((*thePixMap)->rowBytes & 0x3fff);	/* mask out high 3 bits */
	BasePixel = (unsigned char *) (*thePixMap)->baseAddr;
    			
	/* first row inside pixel map */
	
	BasePixel += (rowBytes * y) + (((unsigned long) 8 * x) >> 3);

	/* find first pixel according to offset */
	
	skip_rows		=offset / dx;
	skip_columns	=offset % dx;
	
	BasePixel += (rowBytes * skip_rows);
	
	if (skip_columns != 0)
	{
		long remain;
		
		/* write first incomplete line */
		
		remain=dx-skip_columns;
		
		memcpy(BasePixel+skip_columns,map,remain);
		map+=remain;
		length-=remain;
		BasePixel+=rowBytes;
	}
	
	for (y=0; y<dy; y++)
	{
		/* copy pixels of a single row */
		
		if (dx < length)
		{
			/* at least an entire row is available */
			
			memcpy(BasePixel,map,dx);
			map+=dx;
			length-=dx;
			BasePixel+=rowBytes;
			
		}else
		{
			/* last line */
			
			memcpy(BasePixel,map,length);
			break;
		}
	}
 
	UnlockPixels (thePixMap);
}

void PlotPixelMap (long x,long y,long dx,long dy,long SrcSize,long offset,long length,Ptr map,struct GraphWindow *w)
{
	/* plots a rectangle of pixels in the given window, currently only
	   pixels of size 8 are supported, while the destination window should
	   have an off-chip map with pixels of size 4 or 8 and the rectangle should
	   be aligned on a byte boundary... */

	struct Rect rect;
		
	/* only plot if off-screen pixel map available */
	
	if (w->off_screen == NULL)
		return;
	 
 	switch (SrcSize)
 	{
 		case 8:
 			PlotPixelMap8 (x,y,dx,dy,offset,length,(unsigned char *)map,w);
 			break;
 		default:
 			internal_error("PlotPixelMap");
 			return;
 	}

	/* pixels have been set in off-chip pixel map, copy them to window */
 
 	rect.left	=x;
 	rect.top	=y;
 	rect.right	=x+dx;
 	rect.bottom	=y+dy;
 	
	SetPort(w->window);
	CopyBits ((BitMap *)*(w->off_screen->portPixMap),&(w->window->portBits),&rect,&rect,srcCopy,NULL);
}
