
#include <string.h>
#include <setjmp.h>

#include "ifinterface.h"
#include "cgscanner.h"
#include "cgmemory.h"
#include "cg.h"

/************************************************************************************
 *																					*
 *		global variables															*
 *																					*
 ************************************************************************************/

static FILE	*source_file;
static FILE	*depend_file;
static FILE *assembly_file;
static FILE *object_file;

static jmp_buf		abort_position;

/* variables used by the code generator */

OPTIONS options;

/************************************************************************************
 *																					*
 *		initialisation																*
 *																					*
 ************************************************************************************/

static void init_settings (void)
{
	options.max_errors					=10;

	options.verbose_compile				=false;
	options.verbose_warnings			=false;
	options.verbose_errors				=false;
	options.verbose_messages			=false;

	options.parallel_code				=true;
	options.inline_instructions			=true;
	options.implicit_imports			=true;
	options.ignore_depend				=true;
	options.ignore_defer				=true;
	
	options.output_full_label_names		=false;
	options.output_debug_symbols		=false;
	options.output_settings				=true;

	options.assembler					=PACT_ASSEMBLER;

	options.output_intermediate_code1	=false;
	options.output_intermediate_code2	=false;
	options.output_intermediate_code3	=false;

	options.optimise_arith				=true;
	options.optimise_jumps				=true;
	options.optimise_a_stack			=true;
	options.optimise_b_stack			=true;
	options.optimise_a_parameter		=true;
	options.optimise_stack_checks		=true;

#ifdef _ERIC_
	options.any_proc_eval				=false;
#else
	options.any_proc_eval				=true;
#endif
	options.poll_eval					=false;
	
	options.check_stacks				=true;
	options.check_arithmetic			=false;

	strcpy(options.source_extension,		".abc");
	strcpy(options.assembly_extension,		".s");
	strcpy(options.object_extension,		".o");
	
	strcpy(options.intermediate_extension1,	".im1");
	strcpy(options.intermediate_extension2,	".im2");
	strcpy(options.intermediate_extension3,	".im3");
	
	strcpy(options.depend_directory,	"");
	strcpy(options.source_directory,	"");
	strcpy(options.assembly_directory,	"");
	strcpy(options.object_directory,	"");
	
}

/********************************************************************************
 *																				*
 *		error handling															*
 *																				*
 ********************************************************************************/

void abort_code_generation(void)
{
	longjmp(abort_position,1);
}

void verror (char *mess,va_list argp)
{	
    vfprintf(stderr,mess, argp);    
}

void error (char *mess, ...)
{
	va_list argp;
	
    va_start(argp, mess);
    verror(mess, argp);
    va_end(argp);
}

void internal_error(char *mess, ...)
{	
	va_list argp;
	
	fprintf(stderr,"INTERNAL ERROR:\n");
	
    va_start(argp, mess);
    vfprintf(stderr,mess, argp);
    va_end(argp);
    	
	abort_code_generation();
}

/********************************************************************************
 *																				*
 *		code generation															*
 *																				*
 ********************************************************************************/

#define clear_line()	fprintf(stderr,"\r                                                                                \r")

void vset_file_stage(char *stage,va_list argp)
{
}

void vset_code_stage(char *stage,va_list argp)
{
	if (options.verbose_compile)
	{
		clear_line();
	
		vfprintf(stderr,stage,argp);
		putc('\r',stderr);
	
		fflush(stderr);
	}
}

void set_code_stage (char *stage, ...)
{
	va_list argp;
	
    va_start(argp, stage);
    vset_code_stage(stage, argp);
    va_end(argp);
}

void set_file_stage (char *stage, ...)
{
	va_list argp;
	
    va_start(argp, stage);
    vset_file_stage(stage, argp);
    va_end(argp);
}

void handle_compute_event(void)
{
}

void handle_generate_intermediate(void)
{
}

void handle_write_code(void)
{
}

int main(int argc, char *argv[])
{
	source_file		=NULL;
	depend_file		=NULL;
	assembly_file	=NULL;
	object_file		=NULL;
	
	init_settings();

	if (argc!=2)
	{	fprintf(stderr,"wrong number of arguments\n");
		exit(1);
	}
	
	alloc_code_memory();

	if (setjmp(abort_position)==0)
		generate_code(argv[1]);		/* generate code for first argument */

	close_all_files();
	free_code_memory();
	
	if (options.verbose_compile)
		putc('\n',stderr);
		
	return 0;
}