#ifndef _WINDOWS_
#define _WINDOWS_

#ifndef _TYPES_
#include "cctypes.h"
#endif

/************************************************************************************
 *																					*
 *		constants																	*
 *																					*
 ************************************************************************************/

#define scroll_resolution				1000

/************************************************************************************
 *																					*
 *		global typedefs																*
 *																					*
 ************************************************************************************/

typedef struct GraphWindow
{
	WindowPtr		window;					/* ptr to window structure */
	
	/* if window has text, the text handle != NULL and other fields defined */
	
	TEHandle		text;					/* handle to text record */
	Rect			text_rect;				/* rectangle to display text in */
	int				first_vis_line;			/* first visible line */
	
	/* if handle != NULL, scroll bars exist */
	
	ControlHandle	hbar;					/* control handle to horizontal bar (if any) */
	ControlHandle	vbar;					/* control handle to vertical bar (if any) */
	
	/* If the handle != NULL, an off-screen port exists.
	   This is not allocated by InitGraphWindow, but must be done
	   explicitly by calling AllocWindowOffScreen(graph_window, pixelsize),
	   which will allocate a port with the same size as the window */
	
	CGrafPtr		off_screen;
	
	/* possibly other structures, such as bitmaps for windows with graphics... */
	
} GraphWindow;

/************************************************************************************
 *																					*
 *		exported routines															*
 *																					*
 ************************************************************************************/

extern void AllocWindowOffScreen	(struct GraphWindow *window,int pixelsize);
extern void DisposWindowOffScreen	(struct GraphWindow *window);

extern void InitGraphWindow			(struct GraphWindow *window,int id, WindowPosition *pos, Boolean hscroll, Boolean vscroll, Boolean text, Boolean colour);
extern void DisposeGraphWindow		(struct GraphWindow *window);

extern void GraphWindowGrow			(struct GraphWindow *window,EventRecord *event);
extern void GraphWindowClick		(struct GraphWindow *window,EventRecord *event);
extern void GraphWindowActivate		(struct GraphWindow *window,EventRecord *event);
extern void GraphWindowUpdate		(struct GraphWindow *window,EventRecord *event);

extern void GraphWindowKey			(struct GraphWindow *window, char c);

extern void	GetWindowPosition		(struct GraphWindow *window, WindowPosition *pos);

#define EqualWindowPosition(p1,p2)	((p1)->hpos==(p2)->hpos && (p1)->vpos==(p2)->vpos && (p1)->hsize==(p2)->hsize && (p1)->vsize==(p2)->vsize)

#endif